dnl Check for signal system type
dnl This is not at all complete check yet, but it have to do for the moment.
dnl There is three systems detected, SYSV, 43BSD and POSIX
dnl There is two name collisions not resolved in terms of side-effect, they are
signal (SYSV and 43BSD) and
dnl sigpause (SYSV and 43BSD).

AC_DEFUN(AC_SYS_SIGNAL,
[
   dnl Tell what we are looking for
   AC_MSG_CHECKING(which signal system is being used)

   dnl Make possibly cached check
   AC_CACHE_VAL(ac_cv_sys_signal,
   [
      dnl Default value
      ac_cv_sys_signal=unknown

      dnl Check for SYSV signal (good old unix.... way before things got complex again :)

      dnl Check the SysV signal hook function
      AC_EGREP_HEADER(signal, signal.h, sysv_signal=yes, sysv_signal=no)

      dnl Check the SysV signal hook & mask function
      AC_EGREP_HEADER(sigset, signal.h, sysv_sigset=yes, sysv_sigset=no)

      dnl Check the SysV signal mask helpers
      AC_EGREP_HEADER(sighold, signal.h, sysv_sighold=yes, sysv_sighold=no)
      AC_EGREP_HEADER(sigrelse, signal.h, sysv_sigrelse=yes, sysv_sigrelse=no)
      AC_EGREP_HEADER(sigignore, signal.h, sysv_sigignore=yes, sysv_sigignore=no)
      AC_EGREP_HEADER(sigpause, signal.h, sysv_sigpause=yes, sysv_sigpause=no)

      dnl Summarize Check
      sysv_mask_helpers=no
      if test $sysv_sighold = yes -a $sysv_sigrelse = yes -a $sysv_sigignore = yes -a $sysv_sigpause = yes; then
         sysv_mask_helpers=yes
      fi
      if test $sysv_signal = yes -a $sysv_sigset = yes -a $sysv_mask_helpers = yes; then
         ac_cv_sys_signal=SYSV
      fi

      dnl Check for 4.3BSD signal system

      dnl Check the 4.3BSD signal hook function
      AC_EGREP_HEADER(signal, signal.h, _43bsd_signal=yes, _43bsd_signal=no)

      dnl Check the 4.3BSD signal hook function
      AC_EGREP_HEADER(sigvec, signal.h, _43bsd_sigvec=yes, _43bsd_sigvec=no)

      dnl Check the 4.3BSD signal pause
      AC_EGREP_HEADER(sigpause, signal.h, _43bsd_sigpause=yes, _43bsd_sigpause=no)

      dnl Check the 4.3BSD signal masker
      AC_EGREP_HEADER(sigsetmask, signal.h, _43bsd_sigsetmask=yes, _43bsd_sigsetmask=no)
      AC_EGREP_HEADER(sigblock, signal.h, _43bsd_sigblock=yes, _43bsd_sigblock=no)

      dnl Check the 4.3BSD signal mask helpers
      AC_EGREP_HEADER(sigmask, signal.h, _43bsd_sigmask=yes, _43bsd_sigmask=no)

      dnl Check the 4.3BSD unique features
      AC_EGREP_HEADER(sigstack, signal.h, _43bsd_sigstack=yes, _43bsd_sigstack=no)
      AC_EGREP_HEADER(sigreturn, signal.h, _43bsd_sigreturn=yes, _43bsd_sigreturn=no)

      dnl Summarize Check
      _43bsd_maskers=no
      if test $_43bsd_sigsetmask = yes -a $_43bsd_sigblock = yes -a $_43bsd_sigmask = yes -a $_43bsd_sigstack = yes -a $_43bsd_sigreturn = yes; then
         _43bsd_maskers=yes
      fi
      if test $_43bsd_signal = yes -a $_43bsd_sigvec = yes -a $_43bsd_sigpause = yes -a _43bsd_maskers ; then
         ac_cv_sys_signal=43BSD
      fi

      dnl Check for POSIX signal system

      dnl Check the POSIX signal hook function
      AC_EGREP_HEADER(sigaction, signal.h, posix_sigaction=yes, posix_sigaction=no)

      dnl Check the POSIX signal pause
      AC_EGREP_HEADER(sigsuspend, signal.h, posix_sigsuspend=yes, posix_sigsuspend=no)

      dnl Check the POSIX signal masker
      AC_EGREP_HEADER(sigprocmask, signal.h, posix_sigprocmask=yes, posix_sigprocmask=no)

      dnl Check the POSIX signal mask helpers
      AC_EGREP_HEADER(sigaddset, signal.h, posix_sigaddset=yes, posix_sigaddset=no)
      AC_EGREP_HEADER(sigdelset, signal.h, posix_sigdelset=yes, posix_sigdelset=no)
      AC_EGREP_HEADER(sigemptyset, signal.h, posix_sigemptyset=yes, posix_sigemptyset=no)
      AC_EGREP_HEADER(sigfillset, signal.h, posix_sigfillset=yes, posix_sigfillset=no)
      AC_EGREP_HEADER(sigismember, signal.h, posix_sigismember=yes, posix_sigismember=no)

      dnl Check the POSIX signal unique feature
      AC_EGREP_HEADER(sigpending, signal.h, posix_sigpending=yes, posix_sigpending=no)

      dnl Summarize check
      if test $posix_sigaction = yes -a $posix_sigpending = yes -a $posix_sigprocmask = yes -a $posix_sigsuspend = yes -a $posix_sigaddset = yes -a $posix_sigdelset = yes -a $posix_sigemptyset = yes -a $posix_sigismember = yes; then
         ac_cv_sys_signal=POSIX
      fi
   ])

   dnl Define separate flags
   if test $ac_cv_sys_signal = "43BSD"; then
      AC_DEFINE(HAVE_SIGNAL_43BSD)
   else
      if test $ac_cv_sys_signal = "POSIX"; then
         AC_DEFINE(HAVE_SIGNAL_POSIX)
      else
         if test $ac_cv_sys_signal = "SYSV"; then
            AC_DEFINE(HAVE_SIGNAL_SYSV)
         fi
      fi
   fi

   dnl Define common flag
   AC_DEFINE(HAVE_SIGNAL, $ac_cv_sys_signal)

   dnl Show result
   AC_MSG_RESULT($ac_cv_sys_signal)
])

dnl AC_TRY_CPPI(INCLUDES, [ACTION-IF-TRUE [, ACTION-IF-FALSE]])
AC_DEFUN(AC_TRY_CPPI,
[AC_REQUIRE_CPP()dnl
cat > conftest.$ac_ext <<EOF
[#]line __oline__ "configure"
#include "confdefs.h"
[$1]
EOF
dnl Capture the stderr of cpp.  eval is necessary to expand ac_cpp.
dnl We used to copy stderr to stdout and capture it in a variable, but
dnl that breaks under sh -x, which writes compile commands starting
dnl with ` +' to stderr in eval and subshells.
eval "$ac_cpp $INCLUDES conftest.$ac_ext >/dev/null 2>conftest.out"
ac_err=`grep -v '^ *+' conftest.out`
if test -z "$ac_err"; then
  ifelse([$2], , :, [rm -rf conftest*
  $2])
else
  echo "$ac_err" >&AC_FD_CC
ifelse([$3], , , [  rm -rf conftest*
  $3
])dnl
fi
rm -f conftest*])

dnl AC_FIND_HEADER(HEADER-FILE, DIRS, [ACTION-IF-FOUND [, ACTION-IF-NOT-FOUND]])
AC_DEFUN(AC_FIND_HEADER,
[dnl Do the transliteration at runtime so arg 1 can be a shell variable.
ac_safe=`echo "$1" | tr './\055' '___'`
AC_MSG_CHECKING([for $1])
AC_CACHE_VAL(ac_cv_header_$ac_safe,
[ac_save_INCLUDES="$INCLUDES"
AC_TRY_CPPI([#include <$1>], eval "ac_cv_header_$ac_safe=yes", eval "ac_cv_header_$ac_safe=no")
for ac_dir in $C_INCLUDE_PATH $2
do
 if eval "test \"`echo '$ac_cv_header_'$ac_safe`\" = no"; then
   INCLUDES="-I$ac_dir $ac_save_INCLUDES"
   AC_TRY_CPPI([#include <$1>], eval "ac_cv_header_$ac_safe=$ac_dir")
 fi
done
INCLUDES="$ac_save_INCLUDES"
])
if eval "test \"`echo '$ac_cv_header_'$ac_safe`\" = no"; then
  AC_MSG_RESULT(no)
  ifelse([$4], , , [$4])
else
  AC_MSG_RESULT(yes)
  ifelse([$3], , :, [$3])
  if eval "test \"`echo '$ac_cv_header_'$ac_safe`\" != yes"; then
     eval "INCLUDES=\"$INCLUDES -I`echo '$ac_cv_header_'$ac_safe`\""
  fi
fi
AC_SUBST(INCLUDES)dnl
])

dnl AC_FIND_XLIB(VARIABLE, LIBRARIES, FUNCTION, DIRS, [, ACTION-IF-FOUND
dnl              [, ACTION-IF-NOT-FOUND [, OTHER-LIBRARIES]]])
AC_DEFUN(AC_FIND_XLIB,
[AC_MSG_CHECKING([for $1 library])
changequote(, )dnl
ac_lib_var=`echo $1 | tr '.-/+' '___p'`
changequote([, ])dnl
AC_CACHE_VAL(ac_cv_libdir_$ac_lib_var, [eval "ac_cv_libdir_$ac_lib_var=no"])
AC_CACHE_VAL(ac_cv_lib_$ac_lib_var,
[
for ac_lib in $2
do

 ac_save_LIBS="$LIBS"
 dnl LIBS must come after $7 and XLIBS since that's where -lX11 -lm will be
 dnl XLIBS must come after XLIBDIRS so that dirs can be reused while adding
 dnl   XLIBS at the beginning
 LIBS="$XLIBDIRS -l$ac_lib $7 $XLIBS $LIBS"
 AC_TRY_LINK(, [$3()], eval "ac_cv_lib_$ac_lib_var=$ac_lib",
  eval "ac_cv_lib_$ac_lib_var=no")dnl
 for ac_dir in $LD_LIBRARY_PATH $4
 do
  if eval "test \"`echo '$ac_cv_lib_'$ac_lib_var`\" = no"; then
   if test -r $ac_dir/lib$ac_lib.a; then
    LIBS="$XLIBDIRS -L$ac_dir -l$ac_lib $7 $XLIBS $ac_save_LIBS"
    AC_TRY_LINK(, [$3()], [eval "ac_cv_libdir_$ac_lib_var=$ac_dir"
    eval "ac_cv_lib_$ac_lib_var=$ac_lib"])
   fi
  fi
 done
 LIBS="$ac_save_LIBS"

 if eval "test \"`echo '$ac_cv_lib_'$ac_lib_var`\" != no"; then
   break
 fi
done

])dnl
eval "ac_tmp=`echo '$ac_cv_lib_'$ac_lib_var`"
if test "$ac_tmp" != no; then
  AC_MSG_RESULT($ac_tmp)
  ifelse([$5], ,
[changequote(, )dnl
  ac_tr_lib=HAVE_LIB`echo $1 | tr 'abcdefghijklmnopqrstuvwxyz' 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'`
changequote([, ])dnl
  AC_DEFINE_UNQUOTED($ac_tr_lib)
  if eval "test \"`echo '$ac_cv_libdir_'$ac_lib_var`\" != no"; then
     eval "XLIBDIRS=\"$XLIBDIRS -L`echo '$ac_cv_libdir_'$ac_lib_var`\""
  fi
  eval "XLIBS=\"-l$ac_tmp $XLIBS\""
], [$5])
else
  AC_MSG_RESULT(no)
ifelse([$6], , , [$6
])dnl
fi
AC_SUBST(XLIBDIRS)
AC_SUBST(XLIBS)dnl
])
