/******************************************************************************/
/* BUILTIN COMMAND DEFINITIONS                                                */
/******************************************************************************/

/*
 * Mapping between Tcl commands and C functions
 */
command_t command[]={
 { "mv_tkwait",        Mv_TkwaitCmd, },
 { "mv_dijkstra_tree", do_mark_source, },
 { "mv_varlist",       do_varlist, },
 { "mv_fullname",      do_fullname, },
 { "mv_ipstr",         do_ipstr, },
 { "mv_query",         do_query,        },  /* Get user input */
 { "mv_dump_src_tree", do_dump_src_tree, },
 { "mv_dump_path",     do_dump_path, },   
 { "mv_param",         show_param, },
 { "mv_redraw",        redraw, },
 { "mv_rotate",        rotate, },
 { "mv_flip",          horiz_flip, },
 { "mv_csetx",         csetx, },
 { "mv_csety",         csety, },
#if 0
 { "mv_setage",        set_age,         },/* Change age parameter */
 { "mv_setthresh",     set_link_thresh, },/* Temporarily change threshhold */
 { "mv_metric",        set_metric,     }, /* Update link metric */
 { "mv_declink",       prev_link, },
 { "mv_inclink",       next_link,     },
 { "mv_travlink",      traverse_link, },
 { "mv_disable",       disable_link,   }, /* Temporarily Disable node */
 { "mv_showstats",     compute_stats, },  /* Display stats */
#endif
 { "mv_restore",       restore_node,   }, /* Restore/reload node */
 { "mv_addnode",       do_add_node,    }, /* Add a node */
 { "mv_addlink",       do_add_log_link,}, /* Add a link */
 { "mv_in",            zoom_in  ,    },
 { "mv_out",           zoom_out ,   },
 { "mv_reset",         zoom_reset,  },
 { "mv_toSource",      traverse_towards_source, },
 { "mv_child" ,        traverse_down, },
 { "mv_goto",          jumpto_node,  },
 { "mv_lsvar",         set_lsvar, },
 { "mv_nsvar",         set_nsvar, },
 { "mv_nicond",        set_nicond, },
 { "mv_ndcond",        set_ndcond, },
 { "mv_ldcond",        set_ldcond, },
 { "mv_nhcond",        set_nhcond, },
 { "mv_nodeRed",       set_nodeRed, },
 { "mv_nodeYellow",    set_nodeYellow, },
 { "mv_nodeGreen",     set_nodeGreen, },
 { "mv_linkRed",       set_linkRed, },
 { "mv_linkYellow",    set_linkYellow, },
 { "mv_linkGreen",     set_linkGreen, },
 { "mv_lhcond",        set_lhcond, },
 { "mv_nexthl",        next_hl,     },
 { "mv_find_name",     jumpto_name, },
 { "mv_delete_node",   do_delete_node, },
 { "mv_spread",        do_spring, },
 { "mv_tree",          do_tree, },
 { "mv_center",        do_center, },
 { "mv_nextcloud",     next_cloud, },
 { "mv_dirtoggle",     toggle_dir, },
 { "mv_update_node",   do_update_node, },
 { "mv_force_update_node", force_update_node, },
 { "mv_goto_cloud",    jumpto_cloud, },
 { "mv_markroot",      do_mark_root, },
 { "mv_marksrc",       do_mark_source, },
 { "mv_markdest",      do_mark_dest, },
 { "mv_marklogic",     mark_logic, },
 { "mv_quit",          exit_view, },
 { "mv_save",          save_everything, },
 { "exit",             exit_view, },
 { "mv_dumpnode",      dump_node, },
 { "mv_fixlinks",      fix_links, },
 { "mv_exec",          exec_alias,    },  /* Execute command (overrides tcl exec) */
 { "mv_start",         exec_handler,  },  /* Execute command with handler */
 { "mv_command",       text_cmd,      },  /* Execute command */
 { "mv_delrouters",    delete_routers, }, /* Delete all physical nodes */
 { "mv_hps",           show_handlers, },
 { "mv_schedule",      do_schedule, },
 { "mv_dumpevents",    do_dumpevents, },
 { "mv_tclcmd",        tcl_cmd, },
 { "mv_collect",       do_collect, },
 { "mv_nslookup",      mv_nslookup, },
 { "mv_autosys",       mv_autosys, },
 { "mv_asname",        mv_asname, },
 { "mv_snmpaware",     mv_snmpaware, },
 { "get_value",        tcl_get_value, },
 { "mv_postscript",    mv_postscript, },
 { "set_update_interval", tcl_set_update_interval, },
 { "mv_setCloudVar",   cloud_var, },
 { "is_local",         is_local, },
 { "mv_NodeYellowCrit_text", NodeYellowCrit_text },
 { "mv_NodeGreenCrit_text",  NodeGreenCrit_text },
 { "mv_NodeRedCrit_text",    NodeRedCrit_text },
 { "mv_LinkYellowCrit_text", LinkYellowCrit_text },
 { "mv_LinkGreenCrit_text",  LinkGreenCrit_text },
 { "mv_LinkRedCrit_text",    LinkRedCrit_text },
 { NULL,               NULL }           /* List MUST be NULL terminated */
};

/******************************************************************************/
/* OUTPUT HANDLER DEFINITIONS                                                 */
/******************************************************************************/

/* Output handler function prototypes:
 *
 *   All output handlers take the following arguments:
 *      s     : socket descriptor
 *      arg   : argument (if any) passed via Tcl command
 *      in    : buffer of output text to parse
 *      status: exit status of command
 */
void log_handler        PROTO((int s, char *arg, char *in, int status));
void nslookup           PROTO((int s, char *arg, char *in, int status));
void ping_handler       PROTO((int s, char *arg, char *in, int status));
void traceroute_hlr     PROTO((int s, char *arg, char *in, int status));
void dvmrp_vif_hlr      PROTO((int s, char *arg, char *in, int status));
void pim_if_hlr         PROTO((int s, char *arg, char *in, int status));
void popup              PROTO((int s, char *arg, char *in, int status));
void mrinfo_handler     PROTO((int s, char *arg, char *in, int status));
#if 0
void map_handler        PROTO((int s, char *arg, char *in, int status));
#endif
void snmpaware_hlr      PROTO((int s, char *arg, char *in, int status));
void packetrate_hlr     PROTO((int s, char *arg, char *in, int status));
void mrtree_hlr         PROTO((int s, char *arg, char *in, int status));
void autosys_hlr        PROTO((int s, char *arg, char *in, int status));
void asname_hlr         PROTO((int s, char *arg, char *in, int status));
void mgroups_hlr        PROTO((int s, char *arg, char *in, int status));
void grpsrcs_hlr        PROTO((int s, char *arg, char *in, int status));
void mtrace_hlr         PROTO((int s, char *arg, char *in, int status));
void mtrace_stats_hlr   PROTO((int s, char *arg, char *in, int status));
void ppg_hlr            PROTO((int s, char *arg, char *in, int status));

/* Mapping between output handler names in Tcl and C functions */
handler_name_t handler_name[]={
 { "log",              log_handler, },
 { "nslookup",         nslookup, },
 { "ping",             ping_handler, },
 { "traceroute",       traceroute_hlr, },
 { "dvmrpvif",         dvmrp_vif_hlr, },
 { "pimif",            pim_if_hlr, },
 { "popup",            popup, },
 { "mrinfo",           mrinfo_handler, },
#if 0
 { "map",              map_handler, },
#endif
 { "snmpaware",        snmpaware_hlr, },
 { "packetrate",       packetrate_hlr, },
 { "mcasttree",        mrtree_hlr, },
 { "mtracepath",       mtrace_stats_hlr, },
 { "autosys",          autosys_hlr, },
 { "asname",           asname_hlr, },
 { "mgroups",          mgroups_hlr, },
 { "grpsrcs",          grpsrcs_hlr, },
 { "pktspergroup",     ppg_hlr, },
 { NULL,               NULL           },  /* List MUST be NULL terminated */
};

/******************************************************************************/
/* USER VARIABLE DEFINITIONS                                                  */
/******************************************************************************/

/* Definitions of user variables, fields are:
 *   Variable name
 *   Type (see dict.h)
 *   Size in bytes of value
 *   Maximum array size
 *   Persistent or temporary (see dict.h)
 *   What variable is linked to (see dict.h)
 *   Maximum age before update (or 0 for no updates)
 *   Command to execute for an update (or NULL for no updates)
 *   String of variables this depends on
 */
variable_t variable[]={
{"version",   VT_STRING, 10, MAX_NODES, VF_PERSISTENT, VL_NODE,
    MONTH,  "mv_update_node $exp_index", NULL, },
{"nodestate", VT_INT,     1, MAX_NODES, VF_PERSISTENT, VL_NODE,  
    HOUR,  NULL, "version", },
{"fullname",  VT_STRING, 80, MAX_NODES, VF_PERSISTENT, VL_NODE,
    YEAR,   "mv_nslookup $exp_index", NULL, },
{"country",   VT_STRING,  4, MAX_NODES, VF_PERSISTENT, VL_NODE,
    0, NULL, "fullname", },
{"domain",    VT_STRING, 80, MAX_NODES, VF_PERSISTENT, VL_NODE,
    0, NULL, "fullname", },
{"hostname",  VT_STRING, 80, MAX_NODES, VF_PERSISTENT, VL_NODE,
    0, NULL, "fullname", },
{"autosys",   VT_INT,     4, MAX_NODES, VF_PERSISTENT, VL_NODE,
    MONTH,  "mv_autosys $exp_index",0, },
{"asname",    VT_STRING, 25, MAX_NODES, VF_PERSISTENT, VL_NODE,
    0, NULL, "autosys", },
{"cloud",     VT_INT,     4, MAX_NODES, VF_PERSISTENT, VL_NODE, 0, NULL, NULL,},
{"cloudname", VT_STRING, 25, MAX_NODES, VF_PERSISTENT, VL_CLOUD,0, NULL, NULL,},
{"cloudsize", VT_INT,     4, MAX_NODES, VF_PERSISTENT, VL_CLOUD,0, NULL, NULL},
{"metric",    VT_INT,     1, MAX_LINKS, VF_PERSISTENT, VL_LINK, 0, NULL, NULL,},
{"isGroupSource",VT_INT,  4, MAX_NODES, VF_TEMPORARY,  VL_NODE, 0, NULL, NULL,},
{"threshold", VT_INT,     1, MAX_LINKS, VF_PERSISTENT, VL_LINK, 0, NULL, NULL,},
{"srcip",     VT_STRING, 16, MAX_LINKS, VF_PERSISTENT, VL_LINK, 0, NULL, NULL,},
{"destip",    VT_STRING, 16, MAX_LINKS, VF_PERSISTENT, VL_LINK, 0, NULL, NULL,},
{"linkstate", VT_INT,     1, MAX_LINKS, VF_PERSISTENT, VL_LINK, 0, NULL, "nodestate", },
{"subtreeSize", VT_INT,   4, MAX_NODES, VF_PERSISTENT, VL_NODE, 0, NULL, NULL, },
{"onpath_node", VT_INT,   1, MAX_NODES, VF_TEMPORARY,  VL_NODE, 0, NULL, NULL, },
{"onpath_link", VT_INT,   1, MAX_LINKS, VF_TEMPORARY,  VL_LINK, 0, NULL, NULL, },
{"ontree_node", VT_INT, 1, MAX_NODES, VF_TEMPORARY,  VL_NODE, 0, NULL, NULL, },
{"ontree_link", VT_INT, 1, MAX_LINKS, VF_TEMPORARY,  VL_LINK, 0, NULL, NULL, },
{"snmpaware", VT_INT,     4, MAX_NODES, VF_PERSISTENT, VL_NODE,
    WEEK, "mv_snmpaware $exp_index", NULL, },
{"InPkts",    VT_UINT,     4, MAX_NODES, VF_TEMPORARY,  VL_NODE, MINUTE, "if [ speaks_snmp $exp_index ] { mv_start packetrate($exp_index) list.pkts [mv_ipstr $exp_index]}", "snmpaware", },
{"OutPkts",       VT_UINT,  4, MAX_NODES, VF_TEMPORARY, VL_NODE,0,NULL,"InPkts", },
{"InPktRate",     VT_FLOAT, 4, MAX_NODES, VF_TEMPORARY, VL_NODE,0,NULL,"InPkts", },
{"OutPktRate",    VT_FLOAT, 4, MAX_NODES, VF_TEMPORARY, VL_NODE,0,NULL,"InPkts", },
{"PktTimeStamp",  VT_INT,  4, MAX_NODES, VF_TEMPORARY,  VL_NODE,0,NULL,"InPkts", },
{"linkInPkts",    VT_FLOAT, 4, MAX_LINKS, VF_TEMPORARY, VL_LINK,0,NULL,"InPkts", },
{"linkOutPkts",   VT_FLOAT, 4, MAX_LINKS, VF_TEMPORARY, VL_LINK,0,NULL,"InPkts", },
{"linkInPktRate", VT_FLOAT, 4, MAX_LINKS, VF_TEMPORARY, VL_LINK,0,NULL,"InPkts", },
{"linkOutPktRate",VT_FLOAT, 4, MAX_LINKS, VF_TEMPORARY, VL_LINK,0,NULL,"InPkts", },
{"PktsPerGroup",  VT_UINT,  4, MAX_NODES, VF_TEMPORARY, VL_NODE,MINUTE, "if [ speaks_snmp $exp_index ] { pre_pktspergroup }", "snmpaware", },
{"PktsPerGrpSrc", VT_UINT,  4, MAX_NODES, VF_TEMPORARY, VL_NODE,0,NULL,"PktsPerGroup", },
{"PktsPerGrpSrcUpTime",VT_UINT,4,MAX_NODES,VF_TEMPORARY,VL_NODE,0,NULL,"PktsPerGroup", },
{"PktsPerGrpSrcRate",VT_FLOAT,4,MAX_NODES,VF_TEMPORARY,VL_NODE,0,NULL,"PktsPerGroup", },

{"cloudX",       VT_FLOAT,4,MAXCLOUDS,VF_PERSISTENT,VL_CLOUD,0,NULL,NULL},
{"cloudY",       VT_FLOAT,4,MAXCLOUDS,VF_PERSISTENT,VL_CLOUD,0,NULL,NULL},
{"cloudLink",    VT_INT,  2,MAXCLOUDS,VF_PERSISTENT,VL_CLOUD,0,NULL,NULL},
{"cloudlinkA",   VT_INT,2,MAXCLOUDLINKS,VF_PERSISTENT,VL_CLOUDLINK,0,NULL,NULL},
{"cloudlinkB",   VT_INT,2,MAXCLOUDLINKS,VF_PERSISTENT,VL_CLOUDLINK,0,NULL,NULL},
{"cloudlinkLog", VT_INT,2,MAXCLOUDLINKS,VF_PERSISTENT,VL_CLOUDLINK,0,NULL,NULL},
{"cloudlinkNext",VT_INT,2,MAXCLOUDLINKS,VF_PERSISTENT,VL_CLOUDLINK,0,NULL,NULL},

{NULL, 0, 0, 0, 0, 0, 0, NULL, NULL } /* List MUST be NULL terminated */
};
