/* $Id: builtin.c,v 1.47 1996/07/02 22:40:10 thalerd Exp $
 * -------------------------------------------------------------------------
 *  
 * [C] The Regents of the University of Michigan and Merit Network, Inc. 1996
 * All Rights Reserved 
 *   
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice and this permission notice appear in all 
 * copies of the software and derivative works or modified versions thereof, 
 * and that both the copyright notice and this permission and disclaimer notice 
 * appear in supporting documentation. 
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE REGENTS OF THE 
 * UNIVERSITY OF MICHIGAN AND MERIT NETWORK, INC. DO NOT WARRANT THAT THE 
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR 
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the 
 * University of Michigan and Merit Network, Inc. shall not be liable for any 
 * special, indirect, incidental or consequential damages with respect to any 
 * claim by Licensee or any third party arising from use of the software. 
 *
 * -------------------------------------------------------------------------
 */
/* BUILTIN.C - Built in Tcl commands */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h> /* for sscanf */
#include <stdlib.h> /* for atoi */
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_BSTRING_H
#include <bstring.h>
#endif
#include <netdb.h>
#include <ctype.h> /* for tolower */
#include <math.h>
#include <tcl.h>
#include <tk.h>
#ifndef HAVE_GETTIMEOFDAY
#include <sys/timeb.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h> /* for struct in_addr for arpa/inet.h */
#endif
#include <arpa/inet.h> /* for inet_addr */
#include "struct.h"
#include "globals.h"
#include "database.h"
#include "display.h"
#include "node.h"
#include "builtin.h"
#include "queue.h"
#include "expr.h" /* for get_cond */
#include "dict.h"
#include "high.h" /* for goto_cloud */
#include "driver.h" /* for text_cmd */
#include "cloud.h" /* for set_node_cloud */
#include "spring.h" /* for springembedder */
#include "graph.h" /* for mwatch_file */
#include "lib.h" /* for match */
#include "tree.h" /* for locate_tree */
#include "event.h" /* for schedule_command */
#include "dist_vect.h" /* for source_parent */

#define TK4
#define MINAGE 10 /* Minimum # seconds for age */

int
do_rot(x0,y0,rot,x1,y1)
   double x0, y0, rot, *x1, *y1;
{
   double r;

   rot *= atan(1.0)/45.0; /* convert to rads */
   r   = hypot(x0,y0);

   /* sincos(atan2(y0,x0)-rot, y1, x1);   not POSIX */
   *y1 = sin(atan2(y0,x0)-rot);
   *x1 = cos(atan2(y0,x0)-rot);

   *x1 *= r;
   *y1 *= r;
   return TCL_OK;
}

/* mv_tree : display a tree rooted at the root of the subtree */
int
do_tree(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   locate_tree(curr_node);
   rescale_nodes();
   ndcond.dirty = 1; /* force nodes to be redrawn */
   refresh_display();
   center_node(curr_node);
/*   goto_node(curr_node); */
   return TCL_OK;
}

/* mv_collect [filename]  (default: "out") */
int
do_collect(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char *filename;

   filename = (argc>1)? argv[1] : "out";
   mwatch_file(filename);
   return TCL_OK;
}

int
rotate(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   node_t *node;
   double x0,y0,x1,y1,rot;
   char *str;
   int idx = xy_idx();
   int i, n = num_nodes();

   str = (argc>1)? argv[1] : query("Degrees: ");
   if  ( *str == '\0' )
		return TCL_OK;

   sscanf(str, "%lf", &rot);
   if (fabs(rot) < 0.01)
      return TCL_OK;

   /* Rotate normal nodes */
   for (i=0; i<n; i++) {
      if (is_ndisplayed(i) && !is_ncloud(i)) {
         node = mget_node(i);
         x0 = node->x[idx];
         y0 = node->y[idx];
         do_rot(x0,y0,rot,&x1,&y1);
         node->x[idx]=x1;
         node->y[idx]=y1;
      }
   }
   
   /* Rotate cloud nodes */
   for (i=0; i<num_clouds; i++) {
      if (is_cdisplayed(i)) {
         x0 = cloud_x[i];
         y0 = cloud_y[i];
         do_rot(x0,y0,rot,&x1,&y1);
         cloud_x[i] = x1;
         cloud_y[i] = y1;
      }
   }

   rescale_nodes();

   ndcond.dirty = 1; /* force nodes to be redrawn */
   refresh_display();
   center_node(curr_node);
   return TCL_OK;
}

int
horiz_flip(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   node_t *node;
   int i;
   char *str;
   int idx = xy_idx();
   int n = num_nodes();

   str = (argc>1)? argv[1] : query("Horiz or Vert: ");
   if ( *str == '\0' )
		return TCL_OK;

   if (!match(str, "h_oriz") && !match(str, "v_ert")) {
      add_msg("Aborted.\n");
      return TCL_ERROR;
   }

   /* Flip normal nodes */
   for (i=0; i<n; i++) {
      if (is_ndisplayed(i) && !is_ncloud(i)) {
         node = mget_node(i);
   
         if (tolower(*str)=='v')
            node->y[idx] = 1000 - node->y[idx];
         else
            node->x[idx] = 1000 - node->x[idx];
      }
   }

   /* Flip cloud nodes */
   for (i=0; i<num_clouds; i++) {
      if (is_cdisplayed(i)) {
         if (tolower(*str)=='v')
            cloud_y[i] = 1000 - cloud_y[i];
         else
            cloud_x[i] = 1000 - cloud_x[i];
      }
   }

   ndcond.dirty = 1; /* force nodes to be redrawn */
   refresh_display();
   center_node(curr_node);
   return TCL_OK;
}

/* Change age parameter */
int
set_age(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   int   a;
   char *str;

   str = (argc>1)? argv[1] : query("Maximum age: ");
   if ( *str == '\0' )
    return TCL_OK;
 
   a = atoi(str);
   if (a>=MINAGE)
      maxage = a;
   return TCL_OK;
}

#if 0
/* Change link thresh */
int
set_link_thresh(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   log_link_t *edge;
   node_t     *node;
   char *str;

   str = (argc>1)? argv[1] : query("Threshold: ");
   if ( *str == '\0' )
     return TCL_OK;

   if (atoi(str)) {
      node = mget_node(curr_node);
      edge = get_log_link(node->link[iotype][cure] );
      set_int("threshold", node->link[iotype][cure], atoi(str));
      edge->flags |= ST_DONTSAVE;
   }
   return TCL_OK;
}

/* Change link metric */
int
set_metric(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char       *str;
   log_link_t *edge;
   node_t     *node;

   str = (argc>1)? argv[1] : query("Metric: ");
   if ( *str == '\0' )
     return TCL_OK;

   if (atoi(str)) {
      node = mget_node(curr_node);
      edge = get_log_link(node->link[iotype][cure] );
      set_int("metric", node->link[iotype][cure], atoi(str));
      edge->flags |= ST_DONTSAVE;
   }
   return TCL_OK;
}

/* Temporarily Disable link */
int
disable_link()
{
   node_t     *node;
   log_link_t *edge;

   node = mget_node(curr_node);
   edge = get_log_link(node->link[iotype][cure] );
   edge->flags |= ST_DONTSAVE;
   edge->flags ^= DVMRP_NF_DISABLED;
   return TCL_OK;
}
#endif

/* Restore/reload current node */
int
restore_node()
{
   node_t     *node;
   int         ln;   /* link index */
  
   node = mget_node(curr_node);
   load_node(curr_node, node);

   /* Restore logical links */
   for (ln = node->last_link[LIN]; ln>=0; ln = llink[ln].prev_link[LIN])
      load_log_link(ln, get_log_link(ln));
   for (ln = node->last_link[LOUT]; ln>=0; ln = llink[ln].prev_link[LOUT])
      load_log_link(ln, get_log_link(ln));

   goto_node(curr_node);
   return TCL_OK;
}
 
/* Add a node */
int
do_add_node()
{
    struct hostent *hp;
    u_int32 addr;
    char *str;
    int i;

    str = query("Node name: ");
    if ( *str == '\0' )
       return TCL_OK;

    if ((i=find_node_by_name(str))>=0)
       goto_node(i);
    else {
       hp = gethostbyname(str);
       if (hp) {
          bcopy(hp->h_addr, (char *) &addr, hp->h_length);
          goto_node(add_node(addr));
       } else
          add_msg("Unknown host");
    }
   return TCL_OK;
}

/* Add a link */
int
do_add_log_link()
{
   u_int32 addr;
   node_t *snode, *dnode;
   char   *str;
   int     i;

   snode = mget_node(curr_node);
   str = query("Dest: ");
   if ( *str == '\0' )
     return TCL_OK;

   if ((i=find_node_by_name(str))<0) {
      addr = inet_addr(str);
      if ((i = find_node_by_addr(addr))<0) {
         add_msg("Destination not found.");
         return TCL_OK;
      }
   } else { /* found node */
      dnode = mget_node(i);
      addr = dnode->iface[0];
   }
   add_log_link(curr_node,snode->iface[0], i,addr, ST_DONTSAVE);

   goto_node(curr_node);
   return TCL_OK;
}

/* mv_schedule <period> <command...> */
/* executes a command and schedules it to occur periodically */
int
do_schedule(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   int32 period;
   char command[255];
   int i;

   if (argc<3) {
      add_msg("usage: mv_schedule <period> <command>\n");
      return TCL_ERROR;
   }

   period   = atoi(argv[1]);
   strcpy(command, argv[2]);
   for (i=3; i<argc; i++) {
      strcat(command, " ");
      strcat(command, argv[i]);
   }
   {
#ifdef HAVE_GETTIMEOFDAY
      struct timeval now;
      struct timezone tz;
      gettimeofday(&now, &tz);
      schedule_command(now.tv_sec, now.tv_usec/1000, period, command);
#else
      struct timeb now;
      ftime(&now);
      schedule_command(now.time, now.millitm, period, command);
#endif
   }
   return TCL_OK;
}

/* mv_dumpevents */
int
do_dumpevents(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   dump_timers(VL_NODE); /* was dump_events() but that's not really used */
   return TCL_OK;
}

#if 0
/* This is now obsolete, and users should go to the web docs instead */
/* Display help screen */
int
help()
{
   FILE *fp;
   char buff[255];
   char *str;

   fp = fopen("help","r");
   while (fgets(buff,255,fp)) {
      wtmp_display(buff);
   }
   fclose(fp);

   str = query( "Hit ENTER to continue: ");
   add_msg("");
   return TCL_OK;
}
#endif

int
zoom_in()
{
   zoom *= 2.0;
   ndcond.dirty = TRUE; /* force nodes to be redisplayed */
   refresh_display();
   /* goto_node(curr_node); */
   center_node(curr_node);
   return TCL_OK;
}

int
zoom_out()
{
   zoom/=2.0;
   ndcond.dirty = TRUE; /* force nodes to be redisplayed */
   refresh_display();
   /* goto_node(curr_node); */
   center_node(curr_node);
   return TCL_OK;
}

int
zoom_reset()
{
   char buff[256];
#ifdef TK4
/*
   Tcl_Eval(tcl, sprintf(buff, ".c configure -xscrollincrement 1 -yscrollincrement 1"));
   Tcl_Eval(tcl, sprintf(buff, ".c xview moveto 0"));
   Tcl_Eval(tcl, sprintf(buff, ".c yview moveto 0"));
*/
   sprintf(buff, ".c configure -xscrollincrement 20 -yscrollincrement 1");
   Tcl_Eval(tcl, buff);
#else
/*
   Tcl_Eval(tcl, sprintf(buff, ".c configure -scrollincrement 1"));
   Tcl_Eval(tcl, sprintf(buff, ".c xview 0"));
   Tcl_Eval(tcl, sprintf(buff, ".c yview 0"));
*/
   Tcl_Eval(tcl, sprintf(buff, ".c configure -scrollincrement 20"));
#endif

   zoom=INITIAL_ZOOM;
   refresh_display();
   goto_node(curr_node);
   return TCL_OK;
}

/* Traverse towards source */
int
traverse_towards_source()
{
   int n = source_parent(curr_node);

   if (n==curr_node) {
      if (source_distance(curr_node))
         notify("Error", "Source is unreachable from here");
      else
         notify("Error", "Already at source");
   } else
      goto_node(n);
   return TCL_OK;
}

/* Traverse to child */
int
traverse_down()
{
   goto_node(path_child(curr_node)); 
   return TCL_OK;
}

static void
mv_get_cond(cond, argc, argv)
   cond_t *cond;
   int argc;
   char **argv;
{
   char *p, *q, *str;
   char tmp[80];

   str = (argc>1)? argv[1] : query("Enter condition:");
   get_cond(cond, str);
}

int
set_nhcond(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char buff[256];

   if (argc<2) {
      do_tcl("open_condsel nhcond node mv_nhcond");
   } else {
      stop_updates(nhcond.name);

      mv_get_cond(&nhcond, argc, argv);

      start_updates(nhcond.name);

      nhcond.dirty=1;
      refresh_display();
   }
   return TCL_OK;
}

int
set_nodeRed(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
	if (argc<2) {
		do_tcl("open_condsel nodeRed node mv_nodeRed");
	} else {
		stop_updates(nodeRed.name);

                mv_get_cond(&nodeRed, argc, argv);

		start_updates(nodeRed.name);

		nodeRed.dirty = 1;
		refresh_display();
	}

	text_cmd("set nred_crit_txt [ mv_NodeRedCrit_text ]");
	return TCL_OK;
}

int
set_nodeYellow(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
	if (argc<2) {
		do_tcl("open_condsel nodeYellow node mv_nodeYellow");
	} else {
		stop_updates(nodeYellow.name);

                mv_get_cond(&nodeYellow, argc, argv);

		start_updates(nodeYellow.name);

		nodeYellow.dirty=1;
		refresh_display();
	}

	text_cmd("set nyellow_crit_txt [ mv_NodeYellowCrit_text ]");

	return TCL_OK;
}

int
set_nodeGreen(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
	if (argc<2) {
		do_tcl("open_condsel nodeGreen node mv_nodeGreen");
	} else {
		stop_updates(nodeGreen.name);

                mv_get_cond(&nodeGreen, argc, argv);

		start_updates(nodeGreen.name);

		nodeGreen.dirty=1;
		refresh_display();
	}

	text_cmd("set ngreen_crit_txt [ mv_NodeGreenCrit_text ]");
	return TCL_OK;
}

int
set_linkRed(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
	if (argc < 2) {
		do_tcl("open_condsel linkRed link mv_linkRed");
	} else {
		stop_updates(linkRed.name);

                mv_get_cond(&linkRed, argc, argv);

		start_updates(linkRed.name);

		linkRed.dirty=1;
		refresh_display();
	}
	text_cmd("set lred_crit_txt [ mv_LinkRedCrit_text ]");
	return TCL_OK;
}

int
set_linkYellow(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
	if (argc<2) {
		do_tcl("open_condsel linkYellow link mv_linkYellow");
	} else {
		stop_updates(linkYellow.name);

                mv_get_cond(&linkYellow, argc, argv);

		start_updates(linkYellow.name);

		linkYellow.dirty=1;
		refresh_display();
	}
	text_cmd("set lyellow_crit_txt [ mv_LinkYellowCrit_text ]");
	return TCL_OK;
}

int
set_linkGreen(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
	if ( argc < 2 ) {
		do_tcl("open_condsel linkGreen link mv_linkGreen");
	} else {
		stop_updates(linkGreen.name);

                mv_get_cond(&linkGreen, argc, argv);

		start_updates(linkGreen.name);

		linkGreen.dirty=1;
		refresh_display();
	}
	text_cmd("set lgreen_crit_txt [ mv_LinkGreenCrit_text ]");
	return TCL_OK;
}

int
set_lhcond(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char buff[256];

   if (argc<2) {
      do_tcl("open_condsel lhcond link mv_lhcond");
   } else {
      stop_updates(lhcond.name);

      mv_get_cond(&lhcond, argc, argv);

      start_updates(lhcond.name);

      lhcond.dirty=1;
      refresh_display();
   }
   return TCL_OK;
}

int
set_ldcond(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char buff[256];
   int  upd=0;

   if (argc<2) {
      do_tcl("open_condsel ldcond link mv_ldcond");
   } else {
      stop_updates(ldcond.name);
      mv_get_cond(&ldcond, argc, argv);
      start_updates(ldcond.name);
      ldcond.dirty=1;
      refresh_display();
   }
   return TCL_OK;
}

int
set_nicond(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char buff[256];
   int  upd = 0;

   if (argc<2) {
      do_tcl("open_condsel nicond node mv_nicond");
   } else {
      if (nimode==NI_CUSTOM) {
         stop_updates(nicond.name);
         upd = 1;
      }

      mv_get_cond(&nicond, argc, argv);

      sprintf(buff, "set nimode %d", NI_CUSTOM);
      do_tcl(buff);

      if (upd)
         start_updates(nicond.name);

      nicond.dirty=1;
      /* refresh_display(); not needed for discover condition 
      BUT might need to start discovery XXX
       */
   }
   return TCL_OK;
}

int
set_ndcond(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char buff[256];
   int  upd = 0;

   if (argc<2) {
      do_tcl("open_condsel ndcond node mv_ndcond");
   } else {
      if (ndmode==ND_CUSTOM) {
         stop_updates(ndcond.name);
         upd = 1;
      }

      mv_get_cond(&ndcond, argc, argv);

      sprintf(buff, "set ndmode %d", ND_CUSTOM);
      do_tcl(buff);

      if (upd)
         start_updates(ndcond.name);

      ndcond.dirty=1;
      refresh_display();
   }
   return TCL_OK;
}

int
set_nsvar(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char   buff[256];

   if (argc<3) {
      do_tcl("open_dictvarlist nsvar node");
   } else {
      if (!strcmp(nsvar.name1, argv[1])
       && !strcmp(nsvar.name2, argv[2])) 
         return TCL_OK; /* no change */

      stop_updates(nsvar.name1);
      stop_updates(nsvar.name2);

      strcpy(nsvar.name1, argv[1]);
      strcpy(nsvar.name2, argv[2]);
      if (!strcmp(nsvar.name1, nsvar.name2))
         nsvar.name2[0]='\0';

      start_updates(nsvar.name1);
      start_updates(nsvar.name2);

      nsvar.dirty=1;
      refresh_display();
   }
   return TCL_OK;
}

int
set_lsvar(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char   buff[256];

   if (argc<3) {
      do_tcl("open_dictvarlist lsvar link");
   } else {
      stop_updates(lsvar.name1);
      stop_updates(lsvar.name2);

      strcpy(lsvar.name1, argv[1]);
      strcpy(lsvar.name2, argv[2]);
      if (!strcmp(lsvar.name1, lsvar.name2))
         lsvar.name2[0]='\0';

      start_updates(lsvar.name1);
      start_updates(lsvar.name2);

      lsvar.dirty=1;
      refresh_display();
   }
   return TCL_OK;
}

int
jumpto_cloud(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char   *str;
   int     i;

   str = (argc>1)? argv[1] : query("#: ");
   if ( *str == '\0' )
     return TCL_OK;

   i = atoi(str); /* i is cloudnode_tk index = cloud index */
   goto_cloud(i);
   return TCL_OK;
}

int
jumpto_node(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char   *str;
   int     i;

   str = (argc>1)? argv[1] : query("#: ");
   if ( *str == '\0' )
     return TCL_OK;

   i = atoi(str);
   if (is_ndeleted(i))
      add_msg("Node was deleted.");
   else {
      if (i!=curr_node)
         goto_node(i);
   }
   return TCL_OK;
}

/* Go to next highlighted node */
int
next_hl()
{
   int     i;

   for (i=curr_node+1; ; i++) {
      if (i>=num_nodes()) 
         i-=num_nodes();
      if (i==curr_node) {
         notify("Error", "There are no other highlighted nodes.");
         break;
      }
      if (nhighlight(i)) {
         goto_node(i);
         break;
      }
   }
   return TCL_OK;
}

int
jumpto_name(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char *str;
   int   i;
   char buff[255];

   str = (argc>1)? argv[1] : query("Find: ");
   if ( *str == '\0' )
     return TCL_OK;

   if ((i=find_node_by_name(str))>=0)
      goto_node(i);
   else {
      sprintf(buff, "'%s' not found.", str);
      add_msg(buff);
   }
   return TCL_OK;
}

/* Delete node */
int
do_delete_node(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char *str, buff[256];
   int n = (argc>1)? atoi(argv[1]) : curr_node;
   int next_node= -1;

   if (n==curr_node) {
      int i = n-1;
      if (i<0)
         i += num_nodes();

      /* Find another displayed node */
      while (i!=n && !is_ndisplayed(i)) {
         i--;
         if (i<0)
            i += num_nodes();
      }
      if (i!=n)
         next_node = i;

      /* Find any other node */
      if (next_node < 0) {
         i = n-1;
         if (i<0)
            i += num_nodes();
         while (i!=n) {
            if (!is_ndeleted(i)) {
               next_node = i;
               break;
            }
            i--;
            if (i<0)
               i += num_nodes();
         }
      }

      if (next_node < 0) {
         notify("Error", "Cannot delete the only node.");
         return TCL_OK;
      }
   }

   sprintf(buff, "Node deletion requested.  Really delete %s?",
    fullname(n));
   str = yesno(buff);
   if (!str || tolower(str[0])!='y') {
      add_msg("Deletion aborted.");
      return TCL_OK;
   }
   set_node_cloud(n, 0, NULL);
   delete_node(n);
   if (n==curr_node)
      curr_node = next_node;

   goto_node(curr_node);
   return TCL_OK;
}

int
do_spring(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   if (argc>1) {
      springembedder(0);
      refresh_high();
   } else {
      if (ndmode==ND_INCLOUD)
         springembedder(curr_cloud);
      else if (ndmode==ND_CUSTOM)
         springembedder(1);
      else
         springembedder(curr_cloud);
      ndcond.dirty=1; /* force nodes to be redrawn */
   }
   refresh_display();
   /* goto_node(curr_node); why do we need to do this? */
   center_node(curr_node);
   return TCL_OK;
}

int
do_center(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   int n = (argc>1)? atoi(argv[1]) : curr_node;
   
   center_node(n);
   return TCL_OK;
}

int
next_cloud()
{
   int cloud;
   extern int num_clouds;

   if (num_clouds<3) /* 0 is unused, and must have two clouds to goto next */
      return TCL_OK;

   cloud = curr_cloud;
   do {
      cloud++;
      if (cloud==num_clouds)
         cloud = 1;
      if (cloud == curr_cloud)
         return TCL_ERROR; 
   } while (!get_int("cloudsize", cloud));

   goto_cloud(cloud);
   return TCL_OK;
}

int
toggle_dir()
{
   iotype = 1-iotype;
   refresh_display();
   return TCL_OK;
}

/* Mark source 
int
mark_src()
{
   char buff[256];
   int  i;

   i = dijkstra_tree( curr_node );
   sprintf(buff, "set srcname %s", fullname(curr_node));
   text_cmd(buff);
   sprintf(buff, "set srcip %s", ipstr(curr_node,0));
   text_cmd(buff);
   refresh_display();
   sprintf(buff, "%d nodes reachable", i);
   add_msg(buff);
   return TCL_OK;
}
*/

void
mark_dest(n)
   int n;
{
	char buff[100];
	init_dest(n);

	Tcl_SetVar(tcl, "destname", fullname(n), TCL_GLOBAL_ONLY);
	Tcl_SetVar(tcl, "destip", ipstr(n,0), TCL_GLOBAL_ONLY);

	sprintf(buff, "%s marked as destination", fullname(n));
	add_msg(buff);
}

void
mark_root(n)
   int n;
{
	char buff[100];
	init_dest(n);

	Tcl_SetVar(tcl, "rootname", fullname(n), TCL_GLOBAL_ONLY);
	Tcl_SetVar(tcl, "rootip", ipstr(n,0), TCL_GLOBAL_ONLY);

	sprintf(buff, "%s marked as root", fullname(n));
	add_msg(buff);
}

/* Mark root */
int
do_mark_root(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   int             n;
   char           *str;
   struct hostent *hp;
   u_int32         addr;

   if (argc<2)
      n = curr_node;
   else if (isdigit(argv[1][0]))
      n = find_node_by_addr(inet_addr(argv[1]));
   else
      n = find_node_by_name(argv[1]);

   if (n<0) {
      str = yesno("Not found.  Add to database?");
      if (!str || tolower(str[0])=='n')
         n = curr_node;
      else {
         hp = gethostbyname(argv[1]);
         if (hp) {
            bcopy(hp->h_addr, (char *) &addr, hp->h_length);
            n = add_node(addr);
         } else
            notify("Error", "Unknown host");
      }
   }
   
   mark_root(n);
   return TCL_OK;
}

/* Mark dest */
int
do_mark_dest(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   int             n;
   char           *str;
   struct hostent *hp;
   u_int32         addr;

   if (argc<2)
      n = curr_node;
   else if (isdigit(argv[1][0]))
      n = find_node_by_addr(inet_addr(argv[1]));
   else
      n = find_node_by_name(argv[1]);

   if (n<0) {
      str = yesno("Not found.  Add to database?");
      if (!str || tolower(str[0])=='n')
         n = curr_node;
      else {
         hp = gethostbyname(argv[1]);
         if (hp) {
            bcopy(hp->h_addr, (char *) &addr, hp->h_length);
            n = add_node(addr);
         } else
            notify("Error", "Unknown host");
      }
   }
   
   mark_dest(n);
   refresh_display(); /* in case path is being highlighted */
   return TCL_OK;
}

/* Mark as a logical node */
int
mark_logic()
{
   node_t *node;

   node = mget_node(curr_node);
   node->status |= ST_MROUTER;
   node->timestamp = 0;
   goto_node(curr_node);
   return TCL_OK;
}

int
delete_routers()
{
   node_t *node;
   int i;

   for (i=0; i<num_nodes(); i++) {
      node = mget_node(i);
      if (!(node->status & ST_MROUTER))
         delete_node(i);
   }
   return TCL_OK;
}

int
do_query(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char buff[255], *str;
   int i;

   buff[0]=0;
   for (i=1; i<argc; i++) {
      strcat(buff, argv[i]);
      if (i+1<argc)
         strcat(buff, " ");
   }
   str = query(buff);
   Tcl_SetResult( tcl, str, TCL_VOLATILE );
   return TCL_OK;
}

int
tcl_cmd(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char *str;
   int i;

   str = query(":");
   i=Tcl_Eval(tcl, str);
   add_msg(tcl->result);
   return i;
}

int
mv_nslookup(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   int n;

   n = (argc>1)? atoi(argv[1]) : curr_node;

   push_queue(Q_NSLOOKUP, n);
   return TCL_OK;
}

int
mv_asname(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   int n;

   n = (argc>1)? atoi(argv[1]) : curr_node;

   push_queue(Q_ASNAME, n);
   return TCL_OK;
}

int
mv_autosys(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   int n;

   n = (argc>1)? atoi(argv[1]) : curr_node;

   push_queue(Q_AUTOSYS, n);
   return TCL_OK;
}

int
mv_snmpaware(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   int n;

   n = (argc>1)? atoi(argv[1]) : curr_node;

   push_queue(Q_SNMPAWARE, n);
   return TCL_OK;
}

int
do_update_node(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   int n;

   n = (argc>1)? atoi(argv[1]) : curr_node;

   /* Don't update if outside the area of interest */
   if (!ninfo(n))
      return 0;

   push_queue(Q_MRINFO, n);
   return TCL_OK;
}

int
force_update_node(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   int n;

   n = (argc>1)? atoi(argv[1]) : curr_node;

   update_node(n);
   return TCL_OK;
}

int
tcl_get_value(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   int recnum;

   if (argc<3) {
      notify("Error", "Wrong number of parameters to get_value");
      return TCL_OK;
   }

   recnum = atoi(argv[2]);

   Tcl_SetResult( tcl, get_string(argv[1], recnum), TCL_VOLATILE );
   return TCL_OK;
}

int
tcl_set_update_interval(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   int maxage;

   if (argc<3) {
      notify("Error", "Wrong number of parameters to set_update_interval");
      return TCL_OK;
   }

   maxage = atoi(argv[2]);

   set_update_interval(argv[1], maxage);
   return TCL_OK;
}

int
redraw()
{
   ndcond.dirty = ldcond.dirty = 1;
   refresh_display();
   return TCL_OK;
}

int
csetx(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char buff[256];
   int i;

   strcpy(buff, ".c xview");
   for (i=1; i<argc; i++) {
      strcat(buff, " ");
      strcat(buff, argv[i]);
   }
   Tcl_Eval(tcl, buff);
   Tcl_Eval(tcl,"reset_VW");
   return TCL_OK;
}

int
csety(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char buff[256];
   int i;

   strcpy(buff, ".c yview");
   for (i=1; i<argc; i++) {
      strcat(buff, " ");
      strcat(buff, argv[i]);
   }
   Tcl_Eval(tcl, buff);
   Tcl_Eval(tcl,"reset_VW");
   return TCL_OK;
}

int
NodeYellowCrit_text(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{

	tcl->result = cond_text(&nodeYellow);
	return TCL_OK;
}

int
NodeGreenCrit_text(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{

	tcl->result = cond_text(&nodeGreen);
	return TCL_OK;
}

int
NodeRedCrit_text(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{

	tcl->result = cond_text(&nodeRed);
	return TCL_OK;
}

int
LinkYellowCrit_text(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{

	tcl->result = cond_text(&linkYellow);
	return TCL_OK;
}

int
LinkGreenCrit_text(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{

	tcl->result = cond_text(&linkGreen);
	return TCL_OK;
}

int
LinkRedCrit_text(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{

	tcl->result = cond_text(&linkRed);
	return TCL_OK;
}
