proc estate {n} {
   global yscale Hyscale

   # Resize micro-topology display
   set h [expr 6 * $n]
   set oldh [.c cget -height]
   set yscale [expr $yscale * $h / $oldh]
   .c configure -height $h

   # Resize macro-topology display
   set hh [expr 600 - $h]
   set holdh [.highcanvas cget -height]
   set Hyscale [expr $Hyscale * $hh / $holdh]
   .highcanvas configure -height $hh

   .restate set $n
   mv_redraw
}

proc newCanvas {} {
   scrollbar .scrolly -command "mv_csety"
   scrollbar .scrollx -command "mv_csetx" -orient horizontal
   pack .scrolly -in .map.top.can -side right -fill y
   pack .scrollx -in .map.top.can -side bottom -fill x
   canvas .c \
             -width 400 -height 400 -relief sunken -bd 2\
             -xscrollincrement 20 -yscrollincrement 20 \
             -yscrollcommand ".scrolly set" -xscrollcommand ".scrollx set" 
#            -scrollincrement 20 
#   -scrollregion {0 0 1000 1000} 
   pack .c -in .map.top.can 
#   -fill x -expand true
#  focus .c
#  bind .c c { mv_center }
   .c bind txt <Any-Enter> { .c itemconfigure current -fill red }
   .c bind txt <Any-Leave> { .c itemconfigure current -fill black }

   .c bind txt <Button-2> {
      set origX %x
      set origY %y
      mkOutline [.c find withtag current] %x %y
      set prevX %x
      set prevY %y
   }
   .c bind node <B2-Motion> {
      .c move $outline [expr %x-$prevX] [expr %y-$prevY]
      set prevX %x
      set prevY %y
   }
   .c bind txt <ButtonRelease-2> {
      .c delete $outline
      moveNode [.c find withtag current] [expr %x-$origX] [expr %y-$origY]
   }

   .c bind txt <Button-1> {
      global nodeIdx
      mv_goto $nodeIdx([.c find withtag current])
   }

   .c bind tog <Any-Enter> { .c itemconfigure current -fill red }
   .c bind tog <Any-Leave> { .c itemconfigure current -fill black }

   .c bind tog <Button-2> {
      global togNode
      set origX %x
      set origY %y
      mkOutline $togNode([.c find withtag current]) %x %y
      set prevX %x
      set prevY %y
   }
   .c bind tog <ButtonRelease-2> { 
      global togNode
      .c delete $outline
      moveNode $togNode([.c find withtag current]) [expr %x-$origX] [expr %y-$origY]
   }
   .c bind tog <Button-1> {
      global nodeIdx togNode
      mv_goto $nodeIdx($togNode([.c find withtag current]))
   }

   .c bind box <Any-Enter> { 
      global rectNode
      .c itemconfigure $rectNode([.c find withtag current]) -fill red 
   }
   .c bind box <Any-Leave> { 
      global rectNode
      .c itemconfigure $rectNode([.c find withtag current]) -fill black 
   }
   .c bind box <Button-1> {
      global nodeIdx rectNode
      mv_goto $nodeIdx($rectNode([.c find withtag current]))
   }

   .c bind box <Button-2> {
      global rectNode
      set origX %x
      set origY %y
      mkOutline $rectNode([.c find withtag current]) %x %y
      set prevX %x
      set prevY %y
   }
   .c bind box <ButtonRelease-2> {
      global rectNode
      .c delete $outline
      moveNode $rectNode([.c find withtag current]) [expr %x-$origX] [expr %y-$origY]
   }

   bind .mbar P { mv_param }
   bind .mbar u { mv_force_update_node }
   bind .mbar r { mv_markroot }
   bind .mbar s { mv_marksrc }
   bind .mbar d { mv_markdest }
   bind .mbar ^ { mv_toSource }
   bind .mbar <KeyPress-greater> { mv_rotate -10 }
   bind .mbar <KeyPress-less> { mv_rotate 10 }
#   bind .c : { mv_tclcmd }
#   bind .c d { mv_dumpnode }
#   bind .c F { mv_fixlinks }
#   bind .c T { mv_test }
#   bind .c O { mv_dirtoggle }
#   bind .c <Control-f> { mv_flip }
#   bind .c a { mv_setage }
#   bind .c <KeyPress-numbersign> { mv_goto }
#   bind .c f { mv_exec finger @\[mv_fullname $curr_node\] }

#bind .c z { mv_showstats }
#bind .c q { mv_quit }
#bind .c S { mv_spread }
   newScaleWindow
}

proc dumpPostscript {w} {
   global str

   set bbox [$w bbox all]
   set x0 [lindex $bbox 0]
   set y0 [lindex $bbox 1]
   set x1 [lindex $bbox 2]
   set y1 [lindex $bbox 3]

   prompt .p {User Input} {Filename}
   $w postscript -colormode gray -file $str -x $x0 -y $y0 -height [expr $y1-$y0] -width [expr $x1-$x0]
}

proc mkNode {x y shape str n tog col} {
   global nodeX nodeY edgeFirst edgeSecond nodefont fontsize zoom
   global nodeRect rectNode nodeIdx nodeTog togNode nodeDot

   set m [expr $fontsize / 14.0]
   set w [expr $m*[string length $str]+$m/2]
   set lw 1
   if {$n < 0} {
      set lw 2
   }

   # Draw the node on the canvas
   if {$tog == "0"} {
      set rect [.c create $shape  [expr $x-$w/2] [expr $y-$m] [expr $x-$w/2+$w] [expr $y+$m] -fill $col -tags node -width $lw]
      set togText "-1"
   } else {
      if { $w < [expr $m*[string length $tog]+$m/2] } {
          set w [expr $m*[string length $tog]+$m/2]
      }
      set rect [.c create $shape  [expr $x-$w/2] [expr $y-$m] [expr $x-$w/2+$w] [expr $y+2.5*$m] -fill $col -tags node]
      set togText [.c create text $x [expr $y+1.5*$m] -text $tog -anchor center -font $nodefont]
   }
   set new [.c create text $x $y -text $str -anchor center -font $nodefont]

   # Set tags
   .c itemconfigure $new -tags "txt n$new node"
   .c itemconfigure $rect -tags "box n$new node"
   if { $togText != "-1" } {
      .c itemconfigure $togText -tags "tog n$new node"
   }

   set nodeIdx($new) $n
   set nodeX($new) $x
   set nodeY($new) $y
   set edgeFirst($new) {}
   set edgeSecond($new) {}
   set nodeRect($new) $rect
   set rectNode($rect) $new
   set nodeDot($new) [.s.c create rectangle [expr 5 + $x/10/$zoom] [expr 5 + $y/10/$zoom] [expr 5 + $x/10/$zoom] [expr 5 + $y/10/$zoom]]
   set nodeTog($new) $togText
   set togNode($togText) $new

   rebound
   return $new
}

proc mkOutline {n x y} {
   global nodeRect outline
   set l [.c coords $nodeRect($n)]
   set outline [.c create rectangle [lindex $l 0] [lindex $l 1] [lindex $l 2] [lindex $l 3]]
}

proc rmEdge {edge} {
   global edgeText edgeSrc edgeDest edgeFirst edgeSecond

   # remove it from the edgelists of the 2 nodes
   set src $edgeSrc($edge)
   if { $src >= 0 } {
      set i [lsearch -exact $edgeFirst($src) $edge]
      set edgeFirst($src)   [lreplace $edgeFirst($src) $i $i]
   }
   set dest $edgeDest($edge)
   if { $dest >= 0 } {
      set i [lsearch -exact $edgeSecond($dest) $edge]
      set edgeSecond($dest) [lreplace $edgeSecond($dest) $i $i]
   }

   .c delete $edgeText($edge)
   .c delete $edge
   unset edgeText($edge) edgeSrc($edge) edgeDest($edge)
}

proc rmNode {n} {
   global nodeX nodeY edgeFirst edgeSecond nodeRect nodeIdx togNode nodeTog edgeSrc edgeDest rectNode nodeDot
   .c delete n$n
   .s.c delete $nodeDot($n)
   foreach edge $edgeFirst($n) { 
       set edgeSrc($edge) -1
   }
   foreach edge $edgeSecond($n) { 
       set edgeDest($edge) -1
   }
   unset nodeX($n) nodeY($n) edgeFirst($n) edgeSecond($n) 
   if { $nodeTog($n) >= 0 } {
      unset togNode($nodeTog($n))
   }
   unset rectNode($nodeRect($n))
   unset nodeRect($n) nodeIdx($n) nodeTog($n) nodeDot($n)
   rebound
}

proc setNodeText {node str tog} {
   global nodeX nodeY nodeRect nodefont fontsize nodeTog togNode
   set m [expr $fontsize / 14.0]
   set w [expr $m*[string length $str]+$m/2]
   .c itemconfigure $node -text $str -font $nodefont

   if { ($tog == "0") } {
      .c coords $nodeRect($node) [expr $nodeX($node)-$w/2] [expr $nodeY($node)-$m] [expr $nodeX($node)-$w/2+$w] [expr $nodeY($node)+$m]
      if {($nodeTog($node) != "-1")} {
         .c delete $nodeTog($node)
         unset togNode($nodeTog($node))
         set nodeTog($node) "-1"
      }
   } else {
      if {($w < [expr $m*[string length $tog]+$m/2] )} {
          set w [expr $m*[string length $tog]+$m/2]
      }
      .c coords $nodeRect($node) [expr $nodeX($node)-$w/2] [expr $nodeY($node)-$m] [expr $nodeX($node)-$w/2+$w] [expr $nodeY($node)+2.5*$m]
      if {($nodeTog($node) == "-1")} {
         set nodeTog($node) [.c create text $nodeX($node) [expr $nodeY($node)+1.5*$m ] -text $tog -anchor center -font $nodefont -tags "tog n$node" ]
         set togNode($nodeTog($node)) $node
      } else {
         .c itemconfigure $nodeTog($node) -text $tog -font $nodefont
         .c coords $nodeTog($node) $nodeX($node) [expr $nodeY($node)+1.5*$m]
      }
   }
}

proc setNodeColor {node col} {
   global nodeRect
   .c itemconfigure $nodeRect($node) -fill $col
}

proc setNodeHl {node hl} {
   global nodeRect nodeIdx

   set cl 0
   if {$nodeIdx($node) < 0} {
      set cl 2
   }
   .c itemconfigure $nodeRect($node) \
     -width [expr $cl + 2*$hl+1] -outline [expr {($hl) ? "blue" : "black" }]
}

# Given 2 node ids, a color, and a width,
# make a new edge and return the object number
proc mkEdge { first second col wid str} {

   global nodeX nodeY edgeFirst edgeSecond edgeText nodefont edgeSrc edgeDest

   set edge [.c create line $nodeX($first) $nodeY($first) $nodeX($second) $nodeY($second) -fill $col -width $wid ]
   .c lower $edge
   lappend edgeFirst($first) $edge
   lappend edgeSecond($second) $edge

   set edgeText($edge) [.c create text 0 0 -text $str -anchor sw \
    -font $nodefont -fill purple]
   moveEdgeText $edge

   set edgeSrc($edge) $first
   set edgeDest($edge) $second

   return $edge
}

# Link the source end of an edge to a node
proc setLinkSrc { ln src } {
   global edgeFirst edgeSrc 
   set edgeSrc($ln) $src
   lappend edgeFirst($src) $ln
}

# Link the destination end of an edge to a node
proc setLinkDest { ln dest } {
   global edgeSecond edgeDest
   set edgeDest($ln) $dest
   lappend edgeSecond($dest) $ln
}

proc moveEdgeText { edge } {
   global edgeText

   set text $edgeText($edge)
   set xlow  [lindex [.c coords $edge] 0] 
   set xhigh [lindex [.c coords $edge] 2] 
   set ylow  [lindex [.c coords $edge] 1] 
   set yhigh [lindex [.c coords $edge] 3] 

   # Find a point 1/3 of the way between start and finish of the edge
   set xmid  [expr $xlow + ($xhigh-$xlow)/3]
   set ymid  [expr $ylow + ($yhigh-$ylow)/3]

   .c coords $text $xmid $ymid
}

# We only allow 1 string text here, but mview should allow that text
# to be "X/Y" where X and Y are values of 2 link variables
proc setEdgeText {edge str} {
   global edgeText nodefont
   .c itemconfigure $edgeText($edge) -text $str -font $nodefont
}

# Given a node id, and a delta X,Y
# Move the node by the specified amounts
proc moveNode {node xDist yDist} {
   global nodeX nodeY edgeFirst edgeSecond nodeDot zoom

   .c move n$node $xDist $yDist
   incr nodeX($node) $xDist
   incr nodeY($node) $yDist
   .s.c coords $nodeDot($node) [expr 5+$nodeX($node)/10/$zoom] [expr 5+$nodeY($node)/10/$zoom] [expr 5+$nodeX($node)/10/$zoom] [expr 5+$nodeY($node)/10/$zoom]

   foreach edge $edgeFirst($node) {
      .c coords $edge $nodeX($node) $nodeY($node) \
                      [lindex [.c coords $edge] 2] \
                      [lindex [.c coords $edge] 3]
      moveEdgeText $edge
   }

   foreach edge $edgeSecond($node) {
      .c coords $edge [lindex [.c coords $edge] 0] \
                      [lindex [.c coords $edge] 1] \
                      $nodeX($node) $nodeY($node)
      moveEdgeText $edge
   }
   rebound
}
proc moveNodeTo {node x y} {
   global nodeX nodeY
   moveNode $node [expr $x-$nodeX($node)] [expr $y-$nodeY($node)]
   rebound
}
proc rebound {} {
   global rescaleMode SR zoom

   if {$rescaleMode == "bbox"} {
      # ".c bbox" can only take a tag as an argument, not the "all"
      # keyword so we have to make an equivalent tag (everything)
      .c addtag everything all
      set bbox [.c bbox everything]
      if {$bbox == ""} { return } # in case there's no objects drawn
#puts "bbox=$bbox"
      .c configure -scrollregion $bbox
      .scrollx set [lindex [.c xview] 0] [lindex [.c xview] 1]
      .scrolly set [lindex [.c yview] 0] [lindex [.c xview] 1]
      .s.c coords $SR [expr 5 + [lindex $bbox 0]/10/$zoom] [expr 5 + [lindex $bbox 1]/10/$zoom] [expr 5 + [lindex $bbox 2]/10/$zoom] [expr 5 + [lindex $bbox 3]/10/$zoom]
#puts "SR: [.s.c coords $SR]"
      reset_VW
   }
}
proc rescale_to_all {} {
   global xscale yscale SR rescaleMode

   set rescaleMode all
   set xscale [expr [.c cget -width] / 1000.0]
   set yscale [expr [.c cget -height] / 1000.0]
   .s.c coords $SR 5 5 105 105
   mv_redraw
   reset_VW
}
proc rescale_to_bbox {} {
   global xscale yscale SR rescaleMode zoom

   set rescaleMode bbox
   set bbox [.c bbox everything]
   set x0 [lindex $bbox 0]
   set y0 [lindex $bbox 1]
   set x1 [lindex $bbox 2]
   set y1 [lindex $bbox 3]
   set xscale [expr [.c cget -width] / [expr $x1 - $x0]]
   set yscale [expr [.c cget -height] / [expr $y1 - $y0]]
   .s.c coords $SR [expr 5 + $x0 / 10/$zoom] [expr 5 + $y0 / 10/$zoom] [expr 5 + $x1 / 10/$zoom] [expr 5 + $y1 / 10/$zoom]
   mv_redraw
   reset_VW
}
proc newScaleWindow {} {
   global SR VW
   canvas .s.c -height 110 -width 110
   pack .s.c -in .s
   .s.c create rectangle 5 5 105 105
   set SR [.s.c create rectangle 5 5 105 105 -outline blue]
   set VW [.s.c create rectangle 5 5 105 105 -outline red]
}
proc reset_VW {} {
   global VW zoom
   set xv [.c xview]
   set yv [.c yview]
   set sr [.c cget -scrollregion]
   set xr [expr [lindex $sr 2]-[lindex $sr 0]]
   set yr [expr [lindex $sr 3]-[lindex $sr 1]]
   set x0 [expr [lindex $sr 0] + [lindex $xv 0] * $xr]
   set x1 [expr [lindex $sr 0] + [lindex $xv 1] * $xr]
   set y0 [expr [lindex $sr 1] + [lindex $yv 0] * $yr]
   set y1 [expr [lindex $sr 1] + [lindex $yv 1] * $yr]
#puts "VW: $x0 $y0 $x1 $y1"
   .s.c coords $VW  [expr 5 + $x0 / 10/$zoom] [expr 5 + $y0 / 10/$zoom] [expr 5 + $x1 / 10/$zoom] [expr 5 + $y1 / 10/$zoom]
}

proc centerNode {i} {
   global nodeX nodeY
 
   set bbox [.c bbox everything]
   set x0 [lindex $bbox 0]
   set y0 [lindex $bbox 1]
   set x1 [lindex $bbox 2]
   set y1 [lindex $bbox 3]
#   set xsrlist [.scrollx get]
#   set ysrlist [.scrolly get]
#   set xsrng [expr [lindex $xsrlist 1] - [lindex $xsrlist 0]]
#   set ysrng [expr [lindex $ysrlist 1] - [lindex $ysrlist 0]]
set xsrng [expr 1.0 * [.c cget -width] / ($x1-$x0)]
set ysrng [expr 1.0 * [.c cget -height] / ($y1-$y0)]
if {$xsrng > 1} { set xsrng 1 }
if {$ysrng > 1} { set ysrng 1 }
   set xfrac [expr 1.0 * ($nodeX($i)-$x0) / ($x1 - $x0)]
   set yfrac [expr 1.0 * ($nodeY($i)-$y0) / ($y1 - $y0)]
#puts "node ($nodeX($i),$nodeY($i))  -> ($xfrac,$yfrac)"
   set xsr0 [expr $xfrac - $xsrng/2]
   set ysr0 [expr $yfrac - $ysrng/2]
   set xsr1 [expr $xsr0 + $xsrng]
   set ysr1 [expr $ysr0 + $ysrng]
#puts "srng=$xsrng x $ysrng  sr=$xsr0..$xsr1 x $ysr0..$ysr1"
   if {$xsr0 < 0} { 
      set xsr0 0 
      set xsr1 $xsrng
   }
   if {$ysr0 < 0} { 
      set ysr0 0 
      set ysr1 $ysrng
   }
   if {$xsr1 > 1} {
      set xsr0 [expr 1-$xsrng]
      set xsr1 1
   }
   if {$ysr1 > 1} {
      set ysr0 [expr 1-$ysrng]
      set ysr1 1
   }
   .c xview moveto $xsr0
   .c yview moveto $ysr0
   .scrollx set $xsr0 $xsr1
   .scrolly set $ysr0 $ysr1
#puts "xsr0=$xsr0  ysr0=$ysr0"
   reset_VW
}
