/* $Id: cloud.c,v 1.11 1996/06/20 23:23:51 thalerd Exp $
 * -------------------------------------------------------------------------
 *  
 * [C] The Regents of the University of Michigan and Merit Network, Inc. 1996
 * All Rights Reserved 
 *   
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice and this permission notice appear in all 
 * copies of the software and derivative works or modified versions thereof, 
 * and that both the copyright notice and this permission and disclaimer notice 
 * appear in supporting documentation. 
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE REGENTS OF THE 
 * UNIVERSITY OF MICHIGAN AND MERIT NETWORK, INC. DO NOT WARRANT THAT THE 
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR 
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the 
 * University of Michigan and Merit Network, Inc. shall not be liable for any 
 * special, indirect, incidental or consequential damages with respect to any 
 * claim by Licensee or any third party arising from use of the software. 
 *
 * -------------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <tcl.h>
#include <tk.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include "mview.h"
#include "struct.h"
#include "dict.h"
#include "globals.h"
#include "high.h"
#include "display.h"
#include "database.h" /* for num_nodes */

char *
cloudvar()
{
   char *str;

   str = Tcl_GetVar(tcl, "cloudvar", TCL_GLOBAL_ONLY);
   if (str)
      return str;
   
   return "domain";
}

int num_clouds = 0;
int curr_cloud;
void
init_clouds()
{
   int i, c, n = num_nodes();

   /* Initialize num clouds */
   for (i=0; i<n; i++) {
      c = get_int("cloud", i);
      assert_cloudnode(c);
      if (c >= num_clouds)
         num_clouds = c+1;
   }

   /* Set curr_cloud to cloud of current node */
   curr_cloud = get_int("cloud", curr_node);
}

void
set_node_cloud(node_idx, cloud_idx, name)
   int node_idx;  /* node to put into cloud                  */
   int cloud_idx; /* cloud number to put node into           */
   char *name;    /* name to give cloud if currently unnamed */
{
   int old_cloud, clsize;
   char *str;

   /* Remove from previous cloud */
   old_cloud = get_int("cloud", node_idx);
   if (old_cloud) {
      clsize = get_int("cloudsize", old_cloud);
      set_int("cloudsize", old_cloud, clsize-1);
   }

   if (cloud_idx >= num_clouds) {
      num_clouds = cloud_idx + 1;
      cscond.dirty = 1;
      assert_cloudnode(cloud_idx);
   }

   /* Make sure cloud at least has a name */
   if (name) {
      str=get_string("cloudname", cloud_idx);
      if (!str || !str[0]) 
         set_string("cloudname", cloud_idx, name);
   }
   
   /* Add to new cloud */
   set_int("cloud", node_idx, cloud_idx);
   clsize = get_int("cloudsize", cloud_idx);
   set_int("cloudsize", cloud_idx, clsize+1);

   /* Assert new links if in cloud display mode */
   if (cloud_idx>0) {
      int ln, dn, dcloud, dir;
      node_t *node = mget_node(node_idx);
      
      for (dir=LIN; dir<=LOUT; dir++) {
         for (ln = node->last_link[dir]; ln>=0; ln = llink[ln].prev_link[dir]) {
            if (!(llink[ln].flags & (DVMRP_NF_DOWN|DVMRP_NF_DISABLED))) {
               dn = (dir==LIN)? get_log_link(ln)->src : get_log_link(ln)->dest;
               dcloud = get_int("cloud", dn);
               if (dcloud && dcloud != cloud_idx) 
                  assert_cloudlink(cloud_idx, dcloud, ln);
            }
         }
      }
   }

   /* Change displayed cloud if current node */
   if (ndmode==ND_INCLOUD && curr_node == node_idx)
      goto_node(curr_node);
}

void
assign_node_to_cloud(i)
   int i;
{
   char value[80];
   int c;

   /* Lookup node's value */
   strcpy(value, get_string(cloudvar(), i));
   
   /* Find corresponding cloud */
   for (c=1; c<num_clouds && strcasecmp(get_string("cloudname", c), value); c++);

   /* Assign node */
   set_node_cloud(i, c, value);
}

/* For all nodes whose variable has the given value, put them in a new
 * cloud.
 */
static void
assign_cloud_by_var(var, val)
   char *var;
   char *val;
{
   int n = num_nodes();
   int i;
   int cloud_idx;

   /* Find an unused cloud index */
   for (cloud_idx=1; cloud_idx<num_clouds && get_int("cloudsize", cloud_idx); cloud_idx++);

   for (i=0; i<n; i++) {
      if (!strcasecmp(get_string(var, i), val))
         set_node_cloud(i, cloud_idx, val);
   }
   set_string("cloudname", cloud_idx, val);
}

static void
assign_clouds_by_var(var)
   char *var;
{
   int i,c;
   int n = num_nodes();
   char value[1024];

   stop_updates(cloudvar());
   Tcl_SetVar(tcl, "cloudvar", var, 0);
   start_updates(cloudvar());

   /* Remove all nodes from clouds */
   for (i=0; i<n; i++) 
      set_node_cloud(i, 0, NULL);

   /* Clear all links between clouds */
   remove_cloudlinks();

   /* Remove all cloud names */
   for (i=0; i<num_clouds; i++)
      set_string("cloudname", i, "");

   num_clouds = 0;
   for (i=0; i<n; i++) {
      strcpy(value, get_string(var, i));
      for (c=1; c<num_clouds && strcasecmp(get_string("cloudname", c), value); c++);
      if (c>=num_clouds)
         assign_cloud_by_var(var, value);
   }

   sprintf(value, ".map.bot.hdr configure -text {High-level view of clouds defined by %s}", cloudvar());
   text_cmd(value);
}

/* TCL command: mv_setCloudVar variable */
int
cloud_var(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   if (argc<2)
      return TCL_ERROR;

   assign_clouds_by_var(argv[1]);
   refresh_display();
   return TCL_OK;
}

int
some_node_in_cloud(cloud)
   int cloud;
{
   int n = num_nodes();
   int i;

   for (i=0; i<n && get_int("cloud", i)!=cloud; i++);
   return i;
}
