/* $Id: database.c,v 1.33 1996/06/20 23:23:50 thalerd Exp $
 * -------------------------------------------------------------------------
 *  
 * [C] The Regents of the University of Michigan and Merit Network, Inc. 1996
 * All Rights Reserved 
 *   
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice and this permission notice appear in all 
 * copies of the software and derivative works or modified versions thereof, 
 * and that both the copyright notice and this permission and disclaimer notice 
 * appear in supporting documentation. 
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE REGENTS OF THE 
 * UNIVERSITY OF MICHIGAN AND MERIT NETWORK, INC. DO NOT WARRANT THAT THE 
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR 
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the 
 * University of Michigan and Merit Network, Inc. shall not be liable for any 
 * special, indirect, incidental or consequential damages with respect to any 
 * claim by Licensee or any third party arising from use of the software. 
 *
 * -------------------------------------------------------------------------
 */
/* Takes care of saving and loading data from files */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <errno.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include <tcl.h>
#include <tk.h>
#include <math.h>
#include <ctype.h>
#include <assert.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h> /* for X_OK */
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h> /* for htonl */
#endif
#include "mview.h"
#include "struct.h"
#include "database.h"
#include "node.h"
#include "dict.h"
#include "globals.h"
#include "maplib.h"
#include "queue.h" /* for push_queue */
#include "link.h"
#include "display.h" /* for set_node_display */

static int qdebug = 0;
static node_t dummy;
static log_link_t dumlink;

typedef struct {
   FILE  *fp;
   char   filename[30];
   size_t size;
} file_t;

static file_t file[]={ 
 { 0, "log_link", sizeof(log_link_t), },
 { 0, "phy_link", sizeof(phy_link_t), },
 { 0, "node",     sizeof(node_t), },
};
static node_t node[MAX_NODES];
static int    nodes, start=1;
static int log_links = -1;
       log_link_t llink[MAX_LINKS];
       FILE *log_fp=0;

static int phy_links = -1;
       phy_link_t plink[MAX_LINKS];

/* GET # LOG LINKS */
int
num_log_links() 
{
   assert(log_links>=0);
/*
   struct stat st;
   if (log_links<0) {
      if (stat(file[FLL].filename, &st)<0) {
         if (qdebug)
            printf("stat error: %s\n", file[FLL].filename);
         log_links = 0;
      } else
         log_links = st.st_size / sizeof(log_link_t);
   }
*/
   return log_links;
}

int
is_ndeleted(i)
   int i;
{
   return ((node[i].status & ST_DELETED)!=0);
}

static void
init_database()
{
   int i;
   struct stat st;
   char path[80];

   if (!start)
      return;

   if (access(datadir, X_OK)) {
      sprintf(path, "mkdir -p %s", datadir);
      printf("Creating %s...\n", datadir);
      system(path);
   }
   sprintf(path, "%s/%s", datadir, file[FN].filename);
   if (stat(path, &st)<0) {
      if (qdebug)
         printf("stat error: %s\n", file[FN].filename);
      nodes=0;
   } else
      nodes = st.st_size / sizeof(node_t);
/*printf("Found %d nodes\n", nodes);*/
   for (i=0; i<nodes; i++) 
      load_node(i, &(node[i]));
   if (nodes && !node[nodes-1].iface[0]) {
      while(nodes>0 && !node[nodes-1].iface[0])
         nodes--;
   }

   sprintf(path, "%s/%s", datadir, file[FLL].filename);
   if (stat(path, &st)<0) {
      if (qdebug)
         printf("stat error: %s\n", file[FLL].filename);
      log_links = 0;
   } else
      log_links = st.st_size / sizeof(log_link_t);
   for (i=0; i<log_links; i++)
      load_log_link(i, &(llink[i]));

   sprintf(path, "%s/%s", datadir, file[FPL].filename);
   if (stat(path, &st)<0) {
      if (qdebug)
         printf("stat error: %s\n", file[FPL].filename);
      phy_links = 0;
   } else
      phy_links = st.st_size / sizeof(phy_link_t);
   for (i=0; i<phy_links; i++)
      load_phy_link(i, &(plink[i]));

   start=0;

   /* Create dummy info for 0.0.0.0 */
   for (i=0; i<MAX_IFACE; i++)
      dummy.iface[i]=0;
   dummy.last_link[LOUT] = dummy.last_link[LIN] = -1;
   dummy.status = 0;
   dummy.timestamp = 0;

   /* Create dummy link info just in case */
   dumlink.src = dumlink.dest = -1;
   dumlink.flags = 0;

   log_fp = fopen("mview.log", "a");
   if (!log_fp) 
      log_fp= stdout;
}

/* GET # NODES (phy+log) */
int
num_nodes() 
{ 
   if (start)
      init_database();

   assert(nodes>=0);
   return nodes; 
}

int             /* RETURNS: node #, or -1 if not found */
find_node_by_name(str) /* ARGUMENTS:  */
   char *str;   /*    Hostname */
{
   int i;
   char *ptr;

   assert(nodes>=0);

   if (!str || !strlen(str))
      return -1;
   for (i=0; i<nodes; i++) {
      if (node[i].status & ST_DELETED)
         continue;
      ptr = get_string("fullname", i);
      if (ptr && !strncasecmp(str, ptr, strlen(str)))
         return i;
   }
   return -1;
}

int              /* RETURNS: node #, or -1 if not found */
find_node_by_addr(addr) /* ARGUMENTS:            */
   u_int32 addr;    /*    IP address to find */
{
   int v,j;

   if (!addr)
      return -1;

   /* Look for host */
   for (v=0; v<nodes; v++) {
      if (node[v].status & ST_DELETED) continue;
      for (j=0; j<MAX_IFACE; j++) {
         if (node[v].iface[j]==addr)
            return v;
      }
   }
   return -1;
}

/* Find index of address, adding a new node if not found */
int
assert_node_by_addr(addr)
   u_int32 addr;
{
   int i;

   /* Load in node information from disk */
   if (start) 
      init_database();

   if (!addr) 
      return -1;

   if ((i = find_node_by_addr(addr))>=0)
      return i;

   return add_node(addr);
}

int              /* RETURNS: Index of address in node, -1 if can't add */
iface_idx(n,src) /* ARGUMENTS:                */
   int    n;     /*    Node #                 */
   u_int32 src;   /*    IP address to find/add */
{
   int i;

   if (n<0)
      n = assert_node_by_addr(src);
   for (i=0; i<MAX_IFACE && node[n].iface[i]!=src; i++);
   if (i<MAX_IFACE)
      return i;
   for (i=0; i<MAX_IFACE && node[n].iface[i]; i++);
   if (i==MAX_IFACE) {
abort();
      return -1;
   }
   node[n].iface[i]=src;
   return i;
}

/* Given a physical link #, remove it from both the source and destination */
int            /* RETURNS: 1 on success, 0 on error */
rm_phy_link(ln)
   int ln;     /* Logical link number */
{
   int lsrc, ldest;

   /* Verify validity of argument */
   if (ln<0 || ln>=phy_links) 
      return 0;

   lsrc  = plink[ln].src;
   ldest = plink[ln].dest;

   /* Remove from source */
   rm_link_from_list(ln, lsrc, POUT);

   /* Remove from destination */
   rm_link_from_list(ln, ldest, PIN);

   /* Mark link as deleted */
   plink[ln].src  = -1;
   plink[ln].dest = -1;

   return 1;
}

/* Given a logical link #, remove it from both the source and destination */
int            /* RETURNS: 1 on success, 0 on error */
rm_log_link(ln)
   int ln;     /* Logical link number */
{
   int lsrc, ldest;

   /* Verify validity of argument */
   if (ln<0 || ln>=log_links) 
      return 0;

   lsrc  = llink[ln].src;
   ldest = llink[ln].dest;

   /* Remove from source */
   rm_link_from_list(ln, lsrc, LOUT);

   /* Remove from destination */
   rm_link_from_list(ln, ldest, LIN);

   /* Mark link as deleted */
   llink[ln].src  = -1;
   llink[ln].dest = -1;

   /* We now need to delete it from the display immediately, since
    * we might create another link with the same index right after
    * this and we want it to show up correctly.  -dgt 5/24/96
    */
   set_link_display(ln); /* delete the link */

   return 1;
}

/* XXX need code to go other direction */
int
add_phy_link(ps,pd,ls,ld)
   int ps;  /* physical source node # */
   int pd;  /* physical dest node # */
   int ls;  /* logical source node # */
   int ld;  /* logical dest node # */
{
   node_t *lsnode, *ldnode, *psnode, *pdnode;
   int pl, ll, i;
   int l; /* list index */

   assert(phy_links>=0);
   if (ps<0 || ps>nodes 
    || ls<0 || ls>nodes 
    || pd<0 || pd>nodes 
    || ld<0 || ld>nodes )
      return -1;

   psnode = mget_node(ps);
   pdnode = mget_node(pd);
   lsnode = mget_node(ls);
   ldnode = mget_node(ld);

   ll = find_log_link(ls,ld);

   /* Find deleted link, or add a new one */
   for (pl=0; pl<phy_links && plink[pl].src >=0; pl++) {
 
      /* Add log link to existing phy link */
      if (plink[pl].src==ps && plink[pl].dest==pd) {
         for (i=0; plink[pl].logic[i]>=0 && i<MAX_LOG_PER_PHY_LINK; i++);
         if (i==MAX_LOG_PER_PHY_LINK) 
            abort();
         plink[pl].logic[i] = ll;
         time(&(plink[pl].timestamp));
         break;
      }
   }

   /* New phy link */
   if (pl==phy_links || plink[pl].src<0) {
      plink[pl].src  = ps;
      plink[pl].dest = pd;
      plink[pl].logic[0] = ll;
      for (i=1; i<MAX_LOG_PER_PHY_LINK; i++)
         plink[pl].logic[i]= -1;
      time(&(plink[pl].timestamp));
      for (l=0; l<4; l++)
         llink[pl].prev_link[l] = -1;
   }

   if (pl==phy_links) phy_links++;

   add_link_to_list(pl, ps, POUT); /* Add plink to source pnode */
   add_link_to_list(pl, pd, PIN);  /* Add plink to dest pnode */
   
   fprintf(log_fp,"PLINK %d added between %d and %d for %d\n", pl, ps,pd,ll);
   return pl;
}

int
add_log_link(sn,src,dn,dst,fl)
   int    sn;   /* source node # (optional) or -1 */
   u_int32 src;  /* source IP address (required)   */
   int    dn;   /* dest node # (optional) or -1   */
   u_int32 dst;  /* dest IP address (required)     */
   int    fl;   /* flags                          */
{
   int ln, di, si, f, scloud, dcloud;
   node_t *dnode;
   int l; /* list index */

   assert(log_links>=0);

   if (sn<0)
      sn = assert_node_by_addr(src);

   f = find_node_by_addr(src);
   assert(sn==f || (f==-1 && !src && sn>=0));

   if (dn<0)
      dn = assert_node_by_addr(dst);
   if (dn>=0) {
      dnode =  mget_node(dn);
      if (!(dnode->status & ST_MROUTER)) {
         dnode->status |= ST_MROUTER;
         dnode->timestamp = 0;
         if (qdebug)
            printf("marked %d (%s) as an mrouter\n\r", dn, fullname(dn));
      }
   }

   /* Find deleted link, or add a new one */
   di = iface_idx(dn, dst);
   si = iface_idx(sn, src);
   assert(si>=0 && di>=0);
   for (ln=0; ln<log_links && llink[ln].src >=0; ln++) {
      if (llink[ln].src  == sn && llink[ln].sface == si
       && llink[ln].dest == dn && llink[ln].dface == di) {
fprintf(stderr, "DUP %d exists between #%d (%s) ", ln, 
    sn, inet_fmt(src));
fprintf(stderr, "and %d (%s)\n\r", 
    dn, inet_fmt(dst));
fprintf(log_fp,"DUP %d exists between #%d (%s) ", ln, 
    sn, inet_fmt(src));
fprintf(log_fp,"and %d (%s)\n", 
    dn, inet_fmt(dst));
         return ln;
      }
   }
   if (ln==log_links) log_links++;

   llink[ln].src       = sn;
   llink[ln].sface     = si;
   llink[ln].dest      = dn;
   llink[ln].dface     = di;
   llink[ln].flags     = fl;
   llink[ln].timestamp = 0;
   llink[ln].hops      = -1;
   for (l=0; l<4; l++)
      llink[ln].prev_link[l] = -1;

   if (sn>=0) {
      add_link_to_list(ln, sn, LOUT);
      node[sn].status |= fl; /* for DONTSAVE */
   }

   if (dn>=0) {
      add_link_to_list(ln, dn, LIN);
      node[dn].status |= fl; /* for DONTSAVE */
   }
   
if (qdebug) {
fprintf(stderr, "LOG %d added between #%d (%s) ", ln, 
   sn, inet_fmt(src));
fprintf(stderr, "and %d (%s)\n\r", 
    dn, inet_fmt(dst));
}

fprintf(log_fp,"LOG %d added between #%d (%s) ", ln, 
   sn, inet_fmt(src));
fprintf(log_fp,"and %d (%s)\n", 
    dn, inet_fmt(dst));

   set_string("srcip",  ln, src? inet_fmt(src) : "");
   set_string("destip", ln, dst? inet_fmt(dst) : "");

   /* Now assert high level link between clouds (if in different ones) */
   scloud = get_int("cloud", sn);
   dcloud = get_int("cloud", dn);
   if (scloud && dcloud && scloud!=dcloud)
      assert_cloudlink(scloud, dcloud, ln);

   return ln;
}

/* Return index of link (creating a new one if needed) */
int
assert_log_link(srcaddr, dstaddr, dstmask)
   u_int32 srcaddr; /* source IP address (always a host) */
   u_int32 dstaddr; /* dest IP address   (either a host or a subnet) */
   u_int32 dstmask; /* dest IP mask (0 for host) */
{
   if (!dstmask) { /* tunnel */
      int32 l;
      int sn;    /* src node # */
      int dn,    /* dest node # */
          sface, /* src iface idx */
          dface; /* dest iface idx */

      sn=assert_node_by_addr(srcaddr);
      sface = iface_idx(sn,srcaddr);

      dn=assert_node_by_addr(dstaddr);
      dface = iface_idx(dn,dstaddr);

      for (l = node[sn].last_link[LOUT]; l>=0; l = llink[l].prev_link[LOUT])
         if (llink[l].sface==sface && llink[l].dest==dn && llink[l].dface==dface)
            return l;
      return add_log_link(sn,srcaddr,dn,dstaddr,DVMRP_NF_TUNNEL);
   } else {       /* subnet */
      int subnet;

      subnet = assert_subnet(dstaddr);
      assert_subnet_mask(subnet, dstaddr, dstmask);

      if ((srcaddr & dstmask)==dstaddr)
         return assert_subnet_link(subnet, srcaddr);
      else {
         int sn, i;

         /* Special case, we have a link to a subnet, but the source end
          * isn't on the subnet.  Here, we must assume the source node
          * has an IP address which is on the subnet.  If not, there's
          * a problem.
          */
         sn=assert_node_by_addr(srcaddr);
         for (i=0; i<MAX_IFACE; i++)
            if (node[sn].iface[i] && ((node[sn].iface[i] & dstmask)==dstaddr))
               return assert_subnet_link(subnet, srcaddr);

printf("Couldn't create subnet link!!!\n");
exit(1);
         return -1;
      }
   }
}

/* Given a src/dest spec, check for an existing link there */
int                /* RETURNS: log link # */
find_log_link(sn,dn)  /* ARGUMENTS: */
   int sn;
   int dn;
{
   int32 l;

   for (l = node[sn].last_link[LOUT]; l>=0; l = llink[l].prev_link[LOUT]) {
      if (llink[l].dest==dn)
         return l;
   }

   return -1;
}

/* Find logical link given src & dest node #'s */
log_link_t *
find_ifc(sn,dn)
   int32 sn;
   int32 dn;
{
   int32 l = find_log_link(sn,dn);

   return (l>=0)? &(llink[l]) : 0;
}

node_t *
mget_node(idx)
   int32 idx;
{
   assert(nodes>0);
/*
   if (start) 
      init_database();
*/
   if (idx<0 || idx>nodes) {
      return &dummy;
   }
   return &(node[idx]);
}

log_link_t *
get_log_link(idx)
   int32 idx;
{
   assert(log_links>=0);

   if (idx<0 || idx>log_links)
      return &dumlink;
   return &(llink[idx]);
}

static int /* RETURNS: node index of combined node */
combine_nodes(n1,n2)
   int n1;
   int n2;
{
   int lo, hi, i, k;
   int ln;               /* temporary link index */
   log_link_t *ledge;
   phy_link_t *pedge;

   if (n1==n2) return n1;

   if (n1<n2) {
      lo=n1; hi=n2;
   } else {
      lo=n2; hi=n1;
   }
      
   /* Append hi's info into lo */
   if (fullname(hi)[0]!='?' && fullname(lo)[0]=='?') {
      set_string("fullname", lo, fullname(hi));
      set_string("hostname", lo, hostname(hi));
      set_string("domain", lo, org(hi));
      set_string("country", lo, country(hi));
   }

   /* Combine interfaces */
   for (i=0; i<MAX_IFACE; i++) {
      if (node[hi].iface[i])
         iface_idx(lo, node[hi].iface[i]);
   }

   /* Combine LOUT list */
   while ((ln = node[hi].last_link[LOUT])>=0) {
      ledge = &(llink[ln]);

      /* Modify link */
      k  = iface_idx(lo, node[hi].iface[ledge->sface]);
      ledge->sface = k;
      ledge->src   = lo;

      rm_link_from_list(ln, hi, LOUT);
      add_link_to_list(ln, lo, LOUT);
   }

   /* Combine LIN list */
   while ((ln = node[hi].last_link[LIN])>=0) {
      ledge = &(llink[ln]);

      /* Modify link */
      k  = iface_idx(lo, node[hi].iface[ledge->dface]);
      ledge->dface = k;
      ledge->dest  = lo;

      rm_link_from_list(ln, hi, LIN);
      add_link_to_list(ln, lo, LIN);
   }

   /* Combine POUT list */
   while ((ln = node[hi].last_link[POUT])>=0) {
      pedge = &(plink[ln]);

      /* Modify link */
      k  = iface_idx(lo, node[hi].iface[pedge->sface]);
      pedge->sface = k;
      pedge->src   = lo;

      rm_link_from_list(ln, hi, POUT);
      add_link_to_list(ln, lo, POUT);
   }

   /* Combine PIN list */
   while ((ln = node[hi].last_link[PIN])>=0) {
      pedge = &(plink[ln]);

      /* Modify link */
      k  = iface_idx(lo, node[hi].iface[pedge->dface]);
      pedge->sface  = k;
      pedge->dest   = lo;

      rm_link_from_list(ln, hi, PIN);
      add_link_to_list(ln, lo, PIN);
   }

   node[lo].status |= node[hi].status;

   if (node[lo].timestamp < node[hi].timestamp)
      node[lo].timestamp = node[hi].timestamp;

   combine_node_variables(lo,hi);

   /* Delete hi */
   delete_node(hi);
   return lo;
}

int                       /* RETURNS: new node index */
assert_node_addr(n, addr) /* ARGUMENTS: */
   int     n;             /*    old node index */
   u_int32 addr;          /*    new address */
{
   int dst = find_node_by_addr(addr);
   if (dst>=0 && n != dst)  /* Alternate IP address of existing node */
      return combine_nodes(n,dst);
   
   iface_idx(n,addr);
   return n;
}

int
add_basic_node(addr, mask)
   u_int32 addr;
   u_int32 mask;
{
   int i,n,w;
   node_t *t;
   char buff[256];

   for (n=0; n<nodes && !(node[n].status & ST_DELETED); n++);
   if (n==nodes) nodes++;

   t = &(node[n]);
   t->iface[0]=addr;
   t->iface[1]=mask;
   for (i=2; i<MAX_IFACE; i++)
      t->iface[i]=0;
   t->status = (mask)? ST_SUBNET : 0;
   t->timestamp= 0;

   /* Set display fields */
   w = xy_idx();
   t->x[0] = random()%1000;
   t->y[0] = random()%1000;
   t->x[1] = random()%1000;
   t->y[1] = random()%1000;
   t->x[2] = random()%1000;
   t->y[2] = random()%1000;
   
   for (i=0; i<4; i++)
      t->last_link[i] = -1;

   if (!mask)
      push_queue(Q_NSLOOKUP, n);
   else {
      char buff[256];
      int i;

      for (i=0; i<32 && !(mask & (1<<i)); i++);
      sprintf(buff, "%s/~%d", inet_fmt(addr), 32-i);
      set_string("fullname", n, buff);
      set_string("hostname", n, hostname(n));
      set_string("domain", n, org(n));
      set_string("country", n, country(n));
   }
   update_value("autosys", n);
   update_value("snmpaware", n);

   if (qdebug)
      fprintf(stderr,"%s %d added %s\n\r", (mask)?"SUBNET":"NODE", n, 
       inet_fmt(addr));
   if (mask)
      fprintf(log_fp,"SUBNET %d added %s\n", n, inet_fmts(addr, mask, buff));
   else
      fprintf(log_fp,"NODE %d added %s\n", n, inet_fmt(addr));

   if (!(n%100)) {
      save_everything();
   }

   /* Put it on the display if needed */
   set_node_display(n);

   return n;
}

int
add_node(addr)
   int32 addr;
{
   int n = add_basic_node(addr, 0);
   int s = find_subnet(addr);

   if (s>=0) 
      assert_subnet_link(s, addr);

   return n;
}

int
load_entry(fn, idx, buff) 
   int32  fn;
   int32  idx; /* starting from 0 */
   void *buff;
{
   int i;
   char path[80];

   sprintf(path, "%s/%s", datadir, file[fn].filename);
   if (!file[fn].fp) {
      file[fn].fp = fopen(path,"r+");
      /*printf("%d opened as r+\n", fn);*/
   }
   if (!file[fn].fp) {
      file[fn].fp = fopen(path,"w+");
      /*printf("%d opened as w+\n", fn);*/
   }
   if (!file[fn].fp)
      return 0;
   fflush(file[fn].fp);
   fseek(file[fn].fp, idx*file[fn].size, 0);
   i=fread(buff, file[fn].size, 1, file[fn].fp);
   fflush(file[fn].fp);
   return i;
}

static int
save_entry(fn, idx, buff)
   int32  fn;
   int32  idx;
   void *buff;
{
   int i;
   char path[80];

   errno=0;

/*
   if (!file[fn].fp) {
      file[fn].fp = fopen(file[fn].filename,"r+");
      if (file[fn].fp) 
         printf("[errno %d] %d opened as r+\n", errno, fn);
       else
         errno=0;
   }
*/
   if (!file[fn].fp) {
      sprintf(path, "%s/%s", datadir, file[fn].filename);
      file[fn].fp = fopen(path,"w");
      if (file[fn].fp)
         printf("[errno %d] %ld opened as w\n", errno, fn);
   }
   if (!file[fn].fp)
      return 0;
   if (errno) 
      abort();
   fflush(file[fn].fp);
   if (errno) 
      abort();
   fseek(file[fn].fp, idx*file[fn].size, 0);
   if (errno) 
      abort();
   while (ftell(file[fn].fp) < idx*file[fn].size) /* fill space if needed */
      fputc(0, file[fn].fp);
   if (errno) 
      abort();
if (qdebug)
   printf("file %ld record %ld at %lo ", fn, idx, ftell(file[fn].fp));
   i = fwrite(buff, file[fn].size, 1, file[fn].fp);
   if (errno) 
      abort();
   if (qdebug)
      printf("wrote %d bytes\n", i);
   fflush(file[fn].fp);
   if (errno) 
      abort();
   return i;
}

void
close_files()
{
   int fn;
   for (fn=0; fn<4; fn++) {
      if (file[fn].fp)
         fclose(file[fn].fp);
      file[fn].fp = 0;
   }
   /*printf("files closed\n");*/
}

static void
save_node(idx)
   int idx;
{
   if (!(node->status & ST_DONTSAVE))
      save_entry(FN, idx,&(node[idx]));
}

static void
save_log_link(idx)
   int32 idx;
{
   if (!(llink[idx].flags & ST_DONTSAVE))
      save_entry(FLL, idx,&(llink[idx]));
}

static void
save_phy_link(idx)
   int idx;
{
   save_entry(FPL, idx,&(plink[idx]));
}

static void
save_nodes()
{
   int i;
   char path[80];

   sprintf(path, "%s/%s", datadir, file[FN].filename);
   if (file[FN].fp) {
      fclose(file[FN].fp);
      file[FN].fp = 0;
      unlink(path);
   }
   file[FN].fp = fopen(path,"w");

   for (i=0; i<nodes; i++)
      save_node(i);
}

static void
save_log_links()
{
   int i;

   for (i=0; i<log_links; i++)
      save_log_link(i);
}

static void
save_phy_links()
{
   int i;

   for (i=0; i<phy_links; i++)
      save_phy_link(i);
}

int
save_everything()
{
   save_nodes();
   save_log_links();
   save_phy_links();

   /* Save user data dictionary */
   save_dd();
   save_data();

   return TCL_OK;
}

/*************************************************
 * Code to deal with SUBNETs                     *
 * Subnet nodes have iface[0] = subnet address   *
 *                   iface[1] = subnet mask      *
 *************************************************/

/*
 * Given an IP address, look for a subnet node which would contain
 * that IP address.  Return the node index, or -1 if not found.
 */
int               /* RETURNS: node #, or -1 if not found */
find_subnet(addr) /* ARGUMENTS:            */
   u_int32 addr;   /*    IP address to find subnet for */
{
   int v;

   if (!addr)
      return -1;

   for (v=0; v<nodes; v++) {
      if (!(node[v].status & ST_SUBNET)) continue;
      if (node[v].status & ST_DELETED) continue;
      if ((addr & node[v].iface[1]) == node[v].iface[0])
         return v;
   }
   return -1;
}

/* 
 * Create a subnet with no incoming/outgoing links
 */
static int
add_subnet(nodeaddr)
   u_int32 nodeaddr;
{
   u_int32 addr = nodeaddr;
   u_int32 mask = 0xffffffff;
   int n;

   /* Insure that subnet addr and node addr are different,
    * or that last byte is 0, in which case we assume we were given
    * a subnet address to begin with, rather than a host address.
    */
   while (addr==nodeaddr && (addr & 0xFF)) {
      mask  = htonl(ntohl(mask) << 1);
      addr &= mask;
   }

   n = add_basic_node(addr, mask);
   set_int("nodestate", n, 3); /* assume subnets are always up */
   return n;
}

/*
 * Given an IP address, look for a subnet node which would contain
 * that IP address.  Create one if not found.
 */
int
assert_subnet(addr)
   u_int32 addr;
{
   int n;
   
   n = find_subnet(addr);
   if (n<0)
      n = add_subnet(addr);
   return n;
}

/*
 * After a subnet address/mask has been updated, there may be some nodes
 * which should now be on the subnet.  This procedure does that, as well
 * as changes any previous tunnels with both endpoints on the subnet
 * into subnet adjacencies instead.
 *
 * This routine also detects subnets which are subsumed into others,
 * which can happen if an approx subnet isn't found by an SNMP handler
 * asserting the actual subnet mask (i.e. the duplicate subnet bug).
 * However, this can (should) be avoided by having handlers assert the
 * node's address as the subnet instead (since the host part will be
 * masked out anyway).
 *
 * This procedure now also needs to split subnets when needed, that is,
 * check for nodes with subnet links which are no longer within the mask,
 * and move them to a different subnet.  This happens when a mask is
 * lengthened. XXX
 */
static void
add_nodes_to_subnet(subnet)
   int subnet;
{
   u_int32 addr = node[subnet].iface[0],
          mask = node[subnet].iface[1];
   int src, dest, i, sl,  /* src-to-subnet link */
                  l, dl;  /* dest-to-subnet link */
   int ln;

   for (src=0; src<nodes; src++) {
      if ((src==subnet) || (node[src].status & ST_DELETED))
         continue;
      if (node[src].status & ST_SUBNET) {
         /* See if subnet should be inside this one */
         if ((node[src].iface[0] & mask)==addr) {
            /* Yes, it should, so just delete it.  All of its previous
             * connections will be recreated if necessary because duplicate
             * links might already exist(?)
             */
            delete_node(src);
         }
         continue;
      }

      /* See if node needs to be moved OFF subnet */
      ln = find_log_link(subnet, src);
      if (ln>=0) {
         if ((node[src].iface[llink[ln].dface] & mask) != addr) {
            printf("%d needs to be moved off %s\n", src, fullname(subnet));
            abort();
         }
      }

      /* See if node should be on subnet */
      for (i=0; i<MAX_IFACE; i++) {
         if ((node[src].iface[i] & mask)==addr) {
            sl = assert_subnet_link(subnet, node[src].iface[i]);

            /* Ok, this node is on the subnet, so look for extra tunnels */
            for (l=node[src].last_link[LOUT]; l>=0; l=llink[l].prev_link[LOUT]){
 
               /* See if the other end is a host on this subnet as well */
               dest = llink[l].dest;
               if (dest==subnet) continue;
               if ((node[dest].iface[llink[l].dface] & mask)!=addr) continue;
               dl = assert_subnet_link(subnet,node[dest].iface[llink[l].dface]);

               /*
                * Now we have 2 subnet links from src & dest to subnet, and
                * the original src->dest tunnel.  Need to combine info? XXX
                */
               rm_log_link(l); /* delete the tunnel */

            }
         }
      }
   }
}

void
assert_subnet_mask(subnet, addr, mask)
   int     subnet; /* node index of subnet */
   u_int32 addr;   /* actual subnet address */
   u_int32 mask;   /* actual subnet mask */
{
   char buff[256], buff2[256];

   inet_fmts(node[subnet].iface[0], node[subnet].iface[1], buff2);
   inet_fmts(addr, mask, buff);
/*
   if ((node[subnet].iface[0] & mask) != addr) {
      printf("ERROR: SUBNET %d MASK PROBLEMS!!!!!! %s -> %s\n",
       subnet, buff2, buff);
   }
*/

   /* Overwrite hostname to get rid of tilde */
   set_string("fullname", subnet, buff); 
   set_string("hostname", subnet, hostname(subnet));
     
   if (node[subnet].iface[1] != mask) {
      node[subnet].iface[0] = addr;
      node[subnet].iface[1] = mask;
      if (qdebug)
         fprintf(stderr,"SUBNET %d corrected to %s\n\r", subnet, buff);
      fprintf(log_fp,"SUBNET %d corrected to %s\n\r", subnet, buff);
      add_nodes_to_subnet(subnet);
   }
}

/*
 * Add incoming/outgoing links for an address, and
 * update the mask if needed.
 */
void
add_iface_to_subnet(subnet, addr)
   int     subnet;
   u_int32 addr;
{
   int changed=0;
   int prevaddr;

   if (!addr)
      return;

   /* Update the mask if necessary */
   prevaddr = node[subnet].iface[0];
   while (((addr & node[subnet].iface[1]) != node[subnet].iface[0])
    || (node[subnet].iface[0]==addr)) {
      node[subnet].iface[1]  = htonl(ntohl(node[subnet].iface[1]) << 1);
      node[subnet].iface[0] &= node[subnet].iface[1];
      changed=1;
   }

   if (changed) {
      char buff[256];

      /* Update the name */
      set_string("fullname", subnet, 
       approx_inet_fmts(node[subnet].iface[0], node[subnet].iface[1], buff));
      set_string("hostname", subnet, hostname(subnet));

#if 0
      /* Update all the source interfaces */
      for (l=node[subnet].last_link[LOUT]; l>=0; l=llink[l].prev_link[LOUT])
         if (llink[l].sface==prevaddr)
            llink[l].sface = node[subnet].iface[0];

      /* Update all the destination interfaces */
      for (l=node[subnet].last_link[LIN]; l>=0; l=llink[l].prev_link[LIN])
         if (llink[l].dface==prevaddr)
            llink[l].dface = node[subnet].iface[0];
#endif

      if (qdebug)
         fprintf(stderr,"SUBNET %d changed to %s\n\r", subnet, 
          inet_fmts(node[subnet].iface[0], node[subnet].iface[1], buff));
      fprintf(log_fp,"SUBNET %d changed to %s\n\r", subnet, 
          inet_fmts(node[subnet].iface[0], node[subnet].iface[1], buff));
      add_nodes_to_subnet(subnet);
   }
}

int
assert_subnet_link(subnet, addr)
   int     subnet;  /* node index of subnet */
   u_int32 addr;    /* IP address of host   */
{
   int n;
   int slink = -1, dlink = -1;
   char *str;
   int32 l;
   int sface; /* src iface idx */

   /* Find/create the node for addr */
   n = assert_node_by_addr(addr);

   /* Find/create the node->subnet links */
   sface = iface_idx(n,addr);
   for (l=node[n].last_link[LOUT]; l>=0; l=llink[l].prev_link[LOUT]) {
      if (llink[l].dest==subnet) {
         slink = l;
         break;
      }
   }
   if (slink<0) {
      slink = add_log_link(n,addr,subnet, 0, 0);
      set_int("linkstate", slink, 3); /* assume subnet links are always up */
   }

   /* Find/create the subnet->node reverse link */
   for (l=node[n].last_link[LIN]; l>=0; l=llink[l].prev_link[LIN]) {
      if (llink[l].src==subnet) {
         dlink = l;
         break;
      }
   }
   if (dlink<0) {
      dlink = add_log_link(subnet, 0, n, addr, 0);
      set_int("linkstate", slink, 3); /* assume subnet links are always up */
   }

   /* Put the subnet into the node's domain */
   str = org(n);
   if (str && !isdigit(str[0])) {
      set_string("domain", subnet, str);
      set_string("country", subnet, country(n));
   }

   return slink;
}
