/* $Id: dict.c,v 1.28 1996/06/20 23:23:48 thalerd Exp $
 * -------------------------------------------------------------------------
 *  
 * [C] The Regents of the University of Michigan and Merit Network, Inc. 1996
 * All Rights Reserved 
 *   
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice and this permission notice appear in all 
 * copies of the software and derivative works or modified versions thereof, 
 * and that both the copyright notice and this permission and disclaimer notice 
 * appear in supporting documentation. 
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE REGENTS OF THE 
 * UNIVERSITY OF MICHIGAN AND MERIT NETWORK, INC. DO NOT WARRANT THAT THE 
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR 
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the 
 * University of Michigan and Merit Network, Inc. shall not be liable for any 
 * special, indirect, incidental or consequential damages with respect to any 
 * claim by Licensee or any third party arising from use of the software. 
 *
 * -------------------------------------------------------------------------
 */
/* Supports the user data dictionary */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h> /* for atoi */
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_BSTRING_H
#include <bstring.h>
#endif
#include <math.h>
#include <tcl.h>
#include <tk.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h> /* for X_OK */
#endif
#include "mview.h"
#include "struct.h"
#include "xalloc.h"
#include "globals.h"
#include "display.h" /* for is_ndisplayed */
#include "cloud.h" /* for init_clouds */
#include "node.h" /* for fullname */
#include "database.h" /* for num_log_links */
#include "queue.h"    /* for pop_queue */

static int autoupd=0;

void timerProc     PROTO((ClientData data));
void add_timers    PROTO((int linked, int recnum));
void remove_timers PROTO((int linked, int recnum));

extern char *cloudvar();

static void init_hash();

/*****************************************************************************/
/* First we have code for saving and loading the data dictionary             */
/*****************************************************************************/

static int dd_init = 0;

typedef struct dict_tmp {
   char *name;  /* variable name */
   int   type; 
#define NUM_TYPES 4
#define VT_STRING 0
#define VT_INT    1
#define VT_FLOAT  2
#define VT_UINT   3
   int   len;   /* sizeof variable, in bytes */
   int   arrsize; /* max size of array */
   int   numelts; /* actual size of array */
   int   maxage;  /* max age in seconds */
   char *updcmd;  /* command to update */
   char *depends; /* variable this depends on */
   int   flags;
#define VF_ACTIVE     0x0001 /* always set */
#define VF_PERSISTENT 0x0002
#define VF_DIRTY      0x0004 /* set (in memory) if changed and not displayed */
   int   updating;
   int   linked;
#define VL_NOTHING  0
#define VL_NODE     1  /* MBone node */
#define VL_LINK     2  /* MBone adjacency */
#define VL_CLOUD    5  /* Cloud */
#define VL_CLOUDLINK 6  /* Cloud adjacency */
   char *data;
   struct dict_tmp *next;
} dict_t;

static dict_t *first_dict=NULL;
static char *dd_typestr[]={ "string", "int", "float", "uint", 0 };

static int
dd_typeint(str)
   char *str;
{
   int i;

   for (i=0; dd_typestr[i]; i++) {
      if (!strcmp(dd_typestr[i], str))
         return i;
   }
   return -1;
}

static char *dd_linkstr[]={ "nothing", "node", "link", "obsolete", "obsolete", 
"cloud", "cloudlink", 0 };

static int
dd_linkint(str)
   char *str;
{
   int i;

   for (i=0; dd_linkstr[i]; i++) {
      if (!strcmp(dd_linkstr[i], str))
         return i;
   }
   return -1;
}

typedef struct {
   int   bitmask;
   char  letter;
} dd_flag_t;

static dd_flag_t dd_flag[]={
 { VF_ACTIVE,     'A', },
 { VF_PERSISTENT, 'P', },
 { VF_DIRTY,      'D', },
 { 0, 0 },
};

static int
dd_flagint(str)
   char *str;
{
   int flag=0;
   char *p;
   int   i;

   for (p=str; *p; p++) {
      for (i=0; dd_flag[i].letter; i++) {
         if (dd_flag[i].letter == *p) {
            flag |= dd_flag[i].bitmask;
            break;
         }
      }
   }
   return flag;
}

/* Convert flags from integer to string form */
static char *
dd_flagstr(fl)
   int fl;
{
   int i;
static char buff[256];
   char *p=buff;

   fl |= VF_ACTIVE;

   for (i=0; dd_flag[i].letter; i++) {
      if (fl & dd_flag[i].bitmask)
         *p++ = dd_flag[i].letter;
   }
   *p = '\0';

   return buff;
}

void init_dd();
static dict_t *get_dict PROTO((char *name));

int /* RETURNS: 1 on success, 0 on failure */
create_dd(name, type, len, arrsize, flags, linked)
   char *name;
   int   type;
   int   len;
   int   arrsize;
   int   flags;
   int   linked;
{
   dict_t *new;

   if (!dd_init)
      init_dd();

   if (type<0 || type>=NUM_TYPES)
      return 0;
   if (get_dict(name))
      return 0;

   /* Fill in a new structure */
   new = (dict_t *)xalloc(0,sizeof(dict_t));
   new->name    = xstrdup(name);
   new->type    = type;
   new->len     = len;
   new->arrsize = arrsize;
   new->numelts = 0;
   new->updcmd  = NULL;
   new->depends = NULL;
   new->maxage  = 0;
   new->flags   = flags;
   new->linked  = linked;
   new->data    = (char*)xalloc(0, arrsize * (len+4));

   /* Add to linked list */
   new->next  = first_dict;
   first_dict = new;

   return 1;
}

static void
remove_dict(curr, prev)
   dict_t *curr;
   dict_t *prev;
{
   if (!curr)
      return; /* not found */

   /* Remove from linked list */
   if (prev)
      prev->next = curr->next;
   else
      first_dict = curr->next;
   curr->next = NULL;

   /* Free up space */
   xfree_string(curr->name);
   xfree_string(curr->data);
   xfree_string(curr->updcmd);
   xfree_string(curr->depends);
   xfree_string(curr);
}

void
remove_dd(name)
   char *name;
{
   dict_t *curr, *prev = NULL;

   /* Find entry */
   for (curr=first_dict; curr && strcmp(curr->name, name); prev=curr,curr=curr->next);
   remove_dict(curr, prev);
}

/* Called when exiting to free up memory */
void
remove_all_dd()
{
   while (first_dict)
      remove_dict(first_dict, NULL);
}

int
load_dd()
{
   FILE *fp;
   char name[80], typestr[80], flagstr[80], linkstr[80];
   int len, arrsize;

   sprintf(name, "%s/dict", datadir);
   if ((fp=fopen(name, "r"))==NULL)
      return FALSE;
   while (fscanf(fp, "%s %s %d %d %s %s", name, typestr, &len, &arrsize, 
    flagstr, linkstr)==6)
      create_dd(name, dd_typeint(typestr), len, arrsize, dd_flagint(flagstr),
       dd_linkint(linkstr));
   fclose(fp);
   return TRUE;
}

int
save_dd()
{
   FILE *fp;
   dict_t *dd;
   char file[80];

   sprintf(file, "%s/dict", datadir);
   if ((fp=fopen(file, "w"))==NULL) 
      return FALSE;

   for (dd=first_dict; dd; dd=dd->next) {
      fprintf(fp, "%s %s %d %d %s %s\n", dd->name, dd_typestr[dd->type], 
       dd->len, dd->arrsize, dd_flagstr(dd->flags), dd_linkstr[dd->linked]);
   }
   fclose(fp);
   return TRUE;
}

/*****************************************************************************/
/* Next, we have code for getting and setting variable values                */
/*****************************************************************************/

/* Search for variable in data dict */
static dict_t * /* RETURNS: ptr to dict entry, or NULL if not found */
get_dict(name)
   char *name;
{
   dict_t *dd;

   if (!dd_init)
      init_dd();

   for (dd=first_dict; dd && strcmp(dd->name, name); dd=dd->next);
   return dd;
}

void
dump_dict()
{
   dict_t *dd;
   if (!dd_init)
      init_dd();
   for (dd=first_dict; dd; dd=dd->next) {
      printf("%s\n", dd->name);
   }
}

double       get_float PROTO((char *name, int   recnum));
int          get_int   PROTO((char *name, int   recnum));
unsigned int get_uint  PROTO((char *name, int   recnum));

char *  /* string value, minimally "", never NULL */
get_string(name, recnum)
   char *name;
   int   recnum;
{
static char buff[256];
   dict_t *dict;

   dict = get_dict(name);
   if (!dict) 
      return "";

   if (recnum<0 || recnum >= dict->numelts)
      return "";
      
   switch(dict->type) {
      case VT_STRING:
         return &dict->data[recnum * (dict->len+4)];

      case VT_INT:
         sprintf(buff, "%d", get_int(name, recnum));
         return buff;
 
      case VT_UINT:
         sprintf(buff, "%u", get_int(name, recnum));
         return buff;
 
      case VT_FLOAT:
         sprintf(buff, "%g", get_float(name, recnum));
         return buff;

   }
   return ""; /* should never happen */
}

double
get_float(name, recnum)
   char *name;
   int   recnum;
{
   dict_t *dict;
   char *ptr;
   double dbl;

   dict = get_dict(name);
   if (!dict || recnum<0 || recnum >= dict->numelts) 
      return 0.0;

   switch(dict->type) {
      case VT_STRING:
         return atof(get_string(name, recnum));

      case VT_UINT:
         return (double)get_uint(name, recnum);

      case VT_INT:
         return (double)get_int(name, recnum);
 
      case VT_FLOAT:
         ptr = dict->data + recnum * (dict->len + 4);
         switch(dict->len) {
         case 4: /* float */
            return (double)(*(float *)ptr);
 
         case 8: /* double */
            memcpy(&dbl, ptr, sizeof(double)); /* not nec. on 8 byte bound */
            return dbl;

         default:
            return 0.0; /* should never happen */
         }

   }
   return 0.0; /* should never happen */
}

#if 0
/* Commented out 1/31/96 because it was never called */
static void
set_timestamp(dict, recnum, tm)
   dict_t *dict;
   int     recnum;
   time_t  tm;
{
   char  *ptr;

   if (!dict || recnum<0)
      return;

   if (recnum >= dict->arrsize) {
      printf("Array overflow for %s, rec=%d max=%d\n", dict->name, 
       recnum, dict->arrsize);
      exit(1);
   }
   if (recnum >= dict->numelts)
      dict->numelts = recnum+1;

   ptr = dict->data + recnum * (dict->len + 4) + dict->len;
   memcpy(ptr, &tm, sizeof(tm));
}
#endif

static time_t
get_dict_timestamp(dict, recnum)
   dict_t *dict;
   int     recnum;
{
   char  *ptr;
   time_t tm;

   if (!dict || recnum<0 || recnum >= dict->numelts) 
      return 0;

   ptr = dict->data + recnum * (dict->len + 4) + dict->len;
   memcpy(&tm, ptr, sizeof(tm));
   return tm;
}  

/*
 * Get the timestamp associated with a value of an element of a
 * user variable array.
 */
time_t                      /* RETURNS: timestamp, or 0 if none */
get_timestamp(name, recnum) /* ARGUMENTS:      */
   char *name;              /*   Variable name */
   int   recnum;            /*   Record number */
{
   dict_t *dict;

   dict = get_dict(name);
   if (!dict)
      return 0;
   return get_dict_timestamp(dict, recnum);
}

int
get_numelts(name)
   char *name;
{
   dict_t *dict;

   dict = get_dict(name);
   if (!dict)
      return 0;
   return dict->numelts;
}

int
get_int(name, recnum)
   char *name;
   int   recnum;
{
   dict_t *dict;
   char *ptr;

   dict = get_dict(name);
   if (!dict || recnum<0 || recnum >= dict->numelts) 
      return 0;

   switch(dict->type) {
      case VT_STRING:
         return atof(get_string(name, recnum));

      case VT_FLOAT:
         return (int)floor(0.5 + get_float(name, recnum));
 
      case VT_UINT:
      case VT_INT:
         ptr = dict->data + recnum * (dict->len + 4);
         switch(dict->len) {
         case 1: /* char */
            return (int)(*(char *)ptr);
 
         case 2: /* short */
            return (int)(*(short *)ptr);

         case 4: /* int32 */
            return (int)(*(int32 *)ptr);

         default:
            return 0; /* should never happen */
         }

   }
   return 0; /* should never happen */
}

unsigned int
get_uint(name, recnum)
   char *name;
   int   recnum;
{
   return (unsigned int)get_int(name, recnum);
}

/*
 * Called from display routing, marks a variable as clean.  This
 * means that the display has been updated to reflect any changes.
 */
void 
clean(name)
   char *name;
{
   dict_t *dict;

   if (!name[0])
      return;
   dict = get_dict(name);
   if (!dict)
      return;
   dict->flags &= ~VF_DIRTY;
}

int
is_dirty(name)
   char *name;
{
   dict_t *dict;

   if (!name[0])
      return 0; /* empty string is never dirty */

   dict = get_dict(name);
   if (!dict)
      return 1; /* non variable is always dirty */
   return (dict->flags & VF_DIRTY);
}

void set_float PROTO(( char *name, int recnum, double val));
void set_int   PROTO(( char *name, int recnum, int val));
void set_uint  PROTO(( char *name, int recnum, unsigned int val));

void
set_string(name, recnum, str)
   char *name;
   int   recnum;
   char *str;
{
   dict_t *dict;
   char *ptr;
   time_t tm;

   dict = get_dict(name);
   if (!dict) {
      create_dd(name, VT_STRING, 80, 100, VF_ACTIVE, VL_NOTHING);
      dict = get_dict(name);
      if (!dict)
         return;
   }

   if (recnum >= dict->arrsize) {
      printf("Array overflow for %s, rec=%d max=%d\n", name, 
       recnum, dict->arrsize);
      exit(1);
   }
   if (recnum >= dict->numelts)
      dict->numelts = recnum+1;
   dict->flags |= VF_DIRTY;
   switch(dict->type) {
      case VT_STRING:
         ptr = dict->data + recnum * (dict->len + 4);
         if (strlen(str)+1 > dict->len) {
            printf("String truncated!\n");
            str[dict->len-1]='\0';
         }
         strcpy(ptr, str );
         ptr += dict->len;
         time(&tm);
         memcpy(ptr, &tm, sizeof(tm));
         break;

      case VT_UINT:
      case VT_INT:
         set_int(name, recnum, atoi(str));
         return;
 
      case VT_FLOAT:
         set_float(name, recnum, atof(str));
         return;

   }
 
   if (!strcmp(name, cloudvar()))
      assign_node_to_cloud(recnum);
}

void
set_float(name, recnum, val)
   char *name;
   int recnum;
   double val;
{
static char buff[256];
   dict_t *dict;
   char *ptr;
   time_t tm;

   dict = get_dict(name);
   if (!dict) {
      create_dd(name, VT_FLOAT, 8, 100, VF_ACTIVE, VL_NOTHING);
      dict = get_dict(name);
      if (!dict)
         return;
   }

   if (recnum >= dict->arrsize) {
      printf("Array overflow for %s, rec=%d max=%d\n", name, 
       recnum, dict->arrsize);
      exit(1);
   }
   if (recnum >= dict->numelts)
      dict->numelts = recnum+1;
   dict->flags |= VF_DIRTY;
   switch(dict->type) {
      case VT_STRING:
         sprintf(buff, "%f", val);
         set_string(name, recnum, buff);
         return;

      case VT_UINT:
      case VT_INT:
         set_int(name, recnum, (int)floor(0.5 + val));
         return;
 
      case VT_FLOAT:
         ptr = dict->data + recnum * (dict->len + 4);
         switch(dict->len) {
         case 4: /* float */
            *((float *)ptr) = (float)val;
            break;
 
         case 8: /* double */
            memcpy(ptr, &val, sizeof(double));
            break;
         }
         ptr += dict->len;
         time(&tm);
         memcpy(ptr, &tm, sizeof(tm));
   }

   if (!strcmp(name, cloudvar()))
      assign_node_to_cloud(recnum);
}

/* This function used when deleting a node/link whatever */
static void
unset_dict(dict, recnum)
   dict_t *dict;
   int recnum;
{
#if 0
   time_t  tm   = 0;
   char   *ptr;

   if (!dict) 
      return;

   /* Null out value */
   switch(dict->type) {
      case VT_STRING: set_string(dict->name, recnum, ""); break;
      case VT_INT   : set_int(   dict->name, recnum, 0); break;
      case VT_FLOAT : set_float( dict->name, recnum, 0.0); break;
   }
 
   /* Clear timestamp */
   ptr = dict->data + recnum * (dict->len + 4) + dict->len;
   memcpy(ptr, &tm, sizeof(tm));
#endif

   if (dict)
      bzero(dict->data + recnum*(dict->len+4), dict->len + 4);
}

/* Used to delete a node/link/whatever */
void
unset(linked, recnum)
   int linked; /* Type of variables */
   int recnum; /* Index of record in variable arrays */
{
   dict_t *dict;

   for (dict=first_dict; dict; dict=dict->next) {
      if (dict->linked==linked) {
         unset_dict(dict, recnum);
      }
   }
}

void
unset_all_values(name)
   char *name;
{
   dict_t *dict;

   dict = get_dict(name);
   if (!dict) return;

   bzero(dict->data, dict->arrsize * (dict->len+4));
}

void
set_uint(name, recnum, val)
   char *name;
   int recnum;
   unsigned int val;
{
   set_int(name, recnum, (int)val);
}

void
set_int(name, recnum, val)
   char *name;
   int recnum;
   int val;
{
static char buff[256];
   dict_t *dict;
   char *ptr;
   time_t tm;

   dict = get_dict(name);
   if (!dict) {
      create_dd(name, VT_INT, 4, 100, VF_ACTIVE, VL_NOTHING);
      dict = get_dict(name);
      if (!dict)
         return;
   }

   if (recnum >= dict->arrsize) {
      printf("Array overflow for %s, rec=%d max=%d\n", name, 
       recnum, dict->arrsize);
      exit(1);
   }
   if (recnum >= dict->numelts)
      dict->numelts = recnum+1;
   dict->flags |= VF_DIRTY;
   switch(dict->type) {
      case VT_STRING:
         sprintf(buff, "%d", val);
         set_string(name, recnum, buff);
         return;

      case VT_FLOAT:
         set_float(name, recnum, (double)val);
         return;
 
      case VT_UINT:
      case VT_INT:
         ptr = dict->data + recnum * (dict->len + 4);
         switch(dict->len) {
         case 1: /* char */
            *(char *)ptr = (char)val;
            break;
 
         case 2: /* short */
            *(short *)ptr = (short)val;
            break;

         case 4: /* int32 */
            *(int32 *)ptr = (int32)val;
            break;
         }
         ptr += dict->len;
         time(&tm);
         memcpy(ptr, &tm, sizeof(tm));
         break;
   }
   
   if (!strcmp(name, cloudvar()))
      assign_node_to_cloud(recnum);
}

static void add_dict_timers    PROTO((dict_t *dict));
static void remove_dict_timers PROTO((dict_t *dict));

void
set_update(name, maxage, updcmd, depends)
   char *name;
   int   maxage;
   char *updcmd;
   char *depends;
{
   dict_t *dict;

   dict = get_dict(name);
   if (!dict) {
      create_dd(name, VT_INT, 4, 100, VF_ACTIVE, VL_NOTHING);
      dict = get_dict(name);
      if (!dict)
         return;
   }

   /* Remove old timers if already running */
   if (autoupd && dict->maxage && dict->updcmd)
      remove_dict_timers(dict);

   dict->maxage = maxage;
   if (dict->updcmd)
      xfree_string(dict->updcmd);
   if (updcmd)
      dict->updcmd = xstrdup(updcmd);
   else
      dict->updcmd = NULL;
   if (dict->depends)
      xfree_string(dict->depends);
   if (depends)
      dict->depends = xstrdup(depends);
   else
      dict->depends = NULL;

   /* Set new timers if needed */
   if (autoupd && dict->maxage && dict->updcmd)
      add_dict_timers(dict);
}

void
set_update_interval(name, maxage)
   char *name;
   int   maxage;
{
   dict_t *dict;

   dict = get_dict(name);
   if (!dict || maxage < 0)
      return;

   /* Remove old timers if already running */
   if (autoupd && dict->maxage && dict->updcmd)
      remove_dict_timers(dict);

   dict->maxage = maxage;

   /* Set new timers if needed */
   if (autoupd && dict->maxage && dict->updcmd)
      add_dict_timers(dict);
}

/*****************************************************************************/
/* Now we have code for saving and loading the variable arrays themselves    */
/*****************************************************************************/

void
save_data()
{
   dict_t *dd;
   FILE *fp;
   char filename[256];

   for (dd=first_dict; dd; dd=dd->next) {
      if (!(dd->flags & VF_PERSISTENT)) continue;
      sprintf(filename, "%s/%s", datadir, dd->name);
      if ((fp=fopen(filename, "w"))!=NULL) {
         fwrite(dd->data, dd->len+4, dd->numelts, fp);
         fclose(fp);
      }
   }
}

void
init_data()
{
   dict_t *dd;
   FILE *fp;
   char filename[256];

   if (access(datadir, X_OK)) {
      sprintf(filename, "mkdir -p %s", datadir);
      printf("Creating %s...\n", datadir);
      system(filename);
   }
   for (dd=first_dict; dd; dd=dd->next) {
      if (dd->flags & VF_PERSISTENT) {
         sprintf(filename, "%s/%s", datadir, dd->name);
         if ((fp=fopen(filename, "r"))!=NULL) {
            dd->numelts = fread(dd->data, dd->len+4, dd->arrsize, fp);
            fclose(fp);
         }
      }
   }
}

void
init_dd()
{
   dd_init=1;
   load_dd();
   init_data();
   init_hash();
   init_clouds();
}

#ifdef STANDALONE
main()
{
   char buff[256], var[256], str[256];
   int rec, ival;
   double fval;

   for (;;) {
      printf("> ");
      if (!gets(buff)) break;
      if (sscanf(buff, "sget %s %d", var, &rec)==2) {
         printf("%s[%d]=%s\n", var, rec, get_string(var, rec));
      } else if (sscanf(buff, "iget %s %d", var, &rec)==2) {
         printf("%s[%d]=%d\n", var, rec, get_int(var, rec));
      } else if (sscanf(buff, "fget %s %d", var, &rec)==2) {
         printf("%s[%d]=%f\n", var, rec, get_float(var, rec));
      } else if (sscanf(buff, "sset %s %d %s", var, &rec, str)==3) {
         set_string(var, rec, str);
         printf("%s[%d]=%s\n", var, rec, get_string(var, rec));
      } else if (sscanf(buff, "iset %s %d %d", var, &rec, &ival)==3) {
         set_int(var, rec, ival);
         printf("%s[%d]=%d\n", var, rec, get_int(var, rec));
      } else if (sscanf(buff, "fset %s %d %f", var, &rec, &fval)==3) {
         set_float(var, rec, fval);
         printf("%s[%d]=%f\n", var, rec, get_float(var, rec));
      } else {
         printf("Invalid command: %s\n", buff);
      }
   }

   printf("Saving data...\n");
   save_dd();
   save_data();
}

void add_msg(str) {  printf("%s\n", str); }
#endif

int
do_varlist(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   dict_t *dd;
   int linked= -1;

   if (argc>1) {
      linked = dd_linkint(argv[1]);
   }

   for (dd=first_dict; dd; dd=dd->next) {
      if (linked<0 || dd->linked == linked)
         Tcl_AppendElement(tcl, dd->name);
   }
   return TCL_OK;
}

/******************************************************************************
 * This section handles timers...
 *
 * Data structures:
 *    There are separate hash tables for node timers and link timers
 *    Each hash table uses the record number as the key, and the value
 *      is a pointer to the head of a doubly-linked list of timers
 *      associated with that particular node/link
 ******************************************************************************/

typedef struct timer_tmp {
#if (TK_MINOR_VERSION == 0)
   Tk_TimerToken token;
#else
   Tcl_TimerToken token;
#endif
   dict_t       *dict;
   int           recnum;
   Tcl_HashEntry    *entry;
   struct timer_tmp *next;
   struct timer_tmp *prev;
} mv_timer_t;

static Tcl_HashTable nodeHash, linkHash;

static void
init_hash()
{
   Tcl_InitHashTable(&nodeHash, TCL_ONE_WORD_KEYS);
   Tcl_InitHashTable(&linkHash, TCL_ONE_WORD_KEYS);
}

/*
 * Given a linked type (i.e. node or link), get a pointer to the associated
 * hash table
 */
static Tcl_HashTable *
hashtable(linked)
   int linked;
{
   switch(linked) {
   case VL_NODE: return &nodeHash;
   case VL_LINK: return &linkHash;
   }
   return NULL;
}

/*
 * Add a timer to the hash table/linked lists structures 
 */
static mv_timer_t *
add_timer(dict, recnum)
   dict_t *dict;
   int     recnum;
{
   mv_timer_t *timer, *timerlist;
   Tcl_HashTable *table;
   Tcl_HashEntry *entry;
   int new;

   if (!autoupd)
      return NULL;

   /* Create timer structure */
   timer = (mv_timer_t *)xalloc(0, sizeof(mv_timer_t));
   timer->dict   = dict;
   timer->recnum = recnum;
   timer->token  = 0;
  
   /* Add to lookup table (unless already there) */
   table = hashtable(dict->linked);
   if (!table) 
      return NULL;
   entry = Tcl_FindHashEntry(table, (char *)recnum);
   if (entry)
      new=0;
   else
      entry = Tcl_CreateHashEntry(table, (char *)recnum, &new); 
   timer->entry = entry; /* save ptr back to the hash entry for faster lookup */
   
   /* Add to the head of the linked list */
   timer->next = NULL;
   if (!new) {
      timerlist = (mv_timer_t*) Tcl_GetHashValue(entry);
      if (timerlist) {
         timer->next = timerlist;
         timerlist->prev = timer;
      }
   }
   timer->prev = NULL;
   Tcl_SetHashValue(entry, timer);

   return timer;
}

/* Remove a timer from the linked lists structures 
 * but DON'T remove an empty hash table entry.  This is because
 * it is immediately reset when this is called from timerProc(),
 * and remove_timers() [which deletes the whole list anyway]
 * can delete the entry itself easily enough.
 *
 * Also, don't free up the space, since timerProc needs the
 * structure after this to see if it needs to free the hash table entry.
 */
static void
remove_timer(timer)
   mv_timer_t *timer;
{
   if (!timer || !timer->entry)
      abort();

   /* Remove from linked list */
   if (timer->prev)
      timer->prev->next = timer->next;
   else
      Tcl_SetHashValue(timer->entry, timer->next);
   if (timer->next)
      timer->next->prev = timer->prev;

   timer->next = NULL;
   timer->prev = NULL;
}

static void
schedule_update(dict, recnum, lasttm)
   dict_t *dict;
   int     recnum;
   time_t  lasttm;  /*  Timestamp of last update command issue */
{
   time_t curr_time, sched_time;
   int           ms;
   mv_timer_t      *timer;
#if (TK_MINOR_VERSION == 0)
   Tk_TimerToken token;
#else
   Tcl_TimerToken token;
#endif

   if (!autoupd || dict->maxage<1 || dict->updcmd==NULL 
    || !dict->updating) 
      return;

   time(&curr_time);
   sched_time = lasttm + dict->maxage;
   if (sched_time < curr_time+1)
       sched_time = curr_time+1;
   if (sched_time - curr_time < 24*DAY) { /* i.e. ms fits in int */
      ms = 1000 * (sched_time - curr_time);

      /* assemble data */
      timer = add_timer(dict, recnum);
      if (timer) {
#if (TK_MINOR_VERSION == 0)
         token = Tk_CreateTimerHandler(ms, timerProc, (ClientData)timer);
#else
         token = Tcl_CreateTimerHandler(ms, timerProc, (ClientData)timer);
#endif
         timer->token = token;
      }
   }
}

void
stop_updates(name)
   char *name;
{
   dict_t *dict;

   dict = get_dict(name);
   if (!dict || (!dict->updcmd && !dict->depends))
      return;

   if (!dict->updating)
      return;

   dict->updating--;

   if (dict->depends)
      stop_updates(dict->depends);

   if (!dict->updating && dict->maxage>0 && dict->updcmd)
      remove_dict_timers(dict);
}

/*
 * For variables which have the same updcmd as another one, 
 * we still mark it as updating, even though maxage < 1
 */
void
start_updates(name)
   char *name;
{
   dict_t *dict;

   dict = get_dict(name);
   if (!dict || (!dict->updcmd && !dict->depends))
      return;

   dict->updating++;

   if (dict->depends)
      start_updates(dict->depends);

   if (dict->updating==1 && dict->maxage>0 && dict->updcmd)
      add_dict_timers(dict);

}

void
update_value(name,recnum)
   char *name;
   int recnum;
{
   dict_t *dict;
   char buff[256];

   dict = get_dict(name);
   if (!dict || !dict->updcmd)
      return;

   sprintf(buff, "%d", recnum);
   Tcl_SetVar(tcl, "exp_index", buff, 0);
   do_tcl(dict->updcmd);
}

void
timerProc(data)
   ClientData data;
{
   dict_t *dict;
   int     recnum;
   char buff[255];
   time_t curr_time, sched_time;
   mv_timer_t *timer = (mv_timer_t*) data;

   dict   = timer->dict;
   recnum = timer->recnum;

   remove_timer(timer);

   if (dict->maxage<1 || dict->updcmd==NULL) {

      /* If linked list is NULL, delete the hash table entry */
      if (!Tcl_GetHashValue(timer->entry))
         Tcl_DeleteHashEntry(timer->entry);

      xfree_string(timer);
      return;
   }

   xfree_string(timer);

   time(&curr_time);
   sched_time = get_dict_timestamp(dict, recnum) + dict->maxage;
   if (sched_time <= curr_time) {
      sprintf(buff, "%d", recnum);
      Tcl_SetVar(tcl, "exp_index", buff, 0);
      do_tcl(dict->updcmd);
   }
   schedule_update(dict, recnum, curr_time);
}

/* Called when set_update is done (after a remove_timers) */
static void
add_dict_timers(dict)
   dict_t *dict;
{
   int linked; /* Type of variables */
   int recnum; /* Index of record in variable arrays */
   int numrecs;

   if (!autoupd)
      return;

   linked = dict->linked;
   
   if (!dd_init)
      init_dd();

   switch(linked) {
   case VL_NODE: numrecs = num_nodes();     break;
   case VL_LINK: numrecs = num_log_links(); break;
   default     : numrecs = dict->numelts;     break;
   }
   for (recnum=0; recnum<numrecs; recnum++) {
      if (linked==VL_NODE && !is_ndisplayed(recnum)) continue;
      if (linked==VL_LINK && !is_edisplayed(recnum)) continue;

      if (dict->linked==linked && dict->maxage>0 && dict->updcmd!=NULL
       && dict->updating) {
         schedule_update(dict, recnum, get_dict_timestamp(dict,recnum));
      }
   }
}

/* Called when node 'recnum' is added to the display with linked=VL_NODE */
/* Called when link 'recnum' is added to the display with linked=VL_LINK */
void
add_timers(linked, recnum)
   int linked; /* Type of variables */
   int recnum; /* Index of record in variable arrays */
{
   dict_t *dict;

   if (!autoupd)
      return;
   
   if (!dd_init)
      init_dd();

   for (dict=first_dict; dict; dict=dict->next) {
      if (dict->linked==linked && dict->maxage>0 && dict->updcmd!=NULL
       && dict->updating) {
         schedule_update(dict, recnum, get_dict_timestamp(dict,recnum));
      }
   }
}

/* Called when set_update is done and previously updating */
static void
remove_dict_timers(dict)
   dict_t *dict;
{
   int linked; /* Type of variables */
   int recnum; /* Index of record in variable arrays */
   mv_timer_t *timer, *timerlist;
   Tcl_HashTable *table;
   Tcl_HashEntry *entry;
   int numrecs;

   linked = dict->linked;
   if (!(table = hashtable(linked)))
      return;

   switch(linked) {
   case VL_NODE: numrecs = num_nodes();     break;
   case VL_LINK: numrecs = num_log_links(); break;
   default     : numrecs = dict->numelts;     break;
   }
   for (recnum=0; recnum<numrecs; recnum++) {
      if (linked==VL_NODE && !is_ndisplayed(recnum)) continue;
      if (linked==VL_LINK && !is_edisplayed(recnum)) continue;
      entry     = Tcl_FindHashEntry(table, (char *)recnum);
      if (!entry) continue;
      timerlist = (mv_timer_t*) Tcl_GetHashValue(entry); /* timerlist */

      for (timer = timerlist; timer; timer=timer->next) {
         if (timer->dict==dict) {
#if (TK_MINOR_VERSION == 0)
            Tk_DeleteTimerHandler(timer->token);
#else
            Tcl_DeleteTimerHandler(timer->token);
#endif
            remove_timer(timer);

            /* If linked list is NULL, delete the hash table entry */
            if (!Tcl_GetHashValue(entry))
               Tcl_DeleteHashEntry(entry);

            xfree_string(timer);
            break;
         }
      }
   }
}

/* Called when node 'recnum' is removed from the display with linked=VL_NODE */
/* Called when link 'recnum' is removed from the display with linked=VL_LINK */
void
remove_timers(linked, recnum)
   int linked; /* Type of variables */
   int recnum; /* Index of record in variable arrays */
{
   mv_timer_t *timer, *timerlist;
   Tcl_HashTable *table;
   Tcl_HashEntry *entry;

   /* For all timers which match linked,recnum */
   if (!(table = hashtable(linked)))
      return;
  
   entry     = Tcl_FindHashEntry(table, (char *)recnum);
   if (!entry) return;
   timerlist = (mv_timer_t*) Tcl_GetHashValue(entry); /* timerlist */
   
   /* Delete the whole timerlist */
   for (timer = timerlist; timer; timer=timer->next) {
#if (TK_MINOR_VERSION == 0)
      Tk_DeleteTimerHandler(timer->token);
#else
      Tcl_DeleteTimerHandler(timer->token);
#endif
      remove_timer(timer);
      xfree_string(timer);
   }
   Tcl_DeleteHashEntry(entry); /* the linked list should already be null */
}

/* Called when the "autoupd" tcl variable is changed */
char *
set_autoupd(clientData, interp, name1, name2, flags)
   ClientData  clientData;
   Tcl_Interp *interp;
   char       *name1;
   char       *name2;
   int         flags;
{
   int val = atoi(Tcl_GetVar2(interp,name1,name2,flags));
   int recnum;
   
   if (!autoupd && val) {         /* Turn ON timers  */

      autoupd = val;

      /* Node timers */
      for (recnum=0; recnum<num_nodes(); recnum++)
         if (is_ndisplayed(recnum))
            add_timers(VL_NODE, recnum);

      /* Link timers */
      for (recnum=0; recnum<num_log_links(); recnum++)
         if (is_edisplayed(recnum))
            add_timers(VL_LINK, recnum);
 
   } else if (autoupd && !val) {  /* Turn OFF timers */

      autoupd = val;

      /* Node timers */
      for (recnum=0; recnum<num_nodes(); recnum++)
         if (is_ndisplayed(recnum))
            remove_timers(VL_NODE, recnum);

      /* Link timers */
      for (recnum=0; recnum<num_log_links(); recnum++)
         if (is_edisplayed(recnum))
            remove_timers(VL_LINK, recnum);

      /*
       * Clear out mrinfo queue, but let autosys & snmpaware
       * queues continue for nodes already discovered
       */
      while (pop_queue(Q_MRINFO, 0) >= 0);
   }

   /* Turn on/off node coloring */
   nodeRed.dirty = nodeYellow.dirty = nodeGreen.dirty = 1;
   linkRed.dirty = linkYellow.dirty = linkGreen.dirty = 1;
   refresh_display();

   return NULL;
}

void
dump_timers(linked)
   int linked;
{
   char buff[256];
   Tcl_HashTable *table;
   int recnum; /* Index of record in variable arrays */
   mv_timer_t *timer, *timerlist;
   Tcl_HashEntry *entry;
   int numrecs;
   time_t curr_time, sched_time;

   if (!(table = hashtable(linked)))
      return;

   do_tcl("open_reportwin timers {Node             Period    Left Command}");

   switch(linked) {
   case VL_NODE: numrecs = num_nodes();     break;
   case VL_LINK: numrecs = num_log_links(); break;
   default     : numrecs = 0;               break;
   }

   time(&curr_time);
   for (recnum=0; recnum<numrecs; recnum++) {
      if (linked==VL_NODE && !is_ndisplayed(recnum)) continue;
      if (linked==VL_LINK && !is_edisplayed(recnum)) continue;
      entry     = Tcl_FindHashEntry(table, (char *)recnum);
      if (!entry) continue;
      timerlist = (mv_timer_t*) Tcl_GetHashValue(entry); /* timerlist */

      for (timer = timerlist; timer; timer=timer->next) {

         sched_time = get_dict_timestamp(timer->dict, recnum) 
          + timer->dict->maxage;
         sprintf(buff, ".timers.text insert end {%15s", fullname(recnum));
         sprintf(buff+40, " %7d %7d %s\n}", timer->dict->maxage, 
          (sched_time>curr_time)? sched_time-curr_time:0, timer->dict->updcmd);
         do_tcl(buff);
      }
   }
   do_tcl("close_reportwin timers");
}

/*
 * Here, two nodes are being combined, and we want to combine the node
 * variables somewhat intelligently (e.g. last modified wins).
 */
void
combine_node_variables(lo,hi)
   int lo;
   int hi;
{
   /* XXX */
}
