/* $Id: display.c,v 1.65 1996/07/02 22:42:14 thalerd Exp $
 * -------------------------------------------------------------------------
 *  
 * [C] The Regents of the University of Michigan and Merit Network, Inc. 1996
 * All Rights Reserved 
 *   
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice and this permission notice appear in all 
 * copies of the software and derivative works or modified versions thereof, 
 * and that both the copyright notice and this permission and disclaimer notice 
 * appear in supporting documentation. 
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE REGENTS OF THE 
 * UNIVERSITY OF MICHIGAN AND MERIT NETWORK, INC. DO NOT WARRANT THAT THE 
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR 
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the 
 * University of Michigan and Merit Network, Inc. shall not be liable for any 
 * special, indirect, incidental or consequential damages with respect to any 
 * claim by Licensee or any third party arising from use of the software. 
 *
 * -------------------------------------------------------------------------
 */
/* DISPLAY.C - takes care of displaying windows and such 
 * The display module should be called by other modules, and 
 * shouldn't set any database variables.  The database is responsible
 * for updating the display and not vice versa.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#undef SHOW_EXTRA_LINKS
#include <stdlib.h>
#include <sys/termios.h> /* for rows/cols stty struct */
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <netdb.h>
#include <tcl.h>
#include <tk.h>
#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h> /* for struct in_addr for arpa/inet.h */
#endif
#include <arpa/inet.h> /* for inet_addr */
#include "mview.h"
#include "struct.h"
#include "node.h"
#include "database.h"
#include "display.h"
#include "builtin.h"
#include "globals.h"
#include "lib.h"
#include "xalloc.h"
#include "handler.h"
#include "graph.h"
#include "dict.h"
#include "expr.h"
#include "high.h"
#include "dist_vect.h" /* for link_in_tree */
#include "maplib.h" /* for inet_fmt */
#include "driver.h" /* for text_cmd */
#include "link.h" /* for reverse */

#undef LOG_MESSAGES

/* Tk canvas coordinates can be scrolled, but not rotated or scaled.
 * Thus, we do scaling and rotation ourselves.  xscale and yscale
 * are the scaling constants for the two axes.
 */
/*
       double maxx, maxy, miny, minx, xinc, yinc;
*/
       double zoom=INITIAL_ZOOM;
       double xscale = 1.0;
       double yscale = 1.0;
       int ndmode=ND_INCLOUD;
       int nimode=NI_DISPLAYED; /* was NI_NONE */
       int maxage = 604800;
       int display_ok = 0;
       int iotype=LOUT;
       cond_t ndcond, nhcond, lhcond, ldcond, nicond;
       cond_t nodeRed, nodeYellow, nodeGreen;
       cond_t linkRed, linkYellow, linkGreen;
       var_t  nsvar, lsvar;
       int display_terminal_subnets = 1;

       Tk_Window tk_win[10];
#ifdef LOG_MESSAGES
       FILE *msgsfp=NULL;
#endif

void remove_cloud_display PROTO((int node_idx));

/*
 * In canvas widgets in Tk, each object depicted is assigned a unique
 * number.  The node_tk array contains the object id of the node text itself.
 * Likewise for links and the link_tk array
 *
 * In .mviewrc, there are parallel arrays indexed by the id of the node text:
 *    nodeRect    is the id of the rectangle around the text
 *    nodeX       is the X coord of the text
 *    nodeY       is the Y coord of the text
 *    edgeFirst   is the list of edge id's with this node as the start point
 *    edgeSecond  is the list of edge id's with this node as the end point
 * 
 * All this info is needed by the routines which move nodes around.
 */
static int node_tk[MAX_NODES];
static int cloud_tk[MAX_NODES];
       int cloud_ref[MAX_NODES];
       float cloud_x[MAX_NODES], cloud_y[MAX_NODES];
static int link_tk[MAX_LINKS];

/* Check which coordinate pair to use for display */
int
xy_idx() 
{
   if (ndmode == ND_INCLOUD) 
      return 2;
   return 3; /* temporary */
}

/* Execute and print Tcl command */
char *
do_tcl(str)
   char *str;
{
   if (Tcl_Eval(tcl, str) == TCL_ERROR) {
      if (tcl->result && strlen(tcl->result))
         printf("!%s!%s!\r\n", str, tcl->result);
   }
/* For now, don't print out commands which succeeded
   else
      printf("!%s!\r\n", str);
*/
   fflush(stdout);
   return tcl->result;
}

/* This function brings up a notification window with the given
 * title and text.  An "OK" button destroys the window.
 */
void
notify(title, msg)
   char *title;
   char *msg;
{
   char buff[1024];
   sprintf(buff, "mv_dialog .notify {%s} {%s} {} 0 OK", title, msg);
   text_cmd( buff );
}

/* Check if edge satisfies highlight criteria */
int                   /* RETURNS: 1 if true, 0 if false */
ehighlight(ln)      
   int ln;            /* Index of logical link */
{
   int dn;
   node_t *dnode;
   log_link_t *edge;

   edge = get_log_link(ln);
   if (!edge)
      return 0;

   return testcond(lhcond.name, ln, lhcond.op, &lhcond.value);

#if 0
   case LH_PROPOSED: /* Proposed */
      if (edge->flags & ST_DONTSAVE) return 1;
      break;

   case LH_SRCTREE: /* Source tree */
      return link_in_tree( edge->src, edge->dest );
      break;
   
   case LH_WEIRD: /* Up links to down nodes */
      if (!(edge->flags & (DVMRP_NF_DOWN|DVMRP_NF_DISABLED)))
         return 0;
      dn = (iotype==LOUT)? edge->dest : edge->src;
      dnode = mget_node(dn);
      if (!(dnode->status & ST_MROUTING))
         return 1;
      break;
  
   case LH_NONE:
   default: /* None */
      break;
   }
   return 0;
#endif
}

/* Check if edge satisfies highlight criteria */
int
nhighlight(idx)
   int idx;      /* Index of node */
{
   int ln; /* link index */
   node_t *node;
   time_t tm1;

   /* Don't highlight cloud nodes */
   if (ndmode==ND_INCLOUD && !ndisplay(idx))
      return 0;

   node = mget_node(idx);
   return testcond(nhcond.name, idx, nhcond.op, &nhcond.value);
}

/* Output text into arbitrary text window */
void
xtmp_display( wstr, str )
   char *wstr;
   char *str;
{
   char buff[1024], **line;
   int l, nl=1;

   line = explode(str, "\n");
   nl = (str[strlen(str)-1]=='\n');
   for (l=0; l<xsizeof(line); l++) {
      sprintf(buff, "\".%s.text\" insert end {%s%s}", wstr, line[l],
        ((l==xsizeof(line)-1 && !nl)?"":"\n"));
      text_cmd(buff);
   }
   xfree_array(line);
}

/* Output text into text window ".wtmp" */
void
wtmp_display( str )
   char *str;
{
   char buff[1024], **line;
   int l, nl;
   nl = (str[strlen(str)-1]=='\n');

   line = explode(str, "\n");
   for (l=0; l<xsizeof(line); l++) {
      sprintf(buff, ".wtmp.text insert end {%s%s}", line[l],
        ((l==xsizeof(line)-1 && !nl)?"":"\n"));
      text_cmd(buff);
   }
   xfree_array(line);
}

char *
query(pr)
char *pr;
{
static char str[255];

   sprintf(str, "prompt .p {User Input} {%s}", pr );
   text_cmd(str);
   return Tcl_GetVar(tcl, "str", 0);
}

char *
yesno(pr)
   char *pr;
{
static char str[255];

   sprintf(str, "yesno .p {User Verification} {%s}", pr );
   text_cmd(str);
   return Tcl_GetVar(tcl, "str", TCL_GLOBAL_ONLY );
}

/* Convert canvas to virtual coordinate */
static float /* Virtual X coordinate */
Xc2v(x)
   int x;    /* Canvas X coordinate */
{
#if 0
   int       w   = 500;
   Tk_Window win = Tk_NameToWindow(tcl, ".c", tk_main);

   if (win) w=Tk_Width(win);
   if (w<2) w=500;
   w -= 100;
   return (x-50.0)*xinc/w + minx;
#endif
   return x/xscale/zoom;
}

/* Convert canvas to virtual coordinate */
static float /* Virtual Y coordinate */
Yc2v(y)
   int y;    /* Canvas Y coordinate */
{
#if 0
   int h=300;
   Tk_Window win = Tk_NameToWindow(tcl, ".c", tk_main);
   if (win) h = Tk_Height(win);
   if (h<2) h=300;
   h -= 100;
   return = (y-50.0)*yinc/h + miny;
#endif
   return y/yscale/zoom;
}

/* Convert virtual to canvas coordinate */
static int   /* Canvas X coordinate */
Xv2c(x)
   float x;  /* Virtual X coordinate */
{
#if 0
   int       w   = 500;
   Tk_Window win = Tk_NameToWindow(tcl, ".c", tk_main);

   if (win) w=Tk_Width(win);
   if (w<2) w=500;
   w -= 100;
   return ((int)floor((w*(x-minx))/xinc +50.5));
#endif
   return (int)floor(x*xscale*zoom + 0.5);
}

/* Get canvas coordinate of a node */
static int       /* Canvas X coordinate */
Xn2c(node)
   node_t *node; /* Node */
{
   return Xv2c(node->x[xy_idx()]);
}

int
Xi2c(i)
   int i;
{
   if (ndmode==ND_INCLOUD) {
      int cloud = get_int("cloud", i);
      if (cloud_tk[cloud]>0 && node_tk[i]==cloud_tk[cloud])
         return Xv2c(cloud_x[cloud]);
   }
   return Xn2c(mget_node(i));
}

/* Convert virtual to canvas coordinate */
static int  /* Canvas Y coordinate */
Yv2c(y)
   float y; /* Virtual Y coordinate */
{
#if 0
   Tk_Window win;
   int h=300;
   win = Tk_NameToWindow(tcl, ".c", tk_main);
   if (win) h = Tk_ReqHeight(win);
   if (h<2) h=300;
   h -= 100;
   return ((int)floor((h*(y-miny))/yinc +50.5));
#endif
   return (int)floor(y*yscale*zoom + 0.5);
}

/* Get canvas coordinate of a node */
static int       /* Canvas X coordinate */
Yn2c(node)
   node_t *node; /* Node */
{
   return Yv2c(node->y[xy_idx()]);
}

int
Yi2c(i)
   int i;
{
   if (ndmode==ND_INCLOUD) {
      int cloud = get_int("cloud", i);
      if (cloud_tk[cloud]>0 && node_tk[i]==cloud_tk[cloud])
         return Yv2c(cloud_y[cloud]);
   }
   return Yn2c(mget_node(i));
}

#ifdef __STDC__
/* The GNU compiler doesn't like K&R notation for this function */
void
set_virtual_position(
   int n,     /* Index of node */
   float x,   /* Virtual X coordinate */
   float y    /* Virtual Y coordinate */
)
#else
void
set_virtual_position(n, x, y)
   int   n;   /* Index of node */
   float x;   /* Virtual X coordinate */
   float y;   /* Virtual Y coordinate */
#endif
{
   int cloud;

   if (ndmode==ND_INCLOUD && node_tk[n]>0 
    && node_tk[n]==cloud_tk[cloud=get_int("cloud",n)]) {
      cloud_x[cloud] = x;
      cloud_y[cloud] = y;
   } else {
      int idx = xy_idx();
      node_t *node = mget_node(n);

      node->x[idx] = x;
      node->y[idx] = y;
   }
}

extern void tcl_exec PROTO((const char str[]));

void
init_view()
{
extern const char tcl_display[];
   int i;

#if 0
   nsvar.name1[0]= nsvar.name1[0]='\0';
   lsvar.name2[0]= lsvar.name2[0]='\0';
#endif

#ifdef LOG_MESSAGES
   msgsfp = fopen("mview.msgs", "w");
#endif
   tcl_exec(tcl_display);

   Tk_MakeWindowExist(tk_main);
   display_ok = 1;

/* Now make high level window exist */
/* init_high(); */
/* refresh_high(); */

   for (i=0; i<MAX_NODES; i++) 
      cloud_tk[i] = cloud_ref[i] = 0;
}

int
exit_view()
{
   free_handlers();
   save_everything();
   close_files();
   remove_all_dd();
   xcheck();
   exit(0);
}

/* Convert status into a string */
char *
statstr(st)
   int st;
{
   static char str[256];

   str[0]='\0';
   if (st & ST_RESPONDING) strcat(str, "/ping");
   if (st & ST_MROUTER)    strcat(str, "/mrouted");
   if (st & ST_MROUTING)   strcat(str, "/up");
   return (*str=='/')? str+1 : str;
}

/* Convert edge flags into a string */
void
flagstr(buff,edge)
   char *buff;
   log_link_t *edge;
{
   if (edge->flags & DVMRP_NF_TUNNEL)
      strcat(buff, "/tunnel");
   if (edge->flags & DVMRP_NF_SRCRT)
      strcat(buff, "/srcrt");
   if (edge->flags & DVMRP_NF_QUERIER)
      strcat(buff, "/querier");
   if (edge->flags & DVMRP_NF_DISABLED)
      strcat(buff, "/disabled");
   if (edge->flags & DVMRP_NF_DOWN)
      strcat(buff, "/down");
   if (edge->flags & DVMRP_NF_PIM)
      strcat(buff, "/pim");
   if (edge->flags & DVMRP_NF_LEAF)
      strcat(buff, "/leaf");
}

void
add_msg(msg)
   char *msg;
{
   char buff[1024];
   sprintf(buff, ".r.msg configure -text {%s}", msg);
   Tcl_Eval(tcl, buff);
#ifdef LOG_MESSAGES
   if (msgsfp)
      fprintf(msgsfp, "%s\n", msg);
#endif
}

/*********************************************************/
/* FILL IN INFO IN "CURRENT NODE" SECTION OF THE DISPLAY */
/*********************************************************/
void
show_detail(v)
   int v;
{
   int         dn;
   int         iotmp;
   int         ln; /* link index */
   node_t     *node, 
              *dnode;
   log_link_t *edge;
   char        buff[255], cmd[255];
   time_t      now;

   time(&now);

   /* Clear old info from window */
   text_cmd( ".outlinks delete 0 end" );
   text_cmd( ".inlinks delete 0 end" );

   node = mget_node(v);

   sprintf(cmd,".r.d.label configure -text {%s [%s]}", 
    fullname(v), inet_fmt(node->iface[0]));
   text_cmd(cmd);

   sprintf(buff,"Node #%3d Version %s [%s]", v,
       get_string("version", v), statstr(node->status));
   sprintf(cmd,".nh.line2 configure -text {%s}", buff);
   text_cmd(cmd);

      sprintf(buff, "Seconds since last update:%7d (%s", now-node->timestamp,
       ctime(&(node->timestamp))+4 );
      buff[ strlen(buff)-1 ]='\0';
      strcat(buff, ")");
      sprintf(cmd,".nh.line3 configure -text \"%s\"", buff);
      text_cmd(cmd);

     for (iotmp = iotype & 0xFE; iotmp < (iotype & 0xFE)+2; iotmp++) {
      for (ln=node->last_link[iotmp]; ln>=0; ln=llink[ln].prev_link[iotmp]) {
         edge = get_log_link(ln);
         dn = (iotmp==LOUT)? edge->dest : edge->src;
         if (dn<0) 
            continue;
         dnode = mget_node(dn);
 
         buff[0]='\0';
         if (dn==v)
            sprintf(buff+strlen(buff), "SUBNET iface %s", ipstr(v, edge->sface));
         else
            strcat(buff, fullname(dn));

         sprintf(buff+strlen(buff)," [%d/%d", get_int("metric",ln), 
          get_int("threshold",ln));
         flagstr(buff,edge);
         strcat(buff, "]");

         if (ndmode==ND_INCLOUD) {
            int cl = get_int("cloud", dn);
            if (cl != curr_cloud)
               sprintf(buff+strlen(buff), " (%s=%s)", 
                cloudvar(), (cl)? get_string("cloudname", cl) : "?");
         }

         sprintf(cmd,".%slinks insert end {%s}", (iotmp & 1)?"out":"in",buff);
         text_cmd(cmd);
      }
     }
}

char *
node_color(node_idx)
   int node_idx;
{
   int autoupd = atoi(Tcl_GetVar(tcl, "autoupd", TCL_GLOBAL_ONLY));
   
   /* Don't color cloud nodes, or any nodes in offline mode */
   if (!autoupd || (ndmode==ND_INCLOUD && !ndisplay(node_idx)))
      return (node_idx==curr_node)? "gray" : "white";

   /* Test for red/yellow/green */
   if (testcond(nodeRed.name, node_idx, nodeRed.op, &nodeRed.value))
      return (node_idx==curr_node)? "red3" : "red";
   if (testcond(nodeYellow.name, node_idx, nodeYellow.op, &nodeYellow.value))
      return (node_idx==curr_node)? "yellow3" : "yellow";
   if (testcond(nodeGreen.name, node_idx, nodeGreen.op, &nodeGreen.value))
      return (node_idx==curr_node)? "green3" : "green";

   /* Use the default */
   return (node_idx==curr_node)? "gray" : "white";
}

void
goto_node(i)
   int i;
{
   node_t *node;
   char buff[255];
   int   old_curr = curr_node;
   int   cloud = -1;

   /*
    * If we're going to a "cloud" node (e.g. by clicking on the virtual
    * node), the index will be the negative of the cloud index 
    */
   if (i<0) {
      goto_cloud(-i);
      return;
   }

   /*
    * If the node is in a cloud other than the one displayed,
    * first go to that cloud
    */
   node = mget_node(i);
   if (!node_tk[i] || node_tk[i]==cloud_tk[cloud=get_int("cloud", i)]
    || (ndmode==ND_INCLOUD && cloud!=curr_cloud)) {
       if (ndmode==ND_INCLOUD) {
          if (cloud<0)
             cloud = get_int("cloud", i);
          if (cloud==curr_cloud)
             return;
       }
       if (ndmode==ND_CUSTOM)
          ndmode = ND_INCLOUD;
       goto_cloud(cloud);
       goto_node(i);
       return;
   }

   curr_node = i;

   /* If the old current node is still displayed, unhighlight it */
/*fprintf(stderr,"<GN%d>",i);*/
   if (node_tk[old_curr]) {
      sprintf(buff, ".c itemconfigure $nodeRect(%d) -fill %s", 
       node_tk[old_curr], node_color(old_curr));
      text_cmd( buff);
   }

   /* Set global parameters linked to the current node */
   sprintf(buff, "set name %s", fullname(curr_node));
   text_cmd(buff);
   sprintf(buff, "set ip %s", ipstr(curr_node,0));
   text_cmd(buff);

   /* Highlight and display the new current node */
   show_detail(i);
   sprintf(buff, ".c itemconfigure $nodeRect(%d) -fill %s", 
    node_tk[curr_node], node_color(curr_node));
   text_cmd( buff);
}

char *
node_shape(node_idx)
   int node_idx;
{
   node_t *node;

   node = mget_node(node_idx);
   return (node->status & ST_SUBNET)? "oval" : "rectangle";
}

char *
node_symbol(node_idx)
   int node_idx;
{
static char buff[255];
   char *str = NULL;
   node_t *node;
   int cloud;

   node = mget_node(node_idx);

   if (ndmode==ND_INCLOUD) {
      cloud = get_int("cloud", node_idx);
      if (cloud_tk[cloud] && node_tk[node_idx] == cloud_tk[cloud])
         return get_string("cloudname", cloud);
   }

   if (nsvar.name1[0])
      sprintf(buff, "%s", get_string(nsvar.name1, node_idx));
   else
      buff[0]='\0';
   if (nsvar.name2[0]) {
      if (buff[0])
         strcat(buff, "/");
      sprintf(buff+strlen(buff), "%s", get_string(nsvar.name2, node_idx));
   }
   str = buff;

   if ((node->status & ST_SUBNET) && (!str || !str[0]))
      return fullname(node_idx);

   return str;
}

char *
toggle_text(idx)
   int idx;
{
   return "0";
}

void
set_node_symbol(node_idx)
   int node_idx;
{
   char buff[255];

   if (node_tk[node_idx]) {
      sprintf(buff, "setNodeText %d \"%s\" %s", node_tk[node_idx], 
       node_symbol(node_idx), toggle_text(node_idx));
      text_cmd(buff);
   } 
}
void
set_node_symbols()
{
   int node_idx;
  
   for (node_idx=0; node_idx<num_nodes(); node_idx++) {
      /* if (ndisplay(node_idx)) */
      if (node_tk[node_idx])
         set_node_symbol(node_idx);
   }
}

void
set_node_highlight(node_idx)
   int node_idx;
{
   char buff[255];

   if (node_tk[node_idx]) {
      sprintf(buff, "setNodeHl %d %d", node_tk[node_idx], nhighlight(node_idx));
      Tcl_Eval(tcl, buff);
   }
}

void
set_node_highlights()
{
   int node_idx;
   char *ptr, buff[256];

   ptr = cond_text(&nhcond); 
   sprintf(buff, ".map.top.but.nhcond configure -text {%s}", ptr);
   text_cmd(buff);
  
   for (node_idx=0; node_idx<num_nodes(); node_idx++) {
      if (!node_tk[node_idx]) continue;
      set_node_highlight(node_idx);
   }
}

void
set_node_color(node_idx)
   int node_idx;
{
   char buff[255];

   if (node_tk[node_idx]) {
      sprintf(buff, "setNodeColor %d %s", node_tk[node_idx], node_color(node_idx));
      text_cmd( buff);
   }
}

void
set_node_colors()
{
   int node_idx;
  
   for (node_idx=0; node_idx<num_nodes(); node_idx++) {
      if (!node_tk[node_idx]) continue;
/*      if (!ndisplay(node_idx)) continue; */
      set_node_color(node_idx);
   }
}

int
is_ncloud(idx)
   int idx;
{
   return node_tk[idx] 
    && (node_tk[idx] == cloud_tk[ get_int("cloud",idx)]);
}

int
is_edisplayed(idx)
   int idx;
{
   return (link_tk[idx]>0);
}

int
is_ndisplayed(idx)
   int idx;
{
   return (node_tk[idx]>0);
}

int
is_cdisplayed(idx)
   int idx;
{
   return (cloud_tk[idx]>0);
}

/* Test to see if a node represents a subnet with only one attached host */
static int
is_terminal_subnet(idx)
   int idx;
{
   int lin, lout;
   node_t *node = mget_node(idx);

   if (!(node->status & ST_SUBNET))
      return 0;

   /* Check for multiple incoming links */
   lin = node->last_link[LIN];
   if (lin>=0 && llink[lin].prev_link[LIN]>=0)
      return 0;

   /* Check for multiple outgoing links */
   lout = node->last_link[LOUT];
   if (lout>=0 && llink[lout].prev_link[LOUT]>=0)
      return 0;

   /* Check for different incoming/outgoing links */
   if (lin>=0 && lout>=0 && llink[lin].src != llink[lout].dest)
      return 0;

   return 1;
}

/* Should node be displayed? 
 * Note that if in cloud mode, and node is adjacent to the current cloud,
 * then this returns false.  However a cloud representing the node will
 * still be displayed (by the link display routines)
 */
int /* RETURNS: 1=yes, 0=no */
ndisplay(idx)
   int idx;
{
   if (is_ndeleted(idx))
      return FALSE;

   /* Test for terminal subnet exclusion */
   if (!display_terminal_subnets && idx!=curr_node && is_terminal_subnet(idx))
      return FALSE;

   switch(ndmode) {

   case ND_CUSTOM:
      return testcond(ndcond.name, idx, ndcond.op, &ndcond.value);

   default: 
   case ND_INCLOUD:
      return (get_int("cloud", idx)==curr_cloud);
   }
}

/* Should node have info discovered? */
int /* RETURNS: 1=yes, 0=no */
ninfo(idx)
   int idx;
{
   if (is_ndeleted(idx))
      return 0;

   switch(nimode) {
   case NI_CUSTOM:
      return testcond(nicond.name, idx, nicond.op, &nicond.value);

   case NI_DISPLAYED:
      return ndisplay(idx);

   case NI_NONE:
   default: 
      return 0;
   }
}

/* Is a node to be displayed as a cloud node? */
int
cdisplay(n)
   int n;   /* node index */
{
   int ln;
   node_t *node = mget_node(n);

   if (ndmode!=ND_INCLOUD) 
      return 0;

   if (ndisplay(n)) /* not if its displayed normally! */
      return 0;

   if (!get_int("cloud", n)) /* not if it's in the null cloud */
      return 0;

   /* Test for terminal subnet exclusion */
   if (!display_terminal_subnets && n!=curr_node && is_terminal_subnet(n))
      return FALSE;

   for (ln = node->last_link[LOUT]; ln>=0; ln = llink[ln].prev_link[LOUT])
      if (ndisplay(llink[ln].dest))
         return 1;

   for (ln = node->last_link[LIN]; ln>=0; ln = llink[ln].prev_link[LIN])
      if (ndisplay(llink[ln].src))
         return 1;

   return 0;
}

/* Should node be displayed? */
void
set_node_display(node_idx)
   int node_idx;
{
   int x,y;
   char buff[255];
   node_t *node;

   if (!display_ok)
      return;

   if (ndisplay(node_idx)) {
      node = mget_node(node_idx);
      x = Xn2c(node); /* must use n2c not i2c */
      y = Yn2c(node); /* to avoid cloud stuff */

      /* If previously thrown into cloud node, undo it */
      if (is_ncloud(node_idx)) 
         remove_cloud_display(node_idx);
/*         set_cloudlink_display(node_idx, 0, 0); */

      if (node_tk[node_idx]) {  /* move existing */
         sprintf(buff, "moveNodeTo %d %d %d", node_tk[node_idx], x, y);
         do_tcl(buff);
      } else {                  /* create new */
         sprintf(buff, "mkNode %d %d %s {%s} %d %s %s", x,y, 
          node_shape(node_idx), node_symbol(node_idx), node_idx, 
          toggle_text(node_idx), node_color(node_idx));
         text_cmd( buff);

         node_tk[node_idx] = atoi(tcl->result);
         add_timers(VL_NODE, node_idx);

         set_node_symbol(node_idx);
         set_node_highlight(node_idx);

         ldcond.dirty = TRUE; /* force links to be displayed */
      }
   } else {
      if (node_tk[node_idx] && !is_ncloud(node_idx)) {  /* delete existing */
         sprintf(buff, "rmNode %d", node_tk[node_idx]);
         do_tcl(buff);

         node_tk[node_idx] = 0;
         remove_timers(VL_NODE, node_idx);

         ldcond.dirty = TRUE; /* force links to be undisplayed */
      }
   }
}

char *
cloud_text()
{
static char buff[256];

   switch(ndmode) {
   case ND_CUSTOM:
      return cond_text(&ndcond);

   default:
   case ND_INCLOUD:
      return get_string("cloudname", curr_cloud);
   }
}

void
remove_cloud_display(node_idx)
   int node_idx;
{
   int cloud = get_int("cloud", node_idx);
   char buff[256];

   /* Break connection between node and cloud */
   if (node_tk[node_idx]>0 && node_tk[node_idx]==cloud_tk[cloud]) {
      cloud_ref[cloud]--;
      node_tk[node_idx] = 0;
   }

   /* Delete cloud if no more references */
   if (!cloud_ref[cloud] && cloud_tk[cloud]) {  /* delete existing */
      sprintf(buff, "rmNode %d", cloud_tk[cloud]);
      do_tcl(buff);

      cloud_tk[cloud] = 0;
      ldcond.dirty = TRUE; /* force links to be undisplayed */
   }
}

/*
 * We make adjacent clouds displayed by making all node_tk[] values of
 * adjacent nodes be the same, = cloud_tk[cloud] where cloud is the
 * cloud index common to all the nodes in that cloud.
 *
 * Since one cloud node represents multiple adjacent nodes, a cloud_ref[]
 * counter is maintained.  When the ref count hits 0, the cloud node is
 * removed from the display.  This happens when we decide all adjacent
 * nodes (in that cloud) are no longer displayed.
 */
void
set_cloud_display(node_idx)
   int node_idx;
{
   int cloud, near = -1, ln;
   char buff[256];
   node_t *node;
   log_link_t *edge;
   int x,y;

   if (!display_ok)
      return;

   cloud = get_int("cloud", node_idx);
   if (cdisplay(node_idx)) {

      if (!cloud_ref[cloud]) { /* set coords on first node reference */

         /* Locate a nearby node in curr_cloud */
         node = mget_node(node_idx);
         for (ln = node->last_link[LOUT]; ln>=0; ln=llink[ln].prev_link[LOUT]) {
            edge = get_log_link(ln);
            if (get_int("cloud", edge->dest)==curr_cloud) {
               near = edge->dest;
               break;
            }
         }
         if (ln<0)
          for (ln=node->last_link[LIN]; ln>=0; ln=llink[ln].prev_link[LIN]) {
            edge = get_log_link(ln);
            if (get_int("cloud", edge->src)==curr_cloud) {
               near = edge->src;
               break;
            }
          }
         if (near<0)
            abort();
         
         node = mget_node(near);
         /* xy_idx = 2 for ND_INCLOUD */
         cloud_x[cloud] = node->x[2] -50.0 + (random()%100);
         cloud_y[cloud] = node->y[2] -50.0 + (random()%100);
      }
      x = Xv2c(cloud_x[cloud]);
      y = Yv2c(cloud_y[cloud]);

      /* Now make sure the cloud node is placed on the display */
      if (cloud_tk[cloud]) {  /* move existing */
         sprintf(buff, "moveNodeTo %d %d %d", cloud_tk[cloud], x, y);
         do_tcl(buff);
      } else {                  /* create new */
         sprintf(buff, "mkNode %d %d %s {%s} %d %s %s", x,y, "oval",
          get_string("cloudname", cloud), -cloud, "0", "white");
         text_cmd( buff);

         cloud_tk[cloud] = atoi(tcl->result);
         ldcond.dirty = TRUE; /* force links to be displayed */
      }

      /* Delete old node and links if necessary 
       * XXX why should this occur?   set_node_display() should
       * have already deleted it.
       */
      if (node_tk[node_idx] && node_tk[node_idx]!=cloud_tk[cloud]) {  
         set_node_display(node_idx);
/*
         sprintf(buff, "rmNode %d", node_tk[n]);
         do_tcl(buff);
         remove_timers(VL_NODE, n);
*/
      }

      if (node_tk[node_idx] != cloud_tk[cloud]) {
         node_tk[node_idx] = cloud_tk[cloud];
         cloud_ref[cloud]++;
      }

   } else
      remove_cloud_display(node_idx);
}

void
set_cloud_displays()
{
   int node_idx;

   for (node_idx=0; node_idx<num_nodes(); node_idx++)
      set_cloud_display(node_idx);
}

void
set_node_displays()
{
   int node_idx;
   char cmd[256];
#if 0
   int i;
   float xrng, yrng;
   int ok, idx;
   node_t *node;
#endif

   if (ndmode==ND_INCLOUD) {
      sprintf(cmd, ".map.top.but.ndcond configure -text {%s=%s}", cloudvar(), cloud_text());
   } else
      sprintf(cmd, ".map.top.but.ndcond configure -text {%s}", cloud_text());

   text_cmd(cmd);

#if 0
   /* Calculate the range of coordinates to display */
   ok=0;
   idx = xy_idx();
   for (node_idx=0; node_idx<num_nodes(); node_idx++) {
      if (ndisplay(node_idx)) {
         node = mget_node(node_idx);
         if (!ok || node->x[idx] >maxx) maxx=node->x[idx] ;
         if (!ok || node->y[idx] >maxy) maxy=node->y[idx] ;
         if (!ok || node->x[idx] <minx) minx=node->x[idx] ;
         if (!ok || node->y[idx] <miny) miny=node->y[idx] ;
         ok++;
      } else if (ndmode==ND_INCLOUD) {
         /*
          * We also need to take cloud nodes into account when calculating
          * min and max coords so that clouds won't be pushed off the edge
          * by "spread out".
          */
         i = get_int("cloud", node_idx);
         if (cdisplay(node_idx)) {
            if (cloud_tk[i] && cloud_x[i] > maxx) maxx = cloud_x[i];
            if (cloud_tk[i] && cloud_x[i] < minx) minx = cloud_x[i];
            if (cloud_tk[i] && cloud_y[i] > maxy) maxy = cloud_y[i];
            if (cloud_tk[i] && cloud_y[i] < miny) miny = cloud_y[i];
         }
      }
   } 

   xrng = maxx-minx;
   yrng = maxy-miny;
   if (xrng<1.0) xrng=1.0;
   if (yrng<1.0) yrng=1.0;
   xinc = xrng/zoom;
   yinc = yrng/zoom;
#endif

   for (node_idx=0; node_idx<num_nodes(); node_idx++)
      set_node_display(node_idx);

   /* And redisplay the clouds */
   set_cloud_displays();
}

char *
ecolor(ln,rln)
   int ln;
   int rln;
{
   int autoupd = atoi(Tcl_GetVar(tcl, "autoupd", TCL_GLOBAL_ONLY));

   if (ehighlight(ln) || (rln>=0 && ehighlight(rln)))
      return "blue";

   /* Insert red/yellow/green tests here */
   if (autoupd) {
      if (testcond(linkRed.name,  ln, linkRed.op, &linkRed.value)
       || testcond(linkRed.name, rln, linkRed.op, &linkRed.value))
         return "red";
      if (testcond(linkYellow.name,  ln, linkYellow.op, &linkYellow.value)
       || testcond(linkYellow.name, rln, linkYellow.op, &linkYellow.value))
         return "yellow";
      if (testcond(linkGreen.name,  ln, linkGreen.op, &linkGreen.value)
       || testcond(linkGreen.name, rln, linkGreen.op, &linkGreen.value))
         return "green3";
   }

   return "black";
}

char *
link_symbol(ln)
   int ln;
{
static char buff[256];
/*
   log_link_t *edge = get_log_link(ln);
   node_t *src = mget_node(edge->src);

   Links from subnets have no text? 
   if (src->status & ST_SUBNET)
      return "{}";
*/

   if (lsvar.name1[0])
      sprintf(buff, "%s", get_string(lsvar.name1, ln));
   else
      buff[0]='\0';
   if (lsvar.name2[0]) {
      if (buff[0])
         strcat(buff, "/");
      sprintf(buff+strlen(buff), "%s", get_string(lsvar.name2, ln));
   }

   return (buff[0])? buff : "{}";
}

int
ewidth(ln,rln)
   int ln;
   int rln;
{
/*
   if (ehighlight(ln) || (rln>=0 && ehighlight(rln)))
      return 3;
   return 1;
*/
   log_link_t *edge;

   edge = get_log_link(ln);
   if (edge->flags & DVMRP_NF_TUNNEL)
      return 3;
   else
      return 1;
}

static int
edisplay(ln, rln)
   int ln;
   int rln;
{
   log_link_t *edge;

   /* Limit to edges between displayed nodes */
   edge = get_log_link(ln);

#ifdef SHOW_EXTRA_LINKS
   if (ndmode==ND_INCLOUD) { /* for clouds only, show border links */
      int scloud, dcloud;

      /* Don't display it if neither end is a real (non-cloud) node */
      scloud = get_int("cloud", edge->src);
      dcloud = get_int("cloud", edge->dest);
      if (( !node_tk[edge->src]  || node_tk[edge->src ]==cloud_tk[scloud]) 
       && ( !node_tk[edge->dest] || node_tk[edge->dest]==cloud_tk[dcloud]))
         return 0;

   } else 
#endif
   {
      if (edge->src<0 || edge->dest<0 
       || !node_tk[edge->src] || !node_tk[edge->dest])
         return 0;
   }

   if (edge->src<0 || edge->dest<0 || edge->dest==edge->src)
      return 0; /* self-loop (subnet) */

   return testcond(ldcond.name, ln, ldcond.op, &ldcond.value);
}

void
set_link_display(ln)
   int ln; /* log link # */
{
   char buff[255], *str;
   int rln, sx, sy, dx, dy;
   log_link_t *edge;
   int scloud, dcloud;

   rln = reverse(ln);
   if (edisplay(ln, rln)) {
      edge = get_log_link(ln);

      /*
       * Display clouds for border links.  The adjacent nodes should
       * node be displayed at this point, so node_tk[] will be 0 for them
      if (ndmode==ND_INCLOUD) {
         if (!node_tk[edge->src])
            set_cloudlink_display(edge->src, edge->dest, 1);
         else if (!node_tk[edge->dest])
            set_cloudlink_display(edge->dest, edge->src, 1);
      }
       */

      if (link_tk[ln]) {  /* move existing */
         sprintf(buff, ".c itemconfigure %d -fill %s -width %d",
          link_tk[ln], ecolor(ln,rln), ewidth(ln,rln));
         text_cmd( buff);

         /* Fix edge source in case node display criteria changed */
         sprintf(buff, "edgeSrc(%d)", link_tk[ln]);
         str = Tcl_GetVar(tcl, buff, TCL_GLOBAL_ONLY);
         if (str && !strcmp(str, "-1")) { /* link to cloud */
            sprintf(buff,"setLinkSrc %d %d",link_tk[ln], node_tk[edge->src]);
            text_cmd(buff);
         }

         /* Fix edge destination in case node display criteria changed */
         sprintf(buff, "edgeDest(%d)", link_tk[ln]);
         str = Tcl_GetVar(tcl, buff, TCL_GLOBAL_ONLY);
         if (str && !strcmp(str, "-1")) { /* link to cloud */
           sprintf(buff,"setLinkDest %d %d",link_tk[ln],node_tk[edge->dest]);
           text_cmd(buff);
         }

         /*
          * Update coordinates in case endpoints moved by virtue of
          * the X/Y index changing.
          */
         scloud = get_int("cloud", edge->src);
         if (node_tk[edge->src] && node_tk[edge->src]==cloud_tk[scloud]) {
            sx = Xv2c(cloud_x[scloud]);
            sy = Yv2c(cloud_y[scloud]);
         } else {
            sx = Xn2c(mget_node(edge->src)); /* must use n2c not i2c */
            sy = Yn2c(mget_node(edge->src)); /* to avoid cloud stuff */
         }

         dcloud = get_int("cloud", edge->dest);
         if (node_tk[edge->dest] && node_tk[edge->dest]==cloud_tk[dcloud]) {
            dx = Xv2c(cloud_x[dcloud]);
            dy = Yv2c(cloud_y[dcloud]);
         } else {
            dx = Xn2c(mget_node(edge->dest)); /* must use n2c not i2c */
            dy = Yn2c(mget_node(edge->dest)); /* to avoid cloud stuff */
         }

         sprintf(buff, ".c coords %d %d %d %d %d", link_tk[ln], sx, sy, dx, dy);
         do_tcl(buff);

      } else {            /* create new */
         sprintf(buff, "mkEdge %d %d %s %d %s", node_tk[edge->src],
          node_tk[edge->dest], ecolor(ln,rln), ewidth(ln,rln),
          link_symbol(ln));
         text_cmd( buff);

         link_tk[ln] = atoi(tcl->result);
         add_timers(VL_LINK, ln);
      }

   } else {
      if (link_tk[ln]) {  /* delete existing */

         /* Un-Display clouds for border links 
         if (ndmode==ND_INCLOUD) {
            edge = get_log_link(ln);
            if (cloud_tk[get_int("cloud",edge->src)])
               set_cloudlink_display(edge->src, 0, 0);
            else if (cloud_tk[get_int("cloud",edge->dest)])
               set_cloudlink_display(edge->dest, 0, 0);
         }
         */

         sprintf(buff, "rmEdge %d", link_tk[ln]);
         do_tcl(buff);

         link_tk[ln] = 0;
         remove_timers(VL_LINK, ln);
      }
   }
}

void
set_link_displays()
{
   int ln;
   char *ptr, buff[256];

   ptr = cond_text(&ldcond);
   sprintf(buff, ".map.top.but.ldcond configure -text {%s}", ptr);
   text_cmd(buff);

   for (ln=0; ln<num_log_links(); ln++)
      set_link_display(ln);
}

void
set_link_highlight(ln)
   int ln;
{
   char buff[255];
   int rln;

   if (link_tk[ln]) {
      rln = reverse(ln);
      sprintf(buff, ".c itemconfigure %d -fill %s -width %d", 
       link_tk[ln], ecolor(ln,rln), ewidth(ln,rln));
      text_cmd( buff);
   }
}

void
set_link_highlights()
{
   int ln;
   char *ptr, buff[256];

   ptr = cond_text(&lhcond); 
   sprintf(buff, ".map.top.but.lhcond configure -text {%s}", ptr);
   text_cmd(buff);
  
   for (ln=0; ln<num_log_links(); ln++) {
      if (!link_tk[ln]) continue;
      set_link_highlight(ln);
   }
}

void
set_link_symbol(ln)
   int ln;
{
   char buff[255];

   if (link_tk[ln]) {
      sprintf(buff, "setEdgeText %d %s", link_tk[ln], link_symbol(ln));
      text_cmd( buff);
   }
}

void
set_link_symbols()
{
   int ln;
  
   for (ln=0; ln<num_log_links(); ln++) {
      if (!link_tk[ln]) continue;
      set_link_symbol(ln);
   }
}

/*
 * set_X and set_Y are called when the user moves the node around with
 * the mouse 
 */
char *
set_X(clientData, interp, name1, name2, flags)
   ClientData  clientData;
   Tcl_Interp *interp;
   char       *name1;
   char       *name2;
   int         flags;
{
   /*
    * Don't pass the flags argument into the Tcl_GetVar2 calls below
    * since on an Alpha (at least) the PART1_NOT_PARSED bit is set (0x10000)
    * which causes the get to fail and return NULL.
    */
   int cx = atoi(Tcl_GetVar2(interp,name1,name2,TCL_GLOBAL_ONLY));
   int n = atoi(Tcl_GetVar2(interp, "nodeIdx", name2, TCL_GLOBAL_ONLY));
   float vx = Xc2v(cx);

   /* Check if moving cloud node */
   if (n<0)
      cloud_x[-n] = vx;
   else
      mget_node(n)->x[xy_idx()] = vx;
   return NULL;
}

char *
set_Y(clientData, interp, name1, name2, flags)
   ClientData  clientData;
   Tcl_Interp *interp;
   char       *name1;
   char       *name2;
   int         flags;
{
   int cy = atoi(Tcl_GetVar2(interp,name1,name2,TCL_GLOBAL_ONLY));
   int n = atoi(Tcl_GetVar2(interp, "nodeIdx", name2, TCL_GLOBAL_ONLY));
   float vy = Yc2v(cy);

   /* Check if moving cloud node */
   if (n<0)
      cloud_y[-n] = vy;
   else
      mget_node(n)->y[xy_idx()] = vy;
   return NULL;
}

char *
cb_terminals(clientData, interp, name1, name2, flags)
   ClientData  clientData;
   Tcl_Interp *interp;
   char       *name1;
   char       *name2;
   int         flags;
{
   display_terminal_subnets = atoi(Tcl_GetVar2(interp,name1,name2,flags));
   ndcond.dirty = TRUE; /* force nodes to be rechecked for display */
   refresh_display();
   return NULL;
}

char *
cb_fontsize(clientData, interp, name1, name2, flags)
   ClientData  clientData;
   Tcl_Interp *interp;
   char       *name1;
   char       *name2;
   int         flags;
{
   char buff[256];
   int i = atoi(Tcl_GetVar2(interp,name1,name2,flags));
   nsvar.dirty = TRUE; /* force text to be redisplayed */
   sprintf(buff, "set_font %d", i);
   text_cmd(buff);
   refresh_display();
   return NULL;
}

char *
cb_setsrc(clientData, interp, name1, name2, flags)
   ClientData  clientData;
   Tcl_Interp *interp;
   char       *name1;
   char       *name2;
   int         flags;
{
   char *str = Tcl_GetVar2(interp,name1,name2,flags);
   int n;
   char buff[256];

   if (!strcmp(name1, "srcname")) {
      if ((n=find_node_by_name(str))>=0)
         mark_source(n);
      else {
         sprintf(buff, "'%s' not found.", str);
         add_msg(buff);
         mark_source(curr_node);
      }
   } else if (!strcmp(name1, "srcip")) {
      if ((n=find_node_by_addr(inet_addr(str)))>=0)
         mark_source(n);
      else {
         sprintf(buff, "'%s' not found.", str);
         add_msg(buff);
         mark_source(curr_node);
      }
   } else if (!strcmp(name1, "destname")) {
      if ((n=find_node_by_name(str))>=0)
         mark_dest(n);
      else {
         sprintf(buff, "'%s' not found.", str);
         add_msg(buff);
         mark_dest(curr_node);
      }
   } else if (!strcmp(name1, "destip")) {
      if ((n=find_node_by_addr(inet_addr(str)))>=0)
         mark_dest(n);
      else {
         sprintf(buff, "'%s' not found.", str);
         add_msg(buff);
         mark_dest(curr_node);
      }
   }
   refresh_display();
   return NULL;
}

char *
cb_fillin(clientData, interp, name1, name2, flags)
   ClientData  clientData;
   Tcl_Interp *interp;
   char       *name1;
   char       *name2;
   int         flags;
{
   int i = atoi(Tcl_GetVar2(interp,name1,name2,flags));
   if (!strcmp(name1, "ndmode")) {
      if (i==ND_INCLOUD && ndmode!=i) { /* switching TO clouds */
         curr_cloud = get_int("cloud", curr_node);
         start_updates(cloudvar());
      } else if (ndmode==ND_INCLOUD && i!=ndmode) { /* switching OFF clouds */
         stop_updates(cloudvar());
      }
      ndmode = i;
/* Removed 7/2/96 to preserve macro topology layout of clouds
 *    if (i==ND_INCLOUD)
 *       recalc_high();
 */
   } else if (!strcmp(name1, "nimode")) {
      if (i==NI_CUSTOM && nimode!=i)
         start_updates(nicond.name);
      else if (nimode==NI_CUSTOM && i!=nimode)
         stop_updates(nicond.name);
      nimode = i;
   } 
   refresh_display();
   return NULL;
}

void
refresh_display() 
{
   int v;
   char buff[255];

   /* Make sure the main window is active */
   if (!display_ok)
      return;

   /* Remove any deleted nodes */
   for (v=0; v<num_nodes(); v++) {
      if (node_tk[v]) {
         if (is_ndeleted(v)) {
            sprintf(buff, "rmNode %d", node_tk[v]);
            do_tcl(buff);

            node_tk[v] = 0;
            remove_timers(VL_NODE, v);
            ldcond.dirty = TRUE; /* force links to be undisplayed */
         }
      }
   }

   /* Fill in nodes */

   /* Update node display criteria (no text or hl tho) */
   if (ndmode!=ND_CUSTOM || ndcond.dirty || is_dirty(ndcond.name)) {
      set_node_displays();
   }

   /* Update node text */
   if (nsvar.dirty || is_dirty(nsvar.name1) || is_dirty(nsvar.name2)) {
      set_node_symbols();
      nsvar.dirty = 0;
   }

   /* Update node highlighting */
   if (nhcond.dirty || is_dirty(nhcond.name)) {
      set_node_highlights();
      nhcond.dirty = 0;
   }

   /* Update node colors */
   if (nodeRed.dirty    || is_dirty(nodeRed.name)
    || nodeYellow.dirty || is_dirty(nodeYellow.name)
    || nodeGreen.dirty  || is_dirty(nodeGreen.name)) {
      set_node_colors();
      nodeRed.dirty = 0;
      nodeYellow.dirty = 0;
      nodeGreen.dirty = 0;
   }

   /* Fill in edges */

   /* Update edge display criteria */
   /* Note that we also need to do this when node display criteria changes
    * so that links disappear when nodes do.
    */
   if (ldcond.dirty || is_dirty(ldcond.name)) {
      set_link_displays();
      ldcond.dirty = 0;
   }

   /* Update edge highlighting */
   if (linkRed.dirty    || is_dirty(linkRed.name)
    || linkYellow.dirty || is_dirty(linkYellow.name)
    || linkGreen.dirty  || is_dirty(linkGreen.name)
    || lhcond.dirty     || is_dirty(lhcond.name)) {
      set_link_highlights();
      if (clhmode==CLH_LOGLINK)
         set_cloudlink_highlights();
      lhcond.dirty = 0;
      linkRed.dirty = 0;
      linkYellow.dirty = 0;
      linkGreen.dirty = 0;
   }

   /* Update edge symbols */
   if (lsvar.dirty || is_dirty(lsvar.name1) || is_dirty(lsvar.name2)) {
      set_link_symbols();
      lsvar.dirty = 0;
   }

   /* XXX is there some possible test for dirtyness here? */
   show_detail(curr_node); 

   refresh_high();

   /*
    * Mark variables as clean.  Note that we can't do this in
    * conjuction with the displays above since the variable may be
    * used for multiple things, and it must stay dirty for all of 
    * them.
    */
   ndcond.dirty = 0;

   if (ndmode==ND_CUSTOM)    /* clean node displays */
      clean(ndcond.name);

   clean(nsvar.name1);       /* clean node text */
   clean(nsvar.name2);

   clean(nhcond.name);       /* clean node highlights */
   clean(nodeRed.name);      /* clean node colors */
   clean(nodeYellow.name);
   clean(nodeGreen.name);

   clean(ldcond.name);       /* clean link display */
   clean(lhcond.name);       /* clean link highlighting */
   clean(linkRed.name);      /* clean link colors */
   clean(linkYellow.name);
   clean(linkGreen.name);

   clean(lsvar.name1);       /* clean link text */
   clean(lsvar.name2);
}

#define TK4
/* CHANGE OFFX AND OFFY TO PUT NODE IN CENTER OF CANVAS */
void
center_node(n)
   int n;          /* Index of node */
{
   char buff[256];

#ifdef TK4
   sprintf(buff, "centerNode %d", node_tk[n]);
   do_tcl(buff);
/* Tcl_Eval(tcl, ".c configure -xscrollincrement 20 -yscrollincrement 20"); */
#else
   Tk_Window win = Tk_NameToWindow(tcl, ".c", tk_main);
   int cx = Xi2c(n); /* Canvas X,Y coordinates */
   int cy = Yi2c(n);

   Tcl_Eval(tcl, ".c configure -scrollincrement 1");
   Tcl_Eval(tcl, sprintf(buff, ".c xview %d", cx-Tk_Width(win)/2));
   Tcl_Eval(tcl, sprintf(buff, ".c yview %d", cy-Tk_Height(win)/2));
   Tcl_Eval(tcl, ".c configure -scrollincrement 20");
#endif
}

/* Rescale everything displayed into the [100,900]x[100,900] range */
void
rescale_nodes()
{
   node_t *node;
   int i;
   float minx, maxx, miny, maxy;
   float xrng, yrng;
   int idx = xy_idx();
   int cloud;
   int n = num_nodes(), ok=0;

   i = 0;
   if (ndmode==ND_INCLOUD && node_tk[i]>0
    && node_tk[i]==cloud_tk[cloud=get_int("cloud",i)]) {
      minx = maxx = cloud_x[cloud];
      miny = maxy = cloud_y[cloud];
   } else {
      node = mget_node(i);
      minx = maxx = node->x[idx];
      miny = maxy = node->y[idx];
   }

   /* Find bounding coords of normal nodes */
   for (i=0; i<n; i++) {
      if (!is_ndisplayed(i) || is_ncloud(i)) continue;
      node = mget_node(i);
      if (!ok || node->x[idx] < minx) minx = node->x[idx];
      if (!ok || node->x[idx] > maxx) maxx = node->x[idx];
      if (!ok || node->y[idx] < miny) miny = node->y[idx];
      if (!ok || node->y[idx] > maxy) maxy = node->y[idx];
      ok=1;
   }

   if (!ok)
      return; /* should never happen */

   /* Now include cloud nodes */
   for (i=0; i<num_clouds; i++) {
      if (!is_cdisplayed(i)) continue;
      if (cloud_x[i] < minx) minx = cloud_x[i];
      if (cloud_x[i] > maxx) maxx = cloud_x[i];
      if (cloud_y[i] < miny) miny = cloud_y[i];
      if (cloud_y[i] > maxy) maxy = cloud_y[i];
   }

   /* Compute original range lengths */
   xrng = maxx-minx;
   if (!xrng)
      xrng = 10.0; /* to avoid div by 0 */
   yrng = maxy-miny;
   if (!yrng)
      yrng = 10.0;

   /* Now shift & scale coords for normal nodes */
   for (i=0; i<n; i++) {
      if (!is_ndisplayed(i) || is_ncloud(i)) continue;
      node = mget_node(i);
      node->x[idx] = 100.0 + (node->x[idx]-minx) / xrng * 800.0;
      node->y[idx] = 100.0 + (node->y[idx]-miny) / yrng * 800.0;
   }

   /* And shift and scale cloud node coords */
   for (i=0; i<num_clouds; i++) {
      if (!is_cdisplayed(i)) continue;
      cloud_x[i] = 100.0 + (cloud_x[i]-minx) / xrng * 800.0;
      cloud_y[i] = 100.0 + (cloud_y[i]-miny) / yrng * 800.0;
   }
}
