/* $Id: driver.c,v 1.29 1996/06/20 23:22:13 thalerd Exp $
 * -------------------------------------------------------------------------
 *  
 * [C] The Regents of the University of Michigan and Merit Network, Inc. 1996
 * All Rights Reserved 
 *   
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice and this permission notice appear in all 
 * copies of the software and derivative works or modified versions thereof, 
 * and that both the copyright notice and this permission and disclaimer notice 
 * appear in supporting documentation. 
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE REGENTS OF THE 
 * UNIVERSITY OF MICHIGAN AND MERIT NETWORK, INC. DO NOT WARRANT THAT THE 
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR 
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the 
 * University of Michigan and Merit Network, Inc. shall not be liable for any 
 * special, indirect, incidental or consequential damages with respect to any 
 * claim by Licensee or any third party arising from use of the software. 
 *
 * -------------------------------------------------------------------------
 */
/* DRIVER.C - command interpreter */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h> /* for isspace */
#include <sys/types.h>
#include <sys/socket.h>
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_BSTRING_H
#include <bstring.h>
#endif
#include <tcl.h>
#include <tk.h>
#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h> /* for fork, etc */
#endif
#ifndef HAVE_GETHOSTNAME
#include <sys/systeminfo.h>
#endif
#include "mview.h"
#include "struct.h"
#include "builtin.h"
#include "driver.h"
#include "graph.h"
#include "node.h"
#include "display.h"
#include "handler.h"
#include "lib.h"
#include "xalloc.h"
#include "database.h"
#include "dist_vect.h"
#include "globals.h"
#include "dict.h"
#include "high.h"
#include "maplib.h"
#include "app.h" /* define application arrays */
#include "queue.h" /* for init_queues */
#define SHELL "/bin/sh"

char *datadir = NULL;

typedef struct {
   char   key;
   char   fmt[256];
} alias_t;

#define MAXALIAS 100
alias_t alias[MAXALIAS];
Tcl_Interp *tcl;

char *
debug_nodeIdx(clientData, interp, name1, name2, flags)
   ClientData  clientData;
   Tcl_Interp *interp;
   char       *name1;
   char       *name2;
   int         flags;
{
   printf("UNSET name1=%s name2=%s flags=%d\n", name1, name2, flags);
   return NULL;
}

/*
 * Given an IP address, return true(1) or false(0) depending on whether
 * the IP address is on the local subnet.  This is used by the mtrace 
 * handler.
 */
static u_int32 local_ipaddr;
int
is_local(addr)
   u_int32 addr;
{
   int subnet = find_subnet(addr);
   node_t *node;

   if (subnet<0)
      return addr==local_ipaddr;

   node = mget_node(subnet);
   return ((addr & node->iface[1]) == node->iface[0]);
}

#define MAX_CHAR_NET_INFO 100
int
Mv_Init(interp)
   Tcl_Interp *interp;
{
extern void source_mviewrc();
   char my_domain_name[MAX_CHAR_NET_INFO], *tmp_name;
   struct hostent *hp;
   struct in_addr my_addr;
   int i, local_node;

   check_binaries();

   tcl = interp;

   init_queues();

   /* Initialize commands */
   for (i=0; command[i].func; i++)
      Tcl_CreateCommand(tcl, command[i].text, command[i].func, 0, 0);

   /* Link global variables */
   Tcl_LinkVar( tcl, "curr_cloud",(char *)&curr_cloud,TCL_LINK_INT );
   Tcl_LinkVar( tcl, "curr_node", (char *)&curr_node, TCL_LINK_INT );
   Tcl_LinkVar( tcl, "iotype",    (char *)&iotype   , TCL_LINK_INT );
   Tcl_LinkVar( tcl, "source_node",(char *)&source_node, TCL_LINK_INT );
   Tcl_LinkVar( tcl, "dest_node", (char *)&dest_node, TCL_LINK_INT );
   Tcl_LinkVar( tcl, "maxage",    (char *)&maxage   , TCL_LINK_INT );

   Tcl_LinkVar( tcl, "zoom",      (char *)&zoom     , TCL_LINK_DOUBLE );
   Tcl_LinkVar( tcl, "xscale",    (char *)&xscale   , TCL_LINK_DOUBLE );
   Tcl_LinkVar( tcl, "yscale",    (char *)&yscale   , TCL_LINK_DOUBLE );
   Tcl_LinkVar( tcl, "Hxscale",   (char *)&Hxscale  , TCL_LINK_DOUBLE );
   Tcl_LinkVar( tcl, "Hyscale",   (char *)&Hyscale  , TCL_LINK_DOUBLE );
   Tcl_LinkVar( tcl, "Hzoom",     (char *)&Hzoom    , TCL_LINK_DOUBLE );
#if 0
   Tcl_LinkVar( tcl, "maxx",      (char *)&maxx     , TCL_LINK_DOUBLE );
   Tcl_LinkVar( tcl, "maxy",      (char *)&maxy     , TCL_LINK_DOUBLE );
   Tcl_LinkVar( tcl, "minx",      (char *)&minx     , TCL_LINK_DOUBLE );
   Tcl_LinkVar( tcl, "miny",      (char *)&miny     , TCL_LINK_DOUBLE );
   Tcl_LinkVar( tcl, "xinc",      (char *)&xinc     , TCL_LINK_DOUBLE );
   Tcl_LinkVar( tcl, "yinc",      (char *)&yinc     , TCL_LINK_DOUBLE );
#endif

   Tcl_LinkVar( tcl, "datadir",   (char *)&datadir  , TCL_LINK_STRING );

   Tcl_TraceVar(tcl, "nimode",  TCL_TRACE_WRITES, cb_fillin, NULL );
   Tcl_TraceVar(tcl, "ndmode",  TCL_TRACE_WRITES, cb_fillin, NULL );
   Tcl_TraceVar(tcl, "autoupd", TCL_TRACE_WRITES, set_autoupd, NULL );
   Tcl_TraceVar(tcl, "fontsize", TCL_TRACE_WRITES, cb_fontsize, NULL );
   Tcl_TraceVar(tcl, "terminals",TCL_TRACE_WRITES, cb_terminals, NULL );
/*
   Tcl_TraceVar(tcl, "srcname", TCL_TRACE_WRITES, cb_setsrc, NULL );
   Tcl_TraceVar(tcl, "srcip",   TCL_TRACE_WRITES, cb_setsrc, NULL );
   Tcl_TraceVar(tcl, "destname", TCL_TRACE_WRITES, cb_setsrc, NULL );
   Tcl_TraceVar(tcl, "destip",   TCL_TRACE_WRITES, cb_setsrc, NULL );
*/

   Tcl_TraceVar(tcl, "nodeX", TCL_TRACE_WRITES, set_X, NULL );
   Tcl_TraceVar(tcl, "nodeY", TCL_TRACE_WRITES, set_Y, NULL );
   Tcl_TraceVar(tcl, "HnodeX", TCL_TRACE_WRITES, set_cloudX, NULL );
   Tcl_TraceVar(tcl, "HnodeY", TCL_TRACE_WRITES, set_cloudY, NULL );

   /*
    * Initialize local hostname information and add current node
    * to database as physical node if not already there.
    */
#ifdef HAVE_GETHOSTNAME
   gethostname(my_domain_name, MAX_CHAR_NET_INFO); /* eg quip.eecs.umich.edu */
#else
   sysinfo(SI_HOSTNAME, my_domain_name, sizeof(my_domain_name));
#endif

   hp = gethostbyname(my_domain_name);
   if (!hp) {
      printf("gethostbyname failed\n");
      exit(1);
   }
   bcopy(hp->h_addr, (char *) &my_addr, hp->h_length);
   tmp_name = inet_ntoa(my_addr);
   Tcl_SetVar(tcl, "rootip", tmp_name, TCL_GLOBAL_ONLY);
   Tcl_SetVar(tcl, "rootname", my_domain_name, TCL_GLOBAL_ONLY);

   /*
    * assert_node_by_addr assumes datadir has already been set, so
    * source_mviewrc() must be run before this
    */
   source_mviewrc();
   local_ipaddr = my_addr.s_addr;
   local_node = curr_node = assert_node_by_addr(local_ipaddr);
   if (curr_node<0)
      curr_node = 0;

   /*
    * Create user variables before opening display so that timers
    * start running immediately.  This must be done before adding
    * the local host so that it will go ahead and resolve autosys, etc.
    * However, must source mviewrc file (init_view) to get bindir 
    * global var.  Moved back here and added manual update of autosys.
    */
   for (i=0; variable[i].name; i++) {
      create_dd(variable[i].name, variable[i].type, variable[i].eltsize, 
       variable[i].arrsize, variable[i].flags, variable[i].linked);
      if (variable[i].updcmd || variable[i].maxage || variable[i].depends)
         set_update(variable[i].name, variable[i].maxage, variable[i].updcmd,
          variable[i].depends);
   }

   /* Okay, now we can open the display */
   init_view();
   goto_node(curr_node);

   mark_source(curr_node);
   mark_dest(curr_node);

   if (!get_int("autosys", local_node)) {
      update_value("autosys", local_node);
      update_value("snmpaware", local_node);
   }

   SIGNAL_HOOK(SIGINT, exit_view);
   return TCL_OK;
}

int
text_cmd(str1)
   char *str1;
{
   char *str;
   int i;

   str = (str1)? str1 : query(":");
   i=Tcl_Eval(tcl, str);
/*   add_msg(tcl->result); */
   return i;
}

int
exec_handler(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char *argv[];
{
   char cmd[1024], *hlr, *arg=0, arg2[255];
   char *str;
   int fd[2],s,i;

   if (argc<3)
      return TCL_ERROR;

   /* Put command back together */
   strcpy(cmd, argv[1]);
   if (argc>2) {
      strcat(cmd, " ");
      strcat(cmd, locate_command(argv[2]));
   }
   for (i=3; i<argc; i++) {
      strcat(cmd, " ");
      strcat(cmd, argv[i]);
   }

   /* Find handler and argument */
   /* sscanf won't work since scanf reads until space, not ()  */
   str = cmd;
   while (isspace(*str)) str++;
   for (hlr=str; *str && !isspace(*str) && *str!='('; str++);
   if (*str=='(') {
      *str++ = '\0';

      /* Find argument */
      while (isspace(*str)) str++;    /* skip leading spaces */
      for (arg=str; *str && *str!=')'; str++);
      *str = '\0';
      while (arg[strlen(arg)-1]==' ') /* omit trailing spaces */
         arg[strlen(arg)-1]='\0';
      strcpy(arg2, expand_vars(arg));

   } else
      arg2[0] = '\0';
   *str++ = '\0';
   while (isspace(*str)) str++;

   if (!handler_exists(hlr)) {
      add_msg("No such handler.");
      return TCL_OK;
   }

   /* add_msg(str); */

  if (pipe(fd)<0) 
     return TCL_ERROR;      /* e.g. too many open sockets */
/*  if ((fp=fdopen(fd[0],"r"))==NULL) return;*/

  s=fork();
  if (s) { /* parent */

     if (s<0) {
        wlog(LOG_ERR, 0, "Couldn't fork!");
        exit(1);
        /* return; * error: couldn't fork */
     }
     close(fd[1]);
     /*madd(fd[1],str,O_PIPE);*/
     start_handler( fd[0], hlr, arg2, s, str );

  } else { /* child */
     setuid(getuid());
     setgid(getgid());
     SIGNAL_HOOK(SIGINT, SIG_DFL);
     SIGNAL_HOOK(SIGPIPE,SIG_DFL);
     close(1); /* stdout */
     close(2); /* stderr */
     dup(fd[1]); /* write */
     dup(fd[1]); /* write */
     close(fd[0]); /* read */

     if (strpbrk(str,"<>*?|![]{}~`$&';\\\"") == NULL) {
        char **argv = explode(str," ");
        s    = xsizeof(argv);
        argv = xrealloc_array(argv, (short)(s+1));
        argv[s]=0;
        execvp(argv[0],argv);
        printf("oops: Can't execute \"%s\"!\n",str);
     } else {
        char *shpath,*sh;

        shpath = SHELL;
        sh = strrchr(shpath,'/');
        if (!sh) 
           sh = shpath;
        else 
           sh++;
        execl(shpath,sh,"-c",str,(char *)NULL);
     }
     exit(1);
  }
  return TCL_OK;
}

int
exec_alias(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char *argv[];
{
   char cmd[255];
   int i;

   sprintf(cmd, "mv_start popup %s", argv[1]);
   for (i=2; i<argc; i++) {
      strcat(cmd, " ");
      strcat(cmd, argv[i]);
   }
   text_cmd(cmd);
   return TCL_OK;
}

int
hotkey(c)
   int c;
{
   int i;

   for (i=0; alias[i].key; i++) {
      if (c==alias[i].key) {
         text_cmd( alias[i].fmt );
         return 1;
      }
   }
   return 0;
}

char *
expand_vars(str)
   char *str;
{
static char cmd[512];
   char *p, *q, var[80], buff2[255], *v;

   strcpy(cmd, str);
   while ((p = strchr(cmd, '$'))) {
      for (v=var,q=p+1; *q && *q!=' '; )
         *v++ = *q++; /* find end of word */
      *v = *p = '\0';
      sprintf(buff2, "%s%s%s", cmd, expand_var(var), q);
      strcpy(cmd, buff2);
   }

   return cmd;
}

char *
expand_var(str)
   char *str;
{
/*
   if (match(str, "name"))
      return fullname(curr_node);
   if (match(str, "ip"))
      return ipstr(curr_node,0);
   if (match(str, "srcname"))
      return fullname(source_node);
   if (match(str, "srcip"))
      return ipstr(source_node,0);
   if (match(str, "destname"))
      return fullname(dest_node);
   if (match(str, "destip"))
      return ipstr(dest_node,0);
*/
   if (match(str, "fromname")) {
         return fullname(curr_node);
   }
   if (match(str, "fromip")) {
         return ipstr(curr_node, 0);
   }
   return str;
}
