/* $Id: expr.c,v 1.4 1996/05/24 20:29:29 thalerd Exp $
 * -------------------------------------------------------------------------
 *  
 * [C] The Regents of the University of Michigan and Merit Network, Inc. 1996
 * All Rights Reserved 
 *   
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice and this permission notice appear in all 
 * copies of the software and derivative works or modified versions thereof, 
 * and that both the copyright notice and this permission and disclaimer notice 
 * appear in supporting documentation. 
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE REGENTS OF THE 
 * UNIVERSITY OF MICHIGAN AND MERIT NETWORK, INC. DO NOT WARRANT THAT THE 
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR 
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the 
 * University of Michigan and Merit Network, Inc. shall not be liable for any 
 * special, indirect, incidental or consequential damages with respect to any 
 * claim by Licensee or any third party arising from use of the software. 
 *
 * -------------------------------------------------------------------------
 */
/* General expression parser */
/* Expr syntax: variable op value */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <tcl.h>
#include <math.h> 
#include "mview.h"
#include "struct.h"
#include "expr.h"
#include "dict.h"

static char *
ent2str(ent)
   entity_t *ent;
{
static char buff[32];

   switch(ent->type) {
   case VT_STRING: return ent->v.s;
   case VT_UINT  : sprintf(buff, "%u", ent->v.i); return buff;
   case VT_INT   : sprintf(buff, "%d", ent->v.i); return buff;
   case VT_FLOAT : sprintf(buff, "%f", ent->v.f); return buff;
   }
   return NULL;
}

static entity_t * /* RETURNS: entity, or NULL on error */
get_value(name, idx, type)
   char *name;
   int   idx;
   int   type;
{
   static entity_t ent;

   switch(ent.type = type) {
      case VT_UINT  :
      case VT_INT   : ent.v.i = get_int(name,idx); break;
      case VT_FLOAT : ent.v.f = get_float(name,idx); break;
      case VT_STRING: strcpy(ent.v.s, get_string(name,idx)); break;
      default       : return NULL;
   }
   return &ent;
}

/*
 * Convert an operator string to integer format, which some OR'ed combination
 * of the following flags:
 */
#define OP_GT 0x001
#define OP_EQ 0x010 
#define OP_LT 0x100
int            /* RETURNS: Integer-format operator */
opstr2int(str) /* ARGUMENTS: */
   char *str;  /*    Operator string, e.g. "<=", "==", etc */
{
   char *p;
   int not=0;
   int ret=0;

   for (p=str; *p; p++) {
      switch(*p) {
      case 'T': return 0x111;
      case '=': ret|=OP_EQ; break;
      case '>': ret|=OP_GT; break;
      case '<': ret|=OP_LT; break;
      case '!': not=1-not;  break;
      default : return 0x000;  /* including 'F' */
      }
   }

   if (not)
      ret = 0x111 - ret;
   return ret;
}

static char *
opint2str(op)
   int op;
{
   switch(op) {
      case 0x000: return "F"; /* changed, was true */
      case 0x001: return ">";
      case 0x010: return "=";
      case 0x011: return ">=";
      case 0x100: return "<";
      case 0x101: return "!=";
      case 0x110: return "<=";
      case 0x111: return "T"; /* changed, was false */
      default   : return "?";
   }
}

char *
cond_text(c)
   cond_t *c;
{
   static char buff[256];
   if (c->op!=0x000 && c->op!=0x111)
      sprintf(buff, "%s %s %s", c->name, opint2str(c->op), ent2str(&c->value));
   else
      strcpy(buff, opint2str(c->op));
   return buff;
}

int
testcond(name, idx, op, value)
   char     *name;
   int       idx;
   int       op;
   entity_t *value;
{
   entity_t  varent;

   if (op==0x000 || op==0x111) /* always true or always false */
      return op;

   memcpy(&varent, get_value(name, idx, value->type), sizeof(varent));

   switch(value->type) {
      case VT_UINT:
      case VT_INT: {
         int left = varent.v.i;
         int right = value->v.i;

#if 0
         if (op==0x000) /* single operand */
            return left; 
         if (op==0x111) /* single operand */
            return !left;
#endif
      
         /* Two operands */
         return ((left> right && (op & OP_GT))
              || (left==right && (op & OP_EQ))
              || (left< right && (op & OP_LT)));
      }
 
      case VT_FLOAT: {
         double left = varent.v.f;
         double right = value->v.f;

#if 0
         if (op==0x000) /* single operand */
            return left; 
         if (op==0x111) /* single operand */
            return !left;
#endif
      
         /* Two operands */
         return ((left> right && (op & OP_GT))
              || (left==right && (op & OP_EQ))
              || (left< right && (op & OP_LT)));
      }
  
      case VT_STRING: {
         char *left = varent.v.s;
         char *right = value->v.s;

#if 0
         if (op==0x000) /* single operand */
            return (left != NULL);
         if (op==0x111) /* single operand */
            return (left == NULL);
#endif

         /* Two operands */
         return ((strcmp(left, right)> 0 && (op & OP_GT))
              || (strcmp(left, right)==0 && (op & OP_EQ))
              || (strcmp(left, right)< 0 && (op & OP_LT)));
      }

      default:
         return FALSE;
   }
}

/* Fill in a condition structure by parsing a string */
int               /* returns 1 on success, 0 on failure */
get_cond(cond, q)
   cond_t *cond;  /* condition structure to fill in */
   char   *q;     /* string to parse */
{
   char *p;
   char tmp[80];

   if (!q || !q[0])
      return 0;

   /* Skip whitespace */
   while (isspace(*q)) q++;

   /* Get variable */
   p=cond->name;
   while (*q && (isalnum(*q) || *q=='_'))
      *p++ = *q++;
   *p = '\0';

   /* Skip whitespace */
   while (isspace(*q)) q++;

   /* Get op */
   p = tmp;
   while (strchr("<>=!TF", *q))
      *p++ = *q++;
   *p = '\0';
   cond->op = opstr2int(tmp);

   /* Skip whitespace */
   while (isspace(*q)) q++;

   /* Get value */
   if (*q=='"') {
      p = cond->value.v.s;
      q++;
      while (*q && *q!='"')
         *p++ = *q++;
      *p = '\0';
      cond->value.type = VT_STRING;
   } else if (isdigit(*q) || *q=='-') {
      double f;
      int    i;

      f = atof(q);
      i = (int)floor(f);
      if (i*1.0 == f) {
         cond->value.type = VT_INT;
         cond->value.v.i  = i;
      } else {
         cond->value.type = VT_FLOAT;
         cond->value.v.f  = f;
      }
   } else {
      p = cond->value.v.s;
      while (*q)
         *p++ = *q++;
      *p = '\0';

      /* remove trailing whitespace */
      p--;
      while (p>=cond->value.v.s && *p==' ')
         *p-- = '\0';

      cond->value.type = VT_STRING;
   }

   cond->dirty = 1;
   return 1;
}
