/* $Id: graph.c,v 1.25 1996/07/02 22:40:33 thalerd Exp $
 * -------------------------------------------------------------------------
 *  
 * [C] The Regents of the University of Michigan and Merit Network, Inc. 1996
 * All Rights Reserved 
 *   
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice and this permission notice appear in all 
 * copies of the software and derivative works or modified versions thereof, 
 * and that both the copyright notice and this permission and disclaimer notice 
 * appear in supporting documentation. 
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE REGENTS OF THE 
 * UNIVERSITY OF MICHIGAN AND MERIT NETWORK, INC. DO NOT WARRANT THAT THE 
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR 
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the 
 * University of Michigan and Merit Network, Inc. shall not be liable for any 
 * special, indirect, incidental or consequential damages with respect to any 
 * claim by Licensee or any third party arising from use of the software. 
 *
 * -------------------------------------------------------------------------
 */
/* GRAPH.C - Functions that operate on the entire graph */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <tcl.h>
#include <tk.h>
#include <time.h>
#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h> /* for struct in_addr for arpa/inet.h */
#endif
#include <arpa/inet.h> /* for inet_addr */
#include "mview.h"
#include "struct.h"
#include "database.h"
#include "node.h"
#include "globals.h"
#include "xalloc.h"
#include "lib.h"
#include "display.h"
#include "dict.h"
#include "queue.h" /* for delete_queue */
#include "expr.h" /* for cond_text */
#include "link.h" /* for add_link_to_list */

int
fix_links() 
{
   int i,ln;
   node_t *node;
   log_link_t *edge;
   
   printf("Fixing links...\n");

   /* Initialize all Inlinks */
   for (i=0; i<num_nodes(); i++) {
      node = mget_node(i);
      node->last_link[LIN] = -1;
   }

   /* Step through all links */
   for (ln=0; ln<num_log_links(); ln++) {
      edge = get_log_link(ln);
      if (edge->dest < 0) continue;
      add_link_to_list(ln, edge->dest, LIN);
   }

   printf("Done\n");
   return TCL_OK;
}

int
show_param()
{
   char buff[256];
   Tcl_Eval(tcl, "newwin .wtmp");

   sprintf(buff, "maxage    =%7d   curr_cloud=%7d\n", maxage, curr_cloud); 
   wtmp_display(buff);
   sprintf(buff, "zoom      =%7f   iotype    =%7d\n", zoom, iotype); 
   wtmp_display(buff);
   sprintf(buff, "xscale    =%7f   yscale    =%7f\n", xscale, yscale); 
   wtmp_display(buff);
   sprintf(buff, "nsvar1    =%7s   nsvar2    =%s\n", nsvar.name1, nsvar.name2); 
   wtmp_display(buff);
   sprintf(buff, "nimode    =%7d\n", nimode);
   wtmp_display(buff);
   sprintf(buff, "lsvar1    =%7s   lsvar2    =%s\n", lsvar.name1, lsvar.name2); 
   wtmp_display(buff);
   sprintf(buff, "ndmode    =%7d\n", ndmode); 
   wtmp_display(buff);
   sprintf(buff, "nicond = %s\n", cond_text(&nicond)); 
   wtmp_display(buff);
   sprintf(buff, "ndcond = %s\n", cond_text(&ndcond)); 
   wtmp_display(buff);
   sprintf(buff, "nhcond = %s\n", cond_text(&nhcond));
   wtmp_display(buff);
   sprintf(buff, "ldcond = %s\n", cond_text(&ldcond));
   wtmp_display(buff);
   sprintf(buff, "lhcond = %s\n", cond_text(&lhcond));
   wtmp_display(buff);
   sprintf(buff, "nodeRed = %s\n", cond_text(&nodeRed));
   wtmp_display(buff);
   sprintf(buff, "nodeYellow = %s\n", cond_text(&nodeYellow));
   wtmp_display(buff);
   sprintf(buff, "nodeGreen = %s\n", cond_text(&nodeGreen));
   wtmp_display(buff);
#if 0
   sprintf(buff, "maxx      =%7.2f   maxy      =%7.2f\n", maxx, maxy); 
   wtmp_display(buff);
   sprintf(buff, "minx      =%7.2f   miny      =%7.2f\n", minx, miny); 
   wtmp_display(buff);
   sprintf(buff, "xinc      =%7.2f   yinc      =%7.2f\n", xinc, yinc); 
   wtmp_display(buff);
#endif

   return TCL_OK;
}

/* Get a line of input into (argc,argv) format */
int
parse_line(fp, argc, argv) 
   FILE   *fp;
   int    *argc;
   char ***argv;
{
   char    line[256];   /* buffer holding one line of input */
   char   *tmp;         /* pointer into line buffer */

   if (*argv) {
      xfree_array(*argv);
      *argv = 0;
   }

   fgets(line, sizeof(line), fp);
   if (feof(fp))
      return 0;

   for (tmp=line; *tmp==' '; tmp++); /* skip leading spaces */
   (*argv) = explode(line, " "); /* split into fields */
   (*argc) = xsizeof(*argv);

   return 1;
}

int
mwatch_file(filename)
   char *filename;
{
   int      nodes_read=0;
   FILE    *fp;          /* the input file pointer */
   int      argc;        /* # fields in line buffer */
   char   **argv = NULL; /* (xalloc'ed) array of fields in line buffer */
   u_int32  srcaddr;     /* Address of current node */
   int      src = -1;    /* Node ID of current node */
   u_int32  dstaddr;     /* Address of remote endpoint */
   int      dst;         /* Node ID of remote endpoint */
   node_t  *node = 0;    /* Database entry for current node */
   int      vmaj,vmin;   /* Major/minor version number of node */
   char   **stat;        /* Status info for a link */
   int      f=0;
   u_int32  metric, thresh, status;
   int      done=0;
   char     buff[10];

   if ((fp=fopen(filename, "r"))==NULL)
      return 0;

   done = !parse_line(fp, &argc, &argv);
   while (!done) {
   
      if (argc<2) {
         printf("Error, argc=%d argv0=%s\n", argc, argv[0]);
         break;
      }

      /* Parse current node information: "IP (name) [version #.#]:" */
      srcaddr = inet_addr(argv[0]);
      src     = assert_node_by_addr(srcaddr); 
      remove_outlinks(src);
      node    = mget_node(src);
      if (argc>2)
         argv[1][ strlen(argv[1])-1 ]='\0';
      else
         argv[1][ strlen(argv[1])-3 ]='\0';
      set_string("fullname", src, argv[1]+1);
      set_string("hostname", src, hostname(src));
      set_string("domain", src, org(src));
      set_string("country", src, country(src));
      delete_queue(src);
      if (argc>3)
         sscanf(argv[3], "%d.%d", &vmaj, &vmin);
      else
         vmaj = vmin = 0;
      time(&(node->timestamp)); /* should use file timestamp */
      nodes_read++;
      if (!(nodes_read % 10)) {
         printf("."); fflush(stdout);
      }
      sprintf(buff, "%d.%d", vmaj, vmin);
      set_string("version", src, buff);

      /* Parse link information */
      if (!parse_line(fp, &argc, &argv))
         break;
      while (argc>4) {

         /* Get local endpoint address */
         dstaddr = inet_addr(argv[0]);
         if (!dstaddr)
            dstaddr = srcaddr; /* local convention */
         src = assert_node_addr(src, dstaddr);
         srcaddr = dstaddr;

         /* Get remote endpoint address */
         dstaddr = inet_addr(argv[2]);
         if (!dstaddr)
            dstaddr = srcaddr; /* local convention */
         dst     = assert_node_by_addr(dstaddr);

         /* Get link status info */
         argv[4][ strlen(argv[4])-2 ]='\0'; /* trash "]\n" */
         stat = explode( argv[4]+1, "/" );
         metric = atoi(stat[0]);
         thresh = atoi(stat[1]);
         for (status=0,f=2; f<xsizeof(stat); f++) {
            if (!strcmp(stat[f], "tunnel"))
               status |= DVMRP_NF_TUNNEL;
            else if (!strcmp(stat[f], "down"))
               status |= DVMRP_NF_DOWN;
            else if (!strcmp(stat[f], "disabled"))
               status |= DVMRP_NF_DISABLED;
            else if (!strcmp(stat[f], "querier"))
               status |= DVMRP_NF_QUERIER;
            else if (!strcmp(stat[f], "srcrt"))
               status |= DVMRP_NF_SRCRT;
            else if (!strcmp(stat[f], "leaf"))
               status |= DVMRP_NF_LEAF;
            else if (!strcmp(stat[f], "pim"))
               status |= DVMRP_NF_PIM;
         }

         /*
          * Here's a fix for some more cisco strangeness:
          * Sometimes there are lines in mrinfo to odd-addressed
          * destinations, which aren't marked as tunnels.  Since
          * the low host can only be set for a host, we go ahead
          * and say they're tunnels.
          */
         if (dstaddr & 1)
            status |= DVMRP_NF_TUNNEL;
            
         xfree_array(stat);
         update_link(srcaddr, dstaddr, 0, metric, thresh, status);
     

         if (!parse_line(fp, &argc, &argv)) {
            done = 1;
            break;
         }
      }
   }
   
   fclose(fp);
   xfree_array(argv);

   refresh_display();
   return nodes_read;
}
