/* $Id: handler.c,v 1.70 1996/07/02 22:39:49 thalerd Exp $
 * -------------------------------------------------------------------------
 *  
 * [C] The Regents of the University of Michigan and Merit Network, Inc. 1996
 * All Rights Reserved 
 *   
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice and this permission notice appear in all 
 * copies of the software and derivative works or modified versions thereof, 
 * and that both the copyright notice and this permission and disclaimer notice 
 * appear in supporting documentation. 
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE REGENTS OF THE 
 * UNIVERSITY OF MICHIGAN AND MERIT NETWORK, INC. DO NOT WARRANT THAT THE 
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR 
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the 
 * University of Michigan and Merit Network, Inc. shall not be liable for any 
 * special, indirect, incidental or consequential damages with respect to any 
 * claim by Licensee or any third party arising from use of the software. 
 *
 * -------------------------------------------------------------------------
 */
/* HANDLER.C - functions that provide integration with other utilities */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h> /* for atoi */
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include <ctype.h> /* for isdigit */
#include <tcl.h>
#include <tk.h>
#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif
#include <sys/wait.h>  /* for waitpid */
#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h> /* for struct in_addr for arpa/inet.h */
#endif
#include <arpa/inet.h> /* for inet_addr */
#include "mview.h"
#include "struct.h"
#include "globals.h"
#include "handler.h"
#include "display.h"
#include "database.h"
#include "node.h"
#include "lib.h"
#include "xalloc.h"
#include "queue.h"
#include "link.h" /* for count_links */
#include "maplib.h" /* for inet_fmt */
#include "dict.h" /* for get_string */
#include "builtin.h" /* for set_ndcond */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/stat.h> /* for stat() */

static char **pathdirs = NULL;

#undef LOG_COLLECTION

extern FILE *log_fp;

#define MAXHANDLERS 64
typedef struct {
   void (*func)();
   int    pid;
   char  *cmd;  /* original command */
   char  *arg;  /* argument         */
   char  *buff;
#ifdef LOG_COLLECTION
   FILE  *log;
#endif
} handler_t;
static handler_t handler[MAXHANDLERS] = {{ NULL }};

extern handler_name_t handler_name[];

int
handler_exists(str)
   char *str;
{
   int i;

   /* Locate name in table */
   for (i=0; handler_name[i].name && strcmp(handler_name[i].name, str); i++);
 
   if (handler_name[i].name)
      return i+1;
   return 0;
}

void
start_handler(sd, str, arg, pid, cmd)
   int   sd;
   char *str;
   char *arg;
   int   pid;
   char *cmd;
{
   int i;
   char buff[256];

   if (debug & DB_HANDLERS)
      wlog(LOG_INFO, 0, "Assigning %s to socket %d", str, sd);

   /* Locate name in table */
   for (i=0; handler_name[i].name && strcmp(handler_name[i].name, str); i++);

   /* Store handler */
   handler[sd].func = handler_name[i].func;
   handler[sd].pid  = pid;
   handler[sd].cmd = xstrdup(cmd);
   if (arg)
      handler[sd].arg = xstrdup(arg);
   else
      handler[sd].arg = NULL;
   handler[sd].buff = NULL;
   
#if (TK_MINOR_VERSION == 0)
   Tk_CreateFileHandler(sd, TK_READABLE, (Tk_FileProc *)collect, (ClientData)sd);
#else
{
   Tcl_File handle = Tcl_GetFile((ClientData)sd, TCL_UNIX_FD);
   Tcl_CreateFileHandler(handle, TCL_READABLE, (Tcl_FileProc *)collect, (ClientData)sd);
}
#endif

#ifdef LOG_COLLECTION
   sprintf(fn, "hlog%d", sd);
   handler[sd].log = fopen(fn,"w");
   if(log_fp)
      fprintf(log_fp, "STARTING <%s> <%s> to socket %d (hlog%d)\n", 
       handler[sd].cmd, handler[sd].arg, sd, sd);
#endif

   if (debug & DB_HANDLERS)
      fprintf(stderr, "STARTING <%s> <%s> to socket %d\n\r", 
       handler[sd].cmd, handler[sd].arg, sd);

   /* Add it to the "Activities-in-progress" window */
   sprintf(buff, ".r.hlr.list insert end {%s}", cmd);
   Tcl_Eval(tcl, buff);
}

void
stop_handler(sd)
   int sd;
{
   char buff[256];

   if (debug & DB_HANDLERS) {
      wlog(LOG_INFO, 0, "Removing handler on socket %d", sd);
      fprintf(stderr, "Removing handler on socket %d: %s %s\n\r", sd, 
       handler[sd].cmd, handler[sd].arg);
      fprintf(log_fp, "Removing handler on socket %d: %s %s\n", sd, 
       handler[sd].cmd, handler[sd].arg);
   }

   /* Remove it from the "Activities-in-progress" window */
   sprintf(buff, "rm_aip {%s}", handler[sd].cmd);
   Tcl_Eval(tcl, buff);

   xfree_string(handler[sd].cmd);
   if (handler[sd].arg)
      xfree_string(handler[sd].arg);
   handler[sd].func = NULL;
   if (handler[sd].buff) {
      xfree_string(handler[sd].buff);
      handler[sd].buff = NULL;
   }
   close(sd);
#ifdef LOG_COLLECTION
   fclose(handler[sd].log);
#endif

   /*waitpid(handler[sd].pid, 0, 0);*/

#if (TK_MINOR_VERSION == 0)
   Tk_DeleteFileHandler(sd);
#else
{
   Tcl_File handle = Tcl_GetFile((ClientData)sd, TCL_UNIX_FD);
   Tcl_DeleteFileHandler(handle);
   Tcl_FreeFile(handle);
}
#endif
}

void
set_handlers(numfds, fdset, tval, block)
   int            *numfds;
   fd_set         *fdset;
   void           *tval; /* struct timeval */
   int            *block;
{
   int h;

   for (h=0; h<MAXHANDLERS; h++) {
      if (handler[h].func) {
         FD_SET(h, fdset);
         if (*numfds < h+1)
            *numfds = h+1;
      }
   }
}

int
check_handlers( fdset )
   fd_set         *fdset;
{
   int h;

   for (h=0; h<MAXHANDLERS; h++)
      if (FD_ISSET( h, fdset)) {
         collect(h,0);
         return 1;
      }
   return 0;
}

int
show_handlers()
{
   int   h,i;
   char  buff[255];

   do_tcl("open_reportwin sh {sd handler    PID   command}");
   for (h=0; h<MAXHANDLERS; h++) {
      if (!handler[h].func) continue;
      
      for (i=0; 
           handler_name[i].name && handler_name[i].func!=handler[h].func; 
           i++);
      sprintf(buff, ".sh.text insert end \"%2d %-10s %5d %s\n\"", h, 
       handler_name[i].name, handler[h].pid, handler[h].cmd);
      do_tcl(buff);
   }
   do_tcl("close_reportwin sh");
   return TCL_OK;
}

void
collect(h, mask)
   int h;
   int mask;
{
   int len,plen, status;
   char *b2;
   char  buff[1024];

   len = read(h, buff, sizeof(buff)-1);
   if (len<1) { /* if done reading */
      int pid;

#if (TK_MINOR_VERSION == 0)
      Tk_CreateFileHandler(h, 0, (Tk_FileProc *)collect, (ClientData)h);
#else
{
      Tcl_File handle = Tcl_GetFile((ClientData)h, TCL_UNIX_FD);
      Tcl_CreateFileHandler(handle, 0, (Tcl_FileProc *)collect, (ClientData)h);
}
#endif
      if ((pid=waitpid(handler[h].pid, &status, 0)) != handler[h].pid)
         abort();
      status >>= 8; /* exit status is in 2nd byte */

      /* Make sure buff is at least the empty string */
      if (!handler[h].buff) {
         handler[h].buff = (char *)xalloc(0,1);
         handler[h].buff[0] = '\0';
      }

      handler[h].func(h, handler[h].arg, handler[h].buff, status);
      stop_handler(h);
#if 0
/* removed 10/12/95 because if handler changes node display condition,
 * curr_node may need to change.  This call would reset the condition to
 * in component, which is wrong behavior.
 */
      goto_node(curr_node); /* XXX do we really need this or should
                               the handler do it??? */
#endif
      refresh_display();
   } else {
#ifdef LOG_COLLECTION
if (handler[h].log) {
fwrite(buff, len, 1, handler[h].log);
fflush(handler[h].log);
}
#endif
      if (!handler[h].buff)
         plen = 1;
      else
         plen = xeltlen((char **)handler[h].buff);
      b2 = (char *)xalloc(0,plen+len);
      if (plen>1) {
         memcpy(b2, handler[h].buff, plen-1);
         xfree_string(handler[h].buff);
      }
      memcpy(b2+plen-1, buff, len);
      b2[len+plen-1] = '\0';
      if (debug & DB_HANDLERS) {
         fprintf(stderr,"COLLECT %d: len %d->%d\n\r", h, plen-1, len+plen-1);
         fprintf(log_fp,"COLLECT %d: len %d->%d\n", h, plen-1, len+plen-1);
      }
      handler[h].buff = b2;
   }
}

/*----------------------------------------------------------------------------*/

void
log_handler( sd, arg, buff, status )
   int   sd;
   char *arg, *buff;
   int   status;
{
   add_msg(buff);
}

void
nslookup(sd, arg, buff, status)
   int   sd;
   char *arg;  /* original ip address */
   char *buff;
   int   status;
{
   int n,l, n2;
   char name[80], addr[80];
   char str[255], **line;
   node_t *node;

   /* Parse output to get address and hostname */
   name[0]=addr[0]=0;
   line = explode(buff, "\n");
   for (l=0; l<xsizeof(line); l++) {
      if (!strncmp(line[l], "Name:    ", 9))
         strcpy(name, line[l]+9);
      else if (name[0] && !strncmp(line[l], "Address:  ", 10))
         strcpy(addr, line[l]+10);
   }
   xfree_array(line);

   /* If failed, set hostname equal to IP address string */
   if (!addr[0]) {
      strcpy(addr, arg);
      strcpy(name, arg);
      /* sprintf(name, "[%s]", arg); no [] in tcl */
   }

   /* Save hostname found */
   if (addr[0]) {
      n = assert_node_by_addr(inet_addr(addr)); /* inc. add a new router */
      if (n>=0) {
         node = mget_node(n);
         if (!(node->status & ST_SUBNET)) {
            set_string("fullname", n, name);
            set_string("hostname", n, hostname(n));
            set_string("domain", n, org(n));
            set_string("country", n, country(n));

            /* Update domain for all connected subnets */
            for (l=node->last_link[LOUT]; l>=0; l=llink[l].prev_link[LOUT]) {
               node = mget_node(llink[l].dest);
               if (node->status & ST_SUBNET) {
                  set_string("domain", llink[l].dest, org(n));
                  set_string("country", llink[l].dest, country(n));
               }
            }
         }
      }
      sprintf(str, "%d %s=%s", n, name, addr);
      add_msg(str);
   } 

   /* Update rootname, srcname, and destname if necessary */
   if (!strcmp(addr, Tcl_GetVar(tcl, "srcip", TCL_GLOBAL_ONLY)))
      Tcl_SetVar(tcl, "srcname", name, TCL_GLOBAL_ONLY);
   if (!strcmp(addr, Tcl_GetVar(tcl, "destip", TCL_GLOBAL_ONLY)))
      Tcl_SetVar(tcl, "destname", name, TCL_GLOBAL_ONLY);
   if (!strcmp(addr, Tcl_GetVar(tcl, "rootip", TCL_GLOBAL_ONLY)))
      Tcl_SetVar(tcl, "rootname", name, TCL_GLOBAL_ONLY);

   /* Invoke next nslookup in queue */
   n2 = pop_queue(Q_NSLOOKUP, 0);
/*
   if (n!=n2)
      abort();
*/
}

void
ping_handler(sd,arg, buff, status)
   int   sd;
   char *arg, *buff;
   int   status;
{
   int n;
   char  *s;
   char b2[255];
   node_t *node;

   if (status) /* command failed to run correctly */
      return;

   if (!strncmp(buff, "no answer from ", 15)) {
      n = assert_node_by_addr(inet_addr(buff+15)); /* inc. add a new router */
      if (n>=0) {
         node = mget_node(n);
         node->status &= ~ST_RESPONDING;
         sprintf(b2, "%s is not responding", get_string("fullname", n));
         add_msg(b2);
      }
   } else {
      s = strchr(buff, ' ');
      if (!s) return;
      *s++ = '\0';
      n = assert_node_by_addr(inet_addr(buff)); /* inc. add a new router */
      node = mget_node(n);
      if (!strncmp(s, "is alive", 8)) {
         node->status |= ST_RESPONDING;
         sprintf(b2, "%s is responding", get_string("fullname", n));
         add_msg(b2);
      }
   }
}

void
popup(sd, arg, buff, status)
   int   sd;
   char *arg, *buff;
   int   status;
{
   char str[255], *strp;
   char **line;
   int i;
      
	/* Clobber any \r's. Finger uses them in combo with \n's.  Not sure if
	 * this will goof up any other things using popup(). Don't _think_ so.... 
	 * -ala */
	for( i = 0; buff[i]; i++ )
		if ( buff[i] == '\r' )
			buff[i] = ' ';

   line = explode(buff, "\n");
   sprintf(str, "open_reportwin popup%d {%s}", sd, handler[sd].cmd);
   do_tcl(str);
   for (i=0; i<xsizeof(line); i++) {
      strp = (char*)xalloc(0, strlen(line[i])+80);
      sprintf(strp, ".popup%d.text insert end {%s\n}", sd, line[i]);
      do_tcl(strp);
      xfree_string(strp);
   }
   xfree_array(line);
   sprintf(str, "close_reportwin popup%d", sd);
   do_tcl(str);
}

void
traceroute_hlr(sd, arg, buff, status)
   int   sd;
   char *arg, 
        *buff;
   int   status;
{
   int n = -1, p = -1, hitsrc = 0, l, i;
   int srcn,  /* logical source */
       destn; /* logical dest   */
   char out[255], out2[255], name[256];
   char **fieldv, **argv;
   int    fieldc,   argc;
   char **line, *tmp;
   log_link_t *edge;

   if (status) {
      popup(sd, arg, buff, status);
      /* pop_queue(Q_MRINFO, 0); removed 4/4/96, why was this here? */
      return;	
   }

   do_tcl("open_reportwin traceroute {hop address}");
   line = explode(buff, "\n");

   /* Argument contains bounce point node ip address, if any */
   if (!arg || !arg[0])
      hitsrc = 1;

   /* Get source and destination node numbers */
   destn = assert_node_by_addr(inet_addr(arg));
   argv = explode(line[0], " "); /* split into fields */
   argc = xsizeof(argv);
   if (argc<3)
      return;
   srcn  = assert_node_by_addr(inet_addr(argv[2])); 
   xfree_array(argv);

   /* Process lines in output */
   for (l=1; l<xsizeof(line); l++,p=n) {
      for (tmp=line[l]; *tmp==' '; tmp++); /* skip leading spaces */
      fieldv = explode(tmp, " "); /* split into fields */
      fieldc = xsizeof(fieldv);

      if (!hitsrc && !strcmp(fieldv[1], arg)) 
         hitsrc = l;

#if 0
      n = assert_node_by_addr(inet_addr(fieldv[1])); /* inc. add a new router */
      if (p>=0) 
         add_phy_link(p,n,srcn,destn);
#else
      n = find_node_by_addr(inet_addr(fieldv[1]));
      if (n>=0)
         strcpy(name, fullname(n));
      else
         strcpy(name, fieldv[1]);
#endif

      if (hitsrc) {
         sprintf(out, "%2d %s (%s)", l-hitsrc, name, fieldv[1]);
         for (i=2; i<fieldc; i++) {
            strcat(out, " ");
            strcat(out, fieldv[i]);
         }
         sprintf(out2, ".traceroute.text insert end \"%s\n\"", out);
         do_tcl(out2);

      }
      xfree_array(fieldv);
   }
   xfree_array(line);

   /* Update hop count */
   edge = find_ifc(srcn, destn);
   if (edge)
      edge->hops = l-hitsrc;

   /* Close wtmp */
   do_tcl("close_reportwin traceroute");
}

/*
198.108.60.69 (198.108.60.69) [version 3.6]:
  198.108.60.69 -> 198.108.60.51 (198.108.60.51) [1/1]
  198.108.60.69 -> 198.108.60.97 (198.108.60.97) [1/1]
  198.108.60.69 -> 141.213.10.41 (141.213.10.41) [1/1/tunnel]
  198.108.60.69 -> 141.213.10.45 (141.213.10.45) [1/1/tunnel]
*/

/* Note that if multiple mapper queries are executing simultaneously,
   we may very well get another result in with this one, although we 
   will always get all of our own 
 */
void
mrinfo_handler(sd, arg, buff, status)
   int   sd;
   char *arg, *buff;
   int   status;
{
   int     nlines;
   char  **line;
   int     argc;
   char  **argv = NULL; 
   char   *p;
   u_int32 srcaddr, dstaddr;
   int     src, dst;
   node_t *node;
   int     vmaj, vmin;
   char  **stat;
   int     l, f;
   u_int32  metric, thresh, flags;
   char    tmp[10], cmd[80];
   int   dvmrp_subnets=0, pim_subnets=0;

   if (status || !buff || !buff[0]) {
      /* Node didn't respond to mrinfo query */
      if (arg) {
         src = atoi(arg);
         node = mget_node(src);

         /* test incoming links for up state */
         for (l=node->last_link[LIN]; l>=0; l=llink[l].prev_link[LIN]) {
            if (get_int("linkstate", l)==3) {
               set_int("nodestate", src, 2); /* probably busy */
               break;
            }
         }
         if (l<0)
            set_int("nodestate", src, 1); /* not reachable */

         node = mget_node(src);
         /* mark all outgoing links as unreachable */
         for (l=node->last_link[LOUT]; l>=0; l=llink[l].prev_link[LOUT])
            set_int("linkstate", l, 0);
      }
      pop_queue(Q_MRINFO, 0);
      return;
   }
   line   = explode(buff, "\n");
   nlines = xsizeof(line);

   /* Skip 'mrinfo: got reply from...' lines */
   for (l=0; l<nlines && !strncmp(line[l], "mrinfo", 6); l++);

   for (p=line[l]; *p==' '; p++); /* skip leading spaces */
   argv = explode(line[l], " "); /* split into fields */
   argc = xsizeof(argv);

   if (argc<2) {
      printf("Error, argc=%d argv0=%s\n", argc, argv[0]);
      xfree_array(argv);
      xfree_array(line);
      pop_queue(Q_MRINFO, 0);
      return;
   }

   /* Parse current node information: "IP (name) [version #.#]:" */
   srcaddr = inet_addr(argv[0]);
   src     = assert_node_by_addr(srcaddr);
   remove_outlinks(src);
   node    = mget_node(src);
   delete_queue(src); /* remove from all queues */
   if (argc>3)
      sscanf(argv[3], "%d.%d", &vmaj, &vmin);
   else
      vmaj = vmin = 0;
   time(&(node->timestamp)); /* should use file timestamp */
   sprintf(tmp, "%d.%d", vmaj, vmin);
   set_string("version", src, tmp);
   xfree_array(argv);

   /* Set snmpaware if "snmp" is included in version information */
   if (strstr(argv[3], "snmp") && !get_int("snmpaware", src))
      set_int("snmpaware", src, 1); 

   set_int("nodestate", src, 3); /* up */

   /* Parse link information */
   for (l++; l<nlines; l++) {
      argv = explode(line[l], " "); /* split into fields */
      argc = xsizeof(argv);
      if (argc<5) {
         printf("MRINFO ERROR on node %d line %d: %s\n", src, l, line[l]);
         xfree_array(argv);
         xfree_array(line);
         pop_queue(Q_MRINFO, 0);
         return;
      }

      /* Get local endpoint address */
      dstaddr = inet_addr(argv[0]);
      if (!dstaddr)
         dstaddr = srcaddr; /* local convention */

      src = assert_node_addr(src, dstaddr);
      srcaddr = dstaddr;

      /*
       * Get remote endpoint address.  Some "version 1.0" routers (MOSPF?) 
       * report 224.0.0.4 for subnets instead of 0.0.0.0 like everyone else.
       */
      if (!strcmp(argv[2], "224.0.0.4"))
         dstaddr = 0;
      else
         dstaddr = inet_addr(argv[2]);
      if (!dstaddr)
         dstaddr = srcaddr; /* local convention */
      dst     = assert_node_by_addr(dstaddr);

      /* Get link flags info */
      argv[4][ strlen(argv[4])-1 ]='\0'; /* trash "]" */
      stat = explode( argv[4]+1, "/" );
      xfree_array(argv);
      metric = atoi(stat[0]);
      thresh = atoi(stat[1]);
      for (flags=0,f=2; f<xsizeof(stat); f++) {

         /* Some ciscos, such as:
          *
          * 132.203.2.15 (132.203.2.15) [version 10.2]:
          * 132.203.2.15 -> 0.0.0.0 (local) [1/0/tunnel]
          *
          * report subnets as tunnels.  This is indeed a subnet as can be
          * verified by SNMP queries on ifType.1 and igmpInterfaceStatus.1
          *
          * So, if dstaddr is 0.0.0.0, then ignore the tunnel flag
          */
         if (!strcmp(stat[f], "tunnel") && dstaddr!=srcaddr)
            flags |= DVMRP_NF_TUNNEL;
         else if (!strcmp(stat[f], "down"))
            flags |= DVMRP_NF_DOWN;
         else if (!strcmp(stat[f], "disabled"))
            flags |= DVMRP_NF_DISABLED;
         else if (!strcmp(stat[f], "querier"))
            flags |= DVMRP_NF_QUERIER;
         else if (!strcmp(stat[f], "srcrt"))
            flags |= DVMRP_NF_SRCRT;
         else if (!strcmp(stat[f], "pim"))
            flags |= DVMRP_NF_PIM;
         else if (!strcmp(stat[f], "leaf"))
            flags |= DVMRP_NF_LEAF;

         /* When the IP addresses differ by 1, assume it's point-to-point */
         if (abs(srcaddr-dstaddr)==1)
            flags |= DVMRP_NF_TUNNEL;

         if (!(flags & DVMRP_NF_TUNNEL)) {
            if (flags & DVMRP_NF_PIM)
               pim_subnets++;
            else
               dvmrp_subnets++;
         }
      }
      xfree_array(stat);
      update_link(srcaddr, dstaddr, 0, metric, thresh, flags);
   }
   xfree_array(line);

   /* Update subnet info if needed and possible */
   if (get_int("snmpaware", src)) {
      if (dvmrp_subnets) {
         sprintf(cmd, "mv_start dvmrpvif mstat -ni %s", ipstr(src,0));
         do_tcl(cmd);
      }
      if (pim_subnets) {
         sprintf(cmd, "mv_start pimif(%d) mstat -nI %s", src, ipstr(src,0));
         do_tcl(cmd);
      }
   }

   pop_queue(Q_MRINFO, 0);
}

#if 0
/* Note that if multiple mapper queries are executing simultaneously,
   we may very well get another result in with this one, although we 
   will always get all of our own 
 */
void
map_handler(sd, arg, buff, status)
   int   sd;
   char *arg, *buff;
   int   status;
{
   int n = -1, src = -1, l;
   u_int32 srcaddr=0, dstaddr, metric, thresh, flags;
   char out[255];
   node_t *node = 0;
   char **argv = NULL, **stat;
   int    argc;
   char **line, *tmp;

   line = explode(buff, "\n");
fprintf(stderr, "MH invoked on %d with %d lines\n\r", sd, xsizeof(line));
fprintf(log_fp, "MH invoked on %d with %d lines\n", sd, xsizeof(line));
for (l=0; l<xsizeof(line); l++) {
fprintf(stderr, "line %d: !%s!\n\r", l, line[l]);
fprintf(log_fp, "line %d: !%s!\n", l, line[l]);
}
   for (l=0; l<xsizeof(line); l++,xfree_array(argv)) {
      int f=0;

      for (tmp=line[l]; *tmp==' '; tmp++); /* skip leading spaces */
      argv = explode(line[l], " "); /* split into fields */
      argc = xsizeof(argv);
      if (!argc) continue;
      if (!strcmp(argv[0], "warning")) continue;
fprintf(stderr, "MAP: %s\n\r", line[l]);
fprintf(log_fp, "MAP: %s\n", line[l]);

      if (argv[0][ strlen(argv[0])-1 ]==':') {
          argv[0][ strlen(argv[0])-1 ]='\0';
          srcaddr = inet_addr(argv[0]);
          src = -1;

         if (argc>3 && !strcmp(argv[1],"alias")) {
            dstaddr = inet_addr(argv[3]);
            src     = assert_node_by_addr(srcaddr); /* inc. add a new router */
            src = assert_node_addr(src, dstaddr);

fprintf(stderr, "MH %d: %lu (%s)=", sd, srcaddr, inet_fmt(srcaddr));
fprintf(stderr, "%lu (%s)=#%d\n\r", dstaddr, inet_fmt(dstaddr), src);
fprintf(log_fp, "MH %d: %lu (%s)=", sd, srcaddr, inet_fmt(srcaddr));
fprintf(log_fp, "%lu (%s)=#%d\n", dstaddr, inet_fmt(dstaddr), src);
            continue;
         }
         if (src<0)
            src     = assert_node_by_addr(srcaddr); /* inc. add a new router */
fprintf(log_fp, "MH %d: %lu=(%s)=#%d\n", sd, srcaddr, inet_fmt(srcaddr), src);
fprintf(stderr, "MH %d: %lu=(%s)=#%d\n\r", sd, srcaddr, inet_fmt(srcaddr), src);
         if (!node) {
            remove_outlinks(src);
            node    = mget_node(src);
         }

         if (argc==2 && argv[1][0]=='<') {
            char buff[10];
            int vmaj,vmin;

            sscanf(argv[1], "<v%d.%d>", &vmaj, &vmin);
            sprintf(buff, "%d.%d", vmaj, vmin);
            set_string("version", src, buff);
            continue;
         }
         f++;

      }
      if (f>=argc)
         continue;
      dstaddr = inet_addr(argv[f++]);

      if (argv[f][0]=='[') {
         argv[f][ strlen(argv[f])-1 ]='\0'; /* trash ']' */
         stat = explode( argv[f]+1, "/" );
         metric = atoi(stat[0]);
         thresh = atoi(stat[1]);
         for (flags=0,f=2; f<xsizeof(stat); f++) {
            if (!strcmp(stat[f], "tunnel"))
               flags |= DVMRP_NF_TUNNEL;
            else if (!strcmp(stat[f], "down"))
               flags |= DVMRP_NF_DOWN;
            else if (!strcmp(stat[f], "disabled"))
               flags |= DVMRP_NF_DISABLED;
            else if (!strcmp(stat[f], "querier"))
               flags |= DVMRP_NF_QUERIER;
            else if (!strcmp(stat[f], "srcrt"))
               flags |= DVMRP_NF_SRCRT;
         }
         xfree_array(stat);
         update_link(srcaddr, dstaddr, 0, metric, thresh, flags);
      }
   }
   xfree_array(argv);
   if (node) {
      time(&(node->timestamp));
      sprintf(out, "%s has %d outlinks", fullname(src), count_links(src,LOUT));
      add_msg(out);
   }
fprintf(stderr, "MH completed on %d with %d lines\n\r", sd, xsizeof(line));
   xfree_array(line);

   n = pop_queue(Q_MRINFO, 0);
}
#endif

#if 0
/*
 * buff is expected to contain a list of IP addresses, each one on its
 * own line. get_idx_path looks these addresses up in the mview database, 
 * determines the index of each address in the database, and returns an array 
 * of the indices. If an address does not yet exist in the mview database,
 * then it is added, and a new index is assigned (and subsequently returned).
 *
 * Note that the array returned is dynamically allocated
 * and should be freed by the caller.
 */
static int*
get_idx_path( line )
char **line;
{
	int *array, i, num_nodes, idx;
	u_int32 ip_num;

	num_nodes = xsizeof( line );
	array = (int*) malloc( sizeof( int ) * (num_nodes + 1) );

	for( i = 0; i < num_nodes; i++ ) {

		ip_num = inet_addr( line[ i ] );
		if ( ip_num == -1 ) {
			fprintf(stderr,"%s(%d):inet_addr() failed for string '%s'\n",
			                                      __FILE__, __LINE__, line[i] );
			abort();
		}

		idx = assert_node_by_addr( ip_num );
		if ( idx < 0 ) {
			fprintf(stderr,"%s(%d):assert_node_by_addr() returned %d\n",
			                                      __FILE__, __LINE__, idx );
			abort();
		}

		array[ i ] = idx;

	}

	array[ num_nodes ] = -1;
	return array;

}
#endif

/*
 * Read the tcl exp_index var to see what node we're talking about.
 * Parse the output in buff to see if the node speaks snmp.
 * Set the node value
 */
void
snmpaware_hlr(sd, arg, buff, status)
   int   sd;
   char *arg, *buff;
   int   status;     /* exit status of command */
{
   register int snmpaware;
   char **line;

   int  idx = atoi( arg );

   /* Don't check exit status, since it should always "fail" (status=1)
    * since we're asking for both an old OID and a new OID 
    */

   /* We know that if snmpget returned anythig beginning with 'enterprises'
    * that the mrouter in question speaks snmp.
    */
   if ( !strncmp( buff, "No Response", strlen( "No Response" ) ) )
      snmpaware = 0;
   else {
      line = explode( buff, "\n" );
      snmpaware = 1;
      if (!strcmp(line[ xsizeof(line)-1 ], ".1.3.6.1.3.60.1.1.1.0 = 1"))
	 snmpaware = 2; /* new OIDs */
      xfree_array(line);
   }
   set_int( "snmpaware", idx, snmpaware );
   pop_queue(Q_SNMPAWARE, 0);
}

/*
 * Text in buff assumed to consist of 
 *    One line of column headers (for humans to read)
 *    LocalIP RemoteIP|Subnet SubnetMask InPkts OutPkts
 */

void parse_counts( idx, buff, InPkts, OutPkts, timestamp )
	int idx;
	char *buff;
	u_int32 *InPkts, *OutPkts;
	time_t *timestamp;
{
	char **line,
	     local_ip[30],
	     remote_ip[30],
	     mask[30];
	int i,
	    link_idx,
	    num_intfs,
	    rc;
	u_int32 total_inpkts,
	             total_outpkts,
	             oldLinkInPkts,
	             oldLinkOutPkts,
	             new_inpkts,
	             new_outpkts,
	             times;
	float link_in_pkt_rate,
	      link_out_pkt_rate;
	time_t oldTime;
	

	total_inpkts  = 0;
	total_outpkts = 0;
	oldTime       = (int32) get_int( "PktTimeStamp", idx );

	line = explode( buff, "\n" );

	/* Throw away the first line. It's a column header... */
	num_intfs = xsizeof( line ) - 1;
	for(i = 1; i < num_intfs; i++ ) {
		rc = sscanf(line[i],"%s %s %s %lu %lu %ld", local_ip, remote_ip, mask,
		                                          &new_inpkts, &new_outpkts, 
		                                          &times);
		if ( rc != 6 ) {
			fprintf(stderr,"%s(%d): scanf() did not match ", __FILE__, __LINE__ );
			fprintf(stderr,"correct number of parameters.\n");
			abort();
		}

		link_idx = assert_log_link( inet_addr( local_ip ),
		                            inet_addr( remote_ip ),
		                            inet_addr( mask ) );

		oldLinkInPkts = (u_int32) get_uint( "linkInPkts", link_idx );
		oldLinkOutPkts = (u_int32) get_uint( "linkOutPkts", link_idx );

		link_in_pkt_rate =  100.0*( new_inpkts - oldLinkInPkts )/( times - oldTime );
		link_out_pkt_rate = 100.0*( new_outpkts-oldLinkOutPkts )/( times - oldTime);

		set_uint(  "linkInPkts", link_idx, new_inpkts );
		set_uint(  "linkOutPkts",  link_idx,  new_outpkts );
		set_float( "linkInPktRate", link_idx, link_in_pkt_rate );
		set_float( "linkOutPktRate",   link_idx, link_out_pkt_rate );

		total_inpkts += new_inpkts;
		total_outpkts += new_outpkts;
	
#ifdef never
		fprintf(stderr,"(line %d) new time = %li old time = %li\n", link_idx, times , oldTime);
		fprintf(stderr,"(link %d) old linkInPkts = %lu\n", link_idx,oldLinkInPkts);
		fprintf(stderr,"(link %d) old linkOutPkts = %lu\n", link_idx, oldLinkOutPkts );
		fprintf(stderr,"(link %d) new linkInPkts = %lu\n", link_idx, new_inpkts );
		fprintf(stderr,"(link %d) new linkOutPkts = %lu\n", link_idx, new_outpkts );
		fprintf(stderr,"(link %d) linkInPktRate = %f\n", link_idx,link_in_pkt_rate);
		fprintf(stderr,"(link %d) linkOutPktRate = %f\n", link_idx,link_out_pkt_rate);
#endif
	}

	xfree_array( line );
	*InPkts = total_inpkts;
	*OutPkts = total_outpkts;
	*timestamp = times;
	return;
}

void
packetrate_hlr(sd, arg, buff, status)
   int   sd;
   char *arg, *buff;
   int   status;
{

	int   idx;            /* Rec number of node who's info we're updating */
	u_int32  newInPkts,      /* New In Packet count */
	               newOutPkts,     /* New Out Packt count */
	               oldInPkts,      /* Old In Packet count - duh! */
	               oldOutPkts;     /* Old Out Packet Count - duh! duh! */
	float in_packet_rate, /* Just like it reads... */
	      out_packet_rate;/* Just like it reads.... */
	time_t oldTime,
	       newTime;

	/* We might have a non-zero return code simply because the
	 * node being queried is not snmp-aware.
	 */
	if ( status )
		return;

	idx = atoi( arg );

	parse_counts( idx, buff, &newInPkts, &newOutPkts, &newTime );

	oldOutPkts = (u_int32) get_int( "OutPkts", idx );
	oldInPkts  = (u_int32) get_int( "InPkts", idx );
	oldTime    = (int32) get_int( "PktTimeStamp", idx );
		
	in_packet_rate  = 100.0*(newInPkts - oldInPkts)/(newTime - oldTime);
	out_packet_rate = 100.0*(newOutPkts - oldOutPkts)/(newTime - oldTime);

#ifdef never
	fprintf(stderr,"newInPkts: %lu\n", newInPkts);
	fprintf(stderr,"oldInPkts: %lu\n", oldInPkts);
	fprintf(stderr,"newOutPkts: %lu\n", newOutPkts);
	fprintf(stderr,"oldOutPkts: %lu\n", oldOutPkts);
	fprintf(stderr,"oldOutPkts: %lu\n", oldOutPkts);
	fprintf(stderr,"newTime: %li\n", newTime );
	fprintf(stderr,"oldTime: %li\n", oldTime );
	fprintf(stderr,"in_packet_rate: %f\n", in_packet_rate );
	fprintf(stderr,"out_packet_rate: %f\n", out_packet_rate );
#endif
	set_uint(   "OutPkts", idx, newOutPkts );
	set_uint(   "InPkts", idx,  newInPkts );
	set_int( "PktTimeStamp", idx, newTime );
	set_float( "OutPktRate", idx, out_packet_rate );
	set_float( "InPktRate",  idx, in_packet_rate );
}

int 
extract_info( line, local_ip, ip_num, mask, level, mship_certainty )
   char *line;
   u_int32 *local_ip, *ip_num, *mask;
   int *level, *mship_certainty;
{
	char *tok, *p;
	char ip_str[20];
	int i, mask_len;

	i = strlen( line );
	if ( (line[ i - 1 ] == '?') || (line[ i - 2 ] == '?') )
		*mship_certainty = 0;
	else
		*mship_certainty = 1;

	tok = strtok( line, " ");

	if ( !isdigit( *tok ) )
		return 1;

	*level  = atoi( tok );
	tok = strtok( NULL, " " );

	p = strchr(tok, '-');
	if (p) {
		*p ='\0'; 
		*local_ip = inet_addr(tok);
		tok = p+2; /* skip the "local->" part */
	} else
		*local_ip = inet_addr(tok);

	if ( strcmp( tok, "subnet:" ) ) {
		*ip_num = inet_addr( tok );
		*mask = 0;
		return 0;
	}

	tok = strtok( NULL, "/" );
	/* strcpy and strcat the .0 to get around inet_addr() 'feature' when 
	 * string is in form a.b.c.
	 */
	strcpy( ip_str, tok );
	strcat( ip_str, ".0" );
	*ip_num = inet_addr( ip_str );
	tok = strtok( NULL, " " );
	mask_len = atoi( tok );
	*mask = 0;
	for ( i = 0; i < (32 - mask_len); i++ )
		*mask = (*mask * 2) + 1;
	*mask = ~(*mask);

	return 0;
}

typedef struct {
	u_int32 ip_num;
	u_int32 mask;
	int level;
} StackElem;

#define STACK_DEPTH 100

typedef struct {
	StackElem stack[STACK_DEPTH];
	int sptr;
} Stack;
 
inline void stack_init( stack )
Stack *stack;
{
	stack->sptr = 0;
}


void push( stack, ip_num, mask, level )
Stack *stack;
u_int32 ip_num, mask;
int level; 
{
	int s;

	s = stack->sptr;

	if ( s == STACK_DEPTH ) {
		fprintf(stderr,"%s(%d): Stack overflow.\n", __FILE__, __LINE__ );
		abort();
	}

	stack->stack[ s ].ip_num = ip_num;
	stack->stack[ s ].mask   = mask;
	stack->stack[ s ].level  = level;
	stack->sptr = s + 1;
}

void stack_dump( stack )
Stack *stack;
{
	int i, top, level;
	char ip_num[20], mask[20];
	struct in_addr in1;

	top = stack->sptr;

	for( i = top-1; i >= 0; i-- ) {

		level      = stack->stack[ i ].level;
		in1.s_addr = stack->stack[ i ].ip_num;
		strcpy( ip_num, inet_ntoa( in1 ) );
		in1.s_addr = stack->stack[ i ].mask;
		strcpy( mask, inet_ntoa( in1 ) );
		
		printf("%2d %2d %20s %s\n", i, level, ip_num, mask );
	}
}

inline void pop( stack )
Stack *stack;
{
	int s = stack->sptr;
	if ( s )
		stack->sptr = s - 1;
}

inline u_int32 peek_ip( stack )
Stack *stack;
{
	return ( stack->stack[ (stack->sptr) - 1 ].ip_num );
}

inline u_int32 peek_mask( stack )
Stack *stack;
{
	return ( stack->stack[ stack->sptr - 1 ].mask );
}

inline u_int32 peek_level( stack )
Stack *stack;
{
	return ( stack->stack[ stack->sptr - 1 ].level );
}


void 
set_mcastnode_link( stack, local_ip, remote_ip, mask, mship_certainty )
	Stack *stack;
	u_int32 local_ip, 
 	       remote_ip, 
 	       mask;
	int    mship_certainty;
{
	int rmt_node_idx, local_node_idx,
	    link_idx;

	u_int32 local_ip2;

	char s1[20];
	struct in_addr in1;


	in1.s_addr = local_ip;
	strcpy( s1, inet_ntoa( in1 ) );
	in1.s_addr = remote_ip;
/*
	fprintf(stderr,"**** Link between %s and %s\n", s1, inet_ntoa( in1 ) );
*/

	if ( mask ) {
		/* Here we assert the local IP address rather than the
                 * remote IP address, since both should be on the same
                 * subnet anyway, and this way we can avoid duplicate
                 * subnets if one with a longer mask already exists
                 */
		rmt_node_idx = assert_subnet( local_ip );

	} else
		rmt_node_idx = assert_node_by_addr( remote_ip );

	/* Make database aware that both local IP addresses are the same node */
	local_ip2 = peek_ip( stack );
	if (local_ip != local_ip2 && peek_mask(stack)==0) {
		local_node_idx = assert_node_by_addr(local_ip2);
		assert_node_addr(local_node_idx, local_ip);
	}

	if ( mship_certainty )
		set_int( "ontree_node", rmt_node_idx, 3 );
	else
		set_int( "ontree_node", rmt_node_idx, 2 );
	
	if ( peek_mask( stack ) )
		link_idx = assert_log_link( remote_ip, local_ip2, peek_mask( stack ) );

	else
		link_idx = assert_log_link( local_ip, remote_ip, mask );

	set_int( "ontree_link", link_idx, 3 );
}

void
mrtree_hlr(sd, arg, buff, status)
   int   sd;
   char *arg, *buff;
   int   status;
{

	char **line,
	      *node_argv[3];
	int num_lines,        /* Number of newlines in 'buff' array. */
	    i;
	Stack stack;

	int rc,
	    level,
	    mship_certainty;  /* Are we sure node is on the mcast tree? 1=yes */

	u_int32 local_ip,
	       ip_num,
	       mask;

	if ( status ) {
		popup(sd, arg, buff, status);
		return;	
	}

	/* 'Turn off' the 'snmpmcasttree' variable for all nodes. */
        clear_tree();

	line = explode( buff, "\n" );
	num_lines = xsizeof( line );

	if ( num_lines == 1 ) {
		xfree_array( line );
   	do_tcl( "mv_dialog .notify {No Such Tree} {No such multicast tree exists rooted at the given node.} {} 0 OK" );
		return;
	}

	/* Throw away the first line. It's for humans. And deal with the 2nd line
	 * outside of the loop. Of course, if there isn't a second line, then
	 * it means that there was no tree of the desired flavor.
	 */

	extract_info( line[1], &local_ip, &ip_num, &mask, &level, &mship_certainty );
	stack_init( &stack );
	push( &stack, ip_num, mask, level );
	if ( mship_certainty )
		set_int( "ontree_node", assert_node_by_addr( ip_num ), 3 );
	else
		set_int( "ontree_node", assert_node_by_addr( ip_num ), 2 );
	
	for( i = 2; i < num_lines; i++ ) {
		/* fprintf(stderr,"%d line:%s\n", i-2, line[i] ); */

		/* If extract_info has non-zero return code, it means that the line
	 	 * was for human-consumption. Basically, a comment line.
		 */
		if ( extract_info( line[i], &local_ip, &ip_num, &mask, &level, &mship_certainty ) )
			continue;

		/* stack_dump( &stack ); */

		/*
		 * If the level of the new node is higher than the level of the node on
		 * the top of the stack, then things are simple.  Just push the new node
		 * onto the stack and then go read in the next line.
		 */
		if ( level > peek_level( &stack ) ) {
			set_mcastnode_link( &stack, local_ip, ip_num,mask,mship_certainty);
			push( &stack, ip_num, mask, level );
			continue;
		}


		/*
		 * If we arrive here, then the level of the node on the top of the stack
		 * is greater than or equal to the level of the new node.
		 * So pop the stack until we uncover a node whose level is the same as
		 * or less than the new node's level.
		 *
		 * Then, if the node we're left with on top of the stack is at the same
		 * level as the new node, we must check to see if it's a subnet node or
		 * a router node.  (We can tell because router nodes have a zero mask.)
		 * If it is indeed a router node, then pop it too.
		 */
		while( level < peek_level( &stack ) )
			pop( &stack );
		if ( level == peek_level( &stack ) && !peek_mask( &stack ) )
			pop( &stack );

		/*
		 * The node on the top of the stack is either a node (router or subnet)
		 * whose level is less than the new node's level, OR it's a subnet node
		 * of the same level. If the new node's level is greater than the 
		 * stack-top's level, then it doesn't matter what type of node is on top
		 * of the stack. Just push the new node onto the stack and then go read
		 * the next line. 
		 */
		if ( level > peek_level( &stack ) ) {
			set_mcastnode_link( &stack, local_ip,ip_num,mask,mship_certainty);
			push( &stack, ip_num, mask, level );
			continue;
		}

		/* 
		 * If we arrived here, then the level of the node on the top of the
		 * stack is equal to the level of the new node AND the node on the top
		 * of the stack is a subnet node.
		 * If the new node is a subnet node (ie, it has a non-zero mask) then
		 * pop the old subnet node off the stack and push the new one on.
		 * Then go read the next line.
		 */
		if ( mask ) {
			pop( &stack );
			set_mcastnode_link( &stack, local_ip,ip_num,mask,mship_certainty);
			push( &stack, ip_num, mask, level );
			continue;
		}

		/*
		 * If we arrived here, then the following are true:
		 * a) The node on the top of the stack is a subnet node
		 * b) The node on the top of the stack has the same level as the new node
		 * c) The new node is a router node, not a subnet node
		 * We must now check to see if the new (router) node is on the same subnet 
		 * as the (subnet) node on the top of the stack. If it's not, then pop
		 * the old subnet off. In any event, we end up pushing the new router
		 * onto the stack.
		 */
		
		if ( ( peek_ip( &stack ) & ip_num & peek_mask( &stack ) )
		                  != ( peek_ip( &stack ) & peek_mask( &stack ) ) )
			pop( &stack );

		set_mcastnode_link( &stack, local_ip, ip_num, mask, mship_certainty );
		push( &stack, ip_num, mask, level );
	}
        guess_tree();

	node_argv[0] = "";
	node_argv[1] = "ontree_node > 1";
	node_argv[2] = NULL;

	rc = set_ndcond( NULL, tcl, 2, node_argv );

	node_argv[1] = "ontree_link = 3";
	rc = set_ldcond( NULL, tcl, 2, node_argv );

	xfree_array( line );
	refresh_display();
	do_spring( NULL, NULL, 0, NULL );
}

void
autosys_hlr(sd, arg, buff, status)
   int   sd;
   char *arg, *buff;
   int   status;
{
   char cmd[80];
   int autosys = atoi(buff+2);
   int i = atoi(arg);
   int n = num_nodes();
   register int j;
   char *ptr;

   set_int("autosys", i, autosys);

   /* Resolve asname */
   for (j=0; j<n; j++) {
      if (get_int("autosys", j)==autosys)
         break;
   }
   if (j<n && (ptr=get_string("asname", j))[0])
      set_string("asname", i, ptr);
   else {
      sprintf(cmd, "mv_asname %d", i);
      text_cmd(cmd);
   }

   pop_queue(Q_AUTOSYS, 0);
}

void
asname_hlr(sd, arg, buff, status)
   int   sd;
   char *arg, *buff;
   int   status;
{
   char *p;

   if (buff && buff[0]) {
      if (!strncmp(buff, "whois", 5))
         popup(sd, arg, buff, status);
      else {
         p = buff+strlen(buff)-1;
         if (*p=='\n')
            *p='\0';
         set_string("asname", atoi(arg), buff);
      }
   }
   pop_queue(Q_ASNAME, 0);
}

#if 0
void get_mtrace_idx_path( buff, idx_array, ip_array )
char *buff;
int **idx_array;
u_int32 **ip_array;
{
	int i, j, idx, num_lines;
	int matched;
	u_int32 IPnum1, IPnum2, ip_num;
	char IPStr1[20],IPStr2[20],
	     *tok,
	     **line;

	*idx_array = NULL;
	*ip_array  = NULL;

	line      = explode( buff, "\n" );
	num_lines = xsizeof( line );

	/************************************************************************
	 * This first 'if' statment handles the case when the source and        *
	 * destination handed to mtrace are on the same address.                *
	 ************************************************************************/
	if ( num_lines > 1 && strstr( line[ 1 ], "are directly connected" ) ) {

		matched=sscanf(line[0],"Mtrace from %s to %s via group",IPStr1,IPStr2);

		/* Did sscanf match the right num of params? */
		if ( matched != 2 ) {
			fprintf(stderr,"%s(%d):Line parse failed. matched = %d\n",
			                                      __FILE__, __LINE__, matched);
			abort();
		}

		IPnum1 = inet_addr( IPStr1 );
		IPnum2 = inet_addr( IPStr2 );

		if ( IPnum1 == -1 || IPnum2 == -1 ) {
			fprintf(stderr,"%s(%d):inet_addr() failed ", __FILE__, __LINE__);
			fprintf(stderr,"either for address %s or address %s\n", 
			                                         IPStr1, IPStr2 );
			abort();
		}

		*idx_array = (int*) malloc( sizeof( int ) * 3 );
		*ip_array = (u_int32*) malloc( sizeof( u_int32 ) * 3 );
	
		(*idx_array)[ 0 ] = assert_node_by_addr( IPnum1 );
		(*idx_array)[ 1 ] = assert_node_by_addr( IPnum2 );
		(*idx_array)[ 2 ] = -1;
		(*ip_array)[ 0 ]  = IPnum1;
		(*ip_array)[ 1 ]  = IPnum2;
		(*ip_array)[ 2 ]  = -1;

		xfree_array( line ); 
		return;
	}
			
	*idx_array = (int*) malloc( sizeof( int ) * (num_lines - 2) );
	*ip_array = (u_int32*) malloc( sizeof( u_int32 )*(num_lines - 2));

	/* Throw away the first two lines. They're for humans....
	 * Lines should be of the form 'X IP', where X can be any integer
	 * ( 0, positive or negative ), and IP is a host address.
	 */
	for ( j = 0, i = 2; ; i++ ) {

		tok = strtok( line[ i ], " " ); /* Grab the integer. */
                do {
		   tok = strtok( NULL, " " );      /* Grab the host addr. */
                } while (!strcmp(tok, "*"));
		ip_num = inet_addr( tok );

		if ( ip_num == -1 ) {
			/* The first non-addr we find marks the end of the path */
			(*idx_array)[ j ] = -1; /* Mark the end of the array. */
			(*ip_array)[ j ] = -1;  /* Mark the end of the array. */
			xfree_array( line ); 
			return;
		}

		/* This 'if' prevents redundant nodes from being put in our arrays. */
		if ( (i > 2) && ( ip_num == (*ip_array)[j-1] ) )
			continue;

		idx = assert_node_by_addr( ip_num );

		if ( idx < 0 ) {
			fprintf(stderr,"%s(%d):assert_node_by_addr() returned %d\n",
			                                      __FILE__, __LINE__, idx );
			abort();
		}

		(*idx_array)[ j ] = idx;
		(*ip_array)[ j ]  = ip_num;
		j++;
	}
}

void
mtrace_hlr(sd, arg, buff, status)
   int   sd;
   char *arg, *buff;
   int status;
{

	int  *idx_array,
	     *ip_array,
	      i, rc, max_nodes,
	      max_links,
	      subnet_node_num,
	      link_num,
	      node_argc;
	char *node_argv[3],
	     **line;

	if ( status ) {
		popup(sd, arg, buff, status);
		return;	
	}

	get_mtrace_idx_path( buff, &idx_array, &ip_array );
	line      = explode( buff, "\n" );
        clear_path(); /* initialize onpath_node and onpath_link values */

	node_argc = 2;
	node_argv[0] = "";
	node_argv[1] = "onpath_node > 1";
	node_argv[2] = NULL;

	rc = set_ndcond( NULL, tcl, node_argc, node_argv );

	node_argv[1] = "onpath_link > 1";
	rc = set_ldcond( NULL, tcl, node_argc, node_argv );

	/*
	 * If idx_array contains only 0 or 1 nodes, life isn't very interesting.... 
	 */

	if ( idx_array[ 0 ] == -1 ) {
		free(idx_array);
		xfree_array( line ); 
		return;
	}

	if ( idx_array[ 1 ] == -1 ) {
		set_int( "onpath_node", idx_array[0], 3 );
		free(idx_array);
		xfree_array( line ); 
		return;
	}

	/*
	 * Turn on the onpath variable for all nodes in idx_array, and all the
	 * links connecting them. Note that
	 * some links between nodes my be shared - that is to say 2 nodes may
	 * be adjacent by virtue of being on the same subnet.  If so, we want to
	 * be sure that we highlight the node that mview uses to represent the
	 * subnet.  
	 * Use find_log_link() below to determine if a tunnel exists. If 
	 * find_log_link() says that a tunnel does not exist between the two 
	 * routers, then either
	 *       a) a tunnel really doesn't exist or
	 *       b) the mview database is out of date
	 * To figure out which of (a) and (b) is occurring, we call 
	 * find_subnet(). If find_subnet() returns -1, then we have (b), else we
	 * have (a).
	 */

	set_int( "onpath_node", idx_array[0], 3 );

	for( i = 1; idx_array[i] != -1; i++ ) {

		set_int( "onpath_node", idx_array[i], 3 );

		if ( (link_num = find_log_link( idx_array[i], idx_array[i-1] )) == -1 ) { /* no tunnel... */
			subnet_node_num = find_subnet( ip_array[i] );
			if ( subnet_node_num != -1 ) { /* found subnet... */
				set_int( "onpath_node", subnet_node_num, 3 );

				link_num = find_log_link( idx_array[i-1], subnet_node_num );
				if ( link_num != -1 )
					set_int( "onpath_link", link_num, 3 );

				link_num = find_log_link( subnet_node_num, idx_array[i] );
				if ( link_num != -1 )
					set_int( "onpath_link", link_num, 3 );
			} else {
			   /*
			    * No connection found.  We need to create one,
			    * but should we create a subnet or a tunnel?
			    * (always create a tunnel?)
			    * Note that mtrace only gives the correct endpoints
			    * in the stats section, so this handler needs to
			    * be re-written to get everything from there
			    * instead if we are going to create links.
			    */
			}
		}
		else /* found tunnel... */
			set_int( "onpath_link", link_num, 3 );
	}
	
	free(idx_array);
	free(ip_array);
	xfree_array( line ); 
}
#endif

void
mtrace_stats_hlr(sd, arg, buff, status)
   int   sd;
   char *arg, *buff;
   int   status;
{
   char **line;
   int    max_nodes,
          max_links,
          num_lines,
          link_num,
          i, rc, state,
          sn;
   int   src, dest; /* Node #'s of endpoints of each link */
   int   subnet_node_num; /* Node # of intermediate subnet, if any */
   int   node_argc;
   char *node_argv[3];
   char  oif_addr_str[16];
   char  str[16];

   /* If the command failed, put the output in a popup window */
   if ( status ) {
      popup(sd, arg, buff, status);
      return;   
   }

   if ( strstr( buff, "directly connected" ) ) {
      popup(sd, arg, buff, status);
      return;   
   }

   line      = explode( buff, "\n" );
   num_lines = xsizeof(line);
   clear_path(); /* initialize onpath_node and onpath_link values */

   node_argc = 2;
   node_argv[0] = "";
   node_argv[1] = "onpath_node > 1";
   node_argv[2] = NULL;

   rc = set_ndcond( NULL, tcl, node_argc, node_argv );

   node_argv[1] = "onpath_link > 1";
   rc = set_ldcond( NULL, tcl, node_argc, node_argv );

   /* The stats section starts after the '  Source        Response...' line */
   for (i=0; i<num_lines && strncmp(line[i], "  Source", 8); i++);

   /*
    * Okay, from this point on down until the '  Receiver      Query...' line,
    * the first field is either an IP address or a 'v'.  Only one address
    * is shown if the iif and the oif are the same, otherwise both are
    * shown on consecutive lines.  We process lines using a two-state machine.
    */
   i++;
   while (i<num_lines) {
      sscanf(line[i], "%s", oif_addr_str);
      if (oif_addr_str[0]=='*') {
         i += 2;
         continue;
      }
      if (oif_addr_str[0]=='0' || !isdigit(oif_addr_str[0])) {
         i++;
         continue;
      }
      break;
   }

   if (i>=num_lines) {
      xfree_array(line);
      return;
   }

   sn = src = assert_node_by_addr(inet_addr(oif_addr_str));
   set_int( "onpath_node", src, 3 );
   for (state=0, i++; i<num_lines && state>=0; i++) {
      sscanf(line[i], "%s", str);
      if (!strncmp(line[i], "  Receiver",10))
         break;

      switch(state) {
      case 0: /* Expecting 'v' or oif */
         if (!strcmp(str, "v")) { /* Save link stats here if desired... */
            state = 1; /* expect iif */
         } else if (isdigit(str[0])) { /* Process multi-homed node */

            /* Add IP address to node */
            src = assert_node_addr(src, inet_addr(str));
            set_int( "onpath_node", src, 3 );
            
            strcpy(oif_addr_str, str); /* Save new oif address */
            /* keep same state, since we expect a 'v' now */
         } else
            state = -1;
         break;

      case 1: /* Expecting iif */
         if (!isdigit(str[0]))
            state = -1;
         else { /* Process link */
            u_int32 oif_addr = inet_addr(oif_addr_str);
            u_int32 iif_addr = inet_addr(str);

            /*
             * Turn on the onpath variable for all nodes in idx_array, and all 
             * the links connecting them. Note that nodes pairs may be on the 
             * same subnet.  If so, we want to be sure that we highlight the 
             * node that mview uses to represent the subnet.  
             * Use find_log_link() below to determine if a tunnel exists. If 
             * find_log_link() says that a tunnel does not exist between the 
             * two routers, then either
             *       a) a tunnel really doesn't exist or
             *       b) the mview database is incomplete
             * To figure out which of (a) and (b) is occurring, we call 
             * find_subnet(). If find_subnet() returns -1, then we have (b), 
             * else we have (a).
             */

            /* Check for existing tunnel */
            dest = assert_node_by_addr(iif_addr);
            link_num = find_log_link(src, dest);

            /* Check for existing subnet adjacency */
            if (link_num < 0) {
               subnet_node_num = find_subnet(iif_addr);
               if (subnet_node_num>=0 
                && subnet_node_num==find_subnet(oif_addr)) {
                  link_num = find_log_link(subnet_node_num, dest);
                  set_int( "onpath_link", link_num, 3 );
                  link_num = find_log_link(src, subnet_node_num);
                  set_int( "onpath_node", subnet_node_num, 3 );
               }
            }

            /*
             * Ok, we need to add it.  Guess at whether they're adjacent 
             * by comparing 24-bit prefixes 
             */
            if (link_num < 0 
             && (oif_addr & 0xFFFFFF00)==(iif_addr & 0xFFFFFF00)) {
               subnet_node_num = assert_subnet(oif_addr);
               set_int( "onpath_node", subnet_node_num, 3 );

               link_num = assert_subnet_link(subnet_node_num, iif_addr);
               set_int( "onpath_link", link_num, 3 );

               assert_subnet_link(subnet_node_num, oif_addr);
               link_num = find_log_link(src, subnet_node_num);
            }

            /* Ok, add a tunnel */
            if (link_num < 0) {
               assert_log_link(iif_addr, oif_addr, 0x0); /* reverse link */
               link_num = assert_log_link(oif_addr, iif_addr, 0x0);
            }

            /* Set any link variables now... */
            set_int( "onpath_link", link_num, 3 );

            /*
             * Unless we immediately see another IP address, the oif addr
             * is the same as the iif addr
             */
            strcpy(oif_addr_str, str);
            src = assert_node_by_addr(inet_addr(oif_addr_str));
            set_int( "onpath_node", src, 3 );
            state = 0; /* expect v or oif */
         }
      }
   }
   xfree_array( line ); 

   /* Now change to tree format */
   refresh_display();
   locate_tree(sn);
   rescale_nodes();
   ndcond.dirty = 1; /* force nodes to be redrawn */
}

void
mgroups_hlr(sd, arg, buff, status)
   int   sd;
   char *arg, *buff;
   int   status;
{
   char **line;
   int num_lines,        /* Number of newlines in 'buff' array. */
       i;
   char cmd[80], addr[80];

   if (status || !buff)
      return;

   line = explode( buff, "\n" );
   num_lines = xsizeof( line );

   do_tcl(".w.1.box delete 0 end");
   /* There are two header lines, so start at i=2 below */
   for (i=2; i<num_lines; i++) { 
      sscanf(line[i], "%s%*s", addr); /* get first field */
      sprintf(cmd, ".w.1.box insert end {%s}", addr);
      do_tcl(cmd);
   }
   xfree_array( line );
}

/* This handler uses mstat output:
IP Multicast Route Table for shockwave.merit.edu
Mcast-group     Origin-Subnet     InIf  UpTime Tmr   Pkts     Bytes RpF Proto
224.2.127.255   13.2.116.196/32     0       25 534      1       183   0 dvmrp
224.2.127.255   128.3.112.2/32      0      423  69      1       158   0 dvmrp
*/

void
grpsrcs_hlr(sd, arg, buff, status)
   int   sd;
   char *arg, *buff;
   int   status;
{
	char **line;
	int num_lines,        /* Number of newlines in 'buff' array. */
	    i, idx, max_nodes;
	char cmd[80], grpstr[20], srcstr[32], *maskstr;

	max_nodes = num_nodes();
	for( i = 0; i < max_nodes; i++ )
		set_int( "isGroupSource", i, 0 );

	/* If arg is null, then we want to update the list box. */
	if ( !arg || !*arg)
		do_tcl(".w.1.box delete 0 end");

	if (status || !buff)
		return;

	line = explode( buff, "\n" );
	num_lines = xsizeof( line );

	for (i=2; i<num_lines; i++) {
		sscanf(line[i], "%s%s", grpstr, srcstr);
		maskstr = strchr(srcstr, '/');
		if (maskstr)
			*maskstr++ = '\0';
    
		if (maskstr && !strcmp(maskstr, "32"))  {
			if ( !arg || !*arg)
				sprintf(cmd, ".w.1.box insert end {%s}", srcstr);
			idx = assert_node_by_addr( inet_addr( srcstr ) );
			if ( idx < 0 ) {
				fprintf(stderr,"%s(%d):assert_node_by_addr() return negative\n", 
				                                                 __FILE__, __LINE__ );
				abort();
			}
			set_uint("isGroupSource", idx, 1 );
		}

		else
			printf("source=%s mask=%s\n", srcstr, maskstr);

		if (!arg || !*arg)
			do_tcl(cmd);
	}
	xfree_array( line );
}

/* This handler parses mstat output to update subnet masks
 *  PIM Interface Table for hideout.sms.fi
 *  If Address            Mode   DRouter         Ttl Int
 *   1 194.111.122/24     sparse 194.111.122.254   1  30
 *   2 194.197.125.126/28 sparse 194.197.125.126   1  30
 *   3 192.194.252.178/30 sparse 0.0.0.0           4  30
 */
void
pim_if_hlr(sd, arg, buff, status)
   int   sd;
   char *arg, *buff;
   int status;
{
   char **line, **field, *p, **elt;
   int num_lines, i, masklen, j;
   u_int32 srcaddr, dstaddr, dstmask;

   if ( status )
      return;	

   line = explode( buff, "\n" );
   num_lines = xsizeof( line );
   for (i=2; i<num_lines; i++) {
      field = explode(line[i], " "); /* split into fields */
      if ((xsizeof(field)>5) && (p = strchr(field[1], '/'))) {
         masklen = atoi(p+1);
         srcaddr = inet_addr(field[3]);
         (*p) = '\0';
         elt = explode(field[1], ".");
         dstaddr = 0;
         for (j=0; j<xsizeof(elt); j++)
            ((char*)&dstaddr)[j] = atoi(elt[j]);
         if (!srcaddr && xsizeof(elt)==4)
            srcaddr = inet_addr(field[1]);
         xfree_array(elt);
         dstmask = 0xffffffff << (32-masklen);

         /* Some routers report addresses which don't fit within the mask,
          * e.g. hideout.sms.fi
          */
         dstaddr &= dstmask; 
     
         if (!srcaddr) {
            int     v = atoi(arg);
            node_t *node = mget_node(v);
  
            /* See if we know an IP address on this subnet */
            for (j=0; j<MAX_IFACE; j++) {
               if ((node->iface[j] & dstmask)==dstaddr) {
                  srcaddr = node->iface[j];
                  assert_log_link(srcaddr, dstaddr, dstmask); 
                  break;
               }
            }

            /* Nope, just do the best we can */
            if (j==MAX_IFACE) {
               int subnet = assert_subnet(dstaddr);
               assert_subnet_mask(subnet, dstaddr, dstmask);
            }

         } else
            assert_log_link(srcaddr, dstaddr, dstmask); 
      }
      xfree_array(field);
   }
   xfree_array(line);
}

/* This handler parses mstat output to update subnet masks 
 *  DVMRP Virtual Interface Table for shockwave.merit.edu
 *  Vif Local-Address  Met Thr Rat Type    OpStat AdminState   Remote-Address
 *   0  198.108.60.69    1   1   0 querier up     active       198.108.60/24
 *   1  198.108.60.69    1   1 500 tunnel  down   notInService 141.213.10.41
 *   2  198.108.60.69    1   1 500 tunnel  down   active       141.213.10.45
 */
void
dvmrp_vif_hlr(sd, arg, buff, status)
   int   sd;
   char *arg, *buff;
   int status;
{
   char **line, **field, *p, **elt;
   int num_lines, i, masklen, j;
   u_int32 srcaddr, dstaddr, dstmask, flags=0;

   if ( status )
      return;	

   line = explode( buff, "\n" );
   num_lines = xsizeof( line );
   for (i=2; i<num_lines; i++) {
      field = explode(line[i], " "); /* split into fields */
      if ((xsizeof(field)>7) && (p = strchr(field[8], '/'))) {
         masklen = atoi(p+1);
         srcaddr = inet_addr(field[1]);
         (*p) = '\0';
         elt = explode(field[8], ".");
         dstaddr = 0;
         for (j=0; j<xsizeof(elt); j++)
            ((char*)&dstaddr)[j] = atoi(elt[j]);
         xfree_array(elt);
         dstmask = 0xffffffff << (32-masklen);
         assert_log_link(srcaddr, dstaddr, dstmask); 

         flags = 0;
         if (!strcmp(field[5], "tunnel") && dstaddr!=srcaddr)
            flags |= DVMRP_NF_TUNNEL;
         else if (!strcmp(field[5], "querier"))
            flags |= DVMRP_NF_QUERIER;
         if (!strcmp(field[6], "down"))
            flags |= DVMRP_NF_DOWN;
         if (!strcmp(field[7], "notInService"))
            flags |= DVMRP_NF_DISABLED;

         /* When the IP addresses differ by 1, assume it's point-to-point */
         if (abs(srcaddr-dstaddr)==1)
            flags |= DVMRP_NF_TUNNEL;

         update_link(srcaddr, dstaddr, dstmask, atoi(field[2]), atoi(field[3]),
          flags);
      }
      xfree_array(field);
   }
   xfree_array(line);
}

/* This handler parses mstat output to update values for these node variables:
 *   PktsPerGrpSrcRate   : packets/sec from selected group,source 
 *   PktsPerGrpSrcUpTime : timestamp at last sample
 *   PktsPerGrpSrc       : value at last sample
 *   PktsPerGroup        : total packets (ever) rcvd for selected group
 */
void
ppg_hlr(sd, arg, buff, status)
   int   sd;
   char *arg, *buff;
   int status;
{

	char **line, **args, sourcestr[25], *p;
	int rc, i,
	    num_lines,         /* Num lines in 'buff' array. */
	    idx;               /* Index of node whose timer has expired. */
	u_int32 total_count,/* Total pkts for group reported in buff */
	         last_count,   /* old value of PktsPerGrpSrc */
	         curr_count,   /* new value of PktsPerGrpSrc */
	         last_uptime,  /* last value of PktsPerGrpSrcUpTime */
	         curr_uptime,  /* value of 1 instance of ipMRouteUpTime */
	         delta_count,  /* curr_count - last_count */
	         delta_time;   /* curr_uptime - last_uptime */

	if ( status ) {
		popup(sd, arg, buff, status);
		return;	
	}

	args        = explode( arg, "," );
	idx         = atoi( args[0] );
	line        = explode( buff, "\n" );
	num_lines   = xsizeof( line );
	total_count = 0;

	/*
	 * The first two lines of output contains column headers. Ignore them.
	 */
	for(i=2; i<num_lines; i++) {

		rc=sscanf(line[i],"%*s %s %*s %lu %*s %lu %*s %*s %*s",
                 sourcestr, &curr_uptime, &curr_count);
		if ((p = strchr(sourcestr, '/'))!=NULL)
			*p = '\0'; /* ignore mask */
/*
		fprintf(stderr,"*** sourcestr = %s  pkts = %lu\n", sourcestr, curr_count );
		if ( rc != 3 ) {
			fprintf(stderr,"%s(%d):sscanf matched incorrect number of variables.\n",
			        __FILE__, __LINE__ );
			abort();
		}
*/

		if ( !strcmp( args[1], sourcestr ) ) {

			last_uptime = get_uint("PktsPerGrpSrcUpTime", idx );
			if ( last_uptime >= curr_uptime ) {
				if ( curr_uptime == 0 )
					set_float("PktsPerGrpSrcRate", idx, 0 );
				else
					set_float("PktsPerGrpSrcRate", idx, (float)curr_count/(float)curr_uptime);
			}

			else {
				last_count        = (u_int32) get_uint( "PktsPerGrpSrc", idx );
				delta_count       = curr_count - last_count;
				delta_time        = curr_uptime - last_uptime;
				if ( delta_count == 0 )
					set_float("PktsPerGrpSrcRate", idx, 0 );
				else
					set_float("PktsPerGrpSrcRate", idx, (float)delta_count/(float)delta_time);
			}

			set_uint("PktsPerGrpSrcUpTime", idx, curr_uptime );
			set_uint("PktsPerGrpSrc", idx, curr_count );
		}

		total_count += curr_count;
	}

	set_uint("PktsPerGroup", idx, total_count );

	xfree_array( line ); 
	xfree_array( args );
}

/*
 * Function to free up all malloc'ed space when MView is exited while
 * output handlers are in progress 
 */
void
free_handlers()
{
   int h;

   for (h=0; h<MAXHANDLERS; h++) {
      if (!handler[h].func) continue;
      stop_handler(h);
   }

   xfree_array(pathdirs);
}

/*
 * This following code looks for the standard utilities in the user's path,
 * and then in some standard locations if not found.  It warns about
 * missing utilities and incorrect permissions, and saves the location
 * found for use when invoking the utility (to speed up invocation time).  
 */

typedef struct {
   char *name;
   int   needs_suid;
   int   pathdirindex;
} utility_t;

static utility_t util[] = {
 { "traceroute", 1, -1 },
 { "ping",       1, -1 },
 { "whois",      0, -1 },
 { "nslookup",   0, -1 },
 { "asn",        0, -1 },
 { "asname",     0, -1 },
 { "list.pkts",  0, -1 },
 { "snmpoidget", 0, -1 },
 { "mstat",      0, -1 },
 { "mtrace",     1, -1 },
 { "mrtree",     1, -1 },
 { "mrinfo",     1, -1 },
 { NULL,         0, -1 }
};

static char *extradirs[] = { "/usr/sbin", "/sbin", "/usr/etc", "/usr/ucb", 
  "/usr/local/bin", "/bin", "/usr/bin", NULL
};

void
check_binaries()
{
   char file[256];
static char buff[4096];
   int i,j, nd, found;
   char *path = getenv("PATH");
   struct stat st;

   for (i=0; util[i].name; i++) {
      if (pathdirs)
         xfree_array(pathdirs);
      pathdirs = explode(path, ":");
      nd = xsizeof(pathdirs);

      found = 0;
      for (j=0; !found && j<nd; j++) {
         if (pathdirs[j][ strlen(pathdirs[j])-1 ]=='/')
             pathdirs[j][ strlen(pathdirs[j])-1 ]= '\0';
         sprintf(file, "%s/%s", pathdirs[j], util[i].name);

         if (stat(file, &st) == 0) {
            found = 1;
            util[i].pathdirindex = j;
            if (util[i].needs_suid && !(st.st_mode & S_ISUID))
               printf("WARNING: %s needs to be setuid root\n", file);
         }
      }

      for (j=0; !found && extradirs[j]; j++) {
         sprintf(file, "%s/%s", extradirs[j], util[i].name);

         if (stat(file, &st) == 0) {
            found = 1;
            util[i].pathdirindex = nd;
            sprintf(buff, "PATH=%s:%s", path, extradirs[j]);
            putenv(buff);
            path = buff+5;
         }
      }

      if (!found) {
         printf("WARNING: %s not found\n", util[i].name);
         util[i].pathdirindex = -1;
      }
   }
}

/*
 * Return the full pathname of a command if we know it, otherwise
 * just the command itself.
 */
char *
locate_command(str)
   char *str;
{
   int i;
static char buff[256];
 
   for (i=0; util[i].name; i++) {
      if (!strcmp(str, util[i].name)) {
         if (util[i].pathdirindex < 0)
            return str;
         sprintf(buff, "%s/%s", pathdirs[ util[i].pathdirindex ], str);
         return buff;
      }
   }
   return str;
}
