/* $Id: high.c,v 1.21 1996/07/02 22:42:16 thalerd Exp $
 * -------------------------------------------------------------------------
 *  
 * [C] The Regents of the University of Michigan and Merit Network, Inc. 1996
 * All Rights Reserved 
 *   
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice and this permission notice appear in all 
 * copies of the software and derivative works or modified versions thereof, 
 * and that both the copyright notice and this permission and disclaimer notice 
 * appear in supporting documentation. 
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE REGENTS OF THE 
 * UNIVERSITY OF MICHIGAN AND MERIT NETWORK, INC. DO NOT WARRANT THAT THE 
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR 
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the 
 * University of Michigan and Merit Network, Inc. shall not be liable for any 
 * special, indirect, incidental or consequential damages with respect to any 
 * claim by Licensee or any third party arising from use of the software. 
 *
 * -------------------------------------------------------------------------
 */
/* Functions to display a high-level view of relationships between clouds */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdlib.h> /* for atoi */
#include <tcl.h>
#include <tk.h>
#include <math.h>
#include "mview.h"
#include "struct.h"
#include "queue.h"
#include "database.h"
#include "globals.h"
#include "high.h"
#include "link.h" /* for up_directions */
#include "dict.h" /* for remove_dd */
#include "display.h" /* for refresh_display */
#include "driver.h" /* for text_cmd */
#include "cloud.h" /* for some_node_in_cloud */
#include "expr.h" /* for testcond */

       int csmode=CS_NAME, clhmode=CLH_LOGLINK;
       cond_t chcond, clhcond, cscond;

extern int display_ok;

static int num_cloudlinks=0;
       int num_cloudnodes=0;
static int mark[MAX_NODES];

static int cloudnode_tk[MAX_NODES];
static int cloudlink_tk[MAX_LINKS];
static int start=1;

/*
static float Hxinc, Hyinc;
static float Hminx, Hminy;
static float Hmaxx, Hmaxy;
*/
       double Hxscale = 0.500; /* canvas width = 500 */
       double Hyscale = 0.260; /* canvas height ~= 260 */
       double Hzoom   = 1.0;

static int qdebug = 0;

/* Convert canvas to virtual coordinate */
static float /* Virtual X coordinate */
cloudXc2v(x)
   int x;    /* Canvas X coordinate */
{
   return x/Hxscale/Hzoom - 50.0;
}

/* Convert canvas to virtual coordinate */
static float /* Virtual Y coordinate */
cloudYc2v(y)
   int y;    /* Canvas Y coordinate */
{
   return y/Hyscale/Hzoom - 50.0;
}

/* Convert virtual to canvas coordinate */
static int   /* Canvas X coordinate */
cloudXv2c(x)
   float x;  /* Virtual X coordinate */
{
   return (int)floor((x+50)*Hxscale*Hzoom + 0.5);
}

/* Convert virtual to canvas coordinate */
static int   /* Canvas Y coordinate */
cloudYv2c(y)
   float y;  /* Virtual Y coordinate */
{
   return (int)floor((y+50)*Hyscale*Hzoom + 0.5);
}

/* Get canvas coordinate of a cloud */
int          /* Canvas X coordinate */
cloudXi2c(i)
   int i;    /* Cloud # */
{
   return cloudXv2c(get_float("cloudX", i));
}

/* Get canvas coordinate of a cloud */
int          /* Canvas Y coordinate */
cloudYi2c(i)
   int i;    /* Cloud # */
{
   return cloudYv2c(get_float("cloudY", i));
}

static void
init_high()
{
   int i;

   if (!start)
      return;

   for (i=0; i<MAX_NODES; i++)
      cloudnode_tk[i]=0;
   for (i=0; i<MAX_LINKS; i++)
      cloudlink_tk[i]=0;

   /* Fill in clouds */
   num_cloudnodes = get_numelts("cloudX");    
   num_cloudlinks = get_numelts("cloudlinkA");

   start=0;

   /* Removed 7/2/96 to preserve cloud layoud between runs 
    * calc_high();
    */
}

/*
 * Initialize cloud if needed
 */
int
assert_cloudnode(idx)
   int idx;        /* cloud index */
{
   if (start)
      init_high();

   if (idx < num_cloudnodes)
      return idx;

   set_float("cloudX",  idx, 1.0*(random()%1000));
   set_float("cloudY",  idx, 1.0*(random()%1000));
   set_int("cloudLink", idx, -1);

   if (idx >= num_cloudnodes)
      num_cloudnodes = idx+1;

   return idx;
}

static int           /* RETURNS: cloudlink idx, or -1 if not found */
find_cloudlink(an,bn) /* ARGUMENTS:            */
   int an;           /*    cloud index        */
   int bn;           /*    cloud index        */
{
   int ocl=-1, cl = get_int("cloudLink", an);

   while (cl>=0 && get_int("cloudlinkB", cl)!=bn) {
      ocl=cl;
      cl = get_int("cloudlinkNext", ocl);
      if (ocl==cl) {
         printf("ERROR: cloudlinkNext is bad\n");
         return -1;
      }
   }

   return cl;
}

/*
 * Add a link from cloud A to cloud B
 * This is also called by add_log_link when a new link is discovered
 */
int                      /* RETURNS: index of cloudlink */
assert_cloudlink(a,b,ln) /* ARGUMENTS:                 */
   int a;                /*   source cloud index       */
   int b;                /*   destination cloud index  */
   int ln;               /*   node link index          */
{
   int idx;
   int an, bn, cl;

   if (!a || !b) return 0;

   an = assert_cloudnode(a);
   bn = assert_cloudnode(b);
   cl = find_cloudlink(an,bn);
   if (cl>=0)
      return cl;

   idx = num_cloudlinks++;
   set_int("cloudlinkA",    idx, an);
   set_int("cloudlinkB",    idx, bn);
   set_int("cloudlinkNext", idx, get_int("cloudLink", an));
   set_int("cloudlinkLog",  idx, ln);
   set_int("cloudLink", an, idx);
/* XXX Note that for clouds, there can be a list of cloudLinks... */
   if (qdebug)
      printf("Added link %d from %d to %d\n", idx, an,bn);
   return idx;
}

/* Given a starting node, generate all cloud links reachable from that point.  
*/
static void
generate_cloudlinks(n)
   int n; /* node index */
{
   int i, ln;
   node_t *node, *dnode;
   int origcloud, fromcloud, tocloud;
   log_link_t *ledge;

   /* Initialize queue */
   node = mget_node(n);
   origcloud = get_int("cloud", n);
   assert_cloudnode(origcloud);
   push_queue(Q_CLOUDLINK, n);

   while ((i = pop_queue(Q_CLOUDLINK, 0)) >= 0) {  /* get next node */
      node = mget_node(i);
      mark[i]=1;

      fromcloud = get_int("cloud", i);
   
      /* Process logical links */
      for (ln = node->last_link[LOUT]; ln>=0; ln = llink[ln].prev_link[LOUT]) {
         if (up_directions(ln) < 2)
            continue;
         ledge = get_log_link( ln );
         if (ledge->dest < 0)
            continue;
         if (!mark[ ledge->dest ]) {
            dnode = mget_node(ledge->dest);

            tocloud = get_int("cloud", ledge->dest);

            /* make sure cloud node exists for this cloud */
            if (tocloud > 0)
               assert_cloudnode(tocloud);

            if (tocloud > 0 && tocloud != origcloud) {
               if (fromcloud==origcloud)
                  assert_cloudlink(origcloud, tocloud, ln);
               else {
                  assert_cloudnode(fromcloud);
 
                  /* XXX the "ln" below should really be the first link in the
                   * 'fromcloud' tree adjoining the node in 'origcloud'
                   */
                  assert_cloudlink(origcloud, fromcloud, ln); 

                  assert_cloudlink(fromcloud, tocloud, ln);
               }
            } else
               push_queue(Q_CLOUDLINK, ledge->dest);
         }
      }
   }
}

/* 
 * Find the first unmarked node which is in a cloud
 */
static int                 /* RETURNS: node # */
get_unmarked_node()
{
   int i;
   node_t *node;

   for (i=0; i<num_nodes(); i++) {
      if (!mark[i]) {
         node = mget_node(i);
         if (get_int("cloud", i)>0) 
            break;
      }
   }
   return (i>=num_nodes())? -1 : i;
}

/* Assuming the network is already divided into clouds,
 * Fill in user vars for cloud: cloudX, cloudY, cloudLink
 * For cloud links: cloudlinkA, cloudlinkB, cloudlinkNext
 */
void
calc_high()
{
   int i;

   if (start) {
      init_high();
      return; /* init_high will call call_high itself with start=0 */
   }

   /* Clear old vars */
   remove_cloudlinks(); /* undisplay links and clear cloudlink_tk */
   remove_dd("cloudX");
   remove_dd("cloudY");
   remove_dd("cloudLink");
   remove_dd("cloudlinkA");
   remove_dd("cloudlinkB");
   remove_dd("cloudlinkNext");
   remove_dd("cloudlinkLog");
   num_cloudnodes = num_cloudlinks = 0;

   /* Redeclare cloud variables */
   create_dd("cloudX",        VT_FLOAT,   4, MAXCLOUDS,     VF_PERSISTENT, VL_CLOUD);
   create_dd("cloudY",        VT_FLOAT,   4, MAXCLOUDS,     VF_PERSISTENT, VL_CLOUD);
   create_dd("cloudLink",     VT_INT,     2, MAXCLOUDS,     VF_PERSISTENT, VL_CLOUD);

   /* Declare cloudlink variables */
   create_dd("cloudlinkA",    VT_INT,     2, MAXCLOUDLINKS, VF_PERSISTENT, VL_CLOUDLINK);
   create_dd("cloudlinkB",    VT_INT,     2, MAXCLOUDLINKS, VF_PERSISTENT, VL_CLOUDLINK);
   create_dd("cloudlinkLog",  VT_INT,     2, MAXCLOUDLINKS, VF_PERSISTENT, VL_CLOUDLINK);
   create_dd("cloudlinkNext", VT_INT,     2, MAXCLOUDLINKS, VF_PERSISTENT, VL_CLOUDLINK);

   /* Unmark all nodes */
   for (i=0; i<num_nodes(); i++)
      mark[i]=0;

   /* Generate cloud links */
   while ((i=get_unmarked_node()) >= 0)
      generate_cloudlinks(i);
}

void
recalc_high()
{
   calc_high();
   refresh_high();
   goto_cloud(curr_cloud);
}

char *
set_cloudX(clientData, interp, name1, name2, flags)
   ClientData  clientData;
   Tcl_Interp *interp;
   char       *name1;
   char       *name2;
   int         flags;
{
   /*
    * Don't pass the flags argument into the Tcl_GetVar2 calls below
    * since on some versions of Tcl the PART1_NOT_PARSED bit is set (0x10000)
    * which causes the get to fail and return NULL.
    */
   int cx = atoi(Tcl_GetVar2(interp,name1,name2,TCL_GLOBAL_ONLY));
   int n = atoi(Tcl_GetVar2(interp, "HnodeIdx", name2, TCL_GLOBAL_ONLY));
   float vx = cloudXc2v(cx);

   set_float("cloudX", n, vx);
   return NULL;
}

char *
set_cloudY(clientData, interp, name1, name2, flags)
   ClientData  clientData;
   Tcl_Interp *interp;
   char       *name1;
   char       *name2;
   int         flags;
{
   int cy = atoi(Tcl_GetVar2(interp,name1,name2,TCL_GLOBAL_ONLY));
   int n = atoi(Tcl_GetVar2(interp, "HnodeIdx", name2, TCL_GLOBAL_ONLY));
   float vy = cloudYc2v(cy);

   set_float("cloudY", n, vy);
   return NULL;
}

void
goto_cloud(cloud_idx)
   int cloud_idx;     /* cloud index */
{
   char    buff[255]; /* command buffer */
   int     n;         /* node number */

   if (start)
      init_high();
 
   /* Unhighlight old cloud */
   if (curr_cloud>=0 && cloudnode_tk[curr_cloud]) {
      sprintf(buff, ".highcanvas itemconfigure $HnodeRect(%d) -fill white",
       cloudnode_tk[curr_cloud]);
      text_cmd( buff);
   }

   /* Change current cloud and update display */
   curr_cloud   = cloud_idx;
   refresh_display();

   /* Change current node if the old one is no longer displayed */
   if (get_int("cloud", curr_node)!=curr_cloud) {
      n = some_node_in_cloud(curr_cloud);
      goto_node(n);
   }

   /* Highlight new cloud */
   if (curr_cloud>=0 && cloudnode_tk[curr_cloud]) {
      sprintf(buff, ".highcanvas itemconfigure $HnodeRect(%d) -fill yellow",
       cloudnode_tk[curr_cloud]);
      text_cmd( buff);
   }
}

static char *
cloud_symbol(cloud_idx)
   int cloud_idx;
{
static char buff[255];

   switch(csmode) {
   case CS_NAME:  /* Number */  
      return get_string("cloudname", cloud_idx);

   case CS_USERVAR:  /* User Variable */
      sprintf(buff, "%s", get_string(cscond.name, cloud_idx));
      return buff;
   }
   return NULL;
}

static void
set_cloud_symbol(cloud_idx)
   int cloud_idx;
{
   char buff[255];

   if (start)
      init_high();

   if (cloudnode_tk[cloud_idx]) {
      sprintf(buff, "setCloudText %d \"%s\"", cloudnode_tk[cloud_idx],
       cloud_symbol(cloud_idx));
      text_cmd(buff);
   }
}

static void
set_cloud_symbols()
{
   int cloud_idx;

   if (start)
      init_high();

   for (cloud_idx=0; cloud_idx<num_cloudnodes; cloud_idx++) {
      if (cloudnode_tk[cloud_idx])
         set_cloud_symbol(cloud_idx);
   }
}

static void
set_cloud_highlight(cloud_idx)
   int cloud_idx;
{
   char buff[255];

   if (cloudnode_tk[cloud_idx]) {
      sprintf(buff, "setCloudHl %d %d", cloudnode_tk[cloud_idx], 
       chighlight(cloud_idx));
      text_cmd( buff);
   }
}

static void
set_cloud_highlights()
{
   int cloud_idx;

   if (start)
      init_high();

   for (cloud_idx=0; cloud_idx<num_cloudnodes; cloud_idx++) {
      if (!cloudnode_tk[cloud_idx]) continue;
      set_cloud_highlight(cloud_idx);
   }
}

static void
set_cloud_display(cloud_idx)
   int cloud_idx;
{
   char buff[255];
   int x,y, vis;

   x = cloudXi2c(cloud_idx);
   y = cloudYi2c(cloud_idx);

   /* Cloud is visible if it contains nodes and isn't cloud 0 */
   vis = (cloud_idx>0 && get_int("cloudsize", cloud_idx)>0);

   if (cloudnode_tk[cloud_idx]) {  
      if (vis) {                 /* move existing */
         sprintf(buff, "moveCloudTo %d %d %d", cloudnode_tk[cloud_idx], x, y);
         text_cmd(buff);
      } else {                   /* delete existing */
         sprintf(buff, "rmCloud %d", cloudnode_tk[cloud_idx]);
         text_cmd(buff);
         cloudnode_tk[cloud_idx] = 0;
      }
   } else if (vis) {             /* create new */
      sprintf(buff, "mkCloud %d %d {%s} %d", x,y, cloud_symbol(cloud_idx),
       cloud_idx);
      text_cmd( buff);

      cloudnode_tk[cloud_idx] = atoi(tcl->result);
      if (qdebug)
         printf("cloudnode_tk[%d] = %s\n", cloud_idx, tcl->result);
      add_timers(VL_CLOUD, cloud_idx);

      set_cloud_symbol(cloud_idx);
      set_cloud_highlight(cloud_idx);
   }
}

static void
set_cloud_displays()
{
   int   cloud_idx;

   if (start)
      init_high();

   for (cloud_idx=0; cloud_idx<num_cloudnodes; cloud_idx++)
      set_cloud_display(cloud_idx);
}

int
clwidth(ln)
   int ln;
{
   if (clhighlight(ln))
      return 3;
   return 1;
}

char *
clcolor(ln)
   int ln;
{
   if (clhighlight(ln))
      return "blue";
   return "black";
}

void
remove_cloudlink(ln)
   int ln;
{
   int pl, cl, an;
   char buff[256];

   if (cloudlink_tk[ln]) {  /* delete existing */
      sprintf(buff, "rmCloudEdge %d", cloudlink_tk[ln]);
      text_cmd(buff);
      cloudlink_tk[ln] = 0;
   }

   /* Remove from linked list */
   an = get_int("cloudlinkA", ln);
   if (an>=0) {
      pl = -1;
      cl = get_int("cloudLink", an);
      while (cl>=0 && cl!=ln) {
         pl = cl;
         cl = get_int("cloudlinkNext", cl);
      }
   
      if (pl<0) 
         set_int("cloudLink", an, get_int("cloudlinkNext", ln));
      else
         set_int("cloudlinkNext", pl, get_int("cloudlinkNext", ln));
     
   }
   set_int("cloudlinkA", ln, -1);
   set_int("cloudlinkB", ln, -1);
   set_int("cloudlinkLog",  ln, -1);
   set_int("cloudlinkNext", ln, -1);
}

/* Called from assign_clouds_by_var() when reinitializing cloud assignments */
void
remove_cloudlinks()
{
   int ln;
 
   for (ln=0; ln<num_cloudlinks; ln++) {
      if (!cloudlink_tk[ln]) continue;
      remove_cloudlink(ln);
   }
}

static void
set_cloudlink_display(ln)
   int ln;  /* cloud link # */
{
   char buff[255];
   int vis=1, src, dest;
   
   src  = get_int("cloudlinkA",ln);
   dest = get_int("cloudlinkB",ln);
   if (src<0 || dest<0)
      vis = 0;
   else
      vis = ((cloudnode_tk[src]>0) && (cloudnode_tk[dest]));

   if (vis) {
      if (cloudlink_tk[ln]) {  /* move existing */
         sprintf(buff, ".highcanvas itemconfigure %d -fill %s -width %d",
          cloudlink_tk[ln], clcolor(ln), clwidth(ln));
         text_cmd( buff);
      } else {            /* create new */
         sprintf(buff, "mkCloudEdge %d %d %s %d", 
          cloudnode_tk[get_int("cloudlinkA",ln)],
          cloudnode_tk[get_int("cloudlinkB",ln)], 
          clcolor(ln), clwidth(ln));
         text_cmd( buff);
   
         cloudlink_tk[ln] = atoi(tcl->result);
         add_timers(VL_CLOUDLINK, ln);
      }
   } else {
      remove_cloudlink(ln); /* delete existing */
   }
}

static void
set_cloudlink_displays()
{
   int ln;
 
   for (ln=0; ln<num_cloudlinks; ln++)
      set_cloudlink_display(ln);
}

void
set_cloudlink_highlight(ln)
   int ln;
{
   char buff[255];

   if (cloudlink_tk[ln]) {
      sprintf(buff, ".highcanvas itemconfigure %d -fill %s -width %d",
       cloudlink_tk[ln], clcolor(ln), clwidth(ln));
      text_cmd( buff);
   }
}

void
set_cloudlink_highlights()
{
   int ln;
 
   for (ln=0; ln<num_cloudlinks; ln++) {
      if (!cloudlink_tk[ln]) continue;
      set_cloudlink_highlight(ln);
   }
}

void
refresh_high()
{
   /* Make sure the main window is active */
   if (!display_ok)
      return;

   if (start)
      init_high();

   /* Update node display criteria (no text or hl tho) 
    * set_cloud_displays(); - done by open_high calling mv_redraw
    */
   set_cloud_displays();

   /* Update cloud text */
   if (csmode!=CS_USERVAR || is_dirty(cscond.name))
      set_cloud_symbols();

   /* Update node highlighting */
   if (chcond.dirty || is_dirty(chcond.name)) {
      set_cloud_highlights();
      chcond.dirty = 0;
   }

   /* Fill in edges */

   /* Update edge display criteria */
   set_cloudlink_displays();

   /* Update edge highlighting */
   if (clhmode!=CLH_CUSTOM || clhcond.dirty || is_dirty(clhcond.name)) {
      set_cloudlink_highlights();
      clhcond.dirty = 0;
   }

   /*
    * Now clean stuff.  Can't clean them before since vars must stay
    * dirty for multiple criteria
    */
   if (csmode==CS_USERVAR)
      clean(cscond.name);
   clean(chcond.name);
   if (clhmode==CLH_CUSTOM)
      clean(clhcond.name);
}

int
chighlight(cloud_idx)
   int cloud_idx;
{
   return testcond(chcond.name, cloud_idx, chcond.op, &chcond.value);
}

/* Check if edge satisfies highlight criteria */
int                   /* RETURNS: 1 if true, 0 if false */
clhighlight(ln)
   int ln;
{
   int lln;

   switch(clhmode) {

   case CLH_NONE: /* highlight if normal link is highlighted */
         lln = get_int("cloudlinkLog", ln);
         return (lln>=0) && ehighlight(lln);

   case CLH_CUSTOM:
         return testcond(clhcond.name, ln, clhcond.op, &clhcond.value);

   }
   return 0;
}
