proc open_high {} {
#   toplevel .h
#   wm title .h "High-Level View"
   canvas .highcanvas \
             -width 500 -height 300 -relief sunken -bd 2\
             -xscrollincrement 20 -yscrollincrement 20 \
             -yscrollcommand ".scrolly set" -xscrollcommand ".scrollx set" 
   pack .highcanvas -in .map.bot.can -fill both -expand true
#   focus .highcanvas
   .highcanvas bind txt <Any-Enter> { .highcanvas itemconfigure current -fill red }
   .highcanvas bind txt <Any-Leave> { .highcanvas itemconfigure current -fill black }
   .highcanvas bind txt <Button-2> {
      set HcurX %x
      set HcurY %y
   }
   .highcanvas bind txt <B2-Motion> {
      moveCloud [.highcanvas find withtag current] [expr %x-$HcurX] [expr %y-$HcurY]
      set HcurX %x
      set HcurY %y
   }
   .highcanvas bind txt <Button-1> {
      global HnodeIdx
      mv_goto_cloud $HnodeIdx([.highcanvas find withtag current])
   }

   estate [.restate get]
}
proc close_high {} {
   destroy .highcanvas
}

proc mkCloud {x y str n} {
   global HnodeX HnodeY HedgeFirst HedgeSecond nodefont fontsize HnodeRect HnodeIdx
   set m [expr $fontsize / 14.0]
   set w [expr $m*[string length $str]+$m/2]
   set rect [.highcanvas create oval  [expr $x-$w/2] [expr $y-$m] [expr $x-$w/2+$w] [expr $y+$m] -fill white -tags node]
   set new [.highcanvas create text $x $y -text $str -anchor center -font $nodefont -tags txt]
   set HnodeIdx($new) $n
   set HnodeX($new) $x
   set HnodeY($new) $y
   set HedgeFirst($new) {}
   set HedgeSecond($new) {}
   set HnodeRect($new) $rect
   return $new
}

proc rmCloud {n} {
   global HnodeX HnodeY HedgeFirst HedgeSecond HnodeRect HnodeIdx
   global HedgeSrc HedgeDest
   .highcanvas delete $HnodeRect($n)
   .highcanvas delete $n
   foreach edge $HedgeFirst($n) {
       set HedgeSrc($edge) -1
   }
   foreach edge $HedgeSecond($n) {
       set HedgeDest($edge) -1
   }
   unset HnodeX($n) HnodeY($n) HedgeFirst($n) HedgeSecond($n) 
   unset HnodeRect($n) HnodeIdx($n) 
}

proc setCloudText {node str} {
   global HnodeX HnodeY HnodeRect nodefont fontsize
   set m [expr $fontsize / 14.0]
   set w [expr $m*[string length $str]+$m/2]
   .highcanvas itemconfigure $node -text $str -font $nodefont
   .highcanvas coords $HnodeRect($node) [expr $HnodeX($node)-$w/2] [expr $HnodeY($node)-10] [expr $HnodeX($node)-$w/2+$w] [expr $HnodeY($node)+10]
}

proc setCloudHl {node hl} {
   global HnodeRect
   .highcanvas itemconfigure $HnodeRect($node) \
     -width [expr 2*$hl+1] -outline [expr {($hl) ? "blue" : "black" }]
}

# Given 2 node ids, a color, and a width,
# make a new edge and return the object number
proc mkCloudEdge { first second col wid } {
   global HnodeX HnodeY HedgeFirst HedgeSecond HedgeSrc HedgeDest
   set edge [.highcanvas create line $HnodeX($first) $HnodeY($first) $HnodeX($second) $HnodeY($second) -fill $col -width $wid]
   .highcanvas lower $edge
   lappend HedgeFirst($first) $edge
   lappend HedgeSecond($second) $edge
#puts "XXX firstlist=$HedgeFirst($first)   secondlist=$HedgeSecond($second)"

   set HedgeSrc($edge) $first
   set HedgeDest($edge) $second

   return $edge
}

proc rmCloudEdge {edge} {
   global HedgeFirst HedgeSecond HedgeSrc HedgeDest

   # remove it from the edgelists of the 2 nodes
   set src $HedgeSrc($edge)
   if { $src >= 0 } {
#puts "XXX firstlist=$HedgeFirst($src)"
      set i [lsearch -exact $HedgeFirst($src) $edge]
      set HedgeFirst($src)   [lreplace $HedgeFirst($src) $i $i]
   }
   set dest $HedgeDest($edge)
   if { $dest >= 0 } {
#puts "XXX secondlist=$HedgeSecond($dest)"
      set i [lsearch -exact $HedgeSecond($dest) $edge]
      set HedgeSecond($dest) [lreplace $HedgeSecond($dest) $i $i]
   }

   .highcanvas delete $edge
   unset HedgeSrc($edge) HedgeDest($edge)
}

# Given a node id, and a delta X,Y
# Move the node by the specified amounts
proc moveCloud {node xDist yDist} {
   global HnodeX HnodeY HedgeFirst HedgeSecond HnodeRect
   .highcanvas move $node $xDist $yDist
   .highcanvas move $HnodeRect($node) $xDist $yDist
   incr HnodeX($node) $xDist
   incr HnodeY($node) $yDist
   foreach edge $HedgeFirst($node) {
      .highcanvas coords $edge $HnodeX($node) $HnodeY($node) \
                      [lindex [.highcanvas coords $edge] 2] \
                      [lindex [.highcanvas coords $edge] 3]
   }
   foreach edge $HedgeSecond($node) {
      .highcanvas coords $edge [lindex [.highcanvas coords $edge] 0] \
                      [lindex [.highcanvas coords $edge] 1] \
                      $HnodeX($node) $HnodeY($node)
   }
}
proc moveCloudTo {node x y} {
   global HnodeX HnodeY
   moveCloud $node [expr $x-$HnodeX($node)] [expr $y-$HnodeY($node)]
}

