/* $Id: link.c,v 1.3 1996/02/08 17:29:19 thalerd Exp $
 * -------------------------------------------------------------------------
 *  
 * [C] The Regents of the University of Michigan and Merit Network, Inc. 1996
 * All Rights Reserved 
 *   
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice and this permission notice appear in all 
 * copies of the software and derivative works or modified versions thereof, 
 * and that both the copyright notice and this permission and disclaimer notice 
 * appear in supporting documentation. 
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE REGENTS OF THE 
 * UNIVERSITY OF MICHIGAN AND MERIT NETWORK, INC. DO NOT WARRANT THAT THE 
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR 
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the 
 * University of Michigan and Merit Network, Inc. shall not be liable for any 
 * special, indirect, incidental or consequential damages with respect to any 
 * claim by Licensee or any third party arising from use of the software. 
 *
 * -------------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <tcl.h>
#include <tk.h>
#include "mview.h"
#include "struct.h"
#include "globals.h"
#include "database.h"
#include "display.h"
#include "link.h"

int            /* RETURNS: # potential directions in range [0..2] */
potential_directions(ln)
   int ln;     /* log link # */
{
   int dir=1;

   if (reverse(ln)>=0)
      dir++;
   return dir;
}

/*
 * Determine whether a link is up or down
 * Note: by convention, a self-loop is 0-directional
 */
int            /* RETURNS: # active directions in range [0..2] */
up_directions(ln)
   int ln;     /* log link # */
{
   log_link_t *edge, *redge;
   int dir=0;
   int rln; /* reverse link # */

   edge = get_log_link(ln);
   if (edge->src == edge->dest)
      return 0;

   if (!(edge->flags & (DVMRP_NF_DOWN | DVMRP_NF_DISABLED)))
      dir++;

   rln = reverse(ln);
   if (rln>=0) {
      redge = get_log_link(rln);
      if (!(redge->flags & (DVMRP_NF_DOWN | DVMRP_NF_DISABLED)))
         dir++;
   }

   return dir;
}

int             /* RETURNS: link # of reverse direction, -1 if one-way */
reverse(ln)
   int ln;      /* logical link # */
{
   int rln;    /* reverse link number */
   int dest;

   if (ln<0 || ln>num_log_links())
      return -1;

   dest = llink[ln].dest;
   if (dest<0 || dest>num_nodes())
      return -1;
   for (rln=mget_node(dest)->last_link[LOUT]; rln>=0; rln=llink[rln].prev_link[LOUT]){
      if (llink[rln].dest == llink[ln].src)
         return rln;
   }
   return -1;
}

void
rm_link(ln, n, l)
   int ln;   /* link number */
   int n;    /* node number */
   int l;    /* link type   */
{
   if (l==PIN || l==POUT) {
      rm_phy_link(ln);
   } else {
      rm_log_link(ln);
   }
}

int                /* RETURNS: total # of links in given direction (inc down) */
count_links(n,dir) /* ARGUMENTS:            */
   int n;          /*    Node #             */
   int dir;        /*    Direction (LIN/LOUT/PIN/POUT) */
{
   int     e, l=0;

   if (n<0) return 0;

   for (e = mget_node(n)->last_link[dir]; e>=0; e = llink[e].prev_link[dir]) 
      l++;
   return l;
}

/* Remove a link from a linked list */
int
rm_link_from_list(ln,n,l)
   int ln; /* link to remove */
   int n;  /* node index */
   int l;  /* list index */
{
   int next;
   int e; /* current link # */

   if (n<0 || n>num_nodes())
      return 0;

   next = -1;
   for (e=mget_node(n)->last_link[l]; e>=0; e=llink[e].prev_link[l]) {
      if (e==ln) {
         if (next>=0)
            llink[next].prev_link[l] = llink[e].prev_link[l];
         else
            mget_node(n)->last_link[l] = llink[e].prev_link[l];
         llink[e].prev_link[l] = -1;
      }
      next = e;
   }
   return 1;
}

int
link_in_list(ln, n, l)
   int ln;
   int n;
   int l;
{
   int e; /* current link # */

   for (e=mget_node(n)->last_link[l]; e>=0; e=llink[e].prev_link[l]) {
      if (e==ln) 
         return 1;
   }
   return 0;
}

/* Add a link to a linked list */
void
add_link_to_list(ln, n, l)
   int ln; /* link to remove */
   int n;  /* node index */
   int l;  /* list index */
{
   node_t *node = mget_node(n);

if (link_in_list(ln, n, l))
abort();

   if (l==PIN || l==POUT) {
      if ( plink[ln].prev_link[l] >= 0 )
         abort();

      plink[ln].prev_link[l] = node->last_link[l];
      node->last_link[l]   = ln;

   } else {
      if ( llink[ln].prev_link[l] >= 0 )
         abort();

      llink[ln].prev_link[l] = node->last_link[l];
      node->last_link[l]   = ln;

      if (llink[ln].prev_link[l] == ln) /* infinite loop */
         abort();
   }
}  

void
dump_links(n)
   int n;
{
   int     e;
   int dir;

   for (dir=0; dir<1; dir++) {
      printf("Dir=%d\n", dir);
      for (e = mget_node(n)->last_link[dir]; e>=0; e = llink[e].prev_link[dir]) {
         printf("Link %d -> %ld\n", e, (dir==LOUT)? llink[e].dest : llink[e].src );
      }
   }
}
