/* $Id: main.c,v 1.11 1996/06/20 23:21:33 thalerd Exp $
 * -------------------------------------------------------------------------
 *  
 * [C] The Regents of the University of Michigan and Merit Network, Inc. 1996
 * All Rights Reserved 
 *   
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice and this permission notice appear in all 
 * copies of the software and derivative works or modified versions thereof, 
 * and that both the copyright notice and this permission and disclaimer notice 
 * appear in supporting documentation. 
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE REGENTS OF THE 
 * UNIVERSITY OF MICHIGAN AND MERIT NETWORK, INC. DO NOT WARRANT THAT THE 
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR 
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the 
 * University of Michigan and Merit Network, Inc. shall not be liable for any 
 * special, indirect, incidental or consequential damages with respect to any 
 * claim by Licensee or any third party arising from use of the software. 
 *
 * -------------------------------------------------------------------------
 */
/* MAIN.C - start of MView */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <sys/types.h>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#include <sys/socket.h>
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h> 
#endif
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <tcl.h>
#include <tk.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include "struct.h"
#include "database.h"
#include "display.h"
#include "driver.h" /* for Mv_Init */
#include "node.h"
#include "handler.h"
#include "globals.h"
int debug = 0; /* DB_HANDLERS; */
int flags = 0;

int curr_cloud;
int curr_node;
struct in_addr my_addr;
#define MAX_CHAR_NET_INFO 100

/*
 * Declarations for various library procedures and variables (don't want
 * to include tkInt.h or tkPort.h here, because people might copy this
 * file out of the Tk source directory to make their own modified versions).
 * Note: don't declare "exit" here even though a declaration is really
 * needed, because it will conflict with a declaration elsewhere on
 * some systems.
 */

extern int              isatty _ANSI_ARGS_((int fd));
extern int              read _ANSI_ARGS_((int fd, char *buf, size_t size));
extern char *           strrchr _ANSI_ARGS_((CONST char *string, int c));

/*
 * Global variables used by the main program:
 */

       Tk_Window tk_main;   /* The main window for the application.  If
             * NULL then the application no longer
             * exists. */
/*    char *tcl_RcFileName = NULL;   * Name of a user-specific startup script
             * to source if the application is being run
             * interactively (e.g. "~/.wishrc").  Set
             * by Tcl_AppInit.  NULL means don't source
             * anything ever. */
static int tty;         /* Non-zero means standard input is a
             * terminal-like device.  Zero means it's
             * a file. */
#if (TK_MINOR_VERSION == 0) /* Tk4.0 */
static char errorExitCmd[] = "exit 1";
#endif

/*
 * Command-line options:
 */

static int synchronize = 0;
static char *fileName = NULL;
static char *name = NULL;
static char *display = NULL;
static char *geometry = NULL;

static Tk_ArgvInfo argTable[] = {
    {"-file", TK_ARGV_STRING, (char *) NULL, (char *) &fileName,
   "File from which to read commands"},
    {"-geometry", TK_ARGV_STRING, (char *) NULL, (char *) &geometry,
   "Initial geometry for window"},
    {"-display", TK_ARGV_STRING, (char *) NULL, (char *) &display,
   "Display to use"},
    {"-name", TK_ARGV_STRING, (char *) NULL, (char *) &name,
   "Name to use for application"},
    {"-sync", TK_ARGV_CONSTANT, (char *) 1, (char *) &synchronize,
   "Use synchronous mode for display server"},
    {(char *) NULL, TK_ARGV_END, (char *) NULL, (char *) NULL,
   (char *) NULL}
};

/*
 *----------------------------------------------------------------------
 *
 * Tk_Main --
 *
 *      Main program for Wish and most other Tk-based applications.
 *
 * Results:
 *      None. This procedure never returns (it exits the process when
 *      it's done.
 *
 * Side effects:
 *      This procedure initializes the Tk world and then starts
 *      interpreting commands;  almost anything could happen, depending
 *      on the script being interpreted.
 *
 *----------------------------------------------------------------------
 */

void
Tk_Main(argc, argv, appInitProc)
    int argc;                           /* Number of arguments. */
    char **argv;                        /* Array of argument strings. */
    Tcl_AppInitProc *appInitProc;       /* Application-specific initialization
                                         * procedure to call after most
                                         * initialization but before starting
                                         * to execute commands. */
{
#if (TK_MINOR_VERSION == 0) /* Tk4.0 */
{
    char *args, *msg, *p;
    char buf[20];
    int code;

    tcl = Tcl_CreateInterp();

    /*
     * Parse command-line arguments.
     */

    if (Tk_ParseArgv(tcl, (Tk_Window) NULL, &argc, argv, argTable, 0)
       != TCL_OK) {
   fprintf(stderr, "%s\n", tcl->result);
   exit(1);
    }
    if (name == NULL) {
   if (fileName != NULL) {
       p = fileName;
   } else {
       p = argv[0];
   }
   name = strrchr(p, '/');
   if (name != NULL) {
       name++;
   } else {
       name = p;
   }
    }

    /*
     * If a display was specified, put it into the DISPLAY
     * environment variable so that it will be available for
     * any sub-processes created by us.
     */

    if (display != NULL) {
   Tcl_SetVar2(tcl, "env", "DISPLAY", display, TCL_GLOBAL_ONLY);
    }

    /*
     * Initialize the Tk application.
     */

/*    tk_main = Tk_CreateWindowFromPath(tcl, NULL, ".", display); */
/* #if (TK_MINOR_VERSION == 1)
    tk_main = Tk_MainWindow(tcl);
*/
    tk_main = Tk_CreateMainWindow(tcl, display, name, "mview");
    if (tk_main == NULL) {
   fprintf(stderr, "%s\n", tcl->result);
   exit(1);
    }
    if (synchronize) {
   XSynchronize(Tk_Display(tk_main), True);
    }
    Tk_GeometryRequest(tk_main, 200, 200);

    /*
     * Make command-line arguments available in the Tcl variables "argc"
     * and "argv".  Also set the "geometry" variable from the geometry
     * specified on the command line.
     */

    args = Tcl_Merge(argc-1, argv+1);
    Tcl_SetVar(tcl, "argv", args, TCL_GLOBAL_ONLY);
    ckfree(args);
    sprintf(buf, "%d", argc-1);
    Tcl_SetVar(tcl, "argc", buf, TCL_GLOBAL_ONLY);
    Tcl_SetVar(tcl, "argv0", (fileName != NULL) ? fileName : argv[0],
       TCL_GLOBAL_ONLY);
    if (geometry != NULL) {
   Tcl_SetVar(tcl, "geometry", geometry, TCL_GLOBAL_ONLY);
    }

    /*
     * Set the "tcl_interactive" variable.
     */

    tty = isatty(0);
    Tcl_SetVar(tcl, "tcl_interactive",
       ((fileName == NULL) && tty) ? "1" : "0", TCL_GLOBAL_ONLY);

    /*
     * Invoke application-specific initialization.
     */

    if ((*appInitProc)(tcl) != TCL_OK) {
   fprintf(stderr, "Tcl_AppInit failed: %s\n", tcl->result);
    }

    /*
     * Set the geometry of the main window, if requested.
     */

    if (geometry != NULL) {
   code = Tcl_VarEval(tcl, "wm geometry . ", geometry, (char *) NULL);
   if (code != TCL_OK) {
       fprintf(stderr, "%s\n", tcl->result);
   }
    }

    /*
     * Loop infinitely, waiting for commands to execute.  When there
     * are no windows left, Tk_MainLoop returns and we exit.
     */

    while (tk_NumMainWindows > 0) 
    {
       while (Tk_DoOneEvent(TK_DONT_WAIT|TK_X_EVENTS|TK_IDLE_EVENTS)); 
       Tk_DoOneEvent(0);
    }

    /*
     * Don't exit directly, but rather invoke the Tcl "exit" command.
     * This gives the application the opportunity to redefine "exit"
     * to do additional cleanup.
     */

    Tcl_Eval(tcl, "exit");
    exit(1);

error:
    msg = Tcl_GetVar(tcl, "errorInfo", TCL_GLOBAL_ONLY);
    if (msg == NULL) 
       msg = tcl->result;
    fprintf(stderr, "%s\n", msg);
    Tcl_Eval(tcl, errorExitCmd);
    exit(1);         /* Needed only to prevent compiler warnings. */
}
#else /* Tk4.1 */
{
    char *args;
    char buf[20];
    Tcl_Channel errChannel;

    Tcl_FindExecutable(argv[0]);
    tcl = Tcl_CreateInterp();
#ifdef TCL_MEM_DEBUG
    Tcl_InitMemory(tcl);
#endif

    /* Parse command-line arguments.  */
    if (Tk_ParseArgv(tcl, (Tk_Window) NULL, &argc, argv, argTable, 0)
       != TCL_OK) {
   fprintf(stderr, "%s\n", tcl->result);
   exit(1);
    }

    /*
     * Make command-line arguments available in the Tcl variables "argc"
     * and "argv".
     */

    args = Tcl_Merge(argc-1, argv+1);
    Tcl_SetVar(tcl, "argv", args, TCL_GLOBAL_ONLY);
    ckfree(args);
    sprintf(buf, "%d", argc-1);
    Tcl_SetVar(tcl, "argc", buf, TCL_GLOBAL_ONLY);
    Tcl_SetVar(tcl, "argv0", (fileName != NULL) ? fileName : argv[0],
            TCL_GLOBAL_ONLY);

    /*
     * Set the "tcl_interactive" variable.
     */

    /*
     * For now, under Windows, we assume we are not running as a console mode
     * app, so we need to use the GUI console.  In order to enable this, we
     * always claim to be running on a tty.  This probably isn't the right
     * way to do it.
     */

#ifdef __WIN32__
    tty = 1;
#else
    tty = isatty(0);
#endif
    Tcl_SetVar(tcl, "tcl_interactive",
            ((fileName == NULL) && tty) ? "1" : "0", TCL_GLOBAL_ONLY);

    /*
     * Invoke application-specific initialization.
     */

    if ((*appInitProc)(tcl) != TCL_OK) {
        errChannel = Tcl_GetStdChannel(TCL_STDERR);
        if (errChannel) {
            Tcl_Write(errChannel,
                    "application-specific initialization failed: ", -1);
            Tcl_Write(errChannel, tcl->result, -1);
            Tcl_Write(errChannel, "\n", 1);
        }
    }

    /*
     * Loop infinitely, waiting for commands to execute.  When there
     * are no windows left, Tk_MainLoop returns and we exit.
     */

    while (Tk_GetNumMainWindows() > 0)
    {
       while (Tcl_DoOneEvent(TCL_DONT_WAIT|TCL_WINDOW_EVENTS|TCL_IDLE_EVENTS));
       Tcl_DoOneEvent(0);
    }
    Tcl_Exit(0);
}
#endif
}

#if (TK_MINOR_VERSION > 0)
/* needed to override tk_library lookups */
int
TkPlatformInit(interp)
    Tcl_Interp *interp;
{
   TkCreateXEventSource();
   return TCL_OK;
}
#endif

int
Tcl_AppInit(interp)
   Tcl_Interp *interp;		/* Interpreter for application. */
{
/*
   if (Tcl_Init(interp) == TCL_ERROR)
      return TCL_ERROR;
*/
#if (TK_MINOR_VERSION > 0) /* Tk4.1 */
   if (Tk_Init(interp) == TCL_ERROR)
      return TCL_ERROR;
   Tcl_StaticPackage(interp, "Tk", Tk_Init, (Tcl_PackageInitProc *) NULL);
   tk_main = Tk_MainWindow(interp);
#endif
   if (Mv_Init(interp) == TCL_ERROR)
      return TCL_ERROR;

   return TCL_OK;
}

int
main(argc, argv)
    int argc;                   /* Number of command-line arguments. */
    char **argv;                /* Values of command-line arguments. */
{
    Tk_Main(argc, argv, Tcl_AppInit);
    return 0;                   /* Needed only to prevent compiler warning. */
}

/* We now need a replacement for the tkwait command, since we want
 * to override the Tcl_DoOneEvent order and do display events before file
 * events. -dgt 4/19/96
 */
   /* ARGSUSED */
static char *
WaitVariableProc(clientData, interp, name1, name2, flags)
    ClientData clientData; /* Pointer to integer to set to 1. */
    Tcl_Interp *interp;    /* Interpreter containing variable. */
    char *name1;     /* Name of variable. */
    char *name2;     /* Second part of variable name. */
    int flags;       /* Information about what happened. */
{
    int *donePtr = (int *) clientData;

    *donePtr = 1;
    return (char *) NULL;
}

static void
WaitVisibilityProc(clientData, eventPtr)
    ClientData clientData; /* Pointer to integer to set to 1. */
    XEvent *eventPtr;      /* Information about event (not used). */
{
    int *donePtr = (int *) clientData;
    *donePtr = 1;
}

static void
WaitWindowProc(clientData, eventPtr)
    ClientData clientData; /* Pointer to integer to set to 1. */
    XEvent *eventPtr;      /* Information about event. */
{
    int *donePtr = (int *) clientData;

    if (eventPtr->type == DestroyNotify) {
   *donePtr = 1;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Tk_TkwaitCmd --
 *
 *	This procedure is invoked to process the "tkwait" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Mv_TkwaitCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Main window associated with
				 * interpreter. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
    Tk_Window tkwin = tk_main; /* (Tk_Window) clientData; */
    int c, done;
    size_t length;

    if (argc != 3) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " variable|visibility|window name\"", (char *) NULL);
	return TCL_ERROR;
    }
    c = argv[1][0];
    length = strlen(argv[1]);
    if ((c == 'v') && (strncmp(argv[1], "variable", length) == 0)
	    && (length >= 2)) {
	if (Tcl_TraceVar(interp, argv[2],
		TCL_GLOBAL_ONLY|TCL_TRACE_WRITES|TCL_TRACE_UNSETS,
		WaitVariableProc, (ClientData) &done) != TCL_OK) {
	    return TCL_ERROR;
	}
	done = 0;
	while (!done) {
#if (TK_MINOR_VERSION > 0) /* Tk4.1 */
      if (!Tcl_DoOneEvent(TCL_DONT_WAIT|TCL_WINDOW_EVENTS|TCL_IDLE_EVENTS))
	       Tcl_DoOneEvent(0);
#else /* Tk4.0 */
      if (!Tk_DoOneEvent(TK_DONT_WAIT|TK_X_EVENTS|TK_IDLE_EVENTS))
	       Tk_DoOneEvent(0);
#endif
	}
	Tcl_UntraceVar(interp, argv[2],
		TCL_GLOBAL_ONLY|TCL_TRACE_WRITES|TCL_TRACE_UNSETS,
		WaitVariableProc, (ClientData) &done);
    } else if ((c == 'v') && (strncmp(argv[1], "visibility", length) == 0)
	    && (length >= 2)) {
	Tk_Window window;

	window = Tk_NameToWindow(interp, argv[2], tkwin);
	if (window == NULL) {
	    return TCL_ERROR;
	}
	Tk_CreateEventHandler(window, VisibilityChangeMask,
	    WaitVisibilityProc, (ClientData) &done);
	done = 0;
	while (!done) {
#if (TK_MINOR_VERSION > 0) /* Tk4.1 */
      if (!Tcl_DoOneEvent(TCL_DONT_WAIT|TCL_WINDOW_EVENTS|TCL_IDLE_EVENTS))
	       Tcl_DoOneEvent(0);
#else /* Tk4.0 */
      if (!Tk_DoOneEvent(TK_DONT_WAIT|TK_X_EVENTS|TK_IDLE_EVENTS))
	       Tk_DoOneEvent(0);
#endif
	}
	Tk_DeleteEventHandler(window, VisibilityChangeMask,
	    WaitVisibilityProc, (ClientData) &done);
    } else if ((c == 'w') && (strncmp(argv[1], "window", length) == 0)) {
	Tk_Window window;

	window = Tk_NameToWindow(interp, argv[2], tkwin);
	if (window == NULL) {
	    return TCL_ERROR;
	}
	Tk_CreateEventHandler(window, StructureNotifyMask,
	    WaitWindowProc, (ClientData) &done);
	done = 0;
	while (!done) {
#if (TK_MINOR_VERSION > 0) /* Tk4.1 */
      if (!Tcl_DoOneEvent(TCL_DONT_WAIT|TCL_WINDOW_EVENTS|TCL_IDLE_EVENTS))
	       Tcl_DoOneEvent(0);
#else /* Tk4.0 */
      if (!Tk_DoOneEvent(TK_DONT_WAIT|TK_X_EVENTS|TK_IDLE_EVENTS))
	       Tk_DoOneEvent(0);
#endif
	}
	/*
	 * Note:  there's no need to delete the event handler.  It was
	 * deleted automatically when the window was destroyed.
	 */
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
		"\": must be variable, visibility, or window", (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Clear out the interpreter's result, since it may have been set
     * by event handlers.
     */

    Tcl_ResetResult(interp);
    return TCL_OK;
}
