/* $Id: maplib.c,v 1.10 1996/05/06 19:14:02 thalerd Exp $
 * -------------------------------------------------------------------------
 *  
 * [C] The Regents of the University of Michigan and Merit Network, Inc. 1996
 * All Rights Reserved 
 *   
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice and this permission notice appear in all 
 * copies of the software and derivative works or modified versions thereof, 
 * and that both the copyright notice and this permission and disclaimer notice 
 * appear in supporting documentation. 
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE REGENTS OF THE 
 * UNIVERSITY OF MICHIGAN AND MERIT NETWORK, INC. DO NOT WARRANT THAT THE 
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR 
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the 
 * University of Michigan and Merit Network, Inc. shall not be liable for any 
 * special, indirect, incidental or consequential damages with respect to any 
 * claim by Licensee or any third party arising from use of the software. 
 *
 * -------------------------------------------------------------------------
 */
/* MAPLIB.C - network library functions */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <netdb.h>
#include <sys/types.h>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif
#include <errno.h>
#ifdef HAVE_STRING_H
#include <string.h> /* for strcat */
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h> /* for ntohl */
#endif
#include <tcl.h>
#include "struct.h"
#include "database.h"
#include "node.h"
#include "driver.h" /* for text_cmd */

#undef LOG_MESSAGES

#ifdef LOG_MESSAGES
extern FILE *msgsfp;
#endif
extern int     sys_nerr;
#ifndef linux
#ifndef __FreeBSD__
extern char *     sys_errlist[];
#endif
#endif

/*
 * Convert an IP address in u_int32 (network) format into a printable string.
 */
char *
inet_fmt(addr)
    u_int32 addr;
{
static char s[16];      /* buffers to hold the string representations  */
    register u_char *a;

    a = (u_char *)&addr;
    sprintf(s, "%u.%u.%u.%u", a[0], a[1], a[2], a[3]);
    return (s);
}

char   *
inet_name(addr)
   u_int32  addr;
{
   char buff[256];

   if (!addr) return "?";
   /*dns_start(&addr, sizeof(addr), AF_INET);*/
   sprintf(buff, "mv_start nslookup(%s) nslookup %s", inet_fmt(addr), inet_fmt(addr));
   text_cmd(buff);
   return "?";
/* return e ? e->h_name : "?"; */
}

char *
approx_inet_fmts(addr, mask, s)
    u_int32 addr, mask;
    char *s;
{
    register u_char *a, *m;
    int bits;

    if ((addr == 0) && (mask == 0)) {
   sprintf(s, "default");
   return (s);
    }
    a = (u_char *)&addr;
    m = (u_char *)&mask;
    bits = 33 - ffs(ntohl(mask));

    if      (m[3] != 0) sprintf(s, "%u.%u.%u.%u/~%d", a[0], a[1], a[2], a[3],
                  bits);
    else if (m[2] != 0) sprintf(s, "%u.%u.%u/~%d",    a[0], a[1], a[2], bits);
    else if (m[1] != 0) sprintf(s, "%u.%u/~%d",       a[0], a[1], bits);
    else                sprintf(s, "%u/~%d",          a[0], bits);

    return (s);
}

/*
 * Convert an IP subnet number in u_int32 (network) format into a printable
 * string including the netmask as a number of bits.
 */
char *
inet_fmts(addr, mask, s)
    u_int32 addr, mask;
    char *s;
{
    register u_char *a, *m;
    int bits;

    if ((addr == 0) && (mask == 0)) {
   sprintf(s, "default");
   return (s);
    }
    a = (u_char *)&addr;
    m = (u_char *)&mask;
    bits = 33 - ffs(ntohl(mask));

    if      (m[3] != 0) sprintf(s, "%u.%u.%u.%u/%d", a[0], a[1], a[2], a[3],
                  bits);
    else if (m[2] != 0) sprintf(s, "%u.%u.%u/%d",    a[0], a[1], a[2], bits);
    else if (m[1] != 0) sprintf(s, "%u.%u/%d",       a[0], a[1], bits);
    else                sprintf(s, "%u/%d",          a[0], bits);

    return (s);
}

/*
 * Log errors and other messages to stderr, according to the severity of the
 * message and the current debug level.  For errors of severity LOG_ERR or
 * worse, terminate the program.
 */
void 
wlog(severity, syserr, format, a, b, c, d, e)
	int     severity, syserr;
	char   *format;
	int     a, b, c, d, e;
{
	char    fmt[100];
int debug=3;

#ifdef LOG_MESSAGES
   if (!msgsfp) return;
#endif

   switch (debug) {
   case 0: if (severity > LOG_WARNING) return;
   case 1: if (severity > LOG_NOTICE) return;
   case 2: if (severity > LOG_INFO) return;
   default: fmt[0] = '\0';
      if (severity == LOG_WARNING)
         strcat(fmt, "warning - ");
      strncat(fmt, format, 80);
#ifdef LOG_MESSAGES
      fprintf(msgsfp, fmt, a, b, c, d, e);
#endif
      if (severity <= LOG_ERR)
         fprintf(stderr, fmt, a, b, c, d, e);
#ifdef LOG_MESSAGES
      if (syserr == 0) {
         fprintf(msgsfp, "\n");
      } else if (syserr < sys_nerr) {
         fprintf(msgsfp, ": %s\n", sys_errlist[syserr]);
      } else {
         fprintf(msgsfp, ": errno %d\n", syserr);
      }
      if (severity <= LOG_ERR) {
         if (syserr == 0) {
            fprintf(msgsfp, "\n");
         } else if (syserr < sys_nerr) {
            fprintf(msgsfp, ": %s\n", sys_errlist[syserr]);
         } else {
            fprintf(msgsfp, ": errno %d\n", syserr);
         }
      }
#endif
   }

#if 0
   if (severity <= LOG_ERR)
      return; /* was exit */
#endif
}

/*
 * Tk looks for $tk_library/prolog.ps but of course we don't want to
 * rely on the existence of the tk_library.  So we tell Tk to look
 * in /tmp and write prolog.ps there long enough for Tk to generate
 * a postscript dump.
 */
int
mv_postscript(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   FILE *fp;
   static const char *file="/tmp/prolog.ps";
   extern const char tcl_prolog[];
   char buff[256];

   /* Create /tmp/prolog.ps */
   fp = fopen(file, "w");
   if (fp==NULL) 
      return TCL_OK;
   fprintf(fp, "%s", tcl_prolog);
   fclose(fp);

   Tcl_SetVar(tcl, "tk_library", "/tmp", TCL_GLOBAL_ONLY);
   sprintf(buff, "dumpPostscript %s", argv[1]);
   text_cmd(buff);
   Tcl_UnsetVar(tcl, "tk_library", TCL_GLOBAL_ONLY);
   unlink(file);
   return TCL_OK;
}
