proc set_font {x} {
   global nodefont 
   set nodefont "-*-Courier-Medium-R-Normal--*-${x}-*"
   # cb_fontsize will call refresh_display() for us
}

proc newwin {w} {
   toplevel $w
   wm group $w .
   text "$w.text" -relief raised -bd 2 -yscrollcommand "\"$w.scroll\" set"
   scrollbar "$w.scroll" -command "\"$w.text\" yview"
   pack "$w.scroll" -side right -fill y
   pack "$w.text" -side left
}

proc open_reportwin {w hdr} {
   global font
   toplevel .$w
   wm group .$w
   label .$w.header -font $font -text "$hdr"
# text .$w.text -bd 2 -relief sunken -font $font -yscrollcommand ".$w.scroll set" -state disabled
   text .$w.text -bd 2 -relief sunken -font $font -yscrollcommand ".$w.scroll set" 
   scrollbar .$w.scroll -command ".$w.text yview" -takefocus 0
   button .$w.ok -text OK -command "destroy .$w"
   pack .$w.header -side top -anchor w 
   pack .$w.ok -side bottom -fill x
   pack .$w.scroll -side right -fill y
   pack .$w.text -side left
   bind .$w <KeyPress-Return> "destroy .$w"

   wm withdraw .$w
   update idletasks
   set x [expr [winfo screenwidth .$w]/2 - [winfo reqwidth .$w]/2 \
      - [winfo vrootx [winfo parent .$w]]]
   set y [expr [winfo screenheight .$w]/2 - [winfo reqheight .$w]/2 \
      - [winfo vrooty [winfo parent .$w]]]
   wm geom .$w +$x+$y
   wm deiconify .$w

   set x [winfo width .$w]
   set y [winfo height .$w]

   wm minsize .$w $x $y
   wm maxsize .$w $x $y

}

proc close_reportwin {w} {
   .$w.text configure -state disabled
   grab .$w
   mv_tkwait window .$w
}

proc fixsum {var1 var2} {
   set sum ""
   if {$var1!=""} {
      set sum $var1
   }
   if {$var2!="" && $var2!=$var1} {
      if {( $sum != "" )} {
         append sum "/"
      }
      append sum $var2
   }
   return $sum
}

proc collect {} {
   global str 
   prompt .p {Collect Filename} {Enter the filename containing mcollect output}
   mv_collect $str
}

proc do_traceroute { } {
	global srcip destip

	if {$srcip == $destip} {
		mv_dialog .notify {Invalid traceroute} {Source and destination are the same} {} 0 OK
	} elseif { [string compare "$srcip" "Unspecified"] == 0 || [string compare "$destip" "Unspecified"] == 0 }  {
		mv_dialog .notify {Source and Dest Address} {You must first set source and destination addresses in the mview control window.} {} 0 OK
	} else {
		mv_start traceroute( $srcip ) traceroute -n -g $srcip $destip
	}
}

proc speaks_snmp { node } {
	set x [ get_value snmpaware $node ]
	if { $x == 1 } {
		return 1
	} else {
		return 0
	}
}

proc showMCastTreeForGroup { } {
	global group srcip rootip 

	if { [string compare "$group" "Unspecified"] == 0 || [string compare "$srcip" "Unspecified"] == 0 || [string compare "$rootip" "Unspecified"] == 0 }  {
		mv_dialog .notify {Root, Group and Source Addresses} {You must first set root, group and source addresses in the mview control window.} {} 0 OK
	} else {
		mv_start mcasttree mrtree -ln $rootip $group $srcip
	}
}

proc showMCastTree { } {
	global srcip rootip 

	if { [string compare "$srcip" "Unspecified"] == 0 || [string compare "$rootip" "Unspecified"] == 0 }  {
		mv_dialog .notify { Root and Source Addresses} {You must first set the root and source addresses in the mview control window.} {} 0 OK
	} else {
		mv_start mcasttree mrtree -ln $rootip $srcip
	}
}

proc showMtraceDestSrcPath { } {
	global group srcip destip 

	if { [string compare "$srcip" "Unspecified"] == 0 || [string compare "$destip" "Unspecified"] == 0 }  {
		mv_dialog .notify {Group Address} {You must first set source and destination addresses in the mview control window.  The root address should be receivers address.} {} 0 OK
	} elseif { [string compare "$srcip" "$destip"] == 0 } {
		mv_dialog .notify {Src/Dest Address} {The source and destination addresses must differ.} {} 0 OK
	} else {
		if { [string compare "$group" "Unspecified"] == 0 } {
			mv_start mtracepath mtrace -n -S 1 $srcip $destip 224.2.0.1
		} else {
			mv_start mtracepath mtrace -n -S 1 $srcip $destip $group
		}
	}
}

proc pre_pktspergroup { } {
	global group exp_index srcip 

	if { [string compare "$group" "Unspecified"] == 0 || [string compare "$srcip" "Unspecified"] == 0 }  {
		mv_dialog .notify {Source/Group Address} {You must first set source and group addresses in the mview control window.} {} 0 OK
	} else {
		mv_start pktspergroup($exp_index,$srcip) mstat -n -x $group [mv_ipstr $exp_index]
	}
}

# This procedure displays a prompt box, waits for a string to be entered,
# then returns the index of the selected button.
#
# Arguments:
# w -		Window to use for dialog top-level.
# title -	Title to display in dialog's decorative frame.
# text -	Prompt to display 
proc yesno {w title text} {
    global str

    set font "-adobe-courier-medium-r-normal--14-*-*-*-m-*-*-*"

    # 1. Create the top-level window and divide it into top
    # and bottom parts.

    catch {destroy $w}
    toplevel $w -class Dialog
    wm transient $w .
    wm title $w $title
    wm iconname $w Dialog
    frame $w.top -relief raised -bd 1
    pack $w.top -side top -fill both
    frame $w.bot -relief raised -bd 1
    pack $w.bot -side bottom -fill both

    # 2. Fill the top part with message.

    message $w.msg -width 3i -text $text \
	    -font -Adobe-Times-Medium-R-Normal-*-180-*
    pack $w.msg -in $w.top -side right -expand 1 -fill both -padx 5m -pady 5m
# 3. Create Yes/No buttons

    set str ""
    button $w.yes -text "Yes" -command "set str \"yes\"; destroy $w.bot"
    bind $w.yes <KeyPress-Return> "set str \"yes\"; destroy $w.bot"
    button $w.no -text "No" -command "set str \"no\"; destroy $w.bot"
    bind $w.no <KeyPress-Return> "set str \"no\"; destroy $w.bot"
    pack $w.yes $w.no -in $w.bot -fill x -side left -expand 1

    # 4. Withdraw the window, then update all the geometry information
    # so we know how big it wants to be, then center the window in the
    # display and de-iconify it.

    wm withdraw $w
    update idletasks
    set x [expr [winfo screenwidth $w]/2 - [winfo reqwidth $w]/2 \
	    - [winfo vrootx [winfo parent $w]]]
    set y [expr [winfo screenheight $w]/2 - [winfo reqheight $w]/2 \
	    - [winfo vrooty [winfo parent $w]]]
    wm geom $w +$x+$y
    wm deiconify $w

    # 5. Set a grab and claim the focus too.

    set oldFocus [focus]
    grab $w
    focus $w

    set x [winfo width $w]
    set y [winfo height $w]

    wm minsize $w $x $y
    wm maxsize $w $x $y

    # 6. Wait for the user to respond, then restore the focus and
    # return the index of the selected button.

    mv_tkwait window $w.bot
    destroy $w
    focus $oldFocus
    return $str
}

#
# tk_prompt:
#
# This procedure displays a prompt box, waits for a string to be entered,
# then returns the index of the selected button.
#
# Arguments:
# w -		Window to use for dialog top-level.
# title -	Title to display in dialog's decorative frame.
# text -	Prompt to display 

proc prompt {w title text} {
    global str

    set font "-adobe-courier-medium-r-normal--14-*-*-*-m-*-*-*"

    # 1. Create the top-level window and divide it into three parts

    catch {destroy $w}
    toplevel $w -class Dialog
    wm transient $w .
    wm title $w $title
    wm iconname $w Dialog
    frame $w.top -relief raised -bd 1
    pack $w.top -side top -fill both
    frame $w.mid -relief raised -bd 1
    pack $w.mid -side top -fill both
    frame $w.bot -relief raised -bd 1
    pack $w.bot -side bottom -fill both

    # 2. Fill the top part with message.

    message $w.msg -width 3i -text $text \
	    -font -Adobe-Times-Medium-R-Normal-*-180-*
    pack $w.msg -in $w.top -side right -expand 1 -fill both -padx 5m -pady 5m

    # 3. Create an extry box 

    set str ""
    button $w.ok -text "OK" -command "destroy $w.ok" -width 10
    bind $w.ok <KeyPress-Return> " destroy $w.ok "
    button $w.cancel -text "Cancel" -command "set str \"\"; destroy $w.ok" -width 10
    bind $w.cancel <KeyPress-Return> "set str \"\"; destroy $w.ok "
    entry $w.entry -font $font -textvariable str
    pack $w.entry -in $w.mid -fill x 
    pack $w.ok $w.cancel -in $w.bot -fill x -side left

    # 4. Withdraw the window, then update all the geometry information
    # so we know how big it wants to be, then center the window in the
    # display and de-iconify it.

    wm withdraw $w
    update idletasks
    set x [expr [winfo screenwidth $w]/2 - [winfo reqwidth $w]/2 \
	    - [winfo vrootx [winfo parent $w]]]
    set y [expr [winfo screenheight $w]/2 - [winfo reqheight $w]/2 \
	    - [winfo vrooty [winfo parent $w]]]
    wm geom $w +$x+$y
    wm deiconify $w

    set x [winfo width $w]
    set y [winfo height $w]

    wm minsize $w $x $y
    wm maxsize $w $x $y

    # 5. Set a grab and claim the focus too.

    set oldFocus [focus]
    grab $w
    focus $w

    # 6. Wait for the user to respond, then restore the focus and
    # return the index of the selected button.

    mv_tkwait window $w.ok
    destroy $w
    focus $oldFocus
    return $str
}

#
# Help subsystem
#
proc help {topic} {
   global helpurl
   mv_dialog .notify {MView Help} "Online help is now available at $helpurl" {} 0 OK
}

# Delete command from "Activities-in-progress" window
proc rm_aip { cmd } {
   set len [.r.hlr.list index end]

   # find idx of cmd
   set idx -1
   for {set i 0} {($idx<0) && ($i<$len)} {incr i} {
      if {[.r.hlr.list get $i] == $cmd} {
         set idx $i
      }
   }

   if {$idx>=0} {
      .r.hlr.list delete $idx
   }
}

proc root_check_and_set {} {
	global candidate_root

	set candidate_root [string trim $candidate_root]
	if { [string compare "$candidate_root" ""] == 0 } {
		return
	}

	mv_markroot $candidate_root
}

proc change_root_dlg {} {
	global candidate_root

    # 1. Create the top-level window and divide it into top
    # and bottom parts.

    toplevel .w -class Dialog
    wm transient .w .
    wm title .w "Change Root"
    wm iconname .w Dialog
	wm minsize .w 329 100
	wm maxsize .w 329 100

    frame .w.top -relief raised -bd 1
	pack .w.top -side top -fill both -expand true
    frame .w.bot -relief raised -bd 1
    pack .w.bot -side top -fill x

    # 2. Fill the top part with message and entry widget.

    message .w.top.msg -width 3i -text "Enter root's hostname or IP address"
    entry .w.top.entry -textvariable candidate_root
    pack .w.top.msg -in .w.top -side top
    pack .w.top.entry -in .w.top -side top -fill x 

    button .w.bot.ok_b -text OK -width 20 -command "root_check_and_set; set candidate_root \"\"; destroy .w.bot.ok_b"
    bind .w.bot.ok_b <KeyPress-Return> "root_check_and_set; set candidate_root \"\"; destroy .w.bot.ok_b"
    button .w.bot.cancel_b -text Cancel -width 20 -command "set candidate_root \"\"; destroy .w.bot.ok_b"
    bind .w.bot.cancel_b <KeyPress-Return> "set candidate_root \"\"; destroy .w.bot.ok_b"
    pack .w.bot.ok_b .w.bot.cancel_b -in .w.bot -side left -fill x 

    # 4. Withdraw the window, then update all the geometry information
    # so we know how big it wants to be, then center the window in the
    # display and de-iconify it.

    wm withdraw .w
    update idletasks
    set x [expr [winfo screenwidth .w]/2 - [winfo reqwidth .w]/2 \
	    - [winfo vrootx [winfo parent .w]]]
    set y [expr [winfo screenheight .w]/2 - [winfo reqheight .w]/2 \
	    - [winfo vrooty [winfo parent .w]]]
    wm geom .w +$x+$y
    wm deiconify .w

    # 5. Set a grab and claim the focus too.

    set oldFocus [focus]
    grab .w
    focus .w

    # 6. Wait for the user to respond, then restore the focus and
    # return the index of the selected button.

    mv_tkwait window .w.bot.ok_b
    destroy .w
    focus $oldFocus
}

proc dest_check_and_set {} {
	global candidate_dest

	set candidate_dest [string trim $candidate_dest]
	if { [string compare "$candidate_dest" ""] == 0 } {
		return;
	}

	mv_markdest $candidate_dest
}

proc change_dest_dlg {} {
	global candidate_dest

    toplevel .w -class Dialog
    wm transient .w .
    wm title .w "Change Destination"
    wm iconname .w Dialog
	wm minsize .w 329 100
	wm maxsize .w 329 100

    frame .w.top -relief raised -bd 1
	pack .w.top -side top -fill both -expand true
    frame .w.bot -relief raised -bd 1
    pack .w.bot -side top -fill x

    # 2. Fill the top part with message and entry widget.

    message .w.top.msg -width 3i -text "Enter destination's hostname or IP address"
    entry .w.top.entry -textvariable candidate_dest
    pack .w.top.msg -in .w.top -side top
    pack .w.top.entry -in .w.top -side top -fill x 

    button .w.bot.ok_b -text OK -width 20 -command "dest_check_and_set; set candidate_dest \"\"; destroy .w.bot.ok_b"
    bind .w.bot.ok_b <KeyPress-Return> "dest_check_and_set; set candidate_dest \"\"; destroy .w.bot.ok_b"
    button .w.bot.cancel_b -text Cancel -width 20 -command "set candidate_dest \"\"; destroy .w.bot.ok_b"
    bind .w.bot.cancel_b <KeyPress-Return> "set candidate_dest \"\"; destroy .w.bot.ok_b"
    pack .w.bot.ok_b .w.bot.cancel_b -in .w.bot -side left -fill x 

    wm withdraw .w
    update idletasks
    set x [expr [winfo screenwidth .w]/2 - [winfo reqwidth .w]/2 \
	    - [winfo vrootx [winfo parent .w]]]
    set y [expr [winfo screenheight .w]/2 - [winfo reqheight .w]/2 \
	    - [winfo vrooty [winfo parent .w]]]
    wm geom .w +$x+$y
    wm deiconify .w

    set oldFocus [focus]
    grab .w
    focus .w

    mv_tkwait window .w.bot.ok_b
    destroy .w
    focus $oldFocus
}

proc query_root_for_srcs_cb {} {
	global group rootip

	if { [string compare "$group" "Unspecified"] == 0 } {
		mv_dialog .notify {Group Address Not Set} {You must first set the root and group addresses.} {} 0 OK 
	} elseif { [string compare "$rootip" "Unspecified"] == 0 } {
		mv_dialog .notify {Group Address Not Set} {You must first set the root and group addresses.} {} 0 OK 
	} else {
		mv_start grpsrcs mstat -nx $group $rootip
	}
}

proc change_src_dlg { title } {
	global group

	toplevel .w -class Dialog
	wm minsize .w 300 350
    wm title .w $title
	wm iconname .w $title
    wm transient .w .

	frame .w.0
	frame .w.1 -height 100m -relief raised -bd 1
	frame .w.m -height 20m -relief raised -bd 1
	frame .w.b -relief raised -bd 1
	pack .w.0 -in .w -side top -fill x -expand false -anchor n
	pack .w.1 -in .w -side top -fill both -expand true
	pack .w.m -in .w -side top -fill x -expand false -anchor n
	pack .w.b -in .w -side top -fill x -expand false -anchor n

	message .w.0.l -justify center -aspect 350 -text "Press button to query current root for sources or type hostname or IP address in entry blank below."
	button .w.0.query_b -text "Query Root for Sources mcasting to Group" -command { query_root_for_srcs_cb }
	bind .w.0.query_b <KeyPress-Return> {mv_start grpsrcs mstat -nx $group $rootip}
	pack .w.0.l .w.0.query_b -in .w.0 -side top -fill x

	listbox .w.1.box -relief sunken -bd 2 -height 10 -yscrollcommand ".w.1.sb set"
	bind .w.1.box <KeyPress-Return> { set m [%W curselection]; if {$m!=""} { set candidate_src [%W get $m] }}
	bind .w.1.box <ButtonRelease-1> { set m [%W curselection]; if {$m!=""} { set candidate_src [%W get $m] }}

	scrollbar .w.1.sb -command ".w.1.box yview" -takefocus 0
	pack .w.1.sb -in .w.1 -side right -fill y
	pack .w.1.box -in .w.1 -side left -fill both -expand true

	entry .w.m.e -relief sunken -textvariable candidate_src
	pack .w.m.e -in .w.m -side bottom -fill x -expand true -anchor n

	button .w.b.ok_b -text "OK" -command "src_check_and_set; set candidate_src \"\"; destroy .w.b.ok_b;"
	bind .w.b.ok_b <KeyPress-Return> "src_check_and_set; set candidate_src \"\"; destroy .w.b.ok_b"
	button .w.b.cancel_b -text "Cancel" -command "set candidate_src \"\"; destroy .w.b.ok_b"
	bind .w.b.cancel_b <KeyPress-Return> "set candidate_src \"\"; destroy .w.b.ok_b"
	pack .w.b.ok_b .w.b.cancel_b -in .w.b -side left -fill both -expand true -anchor center

	wm withdraw .w
	update idletasks
	set x [expr [winfo screenwidth .w]/2 - [winfo reqwidth .w]/2 \
		- [winfo vrootx [winfo parent .w]]]
	set y [expr [winfo screenheight .w]/2 - [winfo reqheight .w]/2 \
		- [winfo vrooty [winfo parent .w]]]
	wm geom .w +$x+$y
	wm deiconify .w

	set oldFocus [focus]
	grab .w
	focus .w

	mv_tkwait window .w.b.ok_b
	destroy .w
	focus $oldFocus
}


proc grp_check_and_set {} {
	global candidate_group group rootip

	set candidate_group [string trim $candidate_group]
	if { [string compare "$candidate_group" ""] != 0 } {
		set group $candidate_group
		mv_start grpsrcs($group) mstat -nx $group $rootip
	}
}

proc src_check_and_set {} {
	global candidate_src

	set candidate_src [string trim $candidate_src]
	if { [string compare "$candidate_src" ""] == 0 } {
		return
	}

	mv_marksrc $candidate_src
}

proc query_root_for_groups_cb {} {
	global rootip

	if { [string compare "$rootip" "Unspecified"] == 0 } {
		mv_dialog .notify {Group Address Not Set} {You must first set the root address.} {} 0 OK 
	} else {
		mv_start mgroups mstat -ng $rootip
	}
}

proc change_grp_dlg { title } {
	global rootip

	toplevel .w -class Dialog
	wm minsize .w 300 350
    wm title .w $title
	wm iconname .w $title
    wm transient .w .

	frame .w.0
	frame .w.1 -height 100m -relief raised -bd 1
	frame .w.m -height 20m -relief raised -bd 1
	frame .w.b -relief raised -bd 1
	pack .w.0 -in .w -side top -fill x -expand false -anchor n
	pack .w.1 -in .w -side top -fill both -expand true
	pack .w.m -in .w -side top -fill x -expand false -anchor n
	pack .w.b -in .w -side top -fill x -expand false -anchor n


	if { [string compare "$rootip" "Unspecified"] == 0 } {
		message .w.0.l -justify center -aspect 350 -text "Press button to query current root for groups or type group address in entry blank below.\n(Set the root node from the Control Window to enable the Query button below.)"
		button .w.0.query_b -text "Query Root for Groups" -state disabled
	} else {
		message .w.0.l -justify center -aspect 350 -text "Press button to query current root for groups or type group address in entry blank below."
		button .w.0.query_b -text "Query Root for Groups" -command { query_root_for_groups_cb }
		bind .w.0.query_b <KeyPress-Return> { query_root_for_groups_cb }
	}

	pack .w.0.l .w.0.query_b -in .w.0 -side top -fill x

	listbox .w.1.box -relief sunken -bd 2 -height 10 -yscrollcommand ".w.1.sb set"
	bind .w.1.box <KeyPress-Return> { set m [%W curselection]; if {$m!=""} { set candidate_group [%W get $m] }}
	bind .w.1.box <ButtonRelease-1> { set m [%W curselection]; if {$m!=""} { set candidate_group [%W get $m] }}

	scrollbar .w.1.sb -command ".w.1.box yview" -takefocus 0
	pack .w.1.sb -in .w.1 -side right -fill y
	pack .w.1.box -in .w.1 -side left -fill both -expand true

	entry .w.m.e -relief sunken -textvariable candidate_group
	pack .w.m.e -in .w.m -side bottom -fill x -expand true -anchor n

	button .w.b.ok_b -text "OK" -command "grp_check_and_set; set candidate_group \"\"; destroy .w.b.ok_b"
	bind .w.b.ok_b <KeyPress-Return> "grp_check_and_set; set candidate_group \"\"; destroy .w.b.ok_b"
	button .w.b.cancel_b -text "Cancel" -command "set candidate_group \"\"; destroy .w.b.ok_b"
	bind .w.b.cancel_b <KeyPress-Return> "set candidate_group \"\"; destroy .w.b.ok_b"
	pack .w.b.ok_b .w.b.cancel_b -in .w.b -side left -fill both -expand true -anchor center

	wm withdraw .w
	update idletasks
	set x [expr [winfo screenwidth .w]/2 - [winfo reqwidth .w]/2 \
		- [winfo vrootx [winfo parent .w]]]
	set y [expr [winfo screenheight .w]/2 - [winfo reqheight .w]/2 \
		- [winfo vrooty [winfo parent .w]]]
	wm geom .w +$x+$y
	wm deiconify .w

	set oldFocus [focus]
	grab .w
	focus .w

	mv_tkwait window .w.b.ok_b
	destroy .w
	focus $oldFocus
}

proc mode_callback { widget } {
	global autoupd mode_text

	if { $autoupd==0} {
		set mode_text "Toggle to Online Mode"
# First do the Node Color Criteria
		.m.6.1.lbl config -fg grey
		.m.6.1.1.red_b config -bg #d9d9d9 -fg grey
		.m.6.1.1.red_lbl config -fg grey
		.m.6.1.2.yellow_b config -bg #d9d9d9 -fg grey
		.m.6.1.2.yellow_lbl config -fg grey
		.m.6.1.3.green_b config -bg #d9d9d9 -fg grey
		.m.6.1.3.green_lbl config -fg grey

# Now do the Link Color Criteria
		.m.6.2.lbl config -fg grey
		.m.6.2.1.red_b config -bg #d9d9d9 -fg grey
		.m.6.2.1.red_lbl config -fg grey
		.m.6.2.2.yellow_b config -bg #d9d9d9 -fg grey
		.m.6.2.2.yellow_lbl config -fg grey
		.m.6.2.3.green_b config -bg #d9d9d9 -fg grey
		.m.6.2.3.green_lbl config -fg grey
	} else {
# First do the Node Color Criteria
		set mode_text "Toggle to Offline Mode"
		.m.6.1.lbl config -foreground Black
		.m.6.1.1.red_b config -bg red  -fg black
		.m.6.1.1.red_lbl config -fg Black
		.m.6.1.2.yellow_b config -bg yellow -fg black
		.m.6.1.2.yellow_lbl config -foreground Black
		.m.6.1.3.green_b config -bg green -fg black
		.m.6.1.3.green_lbl config -fg Black

# Now do the Link Color Criteria
		.m.6.2.lbl config -fg Black
		.m.6.2.1.red_b config -bg red -fg black
		.m.6.2.1.red_lbl config -fg Black
		.m.6.2.2.yellow_b config -bg yellow -fg black
		.m.6.2.2.yellow_lbl config -fg Black
		.m.6.2.3.green_b config -bg green -fg black
		.m.6.2.3.green_lbl config -fg Black
	}
	
}

proc gen_control {} {
	global group srcname srcip destname destip rootname rootip
	global candidate_group candidate_dest candidate_src candidate_root
	global mode_text autoupd nyellow_crit_txt ngreen_crit_txt nred_crit_txt
	global lyellow_crit_txt lgreen_crit_txt lred_crit_txt

	set autoupd 0
	set mode_text "Toggle to Online Mode"
	set nyellow_crit_txt [ mv_NodeYellowCrit_text ]
	set ngreen_crit_txt [ mv_NodeGreenCrit_text ]
	set nred_crit_txt [ mv_NodeRedCrit_text ]
	set lyellow_crit_txt [ mv_LinkYellowCrit_text ]
	set lgreen_crit_txt [ mv_LinkGreenCrit_text ]
	set lred_crit_txt [ mv_LinkRedCrit_text ]

	set group "224.2.0.1"
	set candidate_group ""
	
	set destname       "Unspecified"
	set destip         "Unspecified"
	set candidate_dest ""
	
	set srcname        "Unspecified"
	set srcip          "Unspecified"
	set candidate_src  ""
	
	set rootname       "Unspecified"
	set rootip         "Unspecified"
	set candidate_root ""
	
	toplevel .m
	wm title .m "MView Control"
	wm minsize .m 512 272
	wm maxsize .m 512 272
	
	frame .m.1
	frame .m.2
	frame .m.3
	frame .m.4
	frame .m.5 -relief raised -bd 1
	frame .m.6
	frame .m.6.1 -relief raised -bd 1
	frame .m.6.2 -relief raised -bd 1
	frame .m.6.1.1
	frame .m.6.1.2
	frame .m.6.1.3
	frame .m.6.2.1
	frame .m.6.2.2
	frame .m.6.2.3
	
	pack .m.1 .m.2 .m.3 .m.4 -in .m -side top -anchor w
	pack .m.5 -in .m -side top -anchor w -fill x -expand true
	pack .m.6 -in .m -side top -anchor w -fill x -expand true
	pack .m.6.1 .m.6.2 -in .m.6 -side left -anchor w -fill x -expand true
	
	button .m.1.set_root_b -text "Change Root..." -width 12 -command { change_root_dlg }
	label .m.1.lbl -text "Root:"  -width 7 -anchor w
	label .m.1.ip -textvariable rootip  -width 17 -anchor w
	label .m.1.hn -textvariable rootname -width 30 -anchor w
	pack .m.1.set_root_b .m.1.lbl .m.1.ip .m.1.hn -in .m.1 -side left
	
	button .m.2.set_grp_b -text "Change Group..." -width 12 -command { change_grp_dlg "Change Group" }
	label .m.2.lbl -text "Group:"  -width 7 -anchor w
	label .m.2.ip -width 17 -anchor w -textvariable group
	pack .m.2.set_grp_b .m.2.lbl .m.2.ip -in .m.2 -side left
	
	button .m.3.set_src_b -text "Change Source..." -width 12 -command { change_src_dlg "Change Source" }
	label .m.3.lbl -text "Source:"  -width 7 -anchor w
	label .m.3.ip -textvariable srcip -width 17 -anchor w
	label .m.3.hn -textvariable srcname -width 30 -anchor w
	pack .m.3.set_src_b .m.3.lbl .m.3.ip .m.3.hn -in .m.3 -side left
	
	button .m.4.set_dest_b -text "Change Dest..." -width 12 -command { change_dest_dlg }
	label .m.4.lbl -text "Dest:"  -width 7 -anchor w
	label .m.4.ip -textvariable destip -width 17 -anchor w
	label .m.4.hn -textvariable destname -width 30 -anchor w
	pack .m.4.set_dest_b .m.4.lbl .m.4.ip .m.4.hn -in .m.4 -side left
	
	checkbutton .m.5.checkbutton -anchor w -textvariable mode_text -variable autoupd -command "mode_callback .m.5.checkbutton"
	pack .m.5.checkbutton -in .m.5 -side left -fill x -expand true

	label .m.6.1.lbl -text "Node Color Criteria" -fg grey
	pack .m.6.1.lbl -in .m.6.1 -side top -fill x -expand true
	pack .m.6.1.1 .m.6.1.2 .m.6.1.3 -in .m.6.1 -side top -anchor w -fill x -expand true

	button .m.6.1.1.red_b -text "Red..." -width 7 -fg grey -activebackground red -command mv_nodeRed
	label .m.6.1.1.red_lbl -textvariable nred_crit_txt  -anchor w -fg grey
	pack .m.6.1.1.red_b .m.6.1.1.red_lbl -in .m.6.1.1 -side left

	button .m.6.1.2.yellow_b -text "Yellow..." -width 7 -fg grey -activebackground yellow -command mv_nodeYellow
	label .m.6.1.2.yellow_lbl -textvariable nyellow_crit_txt  -anchor w -fg grey
	pack .m.6.1.2.yellow_b .m.6.1.2.yellow_lbl -in .m.6.1.2 -side left

	button .m.6.1.3.green_b -text "Green..." -width 7 -fg grey -activebackground green -command mv_nodeGreen
	label .m.6.1.3.green_lbl -textvariable ngreen_crit_txt  -anchor w -fg grey
	pack .m.6.1.3.green_b .m.6.1.3.green_lbl -in .m.6.1.3 -side left

	label .m.6.2.lbl -text "Link Color Criteria" -fg grey
	pack .m.6.2.lbl -in .m.6.2 -side top -fill x -expand true
	pack .m.6.2.1 .m.6.2.2 .m.6.2.3 -in .m.6.2 -side top -anchor w -fill x -expand true

	button .m.6.2.1.red_b -text "Red..." -width 7 -fg grey -activebackground red -command mv_linkRed
	label .m.6.2.1.red_lbl -textvariable lred_crit_txt  -anchor w -fg grey
	pack .m.6.2.1.red_b .m.6.2.1.red_lbl -in .m.6.2.1 -side left

	button .m.6.2.2.yellow_b -text "Yellow..." -width 7 -fg grey -activebackground yellow -command mv_linkYellow
	label .m.6.2.2.yellow_lbl -textvariable lyellow_crit_txt  -anchor w -fg grey
	pack .m.6.2.2.yellow_b .m.6.2.2.yellow_lbl -in .m.6.2.2 -side left

	button .m.6.2.3.green_b -text "Green..." -width 7 -fg grey -activebackground green -command mv_linkGreen
	label .m.6.2.3.green_lbl -textvariable lgreen_crit_txt  -anchor w -fg grey
	pack .m.6.2.3.green_b .m.6.2.3.green_lbl -in .m.6.2.3 -side left
}


proc mv_dialog {w title text bitmap default args} {
 global tkPriv
 catch {destroy $w}
 toplevel $w -class Dialog
 wm title $w $title
 wm iconname $w Dialog
 wm protocol $w WM_DELETE_WINDOW { }
 wm transient $w [winfo toplevel [winfo parent $w]]
 frame $w.top -relief raised -bd 1
 pack $w.top -side top -fill both
 frame $w.bot -relief raised -bd 1
 pack $w.bot -side bottom -fill both
 label $w.msg -wraplength 3i -justify left -text $text \
 -font -Adobe-Times-Medium-R-Normal--*-180-*-*-*-*-*-*
 pack $w.msg -in $w.top -side right -expand 1 -fill both -padx 3m -pady 3m
 if {$bitmap != ""} {
  label $w.bitmap -bitmap $bitmap
  pack $w.bitmap -in $w.top -side left -padx 3m -pady 3m
 }
 set i 0
 foreach but $args {
  button $w.button$i -text $but -command "set tkPriv(button) $i"
  if {$i == $default} {
   frame $w.default -relief sunken -bd 1
   raise $w.button$i $w.default
   pack $w.default -in $w.bot -side left -expand 1 -padx 3m -pady 2m
   pack $w.button$i -in $w.default -padx 2m -pady 2m
   bind $w <Return> "$w.button$i flash; set tkPriv(button) $i"
  } else {
   pack $w.button$i -in $w.bot -side left -expand 1 \
   -padx 3m -pady 2m
  }
  incr i
 }

 wm withdraw $w
 update idletasks
 set x [expr [winfo screenwidth $w]/2 - [winfo reqwidth $w]/2 \
 - [winfo vrootx [winfo parent $w]]]
 set y [expr [winfo screenheight $w]/2 - [winfo reqheight $w]/2 \
 - [winfo vrooty [winfo parent $w]]]
 wm geom $w +$x+$y
 wm deiconify $w
 set oldFocus [focus]
 grab $w
 mv_tkwait visibility $w
 if {$default >= 0} {
  focus $w.button$default
 } else {
  focus $w
 }

 set x [winfo width $w]
 set y [winfo height $w]

 wm minsize $w $x $y
 wm maxsize $w $x $y

 mv_tkwait variable tkPriv(button)
 destroy $w
 catch {focus $oldFocus}
 return $tkPriv(button)
}
