/* $Id: mviewrc.c,v 1.9 1996/06/20 23:21:29 thalerd Exp $
 * -------------------------------------------------------------------------
 *  
 * [C] The Regents of the University of Michigan and Merit Network, Inc.1996 
 * All Rights Reserved 
 *   
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice and this permission notice appear in all 
 * copies of the software and derivative works or modified versions thereof, 
 * and that both the copyright notice and this permission and disclaimer notice 
 * appear in supporting documentation. 
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE REGENTS OF THE 
 * UNIVERSITY OF MICHIGAN AND MERIT NETWORK, INC. DO NOT WARRANT THAT THE 
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR 
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the 
 * University of Michigan and Merit Network, Inc. shall not be liable for any 
 * special, indirect, incidental or consequential damages with respect to any 
 * claim by Licensee or any third party arising from use of the software. 
 *
 * -------------------------------------------------------------------------
 */
/*
 * This module executes the system mviewrc, 
 * then the user's .mviewrc file
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAVE_STRING_H
#include <string.h> /* for strcat */
#endif
#include <tcl.h>
#include <tk.h>
#include "mview.h"
#include "struct.h"
#include "globals.h"
#include "display.h" /* for do_tcl */
#include "xalloc.h" /* for xalloc */

static int
file_exists(path)
   char *path;
{
   struct stat st;
  
   return !stat(path, &st);
}

/* Execute system mviewrc:
 *
 * Method A) Global system mviewrc (in-house)
 *        B) Compile mviewrc into mview (distribution)
 */
void
tcl_exec(str)
   const char str[];
{
   int code;
   int len = strlen(str)+1;
   char *ptr = (char *)xalloc(0, len);

   strcpy(ptr, str);
   code = Tcl_Eval(tcl, ptr);
   xfree_string(ptr);

   if (code != TCL_OK) {
      ptr = Tcl_GetVar(tcl, "errorInfo", TCL_GLOBAL_ONLY);
      if (ptr == NULL)
         ptr = tcl->result;
      fprintf(stderr, "%s\n", ptr);
   }
}

extern const char tcl_canvas[],
                  tcl_high[],
                  tcl_lib[],
                  tcl_misc[],
                  tcl_vars[],
                  tcl_mviewrc[];

static void 
system_mviewrc()
{
   tcl_exec(tcl_lib);
   tcl_exec(tcl_canvas);
   tcl_exec(tcl_high);
   tcl_exec(tcl_misc);
   tcl_exec(tcl_vars);
   tcl_exec(tcl_mviewrc);
}

/* Execute $HOME/.mviewrc */
static void
user_mviewrc()
{
   char buff[512],path[256], *str;

   /* Initialize path */
   path[0]='\0';
/*
   if ((str=getenv("MVIEW_DIR")))
      strcpy(path, str);
   else 
*/
   if ((str=getenv("HOME")))
      strcpy(path, str);
   if (path[0] && path[strlen(path)-1]!='/')
      strcat(path, "/");
   strcat(path, ".mviewrc");

   /* Execute TCL command */
   if (file_exists(path)) {
      sprintf(buff, "source %s", path);
      text_cmd(buff);
      /* printf("Using %s as user rc file...\n", path); */
   }
}
   
void
source_mviewrc()
{
   system_mviewrc();
   user_mviewrc();
}
