/* $Id: queue.c,v 1.5 1996/07/02 22:38:55 thalerd Exp $
 * -------------------------------------------------------------------------
 *  
 * [C] The Regents of the University of Michigan and Merit Network, Inc. 1996
 * All Rights Reserved 
 *   
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice and this permission notice appear in all 
 * copies of the software and derivative works or modified versions thereof, 
 * and that both the copyright notice and this permission and disclaimer notice 
 * appear in supporting documentation. 
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE REGENTS OF THE 
 * UNIVERSITY OF MICHIGAN AND MERIT NETWORK, INC. DO NOT WARRANT THAT THE 
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR 
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the 
 * University of Michigan and Merit Network, Inc. shall not be liable for any 
 * special, indirect, incidental or consequential damages with respect to any 
 * claim by Licensee or any third party arising from use of the software. 
 *
 * -------------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <tcl.h>
#include <tk.h>
#include "mview.h"
#include "queue.h"
#include "node.h" /* for update_node */

static int qdebug = 0;

/************************************** QUEUES *******************************/
static int update_queue[NUM_QUEUES][MAX_NODES];
static int update_push_idx[NUM_QUEUES];
static int update_pop_idx[NUM_QUEUES];

void
init_queues()
{
   int i;
   for (i=0; i<NUM_QUEUES; i++)
      update_push_idx[i] = update_pop_idx[i] = 0;
}

int /* RETURNS: index in queue, or -1 if not found */
find_in_queue(q,n)
   int q;
   int n;
{
   int i;

   /* Check if already on the queue */
   for (i=update_pop_idx[q]; i!=update_push_idx[q]; i=(i+1)%MAX_NODES) {
      if (update_queue[q][i]==n)
         return i;
   }
   return -1;
}

int /* RETURNS: 1 on success, 0 on failure */
push_queue(q,n)
   int q;
   int n;
{
   int go = 0;
   int i;

   if ((update_push_idx[q]+1)%MAX_NODES == update_pop_idx[q])
      return 0;

   /* Check if already on the queue */
   if (find_in_queue(q,n)>=0)
      return 1;
   
   /* If this will be the first entry, then we go ahead and start the 
    * update 
    */
   if (pop_queue(q,1)<0)
      go = 1;

   update_queue[q][update_push_idx[q]] = n;
   update_push_idx[q] = (update_push_idx[q]+1)%MAX_NODES;

   if (qdebug)
      fprintf(stderr, "pushed %d on Q%d\n", n, q);

   if (go) {
      i = pop_queue(q,1);
      switch(q) {
      case Q_MRINFO   : update_node(i); break;
      case Q_NSLOOKUP : name_lookup(i); break;
      case Q_AUTOSYS  : autosys_lookup(i);   break;
      case Q_SNMPAWARE: snmpaware_lookup(i); break;
      case Q_ASNAME   : asname_lookup(i); break;
      }
   }
   return 1;
}

int /* RETURNS: node to update, or -1 if none */
pop_queue(q,peek)
   int q;
   int peek;
{
   int n;
   int i;

   /* Advance over deleted nodes */
   while (update_pop_idx[q]!=update_push_idx[q] &&
    update_queue[q][update_pop_idx[q]]== -1)
      update_pop_idx[q] = (update_pop_idx[q]+1)%MAX_NODES;
   if (update_pop_idx[q]==update_push_idx[q])
      return -1;

   /* Get the node */
   n = update_queue[q][update_pop_idx[q]];

   if (!peek) {

      /* Do the actual pop, and skip deleted nodes */
      if (qdebug)
         fprintf(stderr, "popped %d off Q%d\n", n, q);
      do {
         update_pop_idx[q] = (update_pop_idx[q]+1)%MAX_NODES;
      } while (update_pop_idx[q]!=update_push_idx[q] &&
       update_queue[q][update_pop_idx[q]]== -1);

      if (update_pop_idx[q]!=update_push_idx[q]) {
         i = update_queue[q][update_pop_idx[q]];

         switch(q) {
         case Q_MRINFO   : update_node(i); break;
         case Q_NSLOOKUP : name_lookup(i); break;
         case Q_AUTOSYS  : autosys_lookup(i);   break;
         case Q_SNMPAWARE: snmpaware_lookup(i); break;
         case Q_ASNAME   : asname_lookup(i); break;
         }
      }
   }
   return n;
}

/* Remove n from queue q */
void
delete_one_queue(q,n)
   int q;
   int n;
{
   int i;

   i = find_in_queue(q,n);
   if (i<0) return;;
   update_queue[q][i]= -1;
}

/* Remove n from all queues */
void
delete_queue(n)
   int n;
{
   int q,i;

   for (q=0; q<NUM_QUEUES; q++) {
      i = find_in_queue(q,n);
      if (i<0) continue;

      update_queue[q][i]= -1;
   }
}

/********************************* END QUEUES ********************************/

