#include <sys/types.h>
#include "mview.h"

/* Information about a logical MBONE link (tunnel) */
typedef struct {
   int32   src;       /* source node # */
   int32   sface;     /* source iface idx */
   int32   dest;      /* destination node # */
   int32   dface;     /* destination iface idx */
   u_int32 flags;
   time_t  timestamp;
   int32   hops;      /* # physical hops in link */
   int32   prev_link[4];
} log_link_t;

/* Information about a physical link */
typedef struct {
   int32   src;     /* source node # */
   int32   sface;   /* source iface idx */
   int32   dest;    /* destination node # */
   int32   dface;   /* destination iface idx */
   u_int32 logic[MAX_LOG_PER_PHY_LINK];
   time_t  timestamp;
   int32   prev_link[4];
} phy_link_t;

#define LIN 0
#define LOUT 1
#define PIN 2
#define POUT 3
#define Inlink link[LIN]
#define Outlink link[LOUT]
#define PInlink link[PIN]
#define POutlink link[POUT]

/* Information about a node (router or mrouted either) */
typedef struct {
   u_int32 iface[MAX_IFACE];        /* IP addresses */
   int32   last_link[4]; /* link #'s */
   u_int32 status;
   u_int32 obsolete; /* was version */
   time_t  timestamp;
                    /* INDEX                         VALUE
                       0 = logical component         -1 = uninitialized
                       1 = logical tree               0 = none
                       2 = physical component        >0 = number is tree/comp
                       3 = physical tree
                     */
   float   x[4],    /* Signed x,y coords for comp (see above for index) */
           y[4]; 
} node_t;

#define Xsub(idx) x[idx]
#define Ysub(idx) y[idx]

typedef struct {
   int   type;
   union {
      int    i;
      double f;
      char   s[80];
   } v;
} entity_t;

typedef struct {
   char     name[80];
   int      op;
   entity_t value;
   int      dirty;  /* dirty? */
} cond_t;

typedef struct {
   char     name1[80];
   char     name2[80];
   int      dirty;  /* dirty? */
} var_t;

typedef struct {
   char  *text;
   int  (*func)();
} command_t;

typedef struct {
   char  *name;
   void (*func)();
} handler_name_t;

typedef struct {
   char *name;
   int   type;
   int   eltsize;
   int   arrsize;
   int   flags;
   int   linked;
   int   maxage;
   char *updcmd;
   char *depends;
} variable_t;
