/* $Id: tree.c,v 1.6 1996/06/20 23:20:32 thalerd Exp $
 * -------------------------------------------------------------------------
 *  
 * [C] The Regents of the University of Michigan and Merit Network, Inc. 1996
 * All Rights Reserved 
 *   
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice and this permission notice appear in all 
 * copies of the software and derivative works or modified versions thereof, 
 * and that both the copyright notice and this permission and disclaimer notice 
 * appear in supporting documentation. 
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE REGENTS OF THE 
 * UNIVERSITY OF MICHIGAN AND MERIT NETWORK, INC. DO NOT WARRANT THAT THE 
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR 
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the 
 * University of Michigan and Merit Network, Inc. shall not be liable for any 
 * special, indirect, incidental or consequential damages with respect to any 
 * claim by Licensee or any third party arising from use of the software. 
 *
 * -------------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <tk.h>
#include "mview.h"
#include "struct.h"
#include "globals.h"
#include "database.h"
#include "display.h"
/* Compute screen locations necessary to draw cloud as a subtree
 * rooted at a given node
 */

int mark[MAX_NODES];

void
locate_subtree(root, y, x, xrange)
   int root;
   float y;
   float x;
   float xrange;
{
   node_t *node;
   int count, n;
   int ln;
   log_link_t *edge;

   /* First set location of root */
   node = mget_node(root);

   set_virtual_position(root, x + xrange/2.0, y);
   mark[root] = 1;

   /* Compute # of links downward */
   count=0;
   for (ln = node->last_link[LOUT]; ln>=0; ln = llink[ln].prev_link[LOUT]) {
      if (is_ndisplayed(llink[ln].dest) && !mark[ llink[ln].dest ]
       && up_directions(ln)>0 )
         count++;
   }

   /* Recursively locate each one */
   n = 0;
   for (ln = node->last_link[LOUT]; ln>=0; ln = llink[ln].prev_link[LOUT]) {
      edge = get_log_link(ln);
      if (is_ndisplayed(edge->dest) && !mark[edge->dest]
       && up_directions(ln)>0 ) {
         locate_subtree(edge->dest, y+100.0, x+n*(xrange/count), 
          xrange/count);
         n++;
      }
   }

   /* If there were loops, try to make things look slightly better by
    * centering the root above its children 
    */
   if (n<count) {
      xrange = n*(xrange/count);
      set_virtual_position(root, x + xrange/2.0, y);
   }
}

void
locate_tree(root)
   int root;      /* node number of root */
{
   int v;
   
   /* Unmark everything */
   for (v=0; v<num_nodes(); v++) {
      mark[v]=0;
   }

   locate_subtree(root, 0.0, 0.0, 1000.0);
}
