proc setCloudVar {} {
   global cloudvar
   toplevel .cloud -class Dialog
   wm title .cloud "Select cloud variable"
   wm transient .cloud .

   frame .cloud.f -relief raised -bd 1
   pack .cloud.f -in .cloud -side top

   # Variable selection
   scrollbar .cloud.scrollbar -command ".cloud.list yview"
   pack .cloud.scrollbar -in .cloud.f -side left -fill y
   listbox .cloud.list -relief sunken -borderwidth 2 -yscrollcommand ".cloud.scrollbar set"
   pack .cloud.list -in .cloud.f -side left -fill y
   bind .cloud.list <Button-1> {set cloudvar [.cloud.list get @%x,%y]}

   # Fill in lists with variable names
   foreach var [lsort [mv_varlist node]] {
      .cloud.list insert end $var
   }

   # Frame around all buttons
   frame .cloud.bot -relief raised -bd 1
   pack .cloud.bot -in .cloud -side bottom -fill both -expand true

   # OK button
   frame .cloud.default -relief sunken -bd 1
   button .cloud.ok -text OK -command {mv_setCloudVar $cloudvar; destroy .cloud}
   button .cloud.cancel -text Cancel -command "destroy .cloud"

   pack .cloud.default .cloud.cancel -in .cloud.bot -side left -padx 2m -pady 2m -expand true
   pack .cloud.ok -in .cloud.default -side left -padx 2m -pady 2m

   bind .cloud <KeyPress-Return> "mv_setCloudVar \$cloudvar; destroy .cloud"

   wm withdraw .cloud
   update idletasks
   set x [expr [winfo screenwidth .cloud]/2 - [winfo reqwidth .cloud]/2 \
     - [winfo vrootx [winfo parent .cloud]]]
   set y [expr [winfo screenheight .cloud]/2 - [winfo reqheight .cloud]/2 \
     - [winfo vrooty [winfo parent .cloud]]]
   wm geom .cloud +$x+$y
   wm deiconify .cloud

   set oldFocus [focus]
   grab .cloud
   focus .cloud

   mv_tkwait window .cloud
   focus $oldFocus
}

proc open_dictvarlist {w typ} {
   global ${w}var1 ${w}var2 ${w}sum
   toplevel .$w -class Dialog
   wm title .$w "Text selection for ${typ}s"
   wm transient .$w .

   set ${w}var1 ""
   set ${w}var2 ""

   #label .$w.label -text "Variables"
   #pack .$w.label -side top -fill x
   frame .$w.f -relief raised -bd 1
   pack .$w.f -in .$w -side top

   # Variable 1 selection
   scrollbar .$w.scrollbar1 -command ".$w.list1 yview"
   pack .$w.scrollbar1 -in .$w.f -side left -fill y
   listbox .$w.list1 -relief sunken -borderwidth 2 -yscrollcommand ".$w.scrollbar1 set"
   pack .$w.list1 -in .$w.f -side left -fill y
   bind .$w.list1 <Button-1> "set ${w}var1 \[.$w.list1 get @%x,%y\]; set ${w}sum \[fixsum \$${w}var1 \$${w}var2\]"

   # Variable 2 selection
   scrollbar .$w.scrollbar2 -command ".$w.list2 yview"
   pack .$w.scrollbar2 -in .$w.f -side left -fill y
   listbox .$w.list2 -relief sunken -borderwidth 2 -yscrollcommand ".$w.scrollbar2 set"
   pack .$w.list2 -in .$w.f -side left -fill y
   bind .$w.list2 <Button-1> "set ${w}var2 \[.$w.list2 get @%x,%y\]; set ${w}sum \[fixsum \$${w}var1 \$${w}var2\]"

   # Fill in lists with variable names
   foreach var [lsort [mv_varlist $typ]] {
      .$w.list1 insert end $var
      .$w.list2 insert end $var
   }

   # The summary info
   frame .$w.sum 
   pack .$w.sum -in .$w -side top
   label .$w.sum.l1 -text "Selected variables: "
   label .$w.sum.l2 -textvariable ${w}sum
   pack .$w.sum.l1 .$w.sum.l2 -in .$w.sum -side left -expand true

   # Frame around all buttons
   frame .$w.bot -relief raised -bd 1
   pack .$w.bot -in .$w -side bottom -fill both -expand true
   
   # OK button
   frame .$w.default -relief sunken -bd 1
   button .$w.ok -text OK -command "mv_$w \$${w}var1 \$${w}var2; destroy .$w"
   button .$w.cancel -text Cancel -command "destroy .$w"

   pack .$w.default .$w.cancel -in .$w.bot -side left -padx 2m -pady 2m -expand true
   pack .$w.ok -in .$w.default -side left -padx 2m -pady 2m

   bind .$w <KeyPress-Return> " mv_$w \$${w}var1 \$${w}var2; destroy .$w"

   wm withdraw .$w
   update idletasks
   set x [expr [winfo screenwidth .$w]/2 - [winfo reqwidth .$w]/2 \
     - [winfo vrootx [winfo parent .$w]]]
   set y [expr [winfo screenheight .$w]/2 - [winfo reqheight .$w]/2 \
     - [winfo vrooty [winfo parent .$w]]]
   wm geom .$w +$x+$y
   wm deiconify .$w

   set x [winfo width .$w]
   set y [winfo height .$w]

   wm minsize .$w $x $y
   wm maxsize .$w $x $y

   set oldFocus [focus]
   grab .$w
   focus .$w

   mv_tkwait window .$w
   focus $oldFocus
}

#
# Open a window for getting a condition
#
proc open_condsel {w typ fcn} {
   global ${w}var ${w}op ${w}val

   toplevel .$w -class Dialog
   wm title .$w "Condition Entry"
   wm transient .$w .

   frame .$w.f -relief raised -bd 1
   pack .$w.f -in .$w -side top

   scrollbar .$w.scrollbar -command ".$w.varlist yview"
   pack .$w.scrollbar -in .$w.f -side left -fill y

   # List of variables
   listbox .$w.varlist -relief sunken -height 8 -yscrollcommand ".$w.scrollbar set"
   pack .$w.varlist -in .$w.f -side left
   foreach var [lsort [mv_varlist $typ]] {
      .$w.varlist insert end $var
   }
   bind .$w.varlist <Button-1> "set ${w}var \[.$w.varlist get @%x,%y\]"
   set ${w}var ""

   # List of operators
   frame .$w.op 
   pack .$w.op -in .$w.f -side left
   set count 0
   foreach op {"=" "!=" "<" "<=" ">" ">="} {
      radiobutton .$w.op.$count -text $op -variable ${w}op -value $op -anchor w
      pack .$w.op.$count -in .$w.op -side top -fill x
      incr count
   }

   # Value entry field
   entry .$w.entry -width 20 -relief sunken -bd 2 -textvariable ${w}val
   pack .$w.entry -in .$w.f -side left

   # Frame around all buttons
   frame .$w.bot -relief raised -bd 1
   pack .$w.bot -in .$w -side bottom -fill both -expand true
   
   # OK button
   frame .$w.default -relief sunken -bd 1
   button .$w.ok -text OK -command "if {\$${w}op!=\"\" && \$${w}val!=\"\" && \$${w}var!=\"\"} {$fcn \[list \$${w}var \$${w}op \$${w}val\]}; destroy .$w"
   button .$w.cancel -text Cancel -command "destroy .$w"

   pack .$w.default .$w.cancel -in .$w.bot -side left -padx 2m -pady 2m -expand 1
   pack .$w.ok -in .$w.default -side left -padx 2m -pady 2m

   bind .$w <KeyPress-Return> "if {\$${w}op!=\"\" && \$${w}val!=\"\" && \$${w}var!=\"\"} {$fcn \[list \$${w}var \$${w}op \$${w}val\]}; destroy .$w" 

   wm withdraw .$w
   update idletasks
   set x [expr [winfo screenwidth .$w]/2 - [winfo reqwidth .$w]/2 \
     - [winfo vrootx [winfo parent .$w]]]
   set y [expr [winfo screenheight .$w]/2 - [winfo reqheight .$w]/2 \
     - [winfo vrooty [winfo parent .$w]]]
   wm geom .$w +$x+$y
   wm deiconify .$w

   set x [winfo width .$w]
   set y [winfo height .$w]

   wm minsize .$w $x $y
   wm maxsize .$w $x $y

   set oldFocus [focus]
   grab .$w
   focus .$w

   mv_tkwait window .$w
   focus $oldFocus
}
