/* $Id: xalloc.c,v 1.6 1996/04/16 16:25:03 thalerd Exp $
 * -------------------------------------------------------------------------
 *  
 * [C] David G. Thaler 1996
 * All Rights Reserved 
 *   
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice and this permission notice appear in all 
 * copies of the software and derivative works or modified versions thereof, 
 * and that both the copyright notice and this permission and disclaimer notice 
 * appear in supporting documentation. 
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE REGENTS OF THE 
 * UNIVERSITY OF MICHIGAN AND MERIT NETWORK, INC. DO NOT WARRANT THAT THE 
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR 
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The author shall not
 * be liable for any special, indirect, incidental or consequential damages 
 * with respect to any claim by Licensee or any third party arising from use 
 * of the software. 
 *
 * -------------------------------------------------------------------------
 */
/* XALLOC.C - memory debugging module */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#define wputs add_msg

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <tcl.h>
#include "mview.h"
#include "struct.h"
#include "xalloc.h"
#include "display.h" /* for add_msg */

#ifdef DEBUG
/* to get flags var for debugging only... */
#include "struct.h"
#include "globals.h"
#endif

#define NUMXST 3
static int xst[NUMXST][2]={ { 0, 0 }, { 0, 0 }, { 0, 0} };
#define XS_ACALLS 0
#define XS_RCALLS 1
#define XS_FCALLS 2

/* Information about each array */
typedef struct mem_tt {
   char          **ptr;
   struct mem_tt *next;
   int            num; /* # of elements */
   int            eltsize; /* size of an element */
} mem_t;
static mem_t *first_arr=0,*first_mem=0;

/******************************************************************************/
/* DUMP ALL ARRAYS CURRENTLY ALLOCATED                                        */
/******************************************************************************/
void    /* RETURNS: (nothing) */
xadump() /* ARGUMENTS: (none)  */
{
   mem_t *this;
   int i;

   printf("ARRAYS:\n");
   for (this=first_arr; this; this = this->next) {
      printf("xadump: ptr %x size %hd*%hd\n",this->ptr,this->num,this->eltsize); 
      for (i=0; i<this->num; i++) 
         if (this->ptr[i]) printf("%4hd: %s\n",i,this->ptr[i]);
   }
   printf("\nOTHER:\n");
   for (this=first_mem; this; this = this->next) 
      printf("xadump: ptr %x size %hd*%hd str !%s!\n",this->ptr,this->num,
       this->eltsize,this->ptr); 
}

/******************************************************************************/
/* GET # OF ELEMENTS IN AN ARRAY                                              */
/******************************************************************************/
int        /* RETURNS: size of array */
xsizeof(arr) /* ARGUMENTS: */
   char **arr;  /*    Array */
{
   mem_t *this;

   if (!arr) return 0;
   for (this=first_arr; this && this->ptr != arr; this = this->next);
   if (!this)
      for (this=first_mem; this && this->ptr != arr; this = this->next);
   if (!this) {
      (void)wputs("Tried to get sizeof unxalloced ptr\n");
      return 0; /* not found */
   }
   return this->num;
}

int
xeltlen(arr)
   char **arr;
{
   mem_t *this;

   if (!arr) return 0;
   for (this=first_arr; this && this->ptr != arr; this = this->next);
   if (!this)
      for (this=first_mem; this && this->ptr != arr; this = this->next);
   if (!this) {
      (void)wputs("Tried to get sizeof unxalloced ptr\n");
      return 0; /* not found */
   }
   return this->eltsize;
}

/******************************************************************************/
/* ALLOCATE A NEW    ARRAY                                                    */
/******************************************************************************/
char **             /* RETURNS: New array */
xalloc(num,eltsize) /* ARGUMENTS: */
   int num;          /*    Size of array to get */
   int eltsize;      /*    Size of each element */
{
   mem_t *this;
   char **mem;

   mem=(char **)calloc((unsigned int)num?num:1,(unsigned int)eltsize);
/*printf("xalloc: %d %d > %x\n",num,eltsize,mem);*/
   this          = (mem_t *)malloc(sizeof(mem_t));
   this->ptr     = mem;
   this->num     = num;
   this->eltsize = eltsize;
   if (num) {
      this->next    = first_arr;
      first_arr     = this;
   } else {
      this->next    = first_mem;
      first_mem     = this;
   }

#ifdef DEBUG
   if (flags & O_DEBUG) printf("xalloc: got %x size %hd*%hd\n",
      mem,num,eltsize);
#endif
   xst[XS_ACALLS][!num]++;
   return mem;
}

/* Get structure with data about string, or NULL if not found */
static mem_t *
find_string(arr, prevp)
   void *arr;
   mem_t **prevp;
{
   mem_t *this;

   if (prevp) {
      mem_t *prev=NULL;

      for (this=first_mem; this && this->ptr != arr; prev=this,this=this->next);
      (*prevp) = prev;
   } else {
      for (this=first_mem; this && this->ptr != arr; this=this->next);
   }
   return this;
}

/* Get structure with data about array, or NULL if not found */
static mem_t *
find_array(arr, prevp)
   void *arr;
   mem_t **prevp;
{
   mem_t *this, *prev=NULL;
   for (this=first_arr; this && this->ptr != arr; prev=this,this = this->next);
   if (prevp)
      (*prevp) = prev;
   return this;
}

/******************************************************************************/
/* CHANGE SIZE OF A STRING                                                    */
/******************************************************************************/
char **           /* RETURNS: new array */
xrealloc_string(arr,num) /* ARGUMENTS: */
   void  *arr;       /*    Old array */
   int    num;       /*    New size (# elts for array, # bytes for string) */
{
   mem_t *this;
   char **mem;

   this = find_string(arr, NULL);
   if (!this) {
      (void)wputs("Tried to xrealloc unxalloced string\n");
      return 0; /* not found */
   }

   mem=(char **)realloc((char *)arr, (unsigned int)num);
   this->eltsize = num;
   this->ptr = mem;

   xst[XS_RCALLS][1]++;

   return mem;
}

/******************************************************************************/
/* CHANGE SIZE OF AN ARRAY                                                    */
/******************************************************************************/
char **           /* RETURNS: new array */
xrealloc_array(arr,num) /* ARGUMENTS: */
   void  *arr;       /*    Old array */
   register int num;       /*    New size (# elts for array, # bytes for string
) */
{
   mem_t *this;
   char **mem, *ptr;
   register int i;

   this = find_array(arr, NULL);
   if (!this) {
      (void)wputs("Tried to xrealloc unxalloced array\n");
      return 0; /* not found */
   }

   /* If making array smaller, free up elts */
   for (i=this->num-1; i>=num; i--) {
      if ((ptr = ((char**)arr)[i]))
         xfree_string(ptr);
   }

   mem=(char **)realloc((char *)arr,(num)?(unsigned int)num * this->eltsize:1);
   this->num = num;
   this->ptr = mem;
   xst[XS_RCALLS][0]++;

   return mem;
}

/******************************************************************************/
/* VERIFY THAT ALL ARRAYS HAVE BEEN FREED                                     */
/******************************************************************************/
void
xcheck()
{
   if (!first_arr && !first_mem) {
#ifdef DEBUG
      if (flags & O_DEBUG)
         puts("xcheck: Everything freed.\n");
#endif
   } else {
      printf("xcheck: Error, failed to xfree the following:\n");
      xadump();
   }
}

/******************************************************************************/
/* FREE A BLOCK OF MEMORY                                                     */
/******************************************************************************/
void              /* RETURNS: (nothing) */
xfree_string(mem) /* ARGUMENTS:         */
   void *mem;     /*    Array to free   */
{
   mem_t *this,*prev=0;

   if (!mem) return;

   /* Find info associated with the block of memory */
   this = find_string(mem,&prev);
   if (!this) {
      (void)wputs("Tried to free unxalloced string\n");
      return; /* not found */
   }

   /* Remove from linked list */
   if (!prev)
      first_mem  = this->next;
   else
      prev->next = this->next;

   /* Count for statistics */
   xst[XS_FCALLS][1]++;

   /* Free the space */
   free((char *)mem);
   free((char *)this);
}

/******************************************************************************/
/* FREE AN ARRAY                                                              */
/******************************************************************************/
void             /* RETURNS: (nothing) */
xfree_array(arr) /* ARGUMENTS:         */
   void *arr;    /*    Array to free   */
{
   mem_t *this,*prev=0;
   int i;

   if (!arr) return;

   /* Find info associated with the block of memory */
   this = find_array(arr, &prev);
   if (!this) {
      (void)wputs("Tried to free unxalloced array\n");
      return; /* not found */
   }

   /* Free the elements of the array */
   for (i=0; i<this->num; i++) {
      if (((char**)arr)[i])
         xfree_string(((char**)arr)[i]);
   }

   /* Remove from linked list */
   if (!prev) {
      first_arr  = this->next;
   } else
      prev->next = this->next;

   /* Count for statistics */
   xst[XS_FCALLS][0]++;

   /* Free up space */
   free((char *)arr);
   free((char *)this);
}

/******************************************************************************/
/* STANDARD strdup() CODE FOR THOSE OS'S WHICH DON'T HAVE IT                  */
/******************************************************************************/
char *      /* RETURNS: New string    */
xstrdup(str) /* ARGUMENTS:             */
char *str;  /*    String to duplicate */
{
   char *ptr;
   ptr = (char*)xalloc(0,strlen(str)+1);
   strcpy(ptr,str);
   return ptr;
}

/******************************************************************************/
/* DUMP ALL ARRAYS CURRENTLY ALLOCATED                                        */
/******************************************************************************/
void    /* RETURNS: (nothing) */
xstat() /* ARGUMENTS: (none)  */
{
   mem_t *this;
   int i;
   int n,s,b,nt=0,st=0,bt=0;

   printf("Category: Number   Size  Bytes Acalls Rcalls Fcalls\n");
   n=s=b=0;
   for (this=first_arr; this; this = this->next) {
      n++;
      s += this->num;
      b += (this->num * this->eltsize);
   }
   printf("Arrays  : %6d %6d %6d",n,s,b);
   for (i=0; i<NUMXST; i++)
      printf(" %6d",xst[i][0]);
   nt+=n; st+=s; bt+=b;

   n=s=b=0;
   for (this=first_mem; this; this = this->next) {
      n++;
      s++;
      b+= this->eltsize;
   }
   printf("\nStrings : %6d %6d %6d",n,s,b);
   for (i=0; i<NUMXST; i++)
      printf(" %6d",xst[i][1]);
   nt+=n; st+=s; bt+=b;

   printf("\nTotals  : %6d %6d %6d",nt,st,bt);
   for (i=0; i<NUMXST; i++)
      printf(" %6d",xst[i][0]+xst[i][1]);
   printf("\n");
}

/******************************************************************************/
/* DUMP ALL ARRAYS CURRENTLY ALLOCATED                                        */
/******************************************************************************/
void    /* RETURNS: (nothing) */
xdump() /* ARGUMENTS: (none)  */
{
   mem_t *this;
   int i;

   printf("ARRAYS:\n");
   for (this=first_arr; this; this = this->next) {
      printf("xdump: ptr %x size %d*%d\n",this->ptr,this->num, this->eltsize);
      for (i=0; i<this->num; i++)
         if (this->ptr[i])
            printf("%4hd: %s\n", i, this->ptr[i]);
   }
   printf("\nOTHER:\n");
   for (this=first_mem; this; this = this->next)
      printf("xdump: ptr %x size %d*%d str !%s!\n",this->ptr,this->num,
       this->eltsize, this->ptr);
}
