/*
 * This tool requests configuration info from a multicast router
 * and prints the reply (if any).  Invoke it as:
 *
 *	mrinfo router-name-or-address
 *
 * Written Wed Mar 24 1993 by Van Jacobson (adapted from the
 * multicast mapper written by Pavel Curtis).
 *
 * The lawyers insist we include the following UC copyright notice.
 * The mapper from which this is derived contained a Xerox copyright
 * notice which follows the UC one.  Try not to get depressed noting
 * that the legal gibberish is larger than the program.
 *
 * Copyright (c) 1993 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ---------------------------------
 * Copyright (c) Xerox Corporation 1992. All rights reserved.
 * 
 * License is granted to copy, to use, and to make and to use derivative works
 * for research and evaluation purposes, provided that Xerox is acknowledged
 * in all documentation pertaining to any such copy or derivative work. Xerox
 * grants no other licenses expressed or implied. The Xerox trade name should
 * not be used in any advertising without its written permission.
 * 
 * XEROX CORPORATION MAKES NO REPRESENTATIONS CONCERNING EITHER THE
 * MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS SOFTWARE FOR
 * ANY PARTICULAR PURPOSE.  The software is provided "as is" without express
 * or implied warranty of any kind.
 * 
 * These notices must be retained in any copies of any part of this software.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Id: get_mrinfo.c,v 1.10 1996/03/20 20:14:10 thalerd Exp $";
#endif

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <netdb.h>
#include <sys/time.h>
#include <sys/types.h>
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#include <arpa/inet.h>
#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include "defs.h"
#include "mrtree.h"

#define DEFAULT_TIMEOUT	4	/* How long to wait before retrying requests */
#define DEFAULT_RETRIES 3	/* How many times to ask each router */

u_int32	our_addr;	/* in NET order */
int     debug = 0;
int     retries = DEFAULT_RETRIES;
int     timeout = DEFAULT_TIMEOUT;
int	target_level = 0;
vifi_t  numvifs;		/* to keep loader happy */
				/* (see COPY_TABLES macro called in kern.c) */

char *			inet_name __P((u_int32 addr));
void			ask __P((u_int32 dst));
void			ask2 __P((u_int32 dst));

/* to shut up -Wstrict-prototypes */
int			main __P((int argc, char *argv[]));

/*
 * Log errors and other messages to stderr, according to the severity of the
 * message and the current debug level.  For errors of severity LOG_ERR or
 * worse, terminate the program.
 */
#ifdef __STDC__
void
system_log(int severity, int syserr, char *format, ...)
{
	va_list ap;
	char    fmt[100];

	va_start(ap, format);
#else
void 
system_log(severity, syserr, format, va_alist)
	int     severity, syserr;
	char   *format;
	va_dcl
{
	va_list ap;
	char    fmt[100];

	va_start(ap);
#endif
	switch (debug) {
	case 0:
		if (severity > LOG_WARNING)
			return;
	case 1:
		if (severity > LOG_NOTICE)
			return;
	case 2:
		if (severity > LOG_INFO)
			return;
	default:
		fmt[0] = '\0';
		if (severity == LOG_WARNING)
			strcat(fmt, "warning - ");
		strncat(fmt, format, 80);
		vfprintf(stderr, fmt, ap);
		if (syserr == 0)
			fprintf(stderr, "\n");
		else if (syserr < sys_nerr)
			fprintf(stderr, ": %s\n", sys_errlist[syserr]);
		else
			fprintf(stderr, ": errno %d\n", syserr);
	}

	if (severity <= LOG_ERR)
		exit(-1);
}

/*
 * Send a neighbors-list request.
 */
void 
ask(dst)
	u_int32  dst;
{
	send_igmp(our_addr, dst, IGMP_DVMRP, DVMRP_ASK_NEIGHBORS,
			htonl(MROUTED_LEVEL), 0);
}

void 
ask2(dst)
	u_int32  dst;
{
	send_igmp(our_addr, dst, IGMP_DVMRP, DVMRP_ASK_NEIGHBORS2,
			htonl(MROUTED_LEVEL), 0);
}


int
find_addr( key, p, datalen, version, packet_type )
	u_int32 key;        /* The address we're looking for. */
	u_char  *p;         /* Pointer to the data */
	int     datalen;
	u_int32 version;
	u_char  packet_type;/* 1 = NEIGHBORS msg, 2 = NEIGHBORS2 msg */
{
	u_char *ep;         /* End Ptr - set to mark end of data pointed to by p */
	u_int broken_cisco; /* well, only possibly_broken_cisco, but that's too */
	                    /* long to type. */
	u_char inc;         /* Amount to increment from local-addr to find ncount. */

	broken_cisco = ((version & 0xffff) == 0x020a); /* 10.2 */
	ep  = p + datalen;

	if ( packet_type == 1 )
		inc = 6;
	else if ( packet_type == 2 )
		inc = 7;
	else {
		fprintf( stderr,"%s(%d):Unknown pkt type:%d\n", __FILE__, __LINE__, packet_type );
		exit(-1);
	}

	while ( p < ep ) {
		register int ncount;
		u_int32 localaddr;

		memcpy( (char*) &localaddr, (char*) p, 4 );
		localaddr = (u_int32) ntohl( localaddr );
		if ( localaddr == key )
			return 1;

		ncount = *(p + inc);
		if ( broken_cisco && ncount == 0 )	/* dumb Ciscos */
			ncount = 1;
		if ( broken_cisco && ncount > 15 )	/* dumb Ciscos */
			ncount = ncount & 0xf;

		p = p + inc + 1 + ( ncount * 4 );
	}
	return 0;
}

/*
 * Process an incoming neighbor-list message.
 */
void
mrtree_accept_neighbors(src, dst, p, datalen, level, neighbors, num_neighbors, veronly)
	u_int32	src, dst, level;
	u_char	*p;
	int     datalen;
	void    *neighbors;
	int     *num_neighbors;
 	int     veronly;
{
	u_char *ep = p + datalen;
#define GET_ADDR(a) (a = ((u_int32)*p++ << 24), a += ((u_int32)*p++ << 16),\
		     a += ((u_int32)*p++ << 8), a += *p++)

   if (veronly) {
	   strcpy((char*)neighbors, "OLD");
      return;
   }
/*
	printf("%s (%s):\n", inet_fmt(src, s1), inet_name(src));
*/
	while (p < ep) {
		register u_int32 laddr;
		register u_char metric;
		register u_char thresh;
		register int ncount;

		GET_ADDR(laddr);
		laddr = htonl(laddr);
		metric = *p++;
		thresh = *p++;
		ncount = *p++;
		while (--ncount >= 0) {
			register u_int32 remoteaddr;
			GET_ADDR(remoteaddr);
			remoteaddr = htonl(remoteaddr);
/*
			printf("  %s -> ", inet_fmt(laddr, s1));
			printf("%s (%s) [%d/%d]\n", inet_fmt(remoteaddr, s1),
			       inet_name(remoteaddr), metric, thresh);
*/

			add_neighbor(laddr, remoteaddr, thresh, unknown,neighbors,num_neighbors);
		}
	}
}

void
mrtree_accept_neighbors2(src, dst, p, datalen, level, neighbors, num_neighbors, veronly)
	u_int32	src, dst, level;
	u_char	*p;
	int     datalen;
	void    *neighbors;
	int     *num_neighbors;
	int     veronly;
{
	u_char *ep = p + datalen;
	u_int broken_cisco = ((level & 0xffff) == 0x020a); /* 10.2 */
	/* well, only possibly_broken_cisco, but that's too long to type. */

   if (veronly) {
	   sprintf((char*)neighbors, "%d.%d", level & 0xff, (level >> 8) & 0xff);
      return;
   }

/*
	printf("%s [version %d.%d]:\n", inet_fmt(src, s1), 
	       level & 0xff, (level >> 8) & 0xff);
*/
	
	while (p < ep) {
		register u_char metric;
		register u_char thresh;
		register u_char flags;
		register int ncount;
		register u_int32 laddr = *(u_int32*)p;
		register on_subnet;

		p += 4;
		metric = *p++;
		thresh = *p++;
		flags = *p++;
		ncount = *p++;
		if (broken_cisco && ncount == 0)	/* dumb Ciscos */
			ncount = 1;
		if (broken_cisco && ncount > 15)	/* dumb Ciscos */
			ncount = ncount & 0xf;
		while (--ncount >= 0 && p < ep) {
			register u_int32 remoteaddr = *(u_int32*)p;
			p += 4;
/*
			printf("  %s -> ", inet_fmt(laddr, s1));
			printf("%s (%s) [%d/%d", inet_fmt(remoteaddr, s1),
			       inet_name(remoteaddr), metric, thresh);
			if (flags & DVMRP_NF_TUNNEL)
				printf("/tunnel");
			if (flags & DVMRP_NF_SRCRT)
				printf("/srcrt");
			if (flags & DVMRP_NF_PIM)
				printf("/pim");
			if (flags & DVMRP_NF_QUERIER)
				printf("/querier");
			if (flags & DVMRP_NF_DISABLED)
				printf("/disabled");
			if (flags & DVMRP_NF_DOWN)
				printf("/down");
			if (flags & DVMRP_NF_LEAF)
				printf("/leaf");
			printf("]\n");
*/

			if (!(flags & DVMRP_NF_DOWN)) {
				on_subnet = ( flags & DVMRP_NF_TUNNEL ? no : yes );
				add_neighbor(laddr,remoteaddr,thresh,on_subnet,neighbors,num_neighbors);
			}
		}
	}
}

int
mrinfo(target_addr, level, neighbor, num_neighbors, veronly)
	u_int32  target_addr; /* in NET order */
	u_int32 *level;
	void    *neighbor;
	int     *num_neighbors;
	int      veronly;
{
	int tries = 0;
	int trynew = 1;
	struct timeval et;

	setlinebuf(stderr);

	/*
	 * New strategy: send 'ask2' for two timeouts, then fall back
	 * to 'ask', since it's not very likely that we are going to
	 * find someone who only responds to 'ask' these days
	 */
	ask2(target_addr);

	gettimeofday(&et, 0);
	et.tv_sec += timeout;

	/* Main receive loop */
	for (;;) {
		fd_set           fds;
		struct timeval   tv, now;
		int              count, recvlen, dummy = 0;
		register u_int32 src, dst, group;
		struct ip        *ip;
		struct igmp      *igmp;
		int              ipdatalen, iphdrlen, igmpdatalen;

		FD_ZERO( &fds );
		FD_SET( igmp_socket, &fds );

		gettimeofday( &now, 0 );
		tv.tv_sec = et.tv_sec - now.tv_sec;
		tv.tv_usec = et.tv_usec - now.tv_usec;

		if (tv.tv_usec < 0) {
			tv.tv_usec += 1000000L;
			--tv.tv_sec;
		}
		if (tv.tv_sec < 0)
			tv.tv_sec = tv.tv_usec = 0;

		count = select(igmp_socket + 1, &fds, 0, 0, &tv);

		if ( count < 0 ) {
			if (errno != EINTR)
				perror("select");
			continue;
		} 

		if ( count == 0 ) {
			system_log(LOG_DEBUG, 0, "Timed out receiving neighbor lists");
			if ( ++tries > retries )
				return 0;
			/* If we've tried ASK_NEIGHBORS2 twice with
			 * no response, fall back to ASK_NEIGHBORS
			 */
			if ( tries == 2 && target_level == 0 )
				trynew = 0;

			if ( target_level == 0 && trynew == 0 )
				ask(target_addr);
			else
				ask2(target_addr);

			continue;
		}

		recvlen = recvfrom( igmp_socket, recv_buf, RECV_BUF_SIZE, 0, NULL, &dummy );

		if ( recvlen <= 0 ) {
			if ( recvlen && errno != EINTR )
				perror("recvfrom");
			continue;
		}

		if ( recvlen < sizeof(struct ip) ) {
			system_log(LOG_WARNING, 0, "packet too short (%u bytes) for IP header", recvlen);
			continue;
		}

		ip = (struct ip *) recv_buf;
		if (ip->ip_p == 0)
			continue;	/* Request to install cache entry */

		src       = ip->ip_src.s_addr;
		dst       = ip->ip_dst.s_addr;
		iphdrlen  = ip->ip_hl << 2;
		ipdatalen = ip->ip_len;

		if (iphdrlen + ipdatalen != recvlen) {
			system_log(LOG_WARNING, 0, 
			           "packet shorter (%u bytes) than hdr+data length (%u+%u)",
			            recvlen, iphdrlen, ipdatalen);
			continue;
		}

		igmp        = (struct igmp *) (recv_buf + iphdrlen);
		group       = igmp->igmp_group.s_addr;
		igmpdatalen = ipdatalen - IGMP_MINLEN;

		if (igmpdatalen < 0) {
			system_log(LOG_WARNING, 0,
			    "IP data field too short (%u bytes) for IGMP, from %s",
			    ipdatalen, inet_fmt(src, s1));
			continue;
		}

		if ( igmp->igmp_type != IGMP_DVMRP )
			continue;

		/*************************************************************************
		 * If src != target_addr, then the address of the router sending the     *
		 * reply doesn't match the address we just sent the request to.  This
		 * can happen for two reasons.
		 * (a) the machine to which the request was sent has multiple interfaces
		 * (b) we are receiving a very late reply from an earlier request sent
		 *     to a different machine.
		 * In the case of (b) we want to reject the reply and keeping waiting
		 * for a reply from the machine we most recently sent the request to.
		 * To discriminate between (a) and (b), we search the local addresses
		 * contained in the packet for an instance of 'src'.  It we find one,
		 * then we know we have (a), else we have (b).
		 *************************************************************************/
		switch (igmp->igmp_code) {

			case DVMRP_NEIGHBORS:
				if ( src != target_addr ) {
/*
					fprintf(stderr, "mrinfo: got reply from %s", inet_fmt(src, s1));
					fprintf(stderr, " instead of %s\n", inet_fmt(target_addr, s1));
*/
					if ( !find_addr( target_addr, (u_char *)(igmp+1), 
					                                    igmpdatalen, ntohl(group), 1 ) ){
						/* fprintf( stderr, "mrinfo: Ignoring it.\n" ); */
						continue;
					}
					/* fprintf( stderr, "mrinfo: OK - host with multiple addrs.\n" ); */
				}


				if (group) {
					/* knows about DVMRP_NEIGHBORS2 msg */
					if (target_level == 0) {
						target_level = ntohl(group);
						ask2(target_addr);
					}
				} else {

					*level = ntohl(group);
					mrtree_accept_neighbors(src, dst, (u_char *)(igmp + 1),
							 igmpdatalen, *level, neighbor, num_neighbors, veronly);
					/*****************************************************************
					 * We need to know later on if two neighbors are on the same     *
					 * subnet in order to help figure out if a neighbor is on the    *
					 * tree we're interested in.                                     *
					 *****************************************************************/
					mark_neighbors_on_common_subnet( neighbor, *num_neighbors );
					return 1;
				}
				break;
	
			case DVMRP_NEIGHBORS2:
				if ( src != target_addr ) {
/*
					fprintf(stderr, "mrinfo: got reply from %s", inet_fmt(src, s1));
					fprintf(stderr, " instead of %s\n", inet_fmt(target_addr, s1));
*/
					if ( !find_addr( target_addr, (u_char *)(igmp+1), 
					                                    igmpdatalen, ntohl(group), 2 ) ){
						/* fprintf( stderr, "mrinfo: Ignoring it.\n" ); */
						continue;
					}
					/* fprintf( stderr, "mrinfo: OK - host with multiple addrs.\n" ); */
				}

				*level = ntohl(group);
				mrtree_accept_neighbors2(src, dst, (u_char *)(igmp + 1),
						  igmpdatalen, *level, neighbor, num_neighbors, veronly);
				/*****************************************************************
				 * We need to know later on if two neighbors are on the same     *
				 * subnet in order to help figure out if a neighbor is on the    *
				 * tree we're interested in.                                     *
				 *****************************************************************/
				mark_neighbors_on_common_subnet( neighbor, *num_neighbors );
				return 1;

			default:
				continue;	/* ignore all other DVMRP messages */
		}
	}
}

/* dummies */
void accept_probe(src, dst, p, datalen, level)
	u_int32 src, dst, level;
	char *p;
	int datalen;
{
}
void accept_group_report(src, dst, group, r_type)
	u_int32 src, dst, group;
	int r_type;
{
}
void accept_neighbor_request2(src, dst)
	u_int32 src, dst;
{
}
void accept_report(src, dst, p, datalen, level)
	u_int32 src, dst, level;
	char *p;
	int datalen;
{
}
void accept_neighbor_request(src, dst)
	u_int32 src, dst;
{
}
void accept_prune(src, dst, p, datalen)
	u_int32 src, dst;
	char *p;
	int datalen;
{
}
void accept_graft(src, dst, p, datalen)
	u_int32 src, dst;
	char *p;
	int datalen;
{
}
void accept_g_ack(src, dst, p, datalen)
	u_int32 src, dst;
	char *p;
	int datalen;
{
}
void add_table_entry(origin, mcastgrp)
	u_int32 origin, mcastgrp;
{
}
void check_vif_state()
{
}
void accept_leave_message(src, dst, group)
	u_int32 src, dst, group;
{
}
void accept_mtrace(src, dst, group, data, no, datalen)
	u_int32 src, dst, group;
	char *data;
	u_int no;
	int datalen;
{
}
void accept_membership_query(src, dst, group, tmo)
	u_int32 src, dst, group;
	int tmo;
{
}
void accept_info_request(src, dst, p, datalen)
   u_int32 src, dst;
   u_char *p;
   int datalen;
{
}
void accept_info_reply(src, dst, p, datalen)
   u_int32 src, dst;
   u_char *p;
   int datalen;
{
}
void accept_neighbors(src, dst, p, datalen, level)
   u_int32 src, dst;
   u_char *p;
   int datalen;
   u_int32 level;
{
}
void accept_neighbors2(src, dst, p, datalen, level)
   u_int32 src, dst;
   u_char *p;
   int datalen;
   u_int32 level;
{
}

void
init(target_addr)
	u_int32 target_addr;
{
	init_igmp();

	{			/* Find a good local address for us. */
		int     udp;
		struct sockaddr_in addr;
		int     addrlen = sizeof(addr);

		addr.sin_family = AF_INET;
#if (defined(BSD) && (BSD >= 199103))
		addr.sin_len = sizeof addr;
#endif
		addr.sin_addr.s_addr = target_addr;
		addr.sin_port = htons(2000);	/* any port over 1024 will do... */
		if ((udp = socket(AF_INET, SOCK_DGRAM, 0)) < 0
		|| connect(udp, (struct sockaddr *) & addr, sizeof(addr)) < 0
		    || getsockname(udp, (struct sockaddr *) & addr, &addrlen) < 0) {
			perror("Determining local address");
			return;
		}
		close(udp);
		our_addr = addr.sin_addr.s_addr;
	}
}
