/* mrtree (c)1996 University of Michigan and Merit Network, Inc.
 * By Dave Thaler <thalerd@eecs.umich.edu> and Andy Adams <ala@merit.edu>
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <memory.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netdb.h>
#ifdef HAVE_STRING_H
#include <string.h>
#else
# ifndef HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
char *strchr();
#endif
#include <arpa/inet.h>
#include <stdlib.h>
#include "../mrouted/defs.h"
#include "../snmplib/snmp.h"
#include "../snmplib/asn1.h"
#include "../snmplib/snmp_impl.h"
#include "../snmplib/snmp_api.h"
#include "../snmplib/snmp_client.h"
#include "../snmplib/view.h"
#include "../snmplib/acl.h"
#include "mrtree.h"
#include "snmp.h"

#define CHECK_MAYBES

static char rev[] = "$Header: /n/home/spool/mbone/master_source/mview/src/utils/mrtree.c,v 1.25 1996/05/04 05:25:35 thalerd Exp $";

/* Version bit */
#define NF_SNMP 0x10 /* I respond to SNMP queries */


#define FORWARDING 2 /* To make vif_states more readable. */
#define BRANCH     2 /* To make dvmrpRouteNextHopType more readable */

#define MAXBITS 25
#define REQSIZE ((MAXBITS+1)/2)

#define sysUpTime                       ".1.3.6.1.2.1.1.3.0"
#define ifType                          ".1.3.6.1.2.1.2.2.1.3"
#define ipAdEntIfIndex                  ".1.3.6.1.2.1.4.20.1.2"
#define new_ipMRouteEnable              ".1.3.6.1.3.60.1.1.1.0"
#define new_ipMRouteUpstreamNeighbor    ".1.3.6.1.3.60.1.1.2.1.4"
#define new_ipMRoutePkts                ".1.3.6.1.3.60.1.1.2.1.8"
#define new_ipMRouteOctets              ".1.3.6.1.3.60.1.1.2.1.10"
#define new_ipMRouteNextHopState        ".1.3.6.1.3.60.1.1.3.1.6"
#define new_pimMessageInterval          ".1.3.6.1.3.61.1.1.1.0"
#define new_pimInterfaceAddress         ".1.3.6.1.3.61.1.1.2.1.2"
#define new_pimInterfaceNetMask         ".1.3.6.1.3.61.1.1.2.1.3"
#define new_pimNeighborIfIndex          ".1.3.6.1.3.61.1.1.3.1.2"
#define dvmrpVersion                    ".1.3.6.1.3.62.1.1.1.0"
#define dvmrpVInterfaceLocalAddress     ".1.3.6.1.3.62.1.1.3.1.4"
#define dvmrpVInterfaceRemoteAddress    ".1.3.6.1.3.62.1.1.3.1.5"
#define dvmrpVInterfaceRemoteSubnetMask ".1.3.6.1.3.62.1.1.3.1.6"
#define dvmrpNeighborVersion            ".1.3.6.1.3.62.1.1.4.1.5"
#define dvmrpRouteUpstreamNeighbor      ".1.3.6.1.3.62.1.1.5.1.3"
#define dvmrpRouteNextHopType           ".1.3.6.1.3.62.1.1.6.1.4"
#define old_ipMRouteEnable              ".1.3.6.1.4.1.9.10.2.1.1.1.0"
#define old_ipMRouteUpstreamNeighbor    ".1.3.6.1.4.1.9.10.2.1.1.2.1.4"
#define old_ipMRoutePkts                ".1.3.6.1.4.1.9.10.2.1.1.2.1.8"
#define old_ipMRouteOctets              ".1.3.6.1.4.1.9.10.2.1.1.2.1.10"
#define old_ipMRouteNextHopState        ".1.3.6.1.4.1.9.10.2.1.1.3.1.6"
#define old_pimMessageInterval          ".1.3.6.1.4.1.9.10.3.1.1.1.0"
#define old_pimInterfaceAddress         ".1.3.6.1.4.1.9.10.3.1.1.2.1.2"
#define old_pimInterfaceNetMask         ".1.3.6.1.4.1.9.10.3.1.1.2.1.3"
#define old_pimNeighborIfIndex          ".1.3.6.1.4.1.9.10.3.1.1.3.1.2"

void do_tree();

char *old_pimNeighbor_vars[] = { 
   old_pimNeighborIfIndex, 0 };
char *new_pimNeighbor_vars[] = { 
   new_pimNeighborIfIndex, 0 };
char *dvmrpNeighbor_vars[] = { 
   dvmrpNeighborVersion, 0 };
char *dvmrpVif_vars[]={
   dvmrpVInterfaceRemoteAddress, dvmrpVInterfaceRemoteSubnetMask, 0 };

#define MAX_NEIGHBORS 100

typedef struct node_tmp {
	char name[80];      /* hostname */
	int protocol;       /* protocol used to talk to parent, in prune.h */
	u_int32 addr;        /* IP address seen by parent */
	u_int32 parentaddr;  /* IP address of parent as seen by this node */
	u_int32 raddr;       /* The route ( CIDR address ) used by this node to reach
	                     * the source specified on the cmd line. */
	u_int32 rmask;       /* The route mask ( CIDR address mask ) used by this node
	                     * to reach the source specified on the cmd line. */
	u_int32 snmp_flags;
#define FL_UNKNOWN      0x0000
#define FL_SNMP_V1      0x0001  /* supports SNMPv1 */
#define FL_SNMP_V2      0x0002  /* supports SNMPv2 */
#define FL_NO_SNMP      0x0004  /* doesn't support SNMP */
#define FL_SNMP_NR      0x0008  /* should support, but not responding to SNMP */
#define FL_MCAST        0x0010  /* supports IP Multicast MIB */
#define FL_PIM          0x0100  /* supports PIM MIB */
#define FL_DVMRP        0x1000  /* supports DVMRP MIB */
/*
   int snmp;
*/
	enum TriVal on_subnet,     /* Is node on subnet with parent?              */
	            shares_subnet; /* If node on subnet with parent, does it      *
	                            * share the subnet with any siblings?         */
	u_int32 version;
   int     ver;               /* 0=uses old OIDs, 1=uses new OIDs */

	struct node_tmp *parent;
	
	struct node_tmp *first_child;
	struct node_tmp *next_sibling;

	u_int32 timestamp;
	u_int32 pkts;
	u_int32 byts;
} node_t;

node_t *root_node;

extern int  errno;
int  addronly = 0;         /* If set, dsply numeric ip addrs instead of names */
int  check_maybes = 1;     /* If set, check 3.6+ routers for SNMP capability */
int  showlocal = 0;        /* If set, shows local IP addresses of links */
int  subnet_leaves = 1;    /* If set, displays subnets */
int  compute_stats = 0;
char hostname[80];         /* Hostname of remote mrouted */
char *prefix   = NULL;     /* Instance prefix */
int  snmp_dump_packet = 0; /* referenced by snmp library */
int  interrupted;
char *community = "public";
int  timeout_flag = 0,
     retries,
     ttl_flag = 0,
     ttl;
extern int timeout;
int  dest_port = 0;
int  verbose = 0;


char*
inet_name(addr)
	u_int32  addr;
{
	struct hostent *e;
	struct sockaddr_in tmp;
	static char buff[80];
	static u_int32 prevaddr = 0;

	if ( addr == prevaddr )
		return buff;

	prevaddr = addr;

	if ( !addronly ) {
		e = gethostbyaddr((char *)&addr, sizeof(addr), AF_INET);
		if (e) {
			strcpy(buff, e->h_name);
			return buff;
		}
	}

	tmp.sin_addr.s_addr = addr;
	strcpy(buff, inet_ntoa(tmp.sin_addr));
	return buff;
}

/****************************************************************************
 * Allocate memory for a new node structure.  Fill it in with addr,         *
 * name, on_subnet, and shares_subnet.  Return a pointer to it both as      *
 * the value of the function and in the node_ptr variable.                  *
 ****************************************************************************/
node_t *
new_node( parentaddr, addr, name, on_subnet, shares_subnet )
	u_int32      parentaddr,
	            addr;
	char        *name; /* optional */
	enum TriVal on_subnet,
	            shares_subnet;
{
	node_t *node;
	struct in_addr tmp;

	node                = (node_t *)malloc(sizeof(node_t));
	node->addr          = addr;
	node->parentaddr    = parentaddr;
	node->on_subnet     = on_subnet;
	node->shares_subnet = shares_subnet;
	node->timestamp     = node->pkts = node->byts = 0;
	node->first_child   = node->next_sibling = node->parent = NULL;
	node->snmp_flags    = FL_UNKNOWN;
	node->version       = node->protocol = 0;

	if ( name )
		strcpy(node->name, name);
	else {
		if ( addronly ) {
			tmp.s_addr = addr;
			strcpy( node->name, inet_ntoa( tmp ) );
		}
		else
			strcpy( node->name, inet_name( addr ));
	}
	return node;
}

void 
add_child(parent, child )
	node_t *parent,
	       *child;
{
	node_t *p;

	if (!parent)  {
		root_node = child;
		return;
	}

	if (!parent->first_child) {
		parent->first_child = child;
		child->parent = parent;
		return;
	}

	for( p = parent->first_child; p->next_sibling; p = p->next_sibling )
		continue;

	p->next_sibling = child;
	child->parent   = parent;
	return;
}

void
remove_node(node)
	node_t *node;
{
	node_t *curr=node->parent->first_child,
	       *prev=NULL;

	/* Remove from parent/sibling linked list */
	while (curr && curr!=node)
		prev=curr,curr=curr->next_sibling; /* Interesting syntax.... -ala */

	if (curr) {
		if (prev)
			prev->next_sibling = node->next_sibling;
		else
			node->parent->first_child = node->next_sibling;
	}
   
	free(node);
}

void
get_ipaddr(buff, var)
   char *buff;
   struct variable_list *var;
{
   u_char *ip = var->val.string;
   sprintf(buff, "%d.%d.%d.%d",ip[0], ip[1], ip[2], ip[3]);
}

void
get_timesecs(buff, var)
   char *buff;
   struct variable_list *var;
{
#if (SIZEOF_LONG == 8)
   sprintf(buff, "%u", *var->val.integer / 100);
#else
   sprintf(buff, "%lu", *var->val.integer / 100);
#endif
}

void
get_counter(buff, var)
   char *buff;
   struct variable_list *var;
{
#if (SIZEOF_LONG == 8)
   sprintf(buff, "%u", *var->val.integer);
#else
   sprintf(buff, "%lu", *var->val.integer);
#endif
}

void
get_integer(buff, var)
   char *buff;
   struct variable_list *var;
{
   char objid[256];
   int i = *var->val.integer;

   sprint_objid(objid, var->name, var->name_length);
   sprintf(buff, "%d", i);
}

/*
 * For DisplayString objects, return the string to display
 */
void
get_string(buf, var)
   char *buf;
   struct variable_list *var;
{
   int  x;
   u_char *cp = var->val.string;

   for (x=0; x<var->val_len; x++)
      *buf++ = *cp++;
   *buf = '\0';
}

void
get_value(buf, var)
   char *buf;
   struct variable_list *var;
{
	switch ( var->type ) {
		case ASN_INTEGER:   get_integer(buf, var);          break;
		case ASN_OCTET_STR: get_string(buf, var);           break;
		case IPADDRESS:     get_ipaddr(buf, var);           break;
		case COUNTER:       get_counter(buf, var);          break;
		case TIMETICKS:     get_timesecs(buf, var);         break;
		default:            sprintf(buf, "<unknown type>"); break;
	}
}

struct snmp_pdu *
snmpget(ss, type, names, num_names, err)
   struct snmp_session *ss; /* A valid session handle. */
   int    type;             /* i.e., a "get" or "get next" */
   char **names;            /* OID's of the vars we want to get */
   int    num_names;        /* The number of vars we want to get */
   int   *err;
{
	struct snmp_pdu *pdu, *response = NULL;
	int count;
	int name_length;
	oid name[MAX_NAME_LEN];
	int status;

	pdu = snmp_pdu_create(type);

	for(count = 0; count < num_names; count++){
		name_length = MAX_NAME_LEN;

		if (!get_objid(names[count], name, &name_length))
			printf("Invalid object identifier: %s\n", names[count]);

		snmp_add_null_var(pdu, name, name_length);
	}

	*err = 0;
	for (;;) {
		status = snmp_synch_response(ss, pdu, &response);
		if (status == STAT_SUCCESS) {
			if (response->errstat == SNMP_ERR_NOERROR)
				return response;

			else if (response->errstat==SNMP_ERR_TOOBIG)
				printf("ERROR: Too big\n");

			*err = response->errstat;

		} else if (status == STAT_TIMEOUT) {
			/* printf("No Response from host\n");*/
		} else  {    /* status == STAT_ERROR */
			printf("An error occurred, Quitting\n");
		}
		break;
	}

	if (response)
		snmp_free_pdu(response);
	return NULL;
}

/* Get an IP address from an OID starting at element n = 0,1,2,... */
int
get_address(name, length, addr, n)
	oid    *name;
	int     length;
	u_int32 *addr;
	int     n;
{
	int i;
	int ok = 1;

	(*addr) = 0;

	if (length < n+4)
		return 0;

	for (i=n; i<n+4; i++) {
		(*addr) <<= 8;
		if (i >= length)
			ok = 0;
		else
			(*addr) |= name[i];
	}

	(*addr) = ntohl(*addr);
	return ok;
}

char s[256];

int
get_oid_version(ss)
   struct snmp_session *ss;
{
   char *vars[2];
   struct snmp_pdu *pdu;
   int err;

   /* Try getting old_ipMRouteEnable.0 */
   vars[0] = old_ipMRouteEnable;
   pdu = snmpget(ss, GET_REQ_MSG, vars, 1, &err);
   if (pdu) { /* ok, it worked */
      snmp_free_pdu(pdu);
      return 0;
   }

   /* Try getting new_ipMRouteEnable.0 */
   vars[0] = new_ipMRouteEnable;
   pdu = snmpget(ss, GET_REQ_MSG, vars, 1, &err);
   if (pdu) { /* ok, it worked */
      snmp_free_pdu(pdu);
      return 1;
   }

   return -1;
}

/* Returns session pointer, or NULL on error */
struct snmp_session *
open_session( remotehost, remoteport, community, timeout, ver )
	char *remotehost;       /* destination */
	int   remoteport;       /* (optional) default 161 */
	char *community;        /* (optional) */
	int   timeout;          /* (optional) */
   int  *ver; /* returns 0 if uses old OID's, or 1 if uses new ones */
{
	struct snmp_session session, *ss;

	memset( (char*)&session, (int) '\0', sizeof( struct snmp_session ) );
	session.peername      = remotehost;

	if (remoteport)
		session.remote_port= remoteport;
	else
		session.remote_port= DEFAULT_PORT;
	session.version       = SNMP_VERSION_1;

	if (community)
		session.community  = (u_char *)community;
	else
		session.community  = (u_char *)"public";

	session.community_len = strlen(session.community);
	session.retries       = retries;

	if (timeout)
		session.timeout    = timeout * 1000000L; /* # microseconds */
	else
		session.timeout    = SNMP_DEFAULT_TIMEOUT;
	session.authenticator = NULL;
	snmp_synch_setup(&session);
	ss = snmp_open(&session);
#ifdef ALTERNATE_PORT
	if (!ss) return NULL;

   (*ver) = get_oid_version(ss);
   if (*ver>=0)
      return ss;

	/* Didn't work.  Set up session on 9161 instead */
	snmp_close(ss);

	session.remote_port = ALTERNATE_PORT;
	snmp_synch_setup(&session);
	ss = snmp_open(&session);
#endif

   if (ss)
      (*ver) = get_oid_version(ss);
	return ss;
}

/* Returns 1 on success, 0 on failure */
int
snmpget_list(buffp, host, objlist, nobjs, err, tm)
	char   buffp[][256];
	char  *host;
	char **objlist;
	int    nobjs;
	int   *err;
	int   *tm;
{
	struct snmp_pdu *pdu;
	struct snmp_session *ss;
	char tmbuff[256];
	char *vars[100];
	int i, st=0;
	struct variable_list *vp, *vpc;
   int ver;

	ss = open_session(host, 0, (char*) NULL, timeout, &ver);
	if (!ss) return 0;
	if (tm) {
		vars[st++] = sysUpTime;
		nobjs++;
	}

	for ( i = st; i < nobjs; i++ )
		vars[i] = objlist[i-st];

	pdu = snmpget(ss, GET_REQ_MSG, vars, nobjs, err);
	if (!pdu) return 0;

	vp = pdu->variables;
	if (tm) {
		get_counter(tmbuff, vp);
		*tm = atoi(tmbuff);
		vp = vp->next_variable;
	}

	for (i=st, vpc=vp; vpc; i++, vpc=vpc->next_variable)
		get_value( buffp[i-st], vpc );

	snmp_free_pdu(pdu);
	snmp_close(ss);
	return 1;
}

char*
snmpget_string( ss, obj, err, tm )
	struct snmp_session *ss;
	char *obj;
	int  *err;
	int  *tm;
{
	struct snmp_pdu *pdu;
	static char buff[4096];
	char tmbuff[256];
	char *vars[3];

	if (!ss) {
		fprintf(stderr,"%s(%d):Bad session handle.\n", __FILE__, __LINE__ );
		exit(-1);
	}

	vars[0] = obj;
	vars[1] = sysUpTime;
	pdu = snmpget(ss, GET_REQ_MSG, vars, 2, err);

	if (!pdu) 
		return NULL;

	get_value(buff, pdu->variables);
	get_counter(tmbuff, pdu->variables);
	if (tm) 
		*tm = atoi(tmbuff);

	snmp_free_pdu(pdu);
	return buff;
}

/* Given a session and a vif, return the local IP address for that vif */
u_int32
vifLocalAddress( ss, vif, pim, ver )
	struct snmp_session *ss; /* A valid session handle. */
	int   vif;
	int   pim;
{
	char buff[256], *str, localaddrstr[20];
	int err;

   if (pim) {
   	sprintf(buff, "%s.%d", 
       (!ver)? old_pimInterfaceAddress : new_pimInterfaceAddress, vif);
   } else {
	   sprintf(buff, "%s.%d", dvmrpVInterfaceLocalAddress, vif);
   }

	str = snmpget_string( ss, buff, &err, (int*) NULL );
	if ( str )
		strcpy( localaddrstr, str );
	else 
		exit(1);

	return inet_addr( localaddrstr );
}

void
show_subnet( ss, vif, level, pim, neighbor, num_neighbors, node, group, src )
	struct snmp_session *ss; /* A valid session handle. */
	int   vif,
	      level,
	      pim;   /* 1 if PIM, 0 otherwise */
	neighbor_t *neighbor;
	int num_neighbors;
	node_t *node;
	u_int32 group,
	       src;
{
	char buff[256];
	char remoteaddrstr[20], maskstr[20], *str;
	u_int32 remoteaddr, remotemask, localaddr;
	node_t *child;
	int i,
	    err;

	if ( pim ) {
		sprintf( buff, "%s.%d", ifType, vif );
		str = snmpget_string( ss, buff, &err, (int*) NULL );
		if ( str && !strcmp(str, "1") ) {

			for ( i = 0; i < num_neighbors && neighbor[i].vif >= 0; i++ )
				continue;

			if ( i >= num_neighbors ) {
				printf("tunnel to DVMRP\n");
				for ( i = 0; i < num_neighbors; i++ ) {
					if ( !neighbor[i].protocol ) {
						child = new_node( neighbor[i].remoteaddr, neighbor[i].localaddr,
						                 (char*) NULL, neighbor[i].on_subnet,
						                  neighbor[i].shares_subnet ); 
						do_tree( node, child, level+1, group, src );
					}
				}
			} 
			else {
				neighbor[i].vif = vif;
				if ( verbose ) 
					printf("Set neighbor %s to vif %d\n", 
					                            inet_fmt(neighbor[i].remoteaddr,s), vif);
			}
			return;
		}
	}

	/* If we arrived here, then either the current node is NOT a pim speaker,
	 * or it is, but doesn't speak SNMP. */
	if ( showlocal )
		localaddr = vifLocalAddress( ss, vif, pim, node->ver );

	/* XXX pim problems with local/remote addr here XXX */

   if (pim) {
	   sprintf(buff, "%s.%d", 
	             (!node->ver)? old_pimInterfaceAddress : new_pimInterfaceAddress, vif);
   } else {
	   sprintf(buff, "%s.%d", 
	             dvmrpVInterfaceRemoteAddress, vif);
   }
	str = snmpget_string( ss, buff, &err, (int*) NULL );

	if ( str )
		strcpy( remoteaddrstr, str );
	else {
		printf(" - no subnet info -\n");
		return;
	}

	remoteaddr = inet_addr( remoteaddrstr );

   if (pim) {
	   sprintf(buff, "%s.%d", 
         (!node->ver)? old_pimInterfaceNetMask : new_pimInterfaceNetMask , vif);
   } else {
	   sprintf(buff, "%s.%d", 
	           dvmrpVInterfaceRemoteSubnetMask, vif);
   }

	str = snmpget_string( ss, buff, &err, (int*) NULL );

	if ( str )
		strcpy( maskstr, str );
	else {
		printf(" - no mask info -\n");
		return;
	}

	remotemask = inet_addr(maskstr);
	/* Mask = 0 means zero-length mask. IE, a host addr, not a net addr. */
	if ( !remotemask )
		return;

	for ( i = 0; i < level; i++ )
		printf("   ");

	if ( showlocal )
		printf("%2d %s->subnet: %s\n", level, inet_fmt(localaddr,s), 
		                                inet_fmts(remoteaddr, remotemask, buff));
	else
		printf("%2d subnet: %s\n", level, inet_fmts(remoteaddr, remotemask, buff));
}

int /* RETURNS: -1 if not found, idx otherwise */
find_neighbor(neighbor, neighbors, num_neighbors)
	u_int32      neighbor;
	neighbor_t *neighbors;
	int         num_neighbors;
{
	int i;
	for ( i = 0; i<num_neighbors && neighbor!=neighbors[i].remoteaddr; i++ )
		continue;

	return ( i < num_neighbors ) ? i : -1;
}

void
add_neighbor( localaddr,remoteaddr,thresh,on_subnet,neighbors,num_neighbors )
	u_int32      localaddr,
	            remoteaddr;
	int         thresh;
	enum TriVal on_subnet;
	neighbor_t  *neighbors;
	int         *num_neighbors;
{
	if (!remoteaddr || find_neighbor(remoteaddr, neighbors, *num_neighbors)>=0)
		return;

	neighbors[ *num_neighbors ].localaddr     = localaddr;
	neighbors[ *num_neighbors ].remoteaddr    = remoteaddr;
	neighbors[ *num_neighbors ].threshold     = thresh;
	neighbors[ *num_neighbors ].on_subnet     = on_subnet;
	neighbors[ *num_neighbors ].vif           = -1; 
	neighbors[ *num_neighbors ].protocol      = 0;       /* unknown? */
	neighbors[ *num_neighbors ].shares_subnet = (on_subnet == no ? no : unknown);
	
	if (verbose)
		printf("added %s as DVMRP neighbor\n", inet_fmt(remoteaddr,s));
	(*num_neighbors)++;
}

/*****************************************************************************
 * This routine looks through the array pointed to by neighbor and figures   *
 * out if any of the neighbors are on the same subnet.  It does this by      *
 * finding neighbors which have the same 'localaddr'. (Remember here that    *
 * these are neighbors of some other node that we know nothing about.        *
 * 'localaddr' means the 'local address' from _node's_ point of view, not    *
 * the neighbor's.)                                                          *
 *                                                                           *
 * When this routine exits, each neighbor's 'shares_subnet' field should     *
 * be definitively known - i.e. 'yes' or 'no'.  None will have the value     *
 * 'unknown.'                                                                *
 *****************************************************************************/
void 
mark_neighbors_on_common_subnet( neighbor, num_neighbors )
neighbor_t *neighbor;      /* Pointer to an array of neighbor structs */
int         num_neighbors; /* Number of neighbor structs in neighbor  */
{
	int i, j;

	for( i = 0; i < num_neighbors; i++ ) {
		if ( neighbor[i].on_subnet == no || neighbor[i].shares_subnet == yes )
			continue;

		for( j = i+1; j < num_neighbors; j++ ) {
			if ( neighbor[j].on_subnet == no || neighbor[j].shares_subnet == yes )
				continue;
			if ( neighbor[i].localaddr == neighbor[j].localaddr )
				neighbor[i].shares_subnet = neighbor[j].shares_subnet = yes;
		}
	}

	/* When a new neighbor entry is created, its 'shares_subnet' field is
	 * set to 'no' if its 'on_subnet' field is 'no', otherwise its 
	 * 'shares_subnet' field is set to 'unknown'.  The nested loops above
	 * change all appropriate shares_subnet fields with a value of 'unknown' 
	 * to 'yes'.  We now reset the remaining 'unknowns' to 'no'.
	 */
	for( i = 0; i < num_neighbors; i++ )
		if ( neighbor[i].on_subnet == yes && neighbor[i].shares_subnet == unknown )
			neighbor[i].shares_subnet = no;
}

int root_pkts=0;

char *
vers2str(flags)
	u_int32 flags;
{
	static char buff[80];
#if (SIZEOF_LONG == 8)
	sprintf(buff,"%d.%d", flags & 0xff, (flags >> 8) & 0xff);
#else
	sprintf(buff,"%ld.%ld", flags & 0xff, (flags >> 8) & 0xff);
#endif
	if ((flags >> 16) & NF_LEAF)   { strcat(buff,",leaf"); }
	if ((flags >> 16) & NF_PRUNE)  { strcat(buff,",prune"); }
	if ((flags >> 16) & NF_GENID)  { strcat(buff,",genid"); }
	if ((flags >> 16) & NF_MTRACE) { strcat(buff,",mtrace"); }
	if ((flags >> 16) & NF_SNMP)   { strcat(buff,",snmp"); }
	return buff;
}

void
print_name(node, level)
	node_t *node;    /* IN : node pointer */
	int     level;   /* IN : Tree level */
{
	int i;

	/* Output start of line for this node */
	for ( i = 0; i < level; i++ )
		printf("   ");

	printf("%2d ", level);
 
	if (showlocal && node->parentaddr)
		printf("%s->", inet_fmt(node->parentaddr,s));
 
	if (addronly)
		/* addronly==1 means node->name contains ip addr. */
		printf("%s", node->name );
	else
		printf("%s (%s)", node->name, inet_fmt(node->addr,s));

	if ( node->version )
		printf(" [ver %s]", vers2str((u_int32) node->version));
	else
		printf(" [ ]");

	fflush( stdout );
}

/*
 * When this exits, the SNMP, MCAST, DVMRP, and PIM info
 * should all be authoritative, never unknown.  Before this,
 * the flags are all 0.
 *
 * The 32-bit "version" field is now split into four independent bytes:
 * Vendor ID
 * Capability bitstring
 * Minor version number
 * Major version number
 *
 * Each vendor should have a different vendor code.  The capability bitstring
 * currently includes bits for the following capabilities: Leaf reporting,
 * Pruning, Mtrace, Generation #'s, and SNMP.
 * 
 * If node doesn't speak snmp at all, then waiting for the snmp queries
 * below to timeout is a killer. We should look into shrinking the timeouts.
 * -ala 18 Jan 96
 */
struct snmp_session*
get_mibs_supported( node )
	node_t *node;
{
	struct snmp_session *ss;
	char *str;
	int   err;
	int   vmaj,  /* Major version number */
	      vmin,  /* Minor version number */
	      vven,  /* Vender ID            */
	      vcap;  /* Capability bitstring */

	/* Split version into pieces */
	vmaj = node->version & 0xFF;
	vmin = ((node->version >> 8) & 0xff);
	vcap = ((node->version >> 16) & 0xff);
	vven = ((node->version >> 24) & 0xff);

	/* Guess at SNMP capability: */
	if (vcap & NF_SNMP || vmaj >= 10) /* cisco */
		node->snmp_flags |= FL_SNMP_V1;
#ifdef CHECK_MAYBES
	else if (check_maybes && vmaj==3 && vmin>=6)
		node->snmp_flags |= FL_SNMP_V1;
#endif
	else {
		node->snmp_flags = FL_NO_SNMP;
		return NULL;
	}

	ss = open_session( inet_fmt(node->addr,s), dest_port, community, timeout, &node->ver);
	if ( !ss ) {
		node->snmp_flags |= FL_SNMP_NR;
		return NULL;
	}

	/* See if it supports Multicast MIB - every SNMP-mcast router should */
   /* We've actually already tried this, and 'node->ver' is the result 
	 * str = snmpget_string( ss, ipMRouteEnable, &err, (int*) NULL);
	 * if (!str || strcmp(str,"1")) 
    */
   if (node->ver < 0) {
		node->snmp_flags |= FL_SNMP_NR;
		snmp_close( ss );
		return NULL; /* Probably no point in further queries. */
	} 

	/* If we got here then we at least know that node supports mcast. */
	node->snmp_flags |= FL_MCAST;

	/* See if it supports DVMRP MIB */
	str = snmpget_string( ss, dvmrpVersion, &err, (int*) NULL );
	if (str)
		node->snmp_flags |= FL_DVMRP;

	/* See if it supports PIM MIB */
	str = snmpget_string( ss, 
    (!node->ver)? old_pimMessageInterval : new_pimMessageInterval, 
    &err, (int*) NULL);
	if (str)
		node->snmp_flags |= FL_PIM;

	return ss;
}

/*****************************************************************************
 * This routine issues SNMP queries to determine the list of PIM neighbors   *
 * for 'node'.  Usually, this list has already been determined through       *
 * from a call to mrinfo, so we essentially just query the PIM mib,          *
 * look up the node in the 'neighbors' array, and mark the neighbor as       *
 * speaking PIM.                                                             *
 *                                                                           *
 * By the way, this routine assumes that 'node' speaks SNMP.                 *
 *****************************************************************************/
void
get_pim_neighbors( ss, neighbor, num_neighbors, ver )
	struct snmp_session *ss;
	neighbor_t           neighbor[MAX_NEIGHBORS];
	int                 *num_neighbors;
   int                  ver; /* 0 = uses old OIDs, 1 = uses new ones */
{
	int idx, i, err;
	unsigned int32 addr;
	char initvar[10][256], var[10][256], *vars[MAXBITS+1], tmp[256];
	struct snmp_pdu *response;
	struct variable_list *vp;


	 /* Initialized the list of variables that we want to query for.  */

   if (!ver) {
   	for (i=0; old_pimNeighbor_vars[i]; i++) {
   		strcpy(initvar[i], old_pimNeighbor_vars[i]);
   		strcpy(var[i], initvar[i]);
   		vars[i] = var[i];
   	}
   } else {
   	for (i=0; new_pimNeighbor_vars[i]; i++) {
   		strcpy(initvar[i], new_pimNeighbor_vars[i]);
   		strcpy(var[i], initvar[i]);
   		vars[i] = var[i];
   	}
   }

	/*******************************************************************
	 * Do SNMP get next's on the PIM neighbors until we step past the  *
	 * last neighbor.                                                  *
	 *******************************************************************/

	for (interrupted = 0; !interrupted; ) {
		/*********************************************************************
		 * Note that the following query is an SNMP getnext, not simply an   *
		 * SNMP get as the function name might imply.                        *
		 *********************************************************************/
		response = snmpget(ss, GETNEXT_REQ_MSG, vars, i, &err);
		if (!response || !response->variables || interrupted ) 
			break;

		vp = response->variables;

		sprint_objid(tmp, vp->name, vp->name_length);
		/* Did we just pass the last variable instance that we are interested in? */
		if ( strncmp( tmp, initvar[0], strlen(initvar[0]) ) ) {
			/*
			 * Any routers which don't support the PIM MIB will
			 * exit here the first time.  So will a PIM router
			 * with no neighbors, but that's kind of an odd case.
			 */
			break;
		}

		/*
		 * In the pimNeighborTable, the Vif is in $0, and
		 * the address is in $A(-4).
		 */

		get_address(vp->name, vp->name_length, &addr, vp->name_length-4);
		idx = find_neighbor(addr, neighbor, *num_neighbors);
		if ( idx < 0 )  /* XXX check local end too? */
			idx = (*num_neighbors)++;
		neighbor[idx].vif        = *vp->val.integer;
		neighbor[idx].remoteaddr = addr;
		neighbor[idx].protocol   = PROTO_PIM;

		if (verbose)
			printf("Saw PIM neighbor %s vif %d\n",
			              inet_fmt(neighbor[idx].remoteaddr,s), neighbor[idx].vif);

		/* We need to use the Vif to resolve local address now */
		if (showlocal) {
			neighbor[idx].localaddr = vifLocalAddress( ss, neighbor[idx].vif, 1, ver );
		}

		/* Save new OID in var[i] */
		for ( i = 0; vp; i++, vp = vp->next_variable )
			sprint_objid(var[i], vp->name, vp->name_length);

		snmp_free_pdu(response);
	}

	/* If we arrive here then something went wrong. */
	if ( response )
		snmp_free_pdu( response );
}

/*****************************************************************************
 * This routine issues SNMP queries to determine the list of DVMRP neighbors *
 * for 'node'.  The list of neigbors should have already been determined     *
 * from a call to mrinfo, so we essentially just query the DVMRP mib,        *
 * look up the node in the 'neighbors' array, and mark the neighbor as       *
 * speaking DVMRP.                                                           *
 *                                                                           *
 * By the way, this routine assumes that 'node' speaks SNMP.                 *
 *****************************************************************************/
void
get_dvmrp_neighbors( ss, neighbor, num_neighbors, ver )
	struct snmp_session *ss;
	neighbor_t neighbor[MAX_NEIGHBORS];
	int       *num_neighbors;
   int                  ver; /* 0 = uses old OIDs, 1 = uses new ones */
{
	int idx, i, err;
	unsigned int32 addr;
	char initvar[10][256], var[10][256], *vars[MAXBITS+1], tmp[256];
	struct snmp_pdu *response;
	struct variable_list *vp;

	for (i=0; dvmrpNeighbor_vars[i]; i++) {
		strcpy(initvar[i], dvmrpNeighbor_vars[i]);
		strcpy(var[i], initvar[i]);
		vars[i] = var[i];
	}

	/*********************************************************************
	 * Do SNMP get next's on the DVMRP neighbors until we step past the  *
	 * last neighbor.                                                    *
	 *********************************************************************/

	interrupted = 0;
	while( !interrupted ) {
		/*********************************************************************
		 * Note that the following query is an SNMP getnext, not simply an   *
		 * SNMP get as the function name might imply.                        *
		 *********************************************************************/
		response = snmpget(ss, GETNEXT_REQ_MSG, vars, i, &err);

		if ( !response || !response->variables || interrupted ) 
			break;

		vp = response->variables;

		sprint_objid(tmp, vp->name, vp->name_length);
		if (strncmp(tmp, initvar[0], strlen(initvar[0]))) {
			/*
			 * Any routers which don't support DVMRP MIB will
			 * exit here the first time.  So will a DVMRP router
			 * with no neighbors, but that's kind of an odd case.
			 */
			break;
		}

		/*
		 * In the dvmrpNeighborTable, the Vif is in $A(-5), and
		 * the address is in $A(-4)
		 */
		get_address(vp->name, vp->name_length, &addr, vp->name_length-4);
		idx = find_neighbor(addr, neighbor, *num_neighbors);
		if (idx<0)  /* XXX check local end too? */
			idx = (*num_neighbors)++;
		neighbor[idx].vif        = vp->name[vp->name_length-5];
		neighbor[idx].remoteaddr = addr;
		neighbor[idx].protocol   = PROTO_DVMRP;
		get_string(neighbor[idx].version, vp); /* XXX */

		if (verbose)
			printf("Saw DVMRP neighbor %s vif %d\n",
			               inet_fmt(neighbor[idx].remoteaddr,s), neighbor[idx].vif);

		/* We need to use the Vif to resolve local address now */
		if (showlocal) {
			neighbor[idx].localaddr = vifLocalAddress( ss, neighbor[idx].vif, 0, ver );
		}

		/* Save new OID in var[i] */
		for (i = 0; vp; i++, vp = vp->next_variable)
			sprint_objid(var[i], vp->name, vp->name_length);

		snmp_free_pdu( response );
	}

	/* If we arrived here then something went wrong. */
	if ( response )
		snmp_free_pdu( response );
}

/*****************************************************************************
 * 
 * This routine attempts to determine if 'node' is a cisco which has a
 * particular bug in its SNMP implementation. The routine returns one of
 * yes, no, unknown.
 *
 * At the time of this writing, some (many?) ciscos have a bug in their
 * SNMP implementation which causes them to behave strangely in response
 * to some snmp "get nexts".  
 *
 * (1) A get next request on an incomplete OID returns the first entry in 
 *     the next table.  For example,
 * 
 * % ./snmpi -f mrouted.defs -c public -a maingate.princeton.edu
 * snmpi>  next ipMRouteNextHopState.224.0.1.2.128.112.92.7
 * ipMRouteNextHopUpTime.224.0.1.1.128.4.1.1.255.255.255.255.6=6 minutes, 22.88 
 * seconds (38288 timeticks)
 *
 * (2) A get next on a non-existent oid returns the first entry in the next
 *     table.
 *
 * So to try an decide if node is a broken cisco, we do an snmp get next on
 * ipMRouteNextHopState.1, an incomplete oid.  If we get back an oid
 * that's not in the NextHopState column, then either
 * (a) the table is empty and get next did the right thing or
 * (b) we have a broken cisco.
 *
 * At the moment, I don't know how to tell the two cases apart, or 
 * alternatively, know another way to check for buggy boxes. (a) is _probably_ 
 * unlikely so the probability of a false positive ( ciscos reported as 
 * buggy that really aren't ) is probably really, really low.
 *****************************************************************************/
enum TriVal
is_broken_cisco( ss, node )
	struct snmp_session *ss; /* An active SNMP session */
	node_t *node;
{
	int err;
	char objID[256], *vars[2], tmp[256];
	struct snmp_pdu *response;
	struct variable_list *vp;     /* SNMP Variable list Pointer */

	if ( !( node->snmp_flags & FL_PIM ) ) /* Is node even a cisco? */
		return no;

	sprintf( objID, "%s.1", (!node->ver)? old_ipMRouteNextHopState : new_ipMRouteNextHopState );
	vars[0] = objID;

	response = snmpget(ss, GETNEXT_REQ_MSG, vars, 1, &err);

	if ( !response )
		return unknown;

	vp = response->variables;

	/***********************************************************************
	 * Check to see if we've gotten back a NextHopState instance. If not,  *
	 * then either the NextHop table is empty or node is a broken          *
	 * cisco.  I don't know at this time how to tell the two cases apart.  *
	 ***********************************************************************/
	sprint_objid(tmp, vp->name, vp->name_length);
   if (!node->ver) {
   	if ( strncmp(tmp, old_ipMRouteNextHopState, strlen( old_ipMRouteNextHopState ) ) )
   			return yes;
   } else {
   	if ( strncmp(tmp, new_ipMRouteNextHopState, strlen( new_ipMRouteNextHopState ) ) )
   			return yes;
   }

	return no;
}

int
get_first_dvmrpRouteNextHop_match( ss, src_key, objID )
	struct snmp_session *ss; /* An active SNMP session */
	u_int32 src_key;          /* Address of source */
	char   *objID;
{
	int err;
	char *vars[10], tmp[256];
	struct snmp_pdu *response;
	struct variable_list *vp;     /* SNMP Variable list Pointer */
	u_int32 src;

	sprintf( objID, "%s", dvmrpRouteNextHopType );
	vars[0] = objID;
	
	interrupted = 0;
	response = snmpget( ss, GETNEXT_REQ_MSG, vars, 1, &err );

	if ( !response || !response->variables || interrupted ) 
		return 0; /* Something went wrong... */

	interrupted = 0;
	while( !interrupted && response && response->variables ) {

		/* If we've left the NextHopState column, then there are no entries
		 * that interest us and it's time to go. */
		vp = response->variables;
		sprint_objid( tmp, vp->name, vp->name_length );
		if ( strncmp( tmp, dvmrpRouteNextHopType, strlen( dvmrpRouteNextHopType )))
			break;
		
		(void) get_address( vp->name, vp->name_length, &src, 11 );

		sprint_objid( objID, vp->name, vp->name_length );
		if ( src == src_key ) {
			snmp_free_pdu( response );
			return 1;
		}
		
		snmp_free_pdu( response );
		response = snmpget( ss, GETNEXT_REQ_MSG, vars, 1, &err );
	}

	/* If we arrive here, then something went wrong. */
	if ( response ) snmp_free_pdu( response );
	return 0;
}


/*****************************************************************************
 * This routine ( along with get_tree_state() ) determines                   *
 * which of node's neighbors are on the _potential_ multicast for src.       *
 * 'broken_cisco' refers to a bug in some cisco code which makes it a bit    *
 * trickier obtain info via SNMP queries.                                    *
 *                                                                           *
 * This routine assumes that node speaks SNMP and supports the DVMRP mib.    *
 * It queries the DvmrpRouteNextHop table in node's DVMRP mib to get the vif *
 * and type (leaf or branch) for each vif that's a next hop for src.  If     *
 * the next hop is a leaf, then that vif is ignored.  If not, then the       *
 * neighbor corresponding to that vif is looked up in the neighbor table     *
 * and we jump to that node to traverse it.                                  *
 *****************************************************************************/
void
get_tree_state_broken_cisco(ss, node, src, level, neighbor, num_neighbors)
   struct snmp_session *ss; /* Active SNMP session with 'node'. */
   node_t *node;            /* Node with which we have active SNMP session. */
   u_int32 src;              /* Address of source */
   int level;               /* Our depth in the tree. The root is level 0. */
   neighbor_t *neighbor;    /* All of node's neighbors. */
   int num_neighbors;       /* Number of entries in 'neighbors'. */
{
	int i, err;
	char objID[256], *vars[10];
	struct snmp_pdu *response;
	struct variable_list *vp;
	int vif, state;
	node_t *child;
	
	/* Find the first row of table with < src > if one exists. */
	if ( !get_first_dvmrpRouteNextHop_match( ss, src, objID ) )
		return;

	/* Get and dump the table */
	interrupted = 0;
	vars[0] = objID;
	while( !interrupted ) {
		response = snmpget( ss,  GETNEXT_REQ_MSG, vars, 1, &err );

		if ( !response || !response->variables || interrupted ) 
			break;

		vp = response->variables;
   
		/* Have we stepped past the end of the NextHopType column? */
		/* Save new OID in objID in order to do another get next on it */
		sprint_objid(objID, vp->name, vp->name_length);
		if ( strncmp(objID, dvmrpRouteNextHopType, strlen( dvmrpRouteNextHopType )))
			break;
   
		vif   = vp->name[19];
		state = *vp->val.integer;
		snmp_free_pdu( response );
  
		if (verbose)
			printf("Saw state for Vif %d = %d\n", vif, state );
  
		/* If the vif doesn't point to a branch, then we aren't interested */
		if ( state != BRANCH )
			continue;
  
		for ( i = 0; i < num_neighbors; i++ ) {
			if ( neighbor[i].vif == vif )  {
				if ( ttl_flag && ( ttl - level ) <= neighbor[i].threshold )
					continue; 
				if ( subnet_leaves )
					show_subnet( ss, vif, level+1, (int) (node->snmp_flags & FL_PIM),
					                   neighbor, num_neighbors, node, (u_int32) 0, src );
				child = new_node( neighbor[i].localaddr, neighbor[i].remoteaddr,
				                  (char*) NULL, neighbor[i].on_subnet,
				                  neighbor[i].shares_subnet ); 
				do_tree( node, child, level+1, (u_int32) 0, src );
			}
		}
  
	}
	/* If we arrived here, then something went wrong. */
	if ( response )
		snmp_free_pdu( response );
	return;
}

/*****************************************************************************
 * This routine ( along with get_tree_state_broken_cisco() ) determines      *
 * which of node's neighbors are on the _potential_ multicast for src.       *
 * It does this by one of two methods.                                       *
 * 1) If node speaks SNMP and supports the DVMRP mib, then the routine       *
 *    queries the DvmrpRouteNextHop table in node's DVMRP mib to get the vif *
 *    and type (leaf or branch) for each vif that's a next hop for src.  If  *
 *    the next hop is a leaf, then that vif is ignored.  If not, then the    *
 *    neighbor corresponding to that vif is looked up in the neighbor table  *
 *    and we jump to that node to traverse it.                               *
 * 2) If node does not support the DVMRP mib, then we traverse every         *
 *    neighbor in the neighbor table via a call to do_tree(). do_tree() will *
 *    eventually call VrfyOnTreeSansGroup() to determine if the neighbor     *
 *    is on the multicast tree in which we're interested.                    *
 *****************************************************************************/
void
get_tree_state(ss, node, src, level, neighbor, num_neighbors)
	struct snmp_session *ss; /* Active SNMP session with 'node'. */
	node_t *node;            /* Node with which we have active SNMP session. */
	unsigned int32 src;       /* Address of source */
	int level;               /* Our depth in the tree. The root is level 0. */
	neighbor_t *neighbor;    /* All of node's neighbors. */
	int num_neighbors;       /* Number of entries in 'neighbors'. */
{
	int i, err;
	char initvar[10][256], var[256], *vars[10];
	struct snmp_pdu *response;
	struct variable_list *vp;
	int vif, state;
	node_t *child;
	enum TriVal is_broken;

	/*************************************************************************
	 * If the node speaks SNMP and supports the DVMRP mib, then query it for *
	 * the next hop information.                                             *
	 *************************************************************************/
	if (node->snmp_flags & FL_DVMRP && !( node->snmp_flags & FL_SNMP_NR ) ) {
     
		is_broken = is_broken_cisco( ss, node );

		if ( is_broken == yes ) {
			get_tree_state_broken_cisco(ss, node, src, level, neighbor,num_neighbors);
			return;
		}

		if ( is_broken == unknown )  /* Something went wrong w/is_broken_cisco */
			return;

		sprintf(initvar[0], "%s.%s.", dvmrpRouteNextHopType, 
		                              inet_fmt(node->raddr,s) );
		strcat(initvar[0], inet_fmt(node->rmask,s));
		strcpy( var, initvar[0] );
		vars[0] = var;
   
		/* Get and dump the table */
		interrupted = 0;
		while ( !interrupted ) {
			response = snmpget(ss, GETNEXT_REQ_MSG, vars, 1, &err);

			if ( !response || !response->variables || interrupted ) 
				return;

			vp = response->variables;

			/****************************************************************
			 * Have we stepped past the last dvmrpRouteNextHopType instance *
			 * of interest? If so, then it's time to leave.                 *
			 ****************************************************************/

			/* Save new OID in var so that we can do another get next on it */
			sprint_objid(var, vp->name, vp->name_length);
			if (strncmp(var, initvar[0], strlen(initvar[0]))) {
				snmp_free_pdu( response );
				return;
			}
   
			vif   = vp->name[19];
			state = *vp->val.integer;
			snmp_free_pdu( response );
  
			if ( verbose )
				printf("Saw state for Vif %d = %d\n", vif, state );
  
			/* If state != BRANCH, then we aren't interested in it. */
			if ( state != BRANCH )
				continue;

			for ( i = 0; i < num_neighbors; i++) {
				if ( neighbor[i].vif == vif )  {
					if ( ttl_flag && ( ttl - level ) <= neighbor[i].threshold )
						continue; 
					if ( subnet_leaves )
						show_subnet( ss, vif, level+1, (int) (node->snmp_flags & FL_PIM),
						                 neighbor, num_neighbors, node, (u_int32) 0, src );
					child = new_node( neighbor[i].localaddr, neighbor[i].remoteaddr,
					                  (char*) NULL, neighbor[i].on_subnet,
					                  neighbor[i].shares_subnet ); 
					do_tree( node, child, level+1, (u_int32) 0, src );
				}
			}
		}

		/* If we arrived here, then something went wrong. */
		if ( response )
			snmp_free_pdu( response );
		return;

	} 
	else {
	 /***********************************************************************
		* If node does not support the DVMRP mib, then traverse all it's      *
		* neighbors an let VrfyOnTreeSansGroup() determine if the neighbor is *
		* on the multicast tree that we're interested in.                     *
		**********************************************************************/
		for ( i = 0; i < num_neighbors; i++ ) {
			if ( !node->parent || neighbor[i].remoteaddr != node->parent->addr ) {
				if ( ttl_flag && ( ttl - level ) <= neighbor[i].threshold )
					continue; 
				child = new_node( neighbor[i].localaddr, neighbor[i].remoteaddr,
						              (char*) NULL, neighbor[i].on_subnet,
				                  neighbor[i].shares_subnet ); 
				do_tree( node, child, level+1, (u_int32) 0, src );
			}
		}
	}
}

int
get_first_ipMRouteNextHop_match( ss, group_key, src_key, objID, ver )
	struct snmp_session *ss; /* An active SNMP session */
	u_int32 group_key;        /* Address of multicast group */
	u_int32 src_key;          /* Address of source */
	char   *objID;
   int     ver;             /* 0=use old OIDs, 1=use new OIDs */
{
	int err;
	char *vars[10], tmp[256];
	struct snmp_pdu *response;
	struct variable_list *vp;     /* SNMP Variable list Pointer */
	u_int32 group,
	       src;

   /* First try specifying a mask of 255.255.255.255.  This should always(?)
    * be enough. -dgt 5/13/96
    */
   sprintf( tmp, "%s.%s.",
    (!ver)? old_ipMRouteNextHopState : new_ipMRouteNextHopState,
    inet_fmt( group_key, s ) );
   strcat( tmp, inet_fmt( src_key, s ) );
   strcat( tmp, ".255.255.255.255");
   strcpy(objID, tmp);
	vars[0] = objID;
	response = snmpget( ss, GETNEXT_REQ_MSG, vars, 1, &err );
	if ( !response || !response->variables || interrupted ) 
		return 0; /* Something went wrong... */
	vp = response->variables;
	sprint_objid( objID, vp->name, vp->name_length );
	if (!strncmp(objID, tmp, strlen(tmp))) { /* ok, it worked */
	   snmp_free_pdu(response);
		return 1;
   }

	if (verbose)
		printf("cisco router is _really_ broken\n");

	/* Ok, that failed, try searching through from the beginning */
	sprintf( objID, "%s", (!ver)? old_ipMRouteNextHopState : new_ipMRouteNextHopState );
	vars[0] = objID;
	
	interrupted = 0;
	response = snmpget( ss, GETNEXT_REQ_MSG, vars, 1, &err );

	if ( !response || !response->variables || interrupted ) 
		return 0; /* Something went wrong... */

	interrupted = 0;
	while( !interrupted && response && response->variables ) {

		/* If we've left the NextHopState column, then there are no entries
		 * that interest us and it's time to go. */
		vp = response->variables;
		sprint_objid( tmp, vp->name, vp->name_length );
      if (!ver) {
   		if ( strncmp( tmp, old_ipMRouteNextHopState, strlen( old_ipMRouteNextHopState )))
   			break;
      } else {
   		if ( strncmp( tmp, new_ipMRouteNextHopState, strlen( new_ipMRouteNextHopState )))
   			break;
      }
		
		(void) get_address( vp->name, vp->name_length, &group, (!ver)? 14: 11);
		(void) get_address( vp->name, vp->name_length, &src, (!ver)? 18 : 15);

		sprint_objid( objID, vp->name, vp->name_length );
		if ( group == group_key && src == src_key ) {
			snmp_free_pdu( response );
			return 1;
		}
		
		snmp_free_pdu(response);
		response = snmpget( ss, GETNEXT_REQ_MSG, vars, 1, &err );
	}

	/* If we arrive here, then something went wrong. */
	if ( response ) snmp_free_pdu( response );
	return 0;
}

/*****************************************************************************
 * This routine ( along with get_prune_state() ) determines                  *
 * which of node's neighbors are on the _actual_ (as opposed to _potential_) *
 * multicast tree for <group,src>.  This routine assumes that node speaks    *
 * SNMP and supports the multicast mib.  It queries the ipMRouteNextHop      *
 * table in node's multicast mib to get the vif and vif state                *
 * ( pruned or forwarding ) for each vif that's a next hop for src.  If the  *
 * next hop is pruned, then that vif is ignored.  If not, then the neighbor  *
 * corresponding to that vif is looked up in the neighbor table and we jump  *
 * to that node to traverse it.                                              *
 *****************************************************************************/
void
get_prune_state_broken_cisco(ss, node, group_key, src_key, level, neighbor, num_neighbors)
   struct snmp_session *ss; /* An active SNMP session */
   node_t *node;            /* The node with which we have an SNMP sesion */
   u_int32 group_key;        /* Address of multicast group */
   u_int32 src_key;          /* Address of source */
   int level;
   neighbor_t *neighbor;
   int num_neighbors;
{
	int i, err;
	char objID[256], *vars[10];
	struct snmp_pdu *response;
	struct variable_list *vp;     /* SNMP Variable list Pointer */
	int vif, vif_state;
	node_t *child;
	u_int32 group,
	       src;

	/* Find the first row of table with < group_key, src_key > if one exists. */
	if ( !get_first_ipMRouteNextHop_match( ss, group_key, src_key, objID, node->ver ) )
		return;

	/* Now step through all rows with < group_key, src_key > */
	interrupted = 0;
	vars[0] = objID;

	response = snmpget(ss, GET_REQ_MSG, vars, 1, &err);
	while( !interrupted ) {

		if ( !response || !response->variables || interrupted ) 
			break;

		vp = response->variables;

		/* Have we stepped past the end of the NextHopState column? */
		/* Save new OID in objID so that we can do a get next on it. */ 
		sprint_objid( objID, vp->name, vp->name_length );
      if (!node->ver) {
		   if ( strncmp( objID, old_ipMRouteNextHopState, strlen( old_ipMRouteNextHopState )))
		   	return;
      } else {
		   if ( strncmp( objID, new_ipMRouteNextHopState, strlen( new_ipMRouteNextHopState )))
		   	return;
      }

		(void) get_address( vp->name, vp->name_length, &group, (!node->ver)? 14: 11);
		(void) get_address( vp->name, vp->name_length, &src, (!node->ver)? 18 : 15);

		if ( group != group_key || src != src_key )  {
			snmp_free_pdu( response );
			return;
		}

		vif       = vp->name[ (!node->ver)? 26 : 23];
		vif_state = *vp->val.integer;
		snmp_free_pdu( response );

		if ( verbose )
			printf("Saw state for Vif %d = %d\n", vif, vif_state );

		/* If the vif isn't forwarding, then we aren't interested. */
		if ( vif_state != FORWARDING ) {
		   response = snmpget(ss, GETNEXT_REQ_MSG, vars, 1, &err);
			continue;
		}

		if ( subnet_leaves ) { /* IE, If we're supposed to display subnets... */
			if ( ttl_flag && ( ttl - level ) <= neighbor[i].threshold ) {
		   	response = snmpget(ss, GETNEXT_REQ_MSG, vars, 1, &err);
				continue; 
			}
			show_subnet( ss, vif, level+1, (int) (node->snmp_flags & FL_PIM), 
			             neighbor, num_neighbors, node, group, src );
		}


		/********************************************************************
		 * Vif information for all neighbors in the neighbor table was      *
		 * filled in by get_dvmrp_neighbors.  Find each neighbor in the     *
		 * table whose vif number matches the one just returned from the    *
		 * query.                                                           *
		 * For each match, if the vif is pruned (i.e. not FORWARDING )      *
		 * ignore it. Otherwise, continue our depth-first search by         *
		 * visiting that neighbor.                                          *
		 * Note that it is possible to get a vif number as the result of    *
		 * an snmp query which is not found in the neighbor table.  This    *
		 * is because the neighbor table may be 'incomplete'. For           *
		 * example, it won't have entries for neighbors that are down.      *
		 ********************************************************************/
		for ( i = 0; i < num_neighbors; i++ ) {
			if ( neighbor[i].vif == vif )  {
				if ( ttl_flag && ( ttl - level ) <= neighbor[i].threshold )
					continue; 
				if ( subnet_leaves )
					show_subnet( ss, vif, level+1, (int) (node->snmp_flags & FL_PIM), 
					                        neighbor, num_neighbors, node, group, src );
				child = new_node( neighbor[i].localaddr, neighbor[i].remoteaddr,
				                  (char*) NULL, neighbor[i].on_subnet,
				                  neighbor[i].shares_subnet ); 
				do_tree( node, child, level+1, group, src );
			}
		}
		response = snmpget(ss, GETNEXT_REQ_MSG, vars, 1, &err);
	}

	/* If we arrived here, then something went wrong. */
	if ( response )
		snmp_free_pdu( response );
}

/*****************************************************************************
 * This routine ( along with get_prune_state_broken_cisco() ) determines     *
 * which of node's neighbors are on the _actual_ (as opposed to _potential_) *
 * multicast tree for <group,src>.  It does this by one of two methods.      *
 * 1) If node speaks SNMP and supports the multicast mib, then the routine   *
 *    queries the ipMRouteNextHop table in node's multicast mib to get the   *
 *    vif and vif state ( pruned or forwarding ) for each vif that's a next  *
 *    hop for src.  If the next hop is pruned, then that vif is ignored.     *
 *    If not, then the neighbor corresponding to that vif is looked up in    *
 *    the neighbor table and we jump to that node to traverse it.            *
 * 2) If node does not support the multicast mib, then we traverse every     *
 *    neighbor in the neighbor table via a call to do_tree(). do_tree() will *
 *    eventually call VrfyOnTreeWithGroup() to determine if the neighbor     *
 *    is on the multicast tree in which we're interested.                    *
 *****************************************************************************/
void
get_prune_state(ss, node, group, src, level, neighbor, num_neighbors)
   struct snmp_session *ss; /* An active SNMP session */
   node_t *node;            /* The node with which we have an SNMP sesion */
   unsigned int32 group;     /* Address of multicast group */
   unsigned int32 src;       /* Address of source */
   int level;
   neighbor_t *neighbor;
   int num_neighbors;
{
	int i, err;
	char initvar[10][256], var[256], *vars[10];
	struct snmp_pdu *response;
	struct variable_list *vp;     /* SNMP Variable list Pointer */
	int vif, vif_state;
	node_t *child;
	enum TriVal is_broken;

	if ( node->snmp_flags & FL_MCAST && !( node->snmp_flags & FL_SNMP_NR ) ) {

		is_broken = is_broken_cisco( ss, node );

		if ( is_broken == yes ) {
			get_prune_state_broken_cisco( ss, node, group, src, 
			                              level, neighbor, num_neighbors );
			return;
		}

		if ( is_broken == unknown ) /* Something went wrong w/the is_broken call */
			return;

		/************************************************************************
		 * This table is indexed by <Group, Source, SourceMask, Index>. Just do *
		 * a get next on < Group, Source > to try and find the first entry we   *
		 * need.                                                                * 
		 ************************************************************************/
		sprintf( initvar[0], "%s.%s.", 
       (!node->ver)? old_ipMRouteNextHopState : new_ipMRouteNextHopState, 
       inet_fmt( group, s ) );
		strcat( initvar[0], inet_fmt( src, s ) );
		strcpy( var, initvar[0] );
		vars[0] = var;
	
	
		/* Get and dump the table */
		interrupted = 0;
		while( !interrupted ) {
			response = snmpget(ss, GETNEXT_REQ_MSG, vars, 1, &err);
	
			if (!response || !response->variables || interrupted ) 
				break;
	
			vp = response->variables;
	
			/****************************************************************
			 * Have we stepped past the last ipMRouteNextHopState instance  *
			 * of interest? If so, then it's time to leave.                 *
			 ****************************************************************/
	
			/* Save new OID in var so that we can do a get next on it. */ 
			sprint_objid( var, vp->name, vp->name_length );
			if ( strncmp( var, initvar[0], strlen(initvar[0]) ) ) {
				snmp_free_pdu( response );
				return;
			}
	
			vif       = vp->name[ (!node->ver)? 26 : 23];
			vif_state = *vp->val.integer;
			snmp_free_pdu( response );
	
			if ( verbose )
				printf("Saw state for Vif %d = %d\n", vif, vif_state );

			/* If the vif isn't forwarding, then we aren't interested in it. */
			if ( vif_state != FORWARDING )
				continue;

			/********************************************************************
			 * Vif information for all of the neighbors in the neighbor table   *
			 * was filled in by get_dvmrp_neighbors.  Find each neighbor in the *
			 * table whose vif number matches the one just returned from the    *
			 * query.                                                           *
			 * For each match, if the vif is pruned ignore it. Otherwise,       *
			 * continue our depth-first search by visiting that neighbor.       *
			 * Note that it is possible to get a vif number as the result of    *
			 * an snmp query which is not found in the neighbor table.  This    *
			 * is because the neighbor table may be 'incomplete'. For           *
			 * example, it won't have entries for neighbors that are down.      *
			 ********************************************************************/
			for ( i = 0; i < num_neighbors; i++ ) {
				if ( neighbor[i].vif == vif )  {
					if ( ttl_flag && ( ttl - level ) <= neighbor[i].threshold )
						continue; 
					if ( subnet_leaves ) /* IE, If we should display subnets... */
						show_subnet( ss, vif, level+1, (int) (node->snmp_flags & FL_PIM), 
						             neighbor, num_neighbors, node, group, src );
					child = new_node( neighbor[i].localaddr, neighbor[i].remoteaddr,
					                  (char*) NULL, neighbor[i].on_subnet,
					                  neighbor[i].shares_subnet ); 
					do_tree( node, child, level+1, group, src );
				}
			}
		}

		/* If we arrived here, then something went wrong. */
		if ( response )
			snmp_free_pdu( response );
	}
	else {
	 /***********************************************************************
		* If node does not support the multicast mib, then traverse all it's  *
		* neighbors an let VrfyOnTreeWithGroup() determine if the neighbor is *
		* on the multicast tree that we're interested in.                     *
		***********************************************************************/
		for ( i = 0; i < num_neighbors; i++ ) {
			if ( !node->parent || neighbor[i].remoteaddr != node->parent->addr ) {
				if ( ttl_flag && ( ttl - level ) <= neighbor[i].threshold )
					continue; 
				child = new_node( neighbor[i].localaddr, neighbor[i].remoteaddr,
						              (char*) NULL, neighbor[i].on_subnet,
				                  neighbor[i].shares_subnet ); 
				do_tree( node, child, level+1, group, src );
			}
		}
	}
}

/****************************************************************************
 * This routine gets attempts to determine the number of packets and 
 * octets being forwarded by node for the given group and source.  It does 
 * this by first checking the node info to see if the node supports the 
 * multicast mib. If so, it then issues SNMP queries to the node for the 
 * ipMRoutePkts and ipMRouteOctets counts for the (group,source) pair.  The 
 * result of these queries is stored in the node's node_t structure.
 *
 * If the node does not support the multicast mib, then essentially
 * nothing happens - no queries, no updates of data structures, etc.
 ****************************************************************************/
void
get_packet_count(node, group, src)
	node_t *node;
	unsigned int32 group;
	unsigned int32 src;
{
	int err;

	/* Get packet count */
	if (node->snmp_flags & FL_MCAST) {
		char pkts[256], byts[256], *arr[2];
		char ans[2][256];

		sprintf(pkts, "%s.%s.", (!node->ver)? old_ipMRoutePkts : new_ipMRoutePkts,
       inet_fmt(group,s));
		strcat(pkts, inet_fmt(src,s));
		strcat(pkts, ".255.255.255.255");
		arr[0] = pkts;

		sprintf(byts, "%s.%s.", (!node->ver)? old_ipMRouteOctets : new_ipMRouteOctets, 
       inet_fmt(group,s));
		strcat(byts, inet_fmt(src,s));
		strcat(byts, ".255.255.255.255");
		arr[1] = byts;

		snmpget_list(ans, inet_fmt(node->addr,s), arr, 2, &err, (int*) &node->timestamp);
#if 0
		if (!snmpget_list(ans, inet_fmt(node->addr,s), arr, 2, &err, &node->timestamp))
         node->snmp_flags &= ~FL_MCAST; /* XXX Prob. aggregates across srcs */
#endif

		if (err==SNMP_ERR_NOSUCHNAME) {
			strcpy(ans[0], "0");
			strcpy(ans[1], "0");
		}
		node->pkts = atoi(ans[0]);
		node->byts = atoi(ans[1]);

		/* Output rest of line for this node */
#if (SIZEOF_LONG == 8)
		printf(", %d pkts", node->pkts);
#else
		printf(", %ld pkts", node->pkts);
#endif
		if (root_pkts)
#if (SIZEOF_LONG == 8)
			printf(" (%d%%)", (100*node->pkts)/root_pkts);
#else
			printf(" (%ld%%)", (100*node->pkts)/root_pkts);
#endif
	}
}

int get_ipMRouteUpstrmNghbr( ss, srcaddr,   groupaddr, 
                             nghbraddr_ptr, mask_ptr, ver )
	struct snmp_session *ss; /* A valid session handle. */
	u_int32 srcaddr,  /* all in network order */
	       groupaddr,
	      *nghbraddr_ptr,
	      *mask_ptr; /* in network order */
   int    ver;      /* 0=use old OIDs, 1=use new OIDs */
{
	char tmp[256], *ans;
	char t1[30], t2[30], t3[30]; /* Temp. arrays. */

	int err;
	u_int32 mask; /* in host order */

	for ( mask = 0xFFFFFFFF; mask > 0; mask <<= 1 ) {
		sprintf(tmp, "%s.%s.%s.%s", 
       (!ver)? old_ipMRouteUpstreamNeighbor : new_ipMRouteUpstreamNeighbor, 
		                            inet_fmt( groupaddr, t1 ),
		                            inet_fmt( srcaddr & htonl(mask), t2 ),
		                            inet_fmt( htonl(mask), t3 ) );
		ans = snmpget_string( ss, tmp, &err, (int*) NULL);
		if (ans)
			break;
	}

	if ( ans ) {
		/* Interesting run-time bug here. Not sure why, but you can't hand
		 * ans to inet_addr. You must copy it into something else and hand that
		 * "something else" off to inet_addr. */
		strcpy( t1, ans );
		*nghbraddr_ptr = inet_addr( t1 );
		*mask_ptr = htonl(mask);
		return 1;
	}

	*nghbraddr_ptr = 0;
	*mask_ptr      = 0;
	return 0;
}


int 
get_dvmrpUpstrmNghbr( ss, srcaddr, nghbraddr_ptr, mask_ptr )
	struct snmp_session *ss; /* A valid session handle. */
	u_int32 srcaddr, /* all in network order */
	       *nghbraddr_ptr,
	       *mask_ptr;        /* Pointer to mask in network order */
{
	char tmp[256], *ans;
	char t1[30], t2[30]; /* Temp. arrays. */
	int err;
	u_int32 mask; /* mask in host order */

	for ( mask = 0xFFFFFFFF; mask > 0; mask <<= 1 ) {
		sprintf(tmp, "%s.%s.%s", dvmrpRouteUpstreamNeighbor,
		                         inet_fmt( srcaddr & htonl(mask), t1 ),
		                         inet_fmt( htonl(mask), t2 ) );

		ans = snmpget_string( ss, tmp, &err, (int*) NULL);
		if (ans)
			break;
	}

	if ( ans ) {
		/* Interesting run-time bug here. Not sure why, but you can't hand
		 * ans to inet_addr. You must copy it into something else and hand that
		 * "something else" off to inet_addr. */
		strcpy( t1, ans );
		*nghbraddr_ptr = inet_addr( t1 );
		*mask_ptr = htonl(mask);
		return 1;
	}

	*nghbraddr_ptr = 0;
	*mask_ptr      = 0;
	return 0;
}


/******************************************************************************
 * This routine tries to figure out the upstream neighbor for <src> alone
 * if group=0, or for <src,group> if group!=0.  If, for some reason, it
 * cannot determine the upstream neighbor (perhaps because one doesn't
 * exist for the <group,src> pair), then the function returns a 0.
 * Otherwise, it returns a 1.
 *
 * In the case when a neighbor is determined, the address of the neighbor 
 * is placed in *neighbor and the mask used to route to src is placed in
 * *mask.  (Remember, src is a net address and not a host address. As such,
 * we can't treat it as a CIDR route without a mask.)
 ******************************************************************************/
int 
get_UpstrmNghbr( ss, node, group, src, neighbor, mask )
	struct snmp_session *ss; /* A valid session handle. */
	node_t *node;
	u_int32 group, src, *neighbor, *mask;
{
	int rc    = 0;
	*neighbor = 0;
	*mask     = 0;

	if ( group ) {
		if ( node->snmp_flags & FL_MCAST )
			rc = get_ipMRouteUpstrmNghbr(ss,src,group, neighbor, mask, node->ver );
	}
	else {
		if ( node->snmp_flags & FL_DVMRP )
			rc = get_dvmrpUpstrmNghbr( ss, src, neighbor, mask );
	}
	return rc;
}


/******************************************************************************
 * This routine is called when we are trying to determine the _potential_
 * multicast tree rooted at a given node for a given source.  The
 * logic in this routine is almost exactly the same as that found in
 * VrfyOnTreeWithGroup() which is used when we're trying to determine the
 * _actual_ multicast tree.  The only difference is that in order to
 * determine if node is on the potential tree, either node or it's parent
 * must support the DVMRP mib.  This is because, of the mibs available,
 * only the DVMRP mib contains info for all _potential_ sources, not just
 * those sources which are _active_.
 *
 * This routine returns 'yes' if 'node' definitely is on the multicast tree,
 * 'no' if it is definitely is not on the tree, and 'unknown' if we can't
 * tell.
 ******************************************************************************/
enum TriVal
VrfyOnTreeSansGroup( ss, node, src)
	struct snmp_session *ss; /* A valid session handle. */
	node_t *node;
	u_int32 src;
{
	u_int32 neighbor = 0,
	       mask     = 0;

	if ( node->parent == NULL ) {
		/* If we arrived here then we know that node is the root. */
		if ( node->snmp_flags & FL_DVMRP && !( node->snmp_flags & FL_SNMP_NR ) ) {
			if ( get_dvmrpUpstrmNghbr( ss, src, &neighbor, &mask ) ) {
				node->raddr = src & mask;
				node->rmask = mask;
				return yes;
			}
			else
				return no;
		}
		else
			return unknown;
	}

	/* If we arrived here then we know that node is not the root. */
	if ( node->snmp_flags & FL_DVMRP && !( node->snmp_flags & FL_SNMP_NR ) ) {
		(void) get_dvmrpUpstrmNghbr( ss, src, &neighbor, &mask );
		if ( neighbor != node->parentaddr )
			return no;

		node->raddr = src & mask;
		node->rmask = mask;
		return yes;
	}

	/* If we arrived here then we know that node is not root AND either
	 * (a) node does not support the dvmrp or multicast mibs or 
	 * (b) node is supposed to support dvmrp or mutlicast mibs but isn't
	 *     responding.
	 */

	if ( node->parent->snmp_flags & FL_DVMRP &&
	              !( node->parent->snmp_flags & FL_SNMP_NR ) ) {
		if ( node->on_subnet == no )
			return yes;

		if ( node->shares_subnet == no )
			return yes;

		return unknown;
	}

	return unknown;
}

/******************************************************************************
 * This routine is called when we are trying to determine the _actual_
 * multicast tree rooted at a given node for a given source and group.  The
 * logic in this routine is almost exactly the same as that found in
 * VrfyOnTreeSansGroup() which is used when we're trying to determine the
 * _potential_ multicast tree.  The only difference is that in order to
 * determine if node is on the actual tree, either node or it's enough
 * if node or its parent supports either the DVMRP mib or the MULTICAST
 * mib.  The info we need to make the determination appears in both mibs.
 *
 * This routine returns 'yes' if 'node' definitely is on the multicast tree,
 * 'no' if it is definitely is not on the tree, and 'unknown' if we can't
 * tell.
 ******************************************************************************/
enum TriVal
VrfyOnTreeWithGroup( ss, node, group, src )
	struct snmp_session *ss; /* A valid session handle. */
	node_t *node;
	u_int32 group,
	       src;
{
	u_int32 neighbor = 0,
	       mask     = 0;

	if ( node->parent == NULL ) {
		/* If we arrived here then we know that node is the root. */
		if ( node->snmp_flags & ( FL_MCAST | FL_DVMRP ) &&
		                !( node->snmp_flags & FL_SNMP_NR ) ) {
			if ( get_UpstrmNghbr( ss, node, group, src, &neighbor, &mask ) ) {
				node->raddr = src & mask;
				node->rmask = mask;
				return yes;
			}
			else
				return no;
		}
		else
			return unknown;
	}

	/* If we arrived here then we know that node is not the root. */
	if ( node->snmp_flags & ( FL_MCAST | FL_DVMRP ) && 
	              !( node->snmp_flags & FL_SNMP_NR ) ) {
		(void) get_UpstrmNghbr( ss, node, group, src, &neighbor, &mask );
		if ( neighbor != node->parentaddr )
			return no;

		node->raddr = src & mask;
		node->rmask = mask;
		return yes;
	}

	/* If we arrived here then we know that node is not root AND either
	 * (a) node does not support the dvmrp or multicast mibs or 
	 * (b) node is supposed to support dvmrp or mutlicast mibs but isn't
	 *     responding.
	 */

	if ( node->parent->snmp_flags & ( FL_MCAST | FL_DVMRP ) &&
	              !( node->parent->snmp_flags & FL_SNMP_NR ) ) {
		if ( node->on_subnet == no )
			return yes;

		if ( node->shares_subnet == no )
			return yes;

		return unknown;
	}

	return unknown;
}

/*****************************************************************************
 * This routine figures out "parent's" neighbors through a call to mrinfo.   *
 * If it turns out that the parent speaks PIM or DVMRP and also speaks       *
 * SNMP, then some information on PIM and DVMRP neighbors are determined     *
 * with an SNMP query.  Along the way, the routine prints out some basic     *
 * things that it discovers about parent.                                    *
 *                                                                           *
 * Note that either VrfyOnTreeWithGroup() or VrfyOnTreeSansGroup() is        *
 * called for every child.  This is in order to determine that child         *
 * is indeed on the multicast tree for our <source,group> ( or <source> ).   *
 * The reason we must perform this step is because when we enter this        *
 * routine, we have not yet determined if child is on the multicast tree     *
 * we're interested in.  This could occur when the parent doesn't speak      *
 * SNMP and so we couldn't query it's mib to determine which of parent's     *
 * neighbors were next hops (aka children) on the tree.  Thus, to make our   *
 * best effort at discovering the tree, we look at each neighbor in turn     *
 * (calling do_tree() for each neighbor) in the hope that they speak SNMP    *
 * and are on the tree of interest.                                          *
 *****************************************************************************/
void
do_tree( parent, child, level, group, src )
	node_t *parent,
	       *child;
	int    level;      /* Tree level. 0 = root */
	u_int32 group,      /* Address of multicast group */
	       src;        /* Address of source */
{
	struct snmp_session *ss;

	neighbor_t neighbor[MAX_NEIGHBORS];
	int        num_neighbors = 0,
	           rc,              /* Return Code */
	           i;
	enum TriVal tree_membership; /* Just a return code from VrfyOnTree */

	/* First do mrinfo: This will give us the version and a list of    *
	 * possibly relevant neighbors (not necessarily _all_ neighbors).  */
	rc = mrinfo(child->addr,(u_int32*) &child->version,neighbor,&num_neighbors,0);

	if ( !rc ) {
		print_name( child, level );
		/* If the child didn't respond to the neighbors msg, we may still know
		 * something about its tree membership.  If its parent speaks
		 * snmp, then we know for sure that it's on the tree.  If its parent
		 * doesn't speak snmp, then we can't know for sure.
		 */
		if ( child->parent ) {
			if ( child->parent->snmp_flags & FL_NO_SNMP )
				printf(" ?\n");
			else
				printf("\n");
		}
		else
			printf(" ?\n");

		for ( i = 0; i <= level; i++ )
			printf("   ");
		printf("... (%s did not respond to DVMRP 'NEIGHBORS' msg)\n", child->name);
		return;
	}

	/* Add node to the tree */
	add_child( parent, child );

	ss = get_mibs_supported( child );
	/* ss may be NULL of child doesn't speak snmp. */


	if ( group ) /* We're determining the _actual_ tree */
		tree_membership = VrfyOnTreeWithGroup( ss, child, group, src );
	else         /* We're determining the _potential_ tree */
		tree_membership = VrfyOnTreeSansGroup( ss, child, src );

	if ( tree_membership == unknown ) {
		print_name( child, level );
		printf(" ?\n");
		if ( ss )
			snmp_close( ss );
		return;
	}

	if ( tree_membership == no ) {
		/* If there is in fact a parent, remove the child. */
		if ( child->parent ) 
			remove_node( child );
		if ( ss )
			snmp_close( ss );
		return;
	}

	/* If we made it this far, then the node must be on the tree. */

	print_name( child, level );
	if (group) {
		get_packet_count( child, group, src );
		if ( level == 0 )
			root_pkts = child->pkts;
	}
	printf("\n");

	if (!(child->snmp_flags & FL_MCAST)) {/* If 'child' doesn't support mcast...*/
		for (i=0; i<=level; i++)
			printf("   ");

		if ( child->snmp_flags & FL_SNMP_NR )
			printf("... (No response from %s)\n", child->name);

		else if ( child->snmp_flags & FL_NO_SNMP )
			printf("... (%s does not support SNMP)\n", child->name);

		else
			printf("... (something else happened at %s)\n", child->name);
	}


	/*
	 * These calls to get_pim_neighbors and get_dvmrp_neighbors are somewhat 
	 * redundant because neighbors were already determined with the call to 
	 * mrinfo.
	 */
	if ( child->snmp_flags & FL_PIM )
		get_pim_neighbors( ss, neighbor, &num_neighbors, child->ver );

	if ( child->snmp_flags & FL_DVMRP )
		get_dvmrp_neighbors( ss, neighbor, &num_neighbors, child->ver );

	if ( group ) /* we're discovering the _actual_ tree */
		/* Traverse node's neighbors if they aren't pruned.... */
		get_prune_state( ss, child, group, src, level, neighbor, num_neighbors );
	else /* we're discovering the _potential_ tree */
		/* Traverse node's neighbors if we can.... */
		get_tree_state( ss, child, src, level, neighbor, num_neighbors );

	if ( ss )
		snmp_close( ss );
}

void
do_stats(node, level, group, src)
	node_t       *node;  /* Node for which we'll gather stats. */
	int           level; /* Tree level */
	unsigned int32 group; /* Address of multicast group */
	unsigned int32 src;   /* Address of source */
{
	node_t *child;
	int i, err;
	int newtime;
	u_int32 newpkts, newbyts;

	/* Get new packet counts */
	if (node->snmp_flags & FL_MCAST) {
		char pkts[256], byts[256], *arr[2];
		char ans[2][256];

		sprintf(pkts, "%s.%s.", (!node->ver)? old_ipMRoutePkts : new_ipMRoutePkts,
       inet_fmt(group,s));
		strcat(pkts, inet_fmt(src,s));
		strcat(pkts, ".255.255.255.255");
		arr[0] = pkts;

		sprintf(byts, "%s.%s.", (!node->ver)? old_ipMRouteOctets : new_ipMRouteOctets, 
       inet_fmt(group,s));
		strcat(byts, inet_fmt(src,s));
		strcat(byts, ".255.255.255.255");
		arr[1] = byts;

		snmpget_list(ans, inet_fmt(node->addr,s), arr, 2, &err, &newtime);
		if ( err == SNMP_ERR_NOSUCHNAME ) {
			strcpy(ans[0], "0");
			strcpy(ans[1], "0");
		}

		newpkts = atoi(ans[0]);
		newbyts = atoi(ans[1]);
	}
  
	/* Output a line for this node */
	for (i=0; i<level; i++)
		printf("   ");

	printf("%2d ", level);
	if (addronly)
		printf("%s", inet_fmt(node->addr,s));
	else
		printf("%s (%s)", node->name, inet_fmt(node->addr,s));

	if (node->snmp_flags & FL_MCAST) {
		int timedelta = (newtime-node->timestamp);
#if (SIZEOF_LONG == 8)
		printf(": %.2f secs, %d pkts (%.2f pps), %ld bytes (%.2f kbps)", 
#else
		printf(": %.2f secs, %ld pkts (%.2f pps), %ld bytes (%.2f kbps)", 
#endif
		            timedelta/100.0,                        newpkts-node->pkts,
		            (100.0*(newpkts-node->pkts))/timedelta, newbyts-node->byts,
		            (0.8*(newbyts-node->byts))/timedelta );
	}

	printf("\n");

	for (child = node->first_child; child; child = child->next_sibling)
		do_stats(child, level+1, group, src);
}

void
usage( prog_name )
char *prog_name;
{
	char *name;

	name = strrchr( prog_name, '/' );
	if ( !name )
		name = prog_name;
	else
		name++;

	fprintf(stderr,"Usage: %s [-nsfdl] [-t ttl] [-T secs] [-c community] [-p port] host\n", name);
	fprintf(stderr,"       %s [-nsfdl] [-t ttl] [-T secs] [-c community] [-p port] host", name );
	fprintf(stderr," source\n");
	fprintf(stderr,"       %s [-nsfdl] [-t ttl] [-T secs] [-c community] [-p port] host", name );
	fprintf(stderr," group source\n");
	fprintf(stderr," -n           Display IP addresses in numeric format\n");
	fprintf(stderr," -s           Include subnets as leaves\n");
	fprintf(stderr," -f           Trust only routed version level when determining SNMP capability\n");
	fprintf(stderr," -d           Compute statistics\n");
	fprintf(stderr," -l           Show local IP addresses of links\n");
	fprintf(stderr," -c community Use alternate community string\n");
	fprintf(stderr," -p port      Use alternate SNMP port number\n");
	fprintf(stderr," -t ttl       View tree for packets with time to live=ttl\n");
	fprintf(stderr," -T seconds   Timeout in seconds for IGMP/SNMP queries (default 2)\n");
	exit(1);
}

int
main(argc, argv)
	int    argc;
	char **argv;
{
	char hostname[80],       /* We're interested in the multicast subtree rooted
	                          * at hostname. */
	     srcname[80];        /* Multicast source that we're interested in. */
	u_int32 group = 0,	 /* IP address of group in network order */
	       src   = 0,        /* IP address of srcname */
	       host  = 0;        /* IP address of hostname */
	int unicast = 0;
	node_t        *first_node;
	int c;
	struct hostent *hostentp;

	extern char *optarg;
	extern int optind;


	timeout = 2; /* Number of _seconds_ before timing out on queries */
	retries = 3; /* Number of times to retry a single query */
	srcname[0] = hostname[0] = '\0';

  while ( ( c = getopt( argc, argv, "t:T:c:p:r:dunlsvfh" ) ) != -1 )
		switch( c ) {
			case 'c':
				community = optarg;
				break;
			case 'd':
				compute_stats++;
				break;
			case 'u':
				unicast++;
				break;
			case 'f':
				check_maybes = 0;
				break;
			case 'n':
				addronly++;
				break;
			case 'l':
				showlocal++;
				break;
			case 'p':
				dest_port = atoi( optarg );
				break;
			case 'r':
				retries = atoi( optarg );
				break;
			case 'T':
				timeout_flag++;
				timeout = atoi( optarg );
				break;
			case 't':
				ttl_flag++;
				ttl = atoi( optarg );
				break;
			case 's':
				subnet_leaves = 0;
				break;
			case 'v':
				verbose++;
				break;
			case 'h':
			case '?':
			default:
				usage( argv[0] ); /* usage() never returns. */
	}

	if ( argc - optind == 0 ) {
		fprintf(stderr,"You must at least specify a destination router.\n");
		usage( argv[0] );
		exit( -1 ); /* Never reached. usage() exits. */
	}

	if ( geteuid() != 0 ) {
      printf("Error: this application must be run as root\n");
      exit( 1 );
	}

	/**************************************************************************
	 * If there was 1 non-flagged option, it must be 'hostname'.  In this     *
	 * mode, hostname is assumed to be both the root of the multicast         *
	 * tree we're interested in, as well as the source we're interested in.   *
	 **************************************************************************/
	if ( argc - optind == 1 ) {
			strcpy( hostname, argv[ optind ] );
			strcpy( srcname, hostname );
			printf("Potential distribution tree from %s...\n", hostname);
	}
	
	/******************************************************************
	 * If there were 2 non-flagged options, they must be hostname     *
	 * and srcname - in that order.                                   *
	 ******************************************************************/
	else if ( argc - optind == 2 ) {
		strcpy( hostname, argv[ optind++ ] );
		strcpy( srcname, argv[ optind ] );
		printf("Potential distribution tree from host %s for source %s...\n", hostname, srcname );
	}

	/*****************************************************************
	 * If there were 3 non-flagged options, they must be hostname,   *
	 * group address and source name - in that order.                *
	 *****************************************************************/
	else if ( argc - optind == 3 ) {
		strcpy( hostname, argv[ optind++ ] );

		if ( ! IN_MULTICAST( ntohl( group = inet_addr( argv[ optind++ ] ) ) ) ) {
			fprintf(stderr,"Error: %s is not a valid multicast address.\n",
			                                              argv[optind] );
			usage( argv[0] ); /* Doesn't return. */
		}

		strcpy( srcname, argv[ optind ] );

		printf("Actual distribution tree rooted ");
		printf("at %s for group %s and source %s...\n", hostname, inet_fmt(group,s),
		                                                srcname);
	}


	/* Translate the hostname into an IP address. */
	if ( isdigit(hostname[0]) ) {
		host = inet_addr( hostname );
		hostentp = gethostbyaddr( (char*) &host, sizeof(host), AF_INET );
		if ( hostentp )
			strcpy( hostname, hostentp->h_name );
	}
	else {
		hostentp = gethostbyname( hostname );
		if ( !hostentp ) {
			printf("unknown host: %s\n", hostname);
			exit( 1 );
		}
		host = *(u_int32 *) hostentp->h_addr_list[0];
	}

	init(host);


	/* Translate the sourcename into an IP address if the addronly 
	 * flag isn't set. */
	if ( isdigit( srcname[0] ) ) {
		src = inet_addr( srcname );
		hostentp = gethostbyaddr( (char*) &src, sizeof(src), AF_INET );
		if ( hostentp )
			strcpy( srcname, hostentp->h_name );
	}
	else {
		hostentp = gethostbyname( srcname );
		if ( !hostentp ) {
			printf("unknown host: %s\n", srcname);
			exit(1);
		}
		src  = *(u_int32 *) hostentp->h_addr_list[0];
	} 

	if ( addronly )
		first_node = new_node( (u_int32) 0, host, (char*) NULL, unknown, unknown );
	else
		first_node = new_node( (u_int32) 0, host, hostname, unknown, unknown );

	/* Remember, host is the subtree root, src is the original root */
	do_tree( (node_t*) NULL, first_node, 0, group, src );
   
	if (compute_stats) {
		printf("\nPausing to compute stats...");
		sleep(5);
		printf("\nStatistics:\n");
		do_stats(root_node, 0, group, src);
	}

	return 0; /* To practice safe computing... and make lint happy. */
}
