#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#include "../snmplib/asn1.h"
#include "../snmplib/snmp_api.h"
#include "util.h"
#ifndef HAVE_GETHOSTNAME
#include <sys/systeminfo.h>
#endif

const int tclFlag = 0;

#define DELAY_TIME 10

void
output_row(str)
   char *str;
{
   printf("%s", str);
}

void
output_header(str)
   char *str;
{
   printf("%s", str);
}

void
setAnswers(vp, cumul, oldcumul,nvars,row)
   struct variable_list *vp;
   char cumul[10][256];
   char oldcumul[10][256];
   int nvars;
   int row;
{
}

void
usage()
{
   int i;
   char buff[256], *p;

   fprintf(stderr, "Usage: mstat [-");
   for (i=0; tables[i].opt; i++)
      putc(tables[i].opt, stderr);
   fprintf(stderr, "nDR] [-C count] [-T delay] [-c community] [-p port] [-x prefix] [host]\n");
   for (i=0; tables[i].opt; i++) {
      strcpy(buff, tables[i].header);
      if ((p = strstr(buff, " for ")))
         *p = '\0';
      fprintf(stderr, " -%c            Dump %s", tables[i].opt, buff);
      if (table==i)
         fprintf(stderr, " (default)\n");
      else
         fprintf(stderr, "\n");
   }
   fprintf(stderr, " -n            Display IP addresses in numeric format\n");
   fprintf(stderr, " -D            Display deltas rather than values\n");
   fprintf(stderr, " -R            Display rates rather than values\n");
   fprintf(stderr, " -C count      Repeat table 'count' times (default 1)\n");
   fprintf(stderr, " -T delay      Delay between iterations (default %d)\n",
    DELAY_TIME);
   fprintf(stderr, " -c community  Use alternate community string\n");
   fprintf(stderr, " -p port       Use alternate SNMP port number\n");
   fprintf(stderr, " -x prefix     Use instance prefix to limit table\n");
   fprintf(stderr, " host          Host to query.  If none given, use localhost.\n");
   exit(1);
}

int
main(argc, argv)
   int    argc;
   char **argv;
{
   char *ap;

   int port_flag = 0;
   int	arg;
   int i;
   int ver; /* 0=use old OIDs, 1=use new OIDs */

   hostname[0] = '\0';

   for(arg = 1; arg < argc; arg++){
      ap = argv[arg];
      if (*ap == '-') {
         while (*++ap) {
            for (i=0; tables[i].opt && tables[i].opt!= *ap; i++);
            if (tables[i].opt) {
               table = i;
               if (tables[i].opt=='g') /* Special case option for groups */
                  maskout = 8;
               continue;
            }
            switch(*ap) {
            case 'D':
               mode = DELTA;
               break;
            case 'R':
               mode = RATE;
               break;
            case 'T':
               if (argc <= arg+1)
                  usage();
               delay_time = atoi(argv[++arg]);
               break;
            case 'C':
               if (argc <= arg+1)
                  usage();
               maxcount = atoi(argv[++arg]);
               break;
            case 'c':
               if (argc <= arg+1)
                  usage();
               community = argv[++arg];
               break;
            case 'm':
               if (argc <= arg+1)
                  usage();
               maskout = atoi(argv[++arg]);
               break;
            case 'n':
               addronly++;
               break;
            case 'p':
               if (argc <= arg+1)
                  usage();
               port_flag++;
               dest_port = atoi(argv[++arg]);
               break;
/*
            case 'r':
               if (argc <= arg+1)
                  usage();
               retransmission = atoi(argv[++arg]);
               break;
            case 't':
               if (argc <= arg+1)
                  usage();
               timeout = atoi(argv[++arg]) * 1000000L;
               break;
*/
            case 'x': /* Specify OID instance prefix */
               if (argc <= arg+1)
                  usage();
               prefix = argv[++arg];
               if (prefix == NULL || *prefix == '-')
                  usage();
/*
               if (inet_addr(prefix) == -1) {
                  fprintf(stderr, "%s: invalid address", prefix);
                  exit(-1);
               }
*/
               break;
            default:
               usage();
            }
            continue;
         }
         continue;
      }
      if (!hostname[0])
         strcpy(hostname, argv[arg]);
   }

   if (!hostname[0])
#ifndef HAVE_GETHOSTNAME
      sysinfo(SI_HOSTNAME, hostname, 80);
#else
      gethostname(hostname,80);
#endif

   ss = open_session(hostname, dest_port, community, retransmission, timeout, &ver);
   display_table(ss, ver);
   snmp_close(ss);

   exit(0);
}
